/* Copyright is licensed under GNU LGPL.                 by I.J.Wang 2005

  Read arithmetic expression from standard input and write along with the
  evaluated result to standard output.

  Build: make a_cal

  Note: A small calculator program deserves reference: e-0.02718.tar.gz,
        http://freshmeat.net/projects/e-expr/
*/
#include "WyNum.h"      // for Wy::_strnum
#include "WyStr.h"
#include "WyRegFile.h"  // for system regular file and Wy::cout
#include <cmath>        // for sin,cos..

/*
 [Parsing Grammar]

 <E>::= [+-] <T> + <U> 
        [+-] <T> [<N>]

 <U>::=  <T> + <U> 
         <T> [<N>]

 <N>::= -<T> + <U> 
        -<T> [<N>] 

 <T>::= <T> * <F> 
        <T> / <F> 
        <F>

 <F>::= <ddd> | '(' <E> ')'
        'log(' <E> ')'
        'sqrt(' <E> ')'
        'sin(' <E> ')'
        'cos(' <E> ')'
        'tan(' <E> ')'
        M_PI
     
 <ddd>::= number string
*/

typedef double NumType;

static NumType E(WyCSeg s);
static NumType U(WyCSeg s);
static NumType N(WyCSeg s);
static NumType T(WyCSeg s);
static NumType F(WyCSeg s);

static bool is_failed(NumType n) 
{ return std::isnan(n);
};

static const NumType fail_num()
{
 static const NumType n=std::log(-1.0);
 return(n);
};

// [Syn] Remove blanks in str
//
// [Ret] Size of the resultant str
//
static size_t no_blanks(WyStr& str)
{
 static const WyCSeg Blanks=" \t\n\r\v\f";
 size_t i,j;
 if(str.findc(i,0,Blanks)==false) {
   return str.size();
 }
 for( j=i; str.findc_not(j,j+1,Blanks); ++i) {
    str[i]=str[j];
 }
 str.resize(i);   // shrinking always succeeds
 return str.size();
};

// [Syn] Check whether fterm denotes a function term, and strip off the bracing
//       symbols. e.g. if name=="tan("  and fterm=="tan(xx)", the resultant
//                     fterm will be "xx" and return true.
//
// [Ret] true= fterm denote a function term and is modified
//      false= otherwise
//
static bool func(const WyCSeg& name, WyCSeg& fterm)
{
 if( (name!=fterm.cseg(0,name.size())) || (fterm.back()!=')') ) {
   return false;
 }
 fterm=fterm.cseg(name.size(),fterm.size()-name.size()-1);
 return true;
};

static NumType E(WyCSeg s)
{
 bool is_neg=false;
 NumType n1,n2;
 if(s.size()==0) {
   return fail_num();
 }
 if(s.front()=='+') {
   s.move_begin(1);
 } else if(s.front()=='-') {
   s.move_begin(1);
   is_neg=true;
 } else {};

 if(is_failed(n1=T(s))==false) {
   return is_neg? -n1 : n1;
 }
 for(size_t idx=1; s.findc(idx,idx,"+-"); ++idx) {
   if(is_failed(n1=T(s.cseg(0,idx)))) {
     continue;
   }
   if(s[idx]=='+') {
     if(is_failed(n2=U(s.cseg(idx+1)))) {
       continue;
     }
   } else {
     if(is_failed(n2=N(s.cseg(idx)))) {
       continue;
     }
   }
   return is_neg? n2-n1 : n1+n2;
 };
 return fail_num();
};

static NumType U(WyCSeg s)
{
 NumType n1,n2;
 if(is_failed(n1=T(s))==false) {
   return n1;
 }
 for(size_t idx=1; s.findc(idx,idx,"+-"); ++idx) {
   if(is_failed(n1=T(s.cseg(0,idx)))) {
     continue;
   }
   if(s[idx]=='+') {
     if(is_failed(n2=U(s.cseg(idx+1)))) {
       continue;
     }
   } else {
     if(is_failed(n2=N(s.cseg(idx)))) {
       continue;
     }
   }
   return n1+n2;
 };
 return fail_num();
};

static NumType N(WyCSeg s)
{
 NumType n1,n2;
 if((s.size()==0)||(s.front()!='-')) {
   return fail_num();
 }
 s.move_begin(1);
 if(is_failed(n1=T(s))==false) {
   return -n1;
 }
 for(size_t idx=1; s.findc(idx,idx,"+-"); ++idx) {
   if(is_failed(n1=T(s.cseg(0,idx)))) {
     continue;
   }
   if(s[idx]=='+') {
     if(is_failed(n2=U(s.cseg(idx+1)))) {
       continue;
     }
   } else {
     if(is_failed(n2=N(s.cseg(idx)))) {
       continue;
     }
   }
   return n2-n1;
 };
 return fail_num();
};

static NumType T(WyCSeg s)
{
 NumType n1,n2;
 if(is_failed(n1=F(s))==false) {
   return n1;
 }
 for(size_t idx=s.size(); s.rfindc(idx,idx,"*/"); ) {
   if(is_failed(n1=T(s.cseg(0,idx)))) {
     continue;
   }
   if(is_failed(n2=F(s.cseg(idx+1)))) {
     continue;
   }
   return (s[idx]=='*')?  n1*n2 : n1/n2;
 };
 return fail_num();
};

static NumType F(WyCSeg s)
{
 NumType n;

 if(s.size()==0) {
   return fail_num();
 }

 if((s.front()=='(')&&(s.back()==')')) {  // Try recognizing '(' <E> ')'
   n= E(s.cseg(1,s.size()-2));
 } else if(func("log(",s)) {   // Try recognizing func(<E>)
   n= std::log10( E(s) );
 } else if(func("sin(",s)) {
   n= std::sin( E(s) );
 } else if(func("cos(",s)) {
   n= std::cos( E(s) );
 } else if(func("tan(",s)) {
   n= std::tan( E(s) );
 } else if(func("sqrt(",s)) {
   n= std::sqrt( E(s) );
 } else if(s=="M_PI") {
   n= M_PI;
 } else {
   //
   // Try recognizing <ddd>
   //
   char d0=Wy::_charnum(s.front());
   if((d0<0)||(d0>9)) {     // Verify the first character of <ddd> is a decimal digit
     return fail_num();
   }
   if(Wy::_strnum(n,(const WyCSeg&)s)!=Ok) {
     return fail_num();
   }
 }
 return n;
};

int main(int, char*[])
try {
 WyRet r;
 WyRegFile efile;

 switch(Wy::cin.ftype()) {
   case S_IFREG: // Enable efile if Wy::cin is a regular file
                 if((r=efile.reset(Wy::cin))!=Ok) {
                   WY_THROW(r);
                 };
                 break;
   case S_IFCHR: break;  // Terminal assumed
   default: Wy::cerr << "Unsupported standard input device\n";
            return(-1);
 }
 for(;;) {
   //
   // Read line into str
   //
   WyStr str;
   if(efile.is_default()) {
     Wy::cin >> str;
   } else {
     size_t n_rd; // dummy
     if((r=efile._read_till(str,256,n_rd,'\n'))!=Ok) {
       WY_THROW(r);
     }
   }
   if(no_blanks(str)==0) { // note: expression like "44 5-2" becomes "445-2"
     return 0;   // eof or str is a blank line
   }

   // Evaluate and display str
   //
   const NumType n= E(str.cseg());
   if(is_failed(n)) {
     Wy::cout << " -- Evaluation Failed : " << str << '\n';
   } else {
     Wy::cout << str << " = " << Wy::wrd(n,10,9) << '\n';  // radix=10, fraction digits=9
   }
 }
 return(0);
}
catch(const WyRet& e) {
 if(e!=Ok) {
   Wy::cerr << Wy::wrd(e) << '\n';
 }
 return e->repcode();
}
catch(...) {
 Wy::cerr << "main caught(...)\n";
 throw;
};
