/* Copyright is licensed under GNU LGPL.                 by I.J.Wang 2007

  Read standard input as (telephone) dial number string, convert to DTMF sound sample
  and send to /dev/dsp?

  Sample rate is assumed 8000 per second

  Build: make a_dtmf
*/
#include "WyByteFlow.h" // for standard I/O
#include "Wy_Array.hh"
#include "WyStr.h"      // for WyStr,WyCSeg and Wy::wrd
#include <limits>
#include <cmath>        // for std::sin

const size_t SamplesPerSecond=8000;

typedef unsigned char SoundSampleType;
typedef Wy_Array<SoundSampleType> SoundSample;

// [Syn] Convert phone digits in str to DTMF sound sample stored into smp
//
void mk_smp(const WyCSeg str, SoundSample& smp)
{
 typedef size_t MacroSecond;
 const MacroSecond TmTDA=40;      // time duration accept
 const MacroSecond TmIDP=45;      // inter-digit pause
 const SoundSampleType Amp= std::numeric_limits<SoundSampleType>::max()/4;

 smp.reset();

 // Loop each character of str for the respective DTMF sound sample, and
 // append it to smp
 //
 for(const char* p= str.begin(); p!=str.end(); ++p) {
   size_t f_lo,f_hi;
   if((*p>='0')&&(*p<='9')) {
     const static size_t Freq_low[10] ={  941,  697,  697,  697,  770, 
                                          770,  770,  852,  852,  852};
     const static size_t Freq_high[10]={ 1336, 1209, 1336, 1477, 1209,
                                         1336, 1477, 1209, 1336, 1477};
     f_lo=Freq_low[*p-'0'];
     f_hi=Freq_high[*p-'0'];
   } else {
     switch(*p) {
       case '*': f_lo=941; f_hi=1209; break;
       case '#': f_lo=941; f_hi=1477; break;
       case 'a':
       case 'A': f_lo=697; f_hi=1633; break;
       case 'b':
       case 'B': f_lo=770; f_hi=1633; break;
       case 'c':
       case 'C': f_lo=852; f_hi=1633; break;
       case 'd':
       case 'D': f_lo=941; f_hi=1633; break;
       default: continue;    // skip to the next loop, unrecognizable digit is ignored
     }
   };

   const size_t NumSmp= (TmTDA*SamplesPerSecond)/1000;
   const double Omega_lo=2.0*M_PI*f_lo;
   const double Omega_hi=2.0*M_PI*f_hi;
   for(size_t i=0; i<NumSmp; ++i) {
     const double t=static_cast<double>(i)*TmTDA/(1000.0*NumSmp);
     smp.push_back( static_cast<SoundSampleType>(
           Amp*( std::sin(Omega_lo*t)+ std::sin(Omega_hi*t)+2.0))
         );
   }

   const size_t NumSmp2= (TmIDP*SamplesPerSecond)/1000;
   for(size_t i=0; i<NumSmp2; ++i) {
     smp.push_back( static_cast<SoundSampleType>(Amp*2) );
   }
 }
};

static const char* syntax=
  "Convert telephone dial number from standard input to DTMF sound signal\n"
  "E.g. $a_dtmf\n\n";

int main(int argc, char* argv[])
try {
 if(argc!=1) {
   Wy::cerr << syntax;
   return(-1);
 }
 WyRet r;

 Wy::cout << "Input dial number (^D to quit)\n";

 size_t n_wr;
 WyStr str;
 for(;;) {

   // Sound device handling file bf is in the loop so that sound sample
   // data is drained before processing the next input string
   //
   WyByteFlow bf;
   if((r=bf.reset("/dev/dsp",O_WRONLY))!=Ok) {
     if(r!=Wym_ENOENT) {
       WY_THROW(r);
     }
     // If dsp does not exist, may be dsp1
     if((r=bf.reset("/dev/dsp1",O_WRONLY))!=Ok) {
       WY_THROW(r);
     }
   }

   Wy::cin >> str;     // Get string from standard input
   if(str.size()==0) {
     break;            // EOF
   }

   SoundSample waveform;
   mk_smp(str.cseg(), waveform);

   // Write waveform to device
   if((r=Wy::full_write(bf,waveform.begin(),waveform.size(),n_wr))!=Ok) {
     WY_THROW(r);
   }

 }
 return(0);
}
catch(const WyRet& e) {
 Wy::cerr << Wy::wrd(e) << '\n';
 return e->repcode();
}
catch(...) {
 Wy::cerr << "main caught...\n";
 throw;
};
