/* Copyright is licensed under GNU LGPL.                 by I.J.Wang 2005

 Replace string of file(s). Source files are not modified. In stead, the 
 output files are the given file pathname appended with ".rtmp".

 Build: make a_frep
*/
#include "WyRegFile.h"     // for accessing reglar file
#include "WyFileStat.h"    // for file status
#include "WyStr.h"
#include "Wy_AtDestroy.hh"

static const WyCSeg OutputExt=".rtmp";

// [Syn] Copy file fname to the file of which the pathname is appended with
//       ".rtmp". And replace occurrence of string fndstr with repstr
//
static WyRet replace_string(const WyStr& fname,
                     const WyStr& fndstr, const WyStr& repstr)
try {
 WyRet r;

 // Check whether file fname is suitable or not
 {
   WyFileStat stt;
   if((r=Wy::stat(fname,stt))!=Ok) {
     WY_RETURN(r);
   }
   if(stt.is_reg()==false) {
     WY_RETURN(Wym_EINVAL);
   }
 }
 WyRegFile rfile(fname,O_RDONLY);

 // Create work file (*.rtmp)
 const WyStr wfname(fname+OutputExt);
 WyRegFile wfile;
 if((r=WyRegFile::create(wfile,wfname,O_WRONLY|O_TRUNC,S_IRUSR|S_IWUSR))!=Ok) {
   WY_RETURN(r);
 }

 // Unlink work file (wfname) in case of failure
 Wy_AtDestroy<WyRet,const WyStr&> rrid(Wy::unlink,wfname);

 bool replace_occurred=false;
 WyStr buf;
 for(;;) {
   const size_t ByteToRead=2048;
   size_t n_rd;

   // Read some bytes into buf
   if((r=rfile.read(buf,ByteToRead,n_rd))!=Ok) {
     WY_RETURN(r);
   }
   if(n_rd==0) { // EOF, drain buf
     wfile << buf;
     break;
   }

   // Find occurrence of fndstr in buf and replace with repstr and
   // write to wfile (some chars may left in buf (less than fndstr.size())
   //
   for(size_t sidx(0);;) {
     //
     // Con: char's offset less than sidx are drained to wfile
     //
     size_t blft( buf.size()-sidx );
     if(blft<fndstr.size()) {
       break;
     }
     size_t fidx, drain_size;
     if(buf.find(fidx,sidx,fndstr.cseg())==false) {
       //
       // drain buf which is known no occurrence of fndstr
       // leaving fndstr.size()-1 char's in buf
       //
       drain_size=blft-fndstr.size()+1;
       wfile << buf.cseg(sidx,drain_size);
       buf.erase(0,sidx+drain_size);
       break;
     }
     replace_occurred=true;
     drain_size=fidx-sidx;
     wfile << buf.cseg(sidx,drain_size);
     wfile << repstr;
     sidx=fidx+fndstr.size();
   }
 }
 
 if(replace_occurred==true) {
   rrid.release();
 } else {
 }
 return(Ok);
}
catch(const WyStr::Reply& e) {
 WY_RETURN(e);
}
catch(const WyRegFile::Reply& e) {
 WY_RETURN(e);
}
catch(const WyByteFlow::Reply& e) {
 WY_RETURN(e);
}
catch(const WyRet& e) {
 WY_THROW( WyRet(e) );
}
catch(...) {
 Wy::cerr << "Unknown throw type from replace_string(..)\n";
 throw;
};

// Command-Line Messages
//
static const char *cmd_syntax=
   "\n"
   " Find string in file and write to file with the string replaced\n"
   " The output file name is the input file name appended with '.rtmp'\n"
   "\n"
   " [Usage]$a_frep [-h] fnds reps file{file}\n"
   " fnds     string to find\n"
   " reps     string to replace with\n"
   " file     pathname of regular files to find fnds\n"
   " -h       print this message\n"
   ;

int main(int argc, char *argv[])
try {

 if(argc<4) {
   Wy::cerr << cmd_syntax;
   return(-1);
 }

 // Command-line options
 //
 {
  const char optstr[]="h";
  int optch;
  
  while((optch=::getopt(argc,argv,optstr))!=-1) {
    switch(optch) {
      case 'h':
	      Wy::cout << cmd_syntax;
	      return(0);
      case ':':     // missing parameter
      case '?':     // unknown option
	      Wy::cerr << "parameter error\n";
	      Wy::cerr << cmd_syntax;
              return(-1);
      default:
	      Wy::cerr << "parameter fault\n";
              return(-1);
    }
  }
  WyRet r;
  const WyStr src_str( argv[optind++] );  // source string for replace
  const WyStr des_str( argv[optind++] );  // destination string for replace
  for( ; optind<argc; optind++) {
    const WyStr fname(argv[optind]);
    if((r=replace_string(fname,src_str,des_str))==Ok) {
      // For semantics of commonly used utility 'replace', code here:
      //  1. rename fname to backup file
      //  2. rename output .rtmp file to fname
    } else {
      Wy::cerr << fname << ": " << Wy::wrd(r) << '\n';
    }
  }
 }

 return(0);
}
catch(const WyRet& e) {
 Wy::cerr << Wy::wrd(e) << '\n';
 return(-1);
}
catch(...) {
 Wy::cerr << "main caught(...)\n";
 throw;
};
