/* Copyright is licensed under GNU LGPL.                 by I.J.Wang 2005

  Simulate shell command more
  Read data from the given file and write to stdout

  Build: make a_more
*/
#include "WyTerminal.h"  // for keyboard
#include "WyTermios.h"   
#include "WyByteFlow.h"  // for Wy::cin,cout,cerr and input file
#include "WyStr.h"
#include <typeinfo>
#include <new>           // for new(std::nothrow)

// Console tty i/o (raw mode terminal)
// RawTerm changes tty attributes so that we can read key stroke without press ENTER
//
// Warning: Many virtual members are not reimplemented for simplicity.
//
class RawTerm : public WyTerminal {
    WyTermios _ptio;       // Previous(original) termios setting
  public:
    RawTerm() {};
    RawTerm(const char*, int);
    ~RawTerm();
    WyRet getch(char& ch);
    RawTerm* _alloc(WyRet& r) const;
} static mtty("/dev/tty",O_RDONLY);

RawTerm::RawTerm(const char* pathname, int flag)
 : WyTerminal(pathname,flag)
{
 WyRet r;
 if((r=WyTerminal::flush(TCIFLUSH))!=Ok) {
   WY_THROW( WyTerminal::Reply(r) );
 }
 if((r=WyTerminal::getattr(_ptio))!=Ok) {
   WY_THROW( WyTerminal::Reply(r) );
 }
  
 try {
   // Set the terminal mode this program uses
   //
   WyTermios ntio(_ptio);
   ntio.iflag_and(~BRKINT);
   ntio.lflag_and(~(ICANON| ECHO| ISIG| TOSTOP)); 
   ntio.cc_VMIN(1);
   ntio.cc_VTIME(0);
   if((r=WyTerminal::setattr(ntio,TCSANOW))!=Ok) {
     WY_THROW( WyTerminal::Reply(r) );
   }

   // Read back check
   WyTermios tmptio;
   if((r=WyTerminal::getattr(tmptio))!=Ok) {
     WY_THROW( WyTerminal::Reply(r) );
   }
   if(tmptio!=ntio) {
     Wy::cerr << "Setting termios is not fully accepted\n";
   }
 }
 catch(const WyTermios::Reply& e) {
   WY_THROW( WyTerminal::Reply(r) );
 };
};

RawTerm::~RawTerm()
{
 // If *this is default, these functions just fail (no effect for any device)
 WyTerminal::flush(TCIFLUSH);
 WyTerminal::setattr(_ptio,TCSANOW);
};

// Get one char from terminal
//
WyRet RawTerm::getch(char& ch)
{
 WyRet r;
 size_t n;
 if((r=WyTerminal::read(&ch,sizeof(ch),n))!=Ok) {
   WY_RETURN(r);
 }
 if(n==0) {
   WY_RETURN(Wym_EPIPE);
 }
 return(Ok);
};

// libwy.a requires deriving classes reimplement this member
//
RawTerm* RawTerm::_alloc(WyRet& r) const
try {
 if(typeid(*this)!=typeid(WyTerminal)) {
   r=Wym_ENOSYS; WY_HERE(r);        // _alloc not overridden
   return(NULL);
 }
 RawTerm* p= new(std::nothrow) RawTerm();
 if(p==NULL) {
   r=Wym_ENOMEM; WY_HERE(r);
 } else {
   r=Ok;
 }
 return(p);
}
catch(const WyTerminal::Reply& e) {
 r=e; WY_HERE(r);
 return(NULL);
};

// [Syn] 'more' for infile
//
static WyRet more_file(WyByteFlow& infile) throw()
try {
 const int PageLines=20;    // one page are fixed to 20 lines (for simplicity)
 const size_t MaxLineLength=2048;  // one line contains at most this number of characters
 size_t ln_wr(PageLines);   // lines to write (initially one page)
 WyRet r;
 WyStr rbuf;

 // Loop for each line
 for(size_t lcnt=0; ;) {
   size_t llen;
   if((r=infile.read(rbuf,MaxLineLength,llen))!=Ok) {  // append read
     WY_RETURN(r);
   }
   if(rbuf.size()==0) {
     break;    // eof
   }
   if(rbuf.findc(llen,0,'\n')) {
     ++llen;   // llen= line length in rbuf
   } else {
     llen=rbuf.size();   // no LF, count rbuf as a line
   }
   ++lcnt;

   Wy::cout << rbuf.cseg(0,llen);
   rbuf.erase(0,llen);

   // Check keyboard input at each page pause
   //
   if(lcnt>=ln_wr) {
     lcnt=0;        // reset lcnt 
     char rply;
     if((r=mtty.getch(rply))!=Ok) {
       WY_RETURN(r);
     }
     if((rply=='q')||(rply=='Q')) {
       return(Ok);
     }
     ln_wr= (rply=='\n')? 1:PageLines;  // lines to write before next pause
   }
 }
 return(Ok);
}
catch(const WyTerminal::Reply&) {
 throw;
}
catch(const WyStr::Reply& e) {
 WY_RETURN(e);
}
catch(const WyByteFlow::Reply& e) {
 WY_RETURN(e);
}
catch(const WyRet& e) {
 WY_TERMINATE("unexpected Reply");
}
catch(...) {
 WY_TERMINATE("unknown throw type");
};

static const char* syntax=
 " Read data from the given file (stdin assumed if not specified) and\n"
 " write to stdout\n\n"
 " [Usage]$./a_more [file]\n"
 "   file= file to do the more.\n";

int main(int argc, char* argv[])
try {
 if(argc>2) {
   Wy::cerr << syntax;
   return(-1);
 }

 WyRet r;
 WyByteFlow infile;
 if(argc==1) {
   if(Wy::cin._isatty()) {
     Wy::cerr << syntax;
     return(-1);
   }
   if((r=infile.reset(Wy::cin))!=Ok) {
     WY_THROW(r);
   }
 } else {
   if((r=infile.reset(argv[1],O_RDONLY))!=Ok) {
     WY_THROW(r);
   }
 }
 if((r=more_file(infile))!=Ok) {
   throw(r);
 }
 return(0);
}
catch(const WyRet& e) {
 if(e!=Ok) {
   Wy::cerr << Wy::wrd(e) << '\n';
 }
 return e->repcode();
}
catch(...) {
 Wy::cerr << "main caught(...)\n";
 throw;
};
