/* Copyright is licensed under GNU LGPL.                 by I.J.Wang 2007

  Read standard input, convert to Morse Code sound (send to /dev/dsp?)

  Build: make a_morse
*/
#include "WyByteFlow.h"
#include "Wy_Array.hh"
#include "WyStr.h"      // for WyCSeg and Wy::wrd
#include <limits>
#include <cmath>        // for std::sin

typedef Wy_Array<unsigned char> SoundSample;

// Dot-Dash sound sample
//
class DDC_Sound {
    static SoundSample _dit,_dah,_blk,_wbk;   // Elementary sould samples
    const static bool _static_setup;          // Setup flag

    SoundSample _sample;

    void _init(const WyCSeg& ddc);
    static bool _setup_cwsound_elements();
    const DDC_Sound& operator =(const DDC_Sound&); // Hide
    DDC_Sound(const DDC_Sound&);                   // Hide
  public:
    DDC_Sound() {};
    DDC_Sound(const WyCSeg& ddc);
    const SoundSample& get_sample() const { return _sample; };
    void reset(const WyCSeg& ddc);
};

bool DDC_Sound::_setup_cwsound_elements()
{
 // The following values are taken ad. hoc.
 //
const double ASpeed=0.15;         // Angle speed (tone)
const int Dit_Samples=500;        // Number of samples for Dit (sound period)
const int Dah_Samples=1200;       // Number of samples for Dah
const int Blk_Samples=300;        // Number of samples for sound between Dit/Dah
const int Wbk_Samples=Blk_Samples+Dah_Samples+Blk_Samples;
const int HalfAmp= std::numeric_limits<unsigned char>::max()/2;

 // Setup Dit sound sample
 for(int i=0; i<Dit_Samples; ++i) {
   _dit.push_back((unsigned char)( HalfAmp*(1.0+std::sin(ASpeed*i)) ) );
 }

 // Setup Dah sound sample
 for(int i=0; i<Dah_Samples; ++i) {
   _dah.push_back((unsigned char)( HalfAmp*(1.0+std::sin(ASpeed*i)) ) );
 }

 // Setup break time between Dit and Dah
 for(int i=0; i<Blk_Samples; ++i) {
   _blk.push_back(HalfAmp);
 }

 // Setup break time between characters
 for(int i=0; i<Wbk_Samples; ++i) {
   _wbk.push_back(HalfAmp);
 }
 return(true);
};

// Static setup of sound samples of DDC_Sound
SoundSample DDC_Sound::_dit, DDC_Sound::_dah, DDC_Sound::_blk, DDC_Sound::_wbk;
const bool DDC_Sound::_static_setup= DDC_Sound::_setup_cwsound_elements();

void DDC_Sound::_init(const WyCSeg& ddc)
{
 for(size_t i=0; i<ddc.size(); ++i) {
   switch(ddc[i]) {
     case '.': _sample.insert(_sample.size(),_dit);
               _sample.insert(_sample.size(),_blk);
               break;
     case '-': _sample.insert(_sample.size(),_dah);
               _sample.insert(_sample.size(),_blk);
               break;
     default:
               _sample.insert(_sample.size(),_wbk);
   }
 }
 _sample.insert(_sample.size(),_wbk);
};

DDC_Sound::DDC_Sound(const WyCSeg& ddc)
{
 this->_init(ddc);
};

void DDC_Sound::reset(const WyCSeg& ddc)
{
 _sample.reset();
 this->_init(ddc);
};

//----------

// Append CW sound sample (converted from argument text) to argument buf
//
void append_cwsound(SoundSample &buf, const WyCSeg& text)
{
 class CWS_tab {
     DDC_Sound _samp[1<<CHAR_BIT];
 
     CWS_tab(const CWS_tab&);
     const CWS_tab& operator=(const CWS_tab&);
   public:
     typedef unsigned int uint;
     CWS_tab() {
       _samp[uint(' ')].reset("");
       _samp[uint('a')].reset(".-");
       _samp[uint('b')].reset("-...");
       _samp[uint('c')].reset("-.-.");
       _samp[uint('d')].reset("-..");
       _samp[uint('e')].reset(".");
       _samp[uint('f')].reset("..-.");
       _samp[uint('g')].reset("--.");
       _samp[uint('h')].reset("....");
       _samp[uint('i')].reset("..");
       _samp[uint('j')].reset(".---");
       _samp[uint('k')].reset("-.-");
       _samp[uint('l')].reset(".-..");
       _samp[uint('m')].reset("--");
       _samp[uint('n')].reset("-.");
       _samp[uint('o')].reset("---");
       _samp[uint('p')].reset(".--.");
       _samp[uint('q')].reset("--.-");
       _samp[uint('r')].reset(".-.");
       _samp[uint('s')].reset("...");
       _samp[uint('t')].reset("-");
       _samp[uint('u')].reset("..-");
       _samp[uint('v')].reset("...-");
       _samp[uint('w')].reset(".--");
       _samp[uint('x')].reset("-..-");
       _samp[uint('y')].reset("-.--");
       _samp[uint('z')].reset("--..");
       _samp[uint('1')].reset(".----");
       _samp[uint('2')].reset("..---");
       _samp[uint('3')].reset("...--");
       _samp[uint('4')].reset("....-");
       _samp[uint('5')].reset(".....");
       _samp[uint('6')].reset("-....");
       _samp[uint('7')].reset("--...");
       _samp[uint('8')].reset("---..");
       _samp[uint('9')].reset("----.");
       _samp[uint('0')].reset("-----");
     };

     const DDC_Sound& operator[](char ch) const
     {
       if( ((ch>='a')&&(ch<='z'))||((ch>='0')&&(ch<='9')) ) {
         return _samp[uint(ch)];   // g++ 4.1.1 warns char subscript
       } else if((ch>='A')&&(ch<='Z')) {
         return _samp[uint(ch|0x20)];
       } else {
         return _samp[uint(ch)];   // g++ 4.1.1 warns char subscript
       };
     };
 } static const CWSamp;

 for(size_t i=0; i<text.size(); ++i) {
   buf.insert(buf.size(),CWSamp[text[i]].get_sample());
 }
};

static const char* syntax=
  "Convert string from standard input to Morse Code sound\n"
  "E.g. $a_morse\n\n"
  "Syntax: a_morse\n\n";

int main(int argc, char* argv[])
try {
 if(argc!=1) {
   Wy::cerr << syntax;
   return(-1);
 }
 WyRet r;
 WyByteFlow bf;

 // Open sound device
 //
 if((r=bf.reset("/dev/dsp",O_WRONLY))!=Ok) {
   if(r!=Wym_ENOENT) {
     WY_THROW(r);
   }
   // If dsp does not exist, may be dsp1
   if((r=bf.reset("/dev/dsp1",O_WRONLY))!=Ok) {
     WY_THROW(r);
   }
 }

 size_t n_wr;
 WyStr str;
 for(;;) {
   Wy::cin >> str;     // Get string from standard input
   if(str.size()==0) {
     break;            // EOF
   }

   SoundSample waveform;
   append_cwsound(waveform,str.cseg());  // Convert to CW sound
   if((r=Wy::full_write(bf,waveform.begin(),waveform.size(),n_wr))!=Ok) { // Write waveform to device
     WY_THROW(r);
   }
 }
 return(0);
}
catch(const WyRet& e) {
 Wy::cerr << Wy::wrd(e) << '\n';
 return e->repcode();
}
catch(...) {
 Wy::cerr << "main caught...\n";
 throw;
};
