/* Copyright is licensed under GNU LGPL.                 by I.J.Wang 2006

 TCP echo server using thread pool

 Build: make a_tcpecho_svr
*/
#include "WySockFile.h"        // for socket
#include "WySockAddrINet4.h"   // for socket address
#include "Wy_Thread.h"         // for thread
#include "WyNum.h"             // for Wy::_strnum

// Thread is a thread class, echoing messages to peer socket
//
class Thread : public Wy_Thread {
    WySockFile _sock;          // system socket file that the thread uses

  public:
    Thread() throw() {};
    ~Thread() throw() 
      { tmain_close(); // Stop a possibly active thread 
      };
    WyRet reset() 
      {
        const WyRet r=Wy_Thread::reset();
        if(r==Ok) {
          WY_RETURN( _sock.reset() ); 
        } else {
          _sock.reset();
          WY_RETURN(r);
        }
      };
    WyRet begin(const WySockFile& sock)
      {
        WyRet r;
        if(_sock.is_default()==false) {
          WY_RETURN( Wym_ENOTEMPTY );
        }
        if((r=_sock.reset(sock))!=Ok) {
          WY_RETURN(r);
        }
        if((r=Wy_Thread::begin())!=Ok) {
          _sock.reset();
          WY_RETURN(r);
        }
        return(Ok);
      };

  protected:
    WyRet tmain() try {
      WyStr rbuf;
      rbuf._reserve(1024);
       
      // Loop until eof
      for(;;) {
        _sock >> rbuf;     // Read data from _sock into rbuf
        if(rbuf.size()==0) {
          break;           // eof
        }
        _sock << rbuf;     // Write rbuf to _sock
      }
      return(Ok);
    }
    catch(const WyRet& e) {
      Wy::cerr << "Child: " << Wy::wrd(e) << '\n';
      return(e);
    }
    catch(...) {
      Wy::cerr << "Child: Unknown thrown type";
      throw;
    };
};

static const char* syntax=
 "[Syntax]$./a_tcpecho_svr port\n"
 " e.g. $a_tcpecho_svr 8888\n";

static const size_t MaxThreads=8;
static Thread thrd_pool[MaxThreads];  // note: elements are not movable

int main(int argc, char* argv[])
try {
 if(argc!=2) {
   Wy::cout << syntax;
   return(-1);
 }
 WyRet r;
 uint16_t port;

 // Convert argv[1] to uint16_t and save to port
 if((r=Wy::_strnum(port,argv[1]))!=Ok) {
   WY_THROW(r);
 }

 // Setup msock, the master socket
 WySockFile msock(PF_INET,SOCK_STREAM,0);
 const WySockAddrINet4 SvrAddr( WySockAddrINet4::IPAddr(INADDR_ANY),port );

 if((r=msock.bind(SvrAddr))!=Ok) {
   WY_THROW(r);
 }
 if((r=msock.listen(0))!=Ok) {
   WY_THROW(r);
 }

 // Endless loop
 for(;;) {
   WySockFile a_sock;
   WySockAddrINet4 a_addr;

   // Accept an incoming socket
   if((r=msock.accept(a_sock,a_addr))!=Ok) {
     WY_THROW(r);
   }

   // Find available entry in thrd_pool[]
   size_t pidx;
   for(pidx=0; pidx<MaxThreads; ++pidx) {
     switch(thrd_pool[pidx].thread_state()) {
       case Wy_Thread::Active: continue;
       case Wy_Thread::Null: break;
       default: thrd_pool[pidx].reset();
     }
     if((r=thrd_pool[pidx].begin(a_sock))==Ok) {
       break;
     }
   }
   if(pidx>=MaxThreads) {
     Wy::cerr << "thrd_pool[] is full. Accepted socket will be abondoned\n";
     Wy::sleep( std::time_t(1) ); // sleep 1 second
   };
 }
 WY_ASSERT("UNREACHABLE");
}
catch(const WyRet& e) {
 if(e!=Ok) {
   Wy::cerr << "main caught: " << Wy::wrd(e) << '\n';
 }
 return e->repcode();
}
catch(...) {
 Wy::cerr << "main caught(...)\n";
 throw;
};
