/* Copyright is licensed under GNU LGPL.                 by I.J.Wang 2007
 
  Setup function qWrd to get QString message from WyReply object, using system locale.

  Using Qt3 library (differences are marked for Qt4)

  Build: make q_wrd
*/
#include <qapplication.h> // for Qt4: change to 'include <QApplication>'
#include <qlabel.h>       // for Qt4: change to 'include <QLabel>'
#include <string.h>       // for ::strerror
#include <locale.h>       // for ::setlocale
#include <WyMutex.h>
#include <WyStr.h>
#include <WyByteFlow.h>
#include <WyIConv.h>

static const char* TgtMsgEnc="UTF16"; // QString uses UTF16 encoding

// Buffer for storing UTF16 string
//
// This buffer should construct earlier so destructed latter
//
static WyStr utf16buf[2];

// [Syn] UsrRepInfo inherits Wy_RepInfo for user defined reply
//
class UsrRepInfo : public Wy_RepInfo {
    const char* _mtid;
    const void* _msg;
    size_t      _mlen;
    mutable WyMutex _mtx;

  public:

    UsrRepInfo(int ec, const char* cstr) : Wy_RepInfo(ec,cstr), _mtid(""),
        _msg(cstr), _mlen(std::strlen(cstr)) {};

    // [Override]
    //
    size_t getmsg(const char*& mtid, void* buf, size_t blen) const {
      WyLock aa(_mtx);
      size_t clen(_mlen);
      mtid=_mtid;
      if(buf!=NULL) {
        if(blen<clen) {
          clen=blen;
        }
        std::memcpy(buf,_msg,clen);
      }
      return clen;
    };

    // [Override]
    //
    void setmsg(const char* mtid, const void* str, size_t slen)
        {
          if((mtid==NULL)||(str==NULL)) {
            return;
          }
          WyLock aa(_mtx);
          _mtid=mtid;
          _msg=str;
          _mlen=slen;
        };
};

static UsrRepInfo ri_UsrErr(11225,"UsrErr 11225");

const WyReply UsrErr(ri_UsrErr);    // user reply for public

// [Singleton] Setup UTF16 reply messages (only UsrErr and Wym_EINVAL are setup
//             for demonstration purpose)
//
class InitQStrReply {
    WyIConv _cvtr;
    void cvrt_rmsg(WyReply arep, WyStr& enc);

  public:
    InitQStrReply();
    ~InitQStrReply();
};

// [Syn] Set the reply message of arep to UTF16. The message is converted from
//       the string ::strerr, and store the result into enc as the copy source.
//
void InitQStrReply::cvrt_rmsg(WyReply arep, WyStr& enc)
{
 WyRet r;
 WyStr buf;
 WyCSeg iseg;
 size_t n_rev;   // dummy

 buf=::strerror(arep->repcode());  // ::strerror is fine here. String is referenced
                                   // in the startup time.
 if(buf.size()==0) {
   buf="::strerror("+Wy::wrd(arep->repcode())+") returned empty string.";
 }
 iseg=buf.cseg();

 if((r=_cvtr.iconv(iseg,enc,n_rev))!=Ok) {
   WY_THROW(r);
 }
 arep->setmsg(TgtMsgEnc, enc.begin(), enc.size());
};

InitQStrReply::InitQStrReply()
{
 // [Singleton] Get locale message encoding, store into src_enc
 //
 struct GetSysLocale {  
     WyStr src_enc;
     GetSysLocale() {
       const WyCSeg loc(::setlocale(LC_MESSAGES,""));
       size_t idx;
       if(loc.rfindc(idx, loc.size(), '.')) {
         src_enc=loc.cseg(idx+1);
       }
     };
 } const sys_locale_info;

 WyRet r;
 
 if((r=_cvtr.reset(TgtMsgEnc,sys_locale_info.src_enc.c_str()))!=Ok) {
   WY_THROW(r);
 }
 
 // Set UTF16 reply message of WyReply objects
 //
 this->cvrt_rmsg(Wym_EINVAL, utf16buf[0]);
 this->cvrt_rmsg(UsrErr, utf16buf[1]);

 // Release resources WyIConv object occupies (no longer used)
 //
 _cvtr.reset();
};

InitQStrReply::~InitQStrReply()
{
 // Set these reply message to their default (constructed) reply message
 //
 UsrErr->setmsg("", UsrErr->repstr(), ::strlen(UsrErr->repstr()) );
 Wym_EINVAL->setmsg("", Wym_EINVAL->repstr(), ::strlen(Wym_EINVAL->repstr()) );
};

// [Syn] Get QString reply message from WyReply object
//
QString qWrd(const WyReply r)
{
 WyStr str;
 if(str._reserve(256)!=Ok) {  // Assume enough capacity for directly copying the reply
                              // message into it. Otherwise, may need to use loop.
   WY_THROW( WyRet(Wym_ENOMEM) );
 }

 QString qstr;
 const char* mtid;
 str._setsize( r->getmsg(mtid,str.begin(),str._capacity()) );

 if((mtid==TgtMsgEnc)||(std::strcmp(TgtMsgEnc,mtid)==0)) {
   if((*reinterpret_cast<const ushort*>(str.begin())==0xfeff)||   // BOM
      (*reinterpret_cast<const ushort*>(str.begin())==0xfffe)) {
      str.erase(0,sizeof(ushort));              // shrink always succeeds
   }
   // for Qt4: use setUtf16(....)
   qstr.setUnicodeCodes( reinterpret_cast<const ushort *>(str.begin()), 
                         str.size()/sizeof(ushort) );
 } else {
   // Assume Latin1 is the default
   // for Qt4: use qstr=QString::fromLatin1(str.begin(),str.size());
   qstr.setLatin1(str.begin(),str.size());  
 };
 return qstr;
};

int main(int argc, char *argv[])
try {

 QApplication app(argc,argv);
 InitQStrReply useqstrreply;

 QLabel *mw= new QLabel(0);    // for Qt4: change to 'new QLabel'
 app.setMainWidget(mw);        // for Qt4: remove this line
 mw->setText( "UsrErr=>"+qWrd(UsrErr) + ", Wym_EINVAL=>"+qWrd(Wym_EINVAL) );
 mw->show();
 app.exec();
 return(0);
}
catch(const WyRet &e) {
 Wy::cerr << Wy::wrd(e) << '\n';
 throw;
}
catch(...) {
 Wy::cerr << "Unidentified thrown type\n";
 throw;
};
