/* Copyright is licensed under GNU LGPL.                 by I.J.Wang 2003
*/

#define WYLIB_SOURCE
#include "WyDate.h"
#include "WyStr.h"
#include "WyMutex.h"
#include "Wy_AtDestroy.hh"
#include <cstdlib>
#include <climits>    // for PATH_MAX
#include <new>        // new and bad_alloc
#include <ctime>      // for mktime, localtime_r

#include <time.h>     // for externals
extern char *tzname[2];
extern long timezone;
extern int daylight;

static const char ENVNAME_TZ[]="TZ";

const char WyDate::class_name[]="WyDate";
const char WyDate::TIMEZONE_UTC[]="UTC";

// constant for speeding up copy/convert/compare of TIMEZONE_UTC
//
static const WyStr::size_type wy_TZLEN= sizeof(WyDate::TIMEZONE_UTC)-1;

class Wy__static_check__ {
  public:
    Wy__static_check__ () 
              {
               // Impl assertions
               if(WyTimeSpec()!=WyTimeSpec( WyTimeSpec::SecondType(0),
                                            WyTimeSpec::NanoSecType(0))) {
                 WY_TERMINATE("");   // assertion fail
               }
               if((std::numeric_limits<WyTimeSpec::SecondType>::max()!=
                   std::numeric_limits<std::time_t>::max())||
                  (std::numeric_limits<WyTimeSpec::SecondType>::min()!=
                   std::numeric_limits<std::time_t>::min())||
                  (std::numeric_limits<WyTimeSpec::NanoSecType>::max()!=
                   std::numeric_limits<signed long int>::max())||
                  (std::numeric_limits<WyTimeSpec::NanoSecType>::min()!=
                   std::numeric_limits<signed long int>::min())) {
                 WY_TERMINATE("");   // assertion fail
               }

               if(wy_TZLEN>WyStr::min_capacity()) {
                 WY_TERMINATE("");   // Impl::reset() may fail
               }
              };
} static const wy_do_not_use_me__;

// Class of environment variable
// (for temporary use, rrid-like)
//
class TmpEnvVar {
    const char* _vname;
    bool  _has_value;   // _value available
    WyStr _value;       // value of _vname indicated environment variable

    // Hidden
    TmpEnvVar();
    TmpEnvVar(const TmpEnvVar&);
    const TmpEnvVar& operator =(const TmpEnvVar&);
    bool operator ==(const TmpEnvVar&);
  public:

    // Construct object to contain value of name (restore at destruction)
    //
    // Note: argument name pointed string should be const through
    //       object life time
    //
    //   has_value= true, if name defined in the environment variable
    //              false otherwise
    //       value= string value of name defined environment variable
    //              undefined if has_value is false
    //
    // [Throw] WyStr::Reply
    //       Wym_ENOMEM
    //
    TmpEnvVar(const char* name) WY__TSPC(WyStr::Reply);

    // Restore environment variable with value in *this and destruct *this
    //
    ~TmpEnvVar();

    bool has_value() const WY__TSPC()
         { return _has_value; };
    const WyStr& value() const WY__TSPC()
         { return _value; };
};

TmpEnvVar::TmpEnvVar(const char* name) WY__TSPC(WyStr::Reply)
 : _vname(name),_has_value(false),_value()
{
 const char* cptr=::getenv(name);
 if(cptr==NULL) {
   return;
 }
 _has_value=true;
 _value=WyCSeg(cptr);
};

TmpEnvVar::~TmpEnvVar()
{
 if(_has_value) {
   if(::setenv(_vname,_value.c_str(),1)!=0) {
     WY_THROW( WyRet() );  // recovering the change failed
                           // (insufficient environment space)
                           // Changed from WY_TERMINATE to WY_THROW
   }
 } else {
   ::unsetenv(_vname);
 }
};

// Mutex for accessing environment TZ
//
static WyMutex& wy_tzmutex()
{
 static WyMutex _tzmtx;
 return(_tzmtx);
};

// [Internal] Get the value of environment TZ
//
// [Ret] Ok 
//         defined=  true, value is set to the value of environment TZ
//         defined= false, TZ not in the environment, value intact
//       Wym_ENOMEM
//
static WyRet getenv_tz(bool& defined, WyStr& value) WY__NOTHROW__
try {
 WyLock aa( wy_tzmutex() );
 WyRet r;

 const char *vp( ::getenv(ENVNAME_TZ) );
 if(vp==NULL) {
   defined=false;
 } else {
   if((r=value.reset(WyCSeg(vp)))!=Ok) {  // ctor of WyCSeg wont throw
     WY_RETURN(r);
   }
   defined=true;
 }
 return(Ok);
}
catch(const WyLock::Reply& e) {
 WY_NDEBUG_MSG(e);
 Wy__Base::terminate();
}
catch(const WyRet& e) {
 WY_NDEBUG_MSG(e);
 Wy__Base::terminate();
}
catch(...) {
 WY_TERMINATE("");
};

class WyDate::Impl {
  public:
    // zone = TIMEZONE_UTC
    // year = 1970
    // month= 1    [1-12]
    // mday = 1    [1-31]
    // hour = 0    [0-23]
    // min  = 0    [0-59]
    // sec  = 0    [0-61]
    // wday = 4    [0-6]
    // isdst= 0
    //
    // [Throw] WyDate::Reply
    //             Wym_ENOMEM
    //
    Impl() WY__TSPC(WyDate::Reply);

    // [Throw] WyDate::Reply
    //             Wym_ENOMEM
    //
    Impl(const Impl&) WY__TSPC(WyDate::Reply);

    // [Syn] Construct object of time zone tzval, UTC time ucs
    //
    // [Throw] WyDate::Reply
    //       Wym_EFAULT
    //       Wym_EINVAL  unrecognized tzval
    //       Wym_ERANGE  conversion failed
    //       Wym_ENOMEM 
    //
    Impl(const char* tzval,const WyTimeSpec& ttm);

    // [Syn] Construct object of time zone tzval, argument given date
    //
    // [Throw] WyDate::Reply
    //       Wym_EFAULT
    //       Wym_EINVAL  unrecognized tzval
    //       Wym_ERANGE  conversion failed
    //       Wym_ENOMEM 
    //
    Impl(const char* tzval,int year,int month,int mday, int hour,int min, int sec);

    bool is_default() const WY__NOTHROW__;
    
    // [Syn] Reconstruct *this to WyDate()
    //
    void reset() WY__NOTHROW__;

    // [Syn] Reconstruct *this to WyDate(const Impl&)
    //
    // [Ret] Ok
    //       Wym_ENOMEM
    //
    WyRet reset(const Impl& src) WY__TSPC();

    // [Syn] Reconstruct *this to WyDate(const WyStr&,const WyTimeSpec&)
    //
    // [Ret] Ok
    //       Wym_EFAULT
    //       Wym_EINVAL  unrecognized tzval
    //       Wym_ERANGE  conversion failed
    //       Wym_ENOMEM 
    //
    WyRet reset(const char* tzval, const WyTimeSpec& ucs);

    // [Syn] Reconstruct *this to WyDate(const WyStr&,int,int,int,int,int,int)
    //
    // [Ret] Ok
    //       Wym_EFAULT
    //       Wym_EINVAL  unrecognized tzval
    //       Wym_ERANGE  conversion failed
    //       Wym_ENOMEM 
    //
    WyRet reset(const char* tzv,int year,int month,int mday, int hour,int min, int sec);

    void swap(Impl& another) WY__TSPC();

    // [Ret] Time zone value (as for the environment string)
    //
    // Note: _SC_TZNAME_MAX is defined on this Linux machine (=6)
    //       for now long limit (OS may use longer name)
    //
    WyCSeg time_zone() const WY__NOTHROW__
              { return _tzv.cseg(); };

    int year() const WY__NOTHROW__
              { return _year; };

    // [Syn] Get the month of a year of the object
    // 
    // [Ret] Month of the object [1-12]
    // 
    int month() const WY__NOTHROW__
              { return _month; };

    // [Syn] Get the day of the month
    // 
    // [Ret] Day of the month [1-31]
    // 
    int mday() const WY__NOTHROW__
              { return _mday; };

    // [Syn] Get the hour in a day
    // 
    // [Ret] The hour [0-23]
    // 
    int hour() const WY__NOTHROW__
              { return _hour; };

    // [Syn] Get the minute in an hour
    // 
    // [Ret] Minute [0-59]
    // 
    int min() const WY__NOTHROW__
              { return _min; };

    // [Syn] Get the second in a minute.
    //
    // [Ret] Second [0-61]
    // 
    int sec() const WY__NOTHROW__
              { return _sec; };

    // [Syn] Get the day of the week
    // 
    // [Ret] Day of the week [0-6]
    // 
    int wday() const WY__NOTHROW__
              { return _wday; };

    // [Syn] Get the day of the year
    // 
    // [Ret] Day of the year [0-365]
    // 
    int yday() const WY__NOTHROW__
              { return _yday; };

    int isdst() const WY__NOTHROW__
              { return _isdst; };

    // [Syn] Get the UTC time of *this
    // 
    // [Ret] UTC time of *this
    //
    WyTimeSpec utc_time() const WY__NOTHROW__
              { return _wyt; };

    // [Syn] Synchronize date to UTC second ucs
    //
    // [Ret] Ok
    //       Wym_ERANGE  conversion failed
    //       Wym_ENOMEM 
    //
    WyRet set_utc(const WyTimeSpec& ucs);

    // [Syn] Set date as the arguments indicate
    //
    // [Ret] Ok
    //       Wym_EINVAL  invalid argument
    //       Wym_ERANGE  conversion failed
    //       Wym_ENOMEM 
    //
    WyRet set_date(int year,int month,int mday, int hour,int min, int sec);

    // [Syn] Add linear time to *this
    //
    // [Ret] Ok    
    //       Wym_ERANGE result not representable
    //       Wym_ENOMEM 
    //
    WyRet add(const WyTimeSpec& tm);

    // [Syn] Reconstruct date as by Impl(const Impl&)
    // 
    // [Throw] WyDate::Reply
    //       Wym_ENOMEM
    // 
    // [Ret] const Reference of this object
    //
    const Impl& operator =(const Impl& rhs) WY__TSPC(WyDate::Reply);

    // [Syn] Convert *this to text representation according to the format, and
    //       reset res with the result
    //
    // Note: This member is the conversion for ::strftime. resultant timezone
    //       string may not be what WyDate otherwise uses
    //       (timezone issue is vague to me).
    //       Locale may be involved.
    // 
    // [Ret] Ok
    //       Wym_EFBIG
    //       Wym_ENOMEM 
    //
    WyRet _strftime(WyStr& str, const char* format) const;

    // [Syn] Add time by rhs
    //
    // [Throw] WyDate::Reply
    //       Wym_ERANGE result not representable
    //       Wym_ENOMEM 
    //
    // [Ret] const reference of this object
    //
    // [Refer] WyTimeSpec::operator+ , reset
    //
    const Impl& operator +=(const WyTimeSpec& rhs);

    // [Syn] Equivalence test
    //
    // [Ret] bool result of the equivalence
    //
    bool operator ==(const Impl& rhs) const WY__NOTHROW__
         { return( (_wyt==rhs._wyt)&&(_tzv==rhs._tzv) ); };
    bool operator !=(const Impl& rhs) const WY__NOTHROW__
         { return( (_wyt!=rhs._wyt)||(_tzv!=rhs._tzv) ); };

  private:
    WyStr _tzv;               // time zone string
    int _isdst;               // daylight saving time

    int _year;                // [A.D. year]
    int _month;               // [1,12]
    int _mday;                // [1,31]
    int _hour;                // [0,23]
    int _min;                 // [0,59]
    int _sec;                 // [0,61]

    // The following members are redundant, but useful.
    // 
    int _wday;                // [0-6] week day
    int _yday;                // [0,365] Days since January 1st

    // Corresopnding UTC time (of the date) 
    //
    WyTimeSpec _wyt;
};

WyDate::Impl::Impl() WY__TSPC(WyDate::Reply)
try : _tzv( WyCSeg(TIMEZONE_UTC,wy_TZLEN) ),_isdst(0),
      _year(Default_Year),_month(Default_Month),_mday(Default_MDay),
     _hour(Default_Hour),_min(Default_Min),_sec(Default_Sec),
     _wday(Default_WDay),_yday(Default_YDay), _wyt()
{
}
catch(const WyStr::Reply& e) {
  WY_THROW( WyDate::Reply(e) );
};

WyDate::Impl::Impl(const WyDate::Impl& src) WY__TSPC(WyDate::Reply)
try : _tzv(src._tzv),_isdst(src._isdst),
      _year(src._year),_month(src._month),_mday(src._mday),
     _hour(src._hour),_min(src._min),_sec(src._sec),
     _wday(src._wday),_yday(src._yday), _wyt(src._wyt)
{
}
catch(const WyStr::Reply& e) {
  WY_THROW( WyDate::Reply(e) );
};

WyDate::Impl::Impl(const char* tzval,const WyTimeSpec& tm)
{
 WyRet r=this->reset(tzval,tm);
 if(r!=Ok) {
   WY_THROW( WyDate::Reply(r) );
 }
};

WyDate::Impl::Impl(const char* tzval,int year,int month,int mday, int hour,int min, int sec)
{
 WyRet r=this->reset(tzval,year,month,mday,hour,min,sec);
 if(r!=Ok) {
   WY_THROW( WyDate::Reply(r) );
 }
};

bool WyDate::Impl::is_default() const WY__NOTHROW__
{
 return(_wyt.is_default()
        &&(_tzv.cseg()==WyCSeg(TIMEZONE_UTC,wy_TZLEN)));
};

void WyDate::Impl::reset() WY__NOTHROW__
{
 if(_tzv.reset( WyCSeg(TIMEZONE_UTC,wy_TZLEN) )!=Ok) {
   WY_TERMINATE("");   // asserted not to fail (enough capacity)
 }
 _isdst=0;
 _wyt.reset();
 _year=Default_Year;
 _month=Default_Month;
 _mday=Default_MDay;
 _hour=Default_Hour;
 _min=Default_Min;
 _sec=Default_Sec;
 _wday=Default_WDay;
 _yday=Default_YDay;
};

WyRet WyDate::Impl::reset(const WyDate::Impl& src) WY__TSPC()
{
 WyRet r=_tzv.reset(src._tzv);
 if(r!=Ok) {
   WY_RETURN(r);
 }
 _isdst=src._isdst;
 _wyt.reset(src._wyt);
 _year=src._year;
 _month=src._month;
 _mday=src._mday;
 _hour=src._hour;
 _min=src._min;
 _sec=src._sec;
 _wday=src._wday;
 _yday=src._yday;
 return(Ok);
};

WyRet WyDate::Impl::reset(const char* tzval,const WyTimeSpec& ucs)
try {
struct tm tm;
std::time_t tt;

  if(tzval==NULL) {
    WY_RETURN(Wym_EFAULT);
  }

  // Warning: Not all unrecognized tzval can be reported
  if(tzval[0]==0) {
    WY_RETURN(Wym_EINVAL);  // empty string not accepted for now
  }
  tt=ucs.second();

  {
    WyLock aa(wy_tzmutex());
    TmpEnvVar rr_tz(ENVNAME_TZ);

    if(::setenv(ENVNAME_TZ,tzval,1)!=0) {
      WY_RETURN(Wym_ENOMEM);  // man. indicates no errno and guess sort of this reason
    }
    ::tzset();  // man. indicated this can be saved, call it however.
    if(::localtime_r(&tt,&tm)==NULL) {
      WY_RETURN(Wym_ERANGE);
    }
    _tzv=tzval;   // WyCSeg(tzval) wont throw
  }

 _wyt.reset(ucs);
 _isdst=tm.tm_isdst;
 _year=tm.tm_year+1900;
 _month=tm.tm_mon+1;
 _mday=tm.tm_mday;
 _hour=tm.tm_hour;
 _min=tm.tm_min;
 _sec=tm.tm_sec;
 _wday=tm.tm_wday;
 _yday=tm.tm_yday;
 return(Ok);
}
catch(const WyStr::Reply& e) {
 return(e);
}
catch(const WyRet& e) {
 WY_NDEBUG_MSG(e);
 WY_THROW( WyRet(e) );
};

WyRet WyDate::Impl::reset(const char* tzval,int year,int month,int mday, int hour,int min, int sec)
try {
  if((month<1)||(mday<1)||(hour<0)||(min<0)||(sec<0)) {
    WY_RETURN(Wym_EINVAL);
  }
  if(tzval==NULL) {
    WY_RETURN(Wym_EFAULT);
  }
  // Warning: Not all unrecognized tzval can be reported
  if(tzval[0]==0) {
    WY_RETURN(Wym_EINVAL);  // empty string not accepted for now
  }

  struct std::tm tm;
  std::time_t tt;
  tm.tm_year=year-1900;
  tm.tm_mon=month-1;
  tm.tm_mday=mday;
  tm.tm_hour=hour;
  tm.tm_min=min;
  tm.tm_sec=sec;
  tm.tm_isdst=-1;

  if(tm.tm_year>=year) {
    WY_RETURN(Wym_ERANGE);  // value conversion failed
  }

  {
    WyLock aa(wy_tzmutex());
    TmpEnvVar rr_tz(ENVNAME_TZ);

    if(::setenv(ENVNAME_TZ,tzval,1)!=0) {
      WY_RETURN(Wym_ENOMEM);  // man. indicates no errno and guess sort of this reason
    }
    ::tzset();
    tt=std::mktime(&tm);
    if(tt==static_cast<std::time_t>(-1)) {
      WY_RETURN(Wym_ERANGE);
    }
    _tzv=tzval;   // WyCSeg(tzval) wont throw
  }

  _wyt.reset(tt,WyTimeSpec::NanoSecType(0));
  _isdst=tm.tm_isdst;
  _year=year;
  _month=month;
  _mday=mday;
  _hour=hour;
  _min=min;
  _sec=sec;
  _wday=tm.tm_wday;
  _yday=tm.tm_yday;
  return(Ok);
}
catch(const WyStr::Reply& e) {
 return(e);
}
catch(const WyRet& e) {
 WY_NDEBUG_MSG(e);
 WY_THROW( WyRet(e) );
};

void WyDate::Impl::swap(Impl& another) WY__TSPC()
{
 _tzv.swap(another._tzv);
 _wyt.swap(another._wyt);
 Wy__Base::valswp(_isdst,another._isdst);
 Wy__Base::valswp(_year,another._year);
 Wy__Base::valswp(_month,another._month);
 Wy__Base::valswp(_mday,another._mday);
 Wy__Base::valswp(_hour,another._hour);
 Wy__Base::valswp(_min,another._min);
 Wy__Base::valswp(_sec,another._sec);
 Wy__Base::valswp(_wday,another._wday);
 Wy__Base::valswp(_yday,another._yday);
};

WyRet WyDate::Impl::set_utc(const WyTimeSpec& rhs)
try {
struct tm tm;
const std::time_t tt( rhs.second() );

  {
    WyLock aa(wy_tzmutex());
    TmpEnvVar rr_tz(ENVNAME_TZ);

    if(::setenv(ENVNAME_TZ,_tzv.c_str(),1)!=0) {
      WY_RETURN(Wym_ENOMEM);  // man. indicates no errno and guess sort of this reason
    }
    ::tzset();  // man. indicated this can be saved, call it however.
    if(::localtime_r(&tt,&tm)==NULL) {
      WY_RETURN(Wym_ERANGE);
    }
  }

 _wyt.reset(rhs);
 _isdst=tm.tm_isdst;
 _year=tm.tm_year+1900;
 _month=tm.tm_mon+1;
 _mday=tm.tm_mday;
 _hour=tm.tm_hour;
 _min=tm.tm_min;
 _sec=tm.tm_sec;
 _wday=tm.tm_wday;
 _yday=tm.tm_yday;
 return(Ok);
}
catch(const WyStr::Reply& e) {
 return(e);
}
catch(const WyRet& e) {
 WY_NDEBUG_MSG(e);
 WY_THROW( WyRet(e) );
};

WyRet WyDate::Impl::set_date(int year,int month,int mday, int hour,int min, int sec)
try {
  if((month<1)||(mday<1)||(hour<0)||(min<0)||(sec<0)) {
    WY_RETURN(Wym_EINVAL);
  }

  struct std::tm tm;
  std::time_t tt;
  tm.tm_year=year-1900;
  tm.tm_mon=month-1;
  tm.tm_mday=mday;
  tm.tm_hour=hour;
  tm.tm_min=min;
  tm.tm_sec=sec;
  tm.tm_isdst=_isdst;

  if(tm.tm_year>=year) {
    WY_RETURN(Wym_ERANGE);  // value conversion failed
  }

  {
    WyLock aa(wy_tzmutex());
    TmpEnvVar rr_tz(ENVNAME_TZ);

    if(::setenv(ENVNAME_TZ,_tzv.c_str(),1)!=0) {
      WY_RETURN(Wym_ENOMEM);  // man. indicates no errno and guess sort of this reason
    }
    ::tzset();
    tt=std::mktime(&tm);
    if(tt==static_cast<std::time_t>(-1)) {
      WY_RETURN(Wym_ERANGE);
    }
  }

  _wyt.reset(tt,WyTimeSpec::NanoSecType(0));
  _isdst=tm.tm_isdst;
  _year=year;
  _month=month;
  _mday=mday;
  _hour=hour;
  _min=min;
  _sec=sec;
  _wday=tm.tm_wday;
  _yday=tm.tm_yday;
  return(Ok);
}
catch(const WyStr::Reply& e) {
 return(e);
}
catch(const WyRet& e) {
 WY_NDEBUG_MSG(e);
 WY_THROW( WyRet(e) );
};

WyRet WyDate::Impl::add(const WyTimeSpec& tm)
try {
 const WyRet r=this->set_utc(tm+_wyt);
 WY_RETURN(r);
}
catch(const WyTimeSpec::Reply& e) {
  return(e);
};

const WyDate::Impl & WyDate::Impl::operator +=(const WyTimeSpec &rhs)
try {
 WyRet r=this->set_utc(_wyt+rhs);
 if(r!=Ok) {
   WY_THROW( WyDate::Reply(r) );
 }
 return *this;
}
catch(const WyDate::Reply&) {
  throw;
}
catch(const WyTimeSpec::Reply& e) {
  WY_THROW( WyDate::Reply(e) );
};

const WyDate::Impl& WyDate::Impl::operator =(const WyDate::Impl& src) WY__TSPC(WyDate::Reply)
{
 WyRet r=_tzv.reset(src._tzv);
 if(r!=Ok) {
   WY_THROW( WyDate::Reply(r) );
 }
 _wyt.reset(src._wyt);
 _isdst=src._isdst;
 _year=src._year;
 _month=src._month;
 _mday=src._mday;
 _hour=src._hour;
 _min=src._min;
 _sec=src._sec;
 _wday=src._wday;
 _yday=src._yday;
 return(*this);
};

WyRet WyDate::Impl::_strftime(WyStr& res, const char* format) const
try {
 WyRet r;
 struct tm tm;
 {
   std::time_t tt(_wyt.second());
   WyLock aa(wy_tzmutex());
   TmpEnvVar rr_tz(ENVNAME_TZ);

   if(::setenv(ENVNAME_TZ,_tzv.c_str(),1)!=0) {
     WY_RETURN(Wym_ENOMEM);  // man. indicates no errno and guess sort of this reason
   }
   ::tzset();  // man. indicated this can be saved, call it however.
   if(::localtime_r(&tt,&tm)==NULL) {
     WY_THROW( WyRet() );
   }
 }

 WyStr buf;
 for(size_t bsize=256; ; bsize+=256) {
   if((r=buf.resize(bsize))!=Ok) {
     WY_RETURN(r);  // only Wym_ENOMEM expected
   }

   // Note: expr uses WyStr internal buffer directly
   //
   buf.resize( ::strftime(&buf[0],bsize,format,&tm) );

   // The algorithm detecting that buf space is fully used for insufficient
   // bsize, because ::strftime may return 0 which is valid.
   //
   if(buf.size()<bsize-1) {
     break;
   }
   if(bsize>=1024*128) {
     WY_RETURN(Wym_ENOMEM);  // too big
   }
 }

 buf.swap(res);
 return(Ok);
}
catch(const WyStr::Reply& e) {
 return(e);
}
catch(const WyRet& e) {
 WY_NDEBUG_MSG(e);
 WY_THROW( WyRet(e) );
};

//---------------------------------------------------------------------
WyDate::WyDate() WY__TSPC(WyDate::Reply)
try : _dptr(new Impl())
{
}
catch(const std::bad_alloc&) {
  WY_THROW( WyDate::Reply(Wym_ENOMEM) );
};

WyDate::WyDate(const WyDate &src) WY__TSPC(WyDate::Reply)
try : _dptr(new Impl(*src._dptr))
{
}
catch(const std::bad_alloc&) {
  WY_THROW( WyDate::Reply(Wym_ENOMEM) );
};

WyDate::WyDate(const char* tzval,const WyTimeSpec& ucs)
try : _dptr(new Impl(tzval,ucs))
{
}
catch(const std::bad_alloc&) {
  WY_THROW( WyDate::Reply(Wym_ENOMEM) );
};

WyDate::WyDate(const WyStr& tzval,const WyTimeSpec& ucs)
try : _dptr(new Impl(tzval.c_str(),ucs))
{
}
catch(const std::bad_alloc&) {
  WY_THROW( WyDate::Reply(Wym_ENOMEM) );
};

WyDate::WyDate(const char* tzval,int year,int month,int mday, int hour,int min, int sec)
try : _dptr(new Impl(tzval,year,month,mday,hour,min,sec))
{
}
catch(const std::bad_alloc&) {
  WY_THROW( WyDate::Reply(Wym_ENOMEM) );
};

WyDate::WyDate(const WyStr& tzval,int year,int month,int mday, int hour,int min, int sec)
try : _dptr(new Impl(tzval.c_str(),year,month,mday,hour,min,sec))
{
}
catch(const std::bad_alloc&) {
  WY_THROW( WyDate::Reply(Wym_ENOMEM) );
};

WyDate::WyDate(WyDate& src, Wy::ByMove_t) WY__TSPC()
 : _dptr(src._dptr)
{
 src._dptr=NULL;
};

WyDate::~WyDate() WY__TSPC()
{
 Wy__Base::NoCancel noc;
 delete _dptr; _dptr=0;
};

bool WyDate::is_default() const WY__TSPC()
{
 return _dptr->is_default();
};

void WyDate::reset() WY__TSPC()
{
 _dptr->reset();
};

WyRet WyDate::reset(const WyDate &rhs) WY__TSPC()
{
 return _dptr->reset(*rhs._dptr);
};

WyRet WyDate::reset(const char* tzv,const WyTimeSpec& ucs)
{
 return _dptr->reset(tzv,ucs);
};

WyRet WyDate::reset(const WyStr& tzv,const WyTimeSpec& ucs)
{
 return _dptr->reset(tzv.c_str(),ucs);
};

WyRet WyDate::reset(const char* tzv, int year,int month,int mday,
                 int hour,int min, int sec)
{
 return _dptr->reset(tzv,year,month,mday,hour,min,sec);
};

WyRet WyDate::reset(const WyStr& tzv, int year,int month,int mday,
                 int hour,int min, int sec)
{
 return _dptr->reset(tzv.c_str(),year,month,mday,hour,min,sec);
};

void WyDate::swap(WyDate& ano) WY__TSPC()
{
 // For time_zone() outputs WyStr reference, imp swap instead of
 // pointer swap
 // Wy__Base::valswp(_dptr,ano._dptr);
 //
 _dptr->swap(*ano._dptr);
};

// Not quit sure cstr/WyCSeg is better for this member to return?
//
WyCSeg WyDate::time_zone() const WY__TSPC()
{
 return _dptr->time_zone();
};

int WyDate::year() const WY__TSPC()
{
 return _dptr->year();
};

int WyDate::month() const WY__TSPC()
{
 return _dptr->month();
};

int WyDate::mday() const WY__TSPC()
{
 return _dptr->mday();
};

int WyDate::hour() const WY__TSPC()
{
 return _dptr->hour();
};

int WyDate::min() const WY__TSPC()
{
 return _dptr->min();
};

int WyDate::sec() const WY__TSPC()
{
 return _dptr->sec();
};

int WyDate::wday() const WY__TSPC()
{
 return _dptr->wday();
};

int WyDate::yday() const WY__TSPC()
{
 return _dptr->yday();
};

int WyDate::isdst() const WY__TSPC()
{
 return _dptr->isdst();
};

WyTimeSpec WyDate::utc_time() const WY__TSPC()
{
 return _dptr->utc_time();
};

WyRet WyDate::set_utc(const WyTimeSpec& uct)
{
 return _dptr->set_utc(uct);
};

WyRet WyDate::set_date(int year,int month,int mday, int hour,int min, int sec)
{
 return _dptr->set_date(year,month,mday,hour,min,sec);
};

WyRet WyDate::add(const WyTimeSpec &tm)
{
 return _dptr->add(tm);
};

const WyDate& WyDate::operator =(const WyDate &rhs) WY__TSPC(Reply)
{
 _dptr->operator =(*rhs._dptr);
 return(*this);
};

const WyDate& WyDate::operator +=(const WyTimeSpec& rhs)
{
 _dptr->operator+=(rhs);
 return(*this);
};

bool WyDate::operator ==(const WyDate& rhs) const WY__TSPC()
{ 
 return _dptr->operator ==(*rhs._dptr);
};

bool WyDate::operator !=(const WyDate& rhs) const WY__TSPC()
{ 
 return _dptr->operator !=(*rhs._dptr);
};

//---------------------------------------------------------------------
WyStr Wy::wrd(const WyDate& date)
try {
 WyRet r;
 WyStr str;

 if((r=Wy::_mkstr(str,date.year(), 10))!=Ok) {
   WY_THROW(WyStr::Reply(r));
 }
 str+='-';

 {
  if(date.month()<10) {
    str+='0';
    str+=char(char('0')+char(date.month()%10));
  } else {
    str+=char(char('0')+char(date.month()/10));
    str+=char(char('0')+char(date.month()%10));
  }
 }
 str+='-';

 {
  if(date.mday()<10) {
    str+='0';
    str+=char(char('0')+char(date.mday()%10));
  } else {
    str+=char(char('0')+char(date.mday()/10));
    str+=char(char('0')+char(date.mday()%10));
  }
 }
 str+=' ';

 {
  if(date.hour()<10) {
    str+='0';
    str+=char(char('0')+char(date.hour()%10));
  } else {
    str+=char(char('0')+char(date.hour()/10));
    str+=char(char('0')+char(date.hour()%10));
  }
 }
 str+=':';

 {
  if(date.min()<10) {
    str+='0';
    str+=char(char('0')+char(date.min()%10));
  } else {
    str+=char(char('0')+char(date.min()/10));
    str+=char(char('0')+char(date.min()%10));
  }
 }
 str+='\'';

 {
  if(date.sec()<10) {
    str+='0';
    str+=char(char('0')+char(date.sec()%10));
  } else {
    str+=char(char('0')+char(date.sec()/10));
    str+=char(char('0')+char(date.sec()%10));
  }
 }
 str+= "\"(";
 str+= char(date.wday()+'0');
 str+=')';

 // add time zone information
 str+=' ';
 str+=date.time_zone();

 return(str);
}
catch(const WyStr::Reply& e) {
 if(e==Wym_ENOMEM) {
   throw;
 }
 WY_THROW( WyRet(e) );
}
catch(const std::bad_alloc&) {
  WY_THROW(WyStr::Reply(Wym_ENOMEM));
};

WyRet Wy::_strftime(WyStr& res, const char* format, const WyDate& date)
{
 return( date.wy_impl()->_strftime(res,format) );
};

WyDate operator +(const WyDate& dt,const WyTimeSpec& tm)
try {
 WyDate v(dt);
 v+=tm;
 return(v);
}
catch(const WyDate::Reply&) {
  throw;
} 
catch(const WyTimeSpec::Reply& e) {
  WY_THROW( WyDate::Reply(e) );
};

WyDate operator +(const WyTimeSpec& tm,const WyDate& dt)
try {
 WyDate v(dt);
 v+=tm;
 return(v);
}
catch(const WyDate::Reply&) {
  throw;
} 
catch(const WyTimeSpec::Reply& e) {
  WY_THROW( WyDate::Reply(e) );
};

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <ctype.h>

// [Internal] algorithm#1: 
//            Get the time zone string from symbolic link LocalTimeZone 
//            Reset res with the result
//
// [Ret] Ok
//       Wym_ESRCH       Search failed for local time zone name
//       Wym_ENOMEM      Not enough memory
//
static WyRet wy_local_tzname_1(WyStr& res)
try {
 // Information from ::tzname[2] is not the expected.
 // 1. Few won't work and (daylight information seemd required)
 // 2. less readable
 // 
 
 // It seemd ok to derive the tz data file from the time zone link file
 // This seemd ok for RH 9.0, not for Fedora
 // 
 static const char LocalTimeZone[]="/etc/localtime";
 static const char SrchPtn[]="zoneinfo/";
 const size_t PBufSize=PATH_MAX+1;
 char* pbuf( (char*)std::malloc(PBufSize) );
 if(pbuf==NULL) {
   WY_RETURN(Wym_ENOMEM);
 }
 Wy_AtDestroy<void,char*> rrid(reinterpret_cast<void(*)(char*)>(&std::free),pbuf);

 const int v=::readlink(LocalTimeZone,pbuf,PBufSize);
 if(v<0) {
   if(errno==ENOMEM) {
     WY_RETURN(Wym_ENOMEM);
   }
   WY_RETURN(Wym_ESRCH);
 }
 if(size_t(v)>=PBufSize) {
   WY_THROW( WyRet() );
 }

 WyRet r;
 WyStr str( WyCSeg(pbuf,v) );
 size_t fidx;
 std::free(pbuf);
 rrid.release();

 if(str.find(fidx,size_t(0),SrchPtn)==false) {
   WY_RETURN(Wym_ESRCH);
 }
 {
   WyStr nbuf(":");
   nbuf+=str.cseg(fidx+sizeof(SrchPtn)-1);
   if((r=res.reset(nbuf))!=Ok) {
     WY_RETURN(Wym_ENOMEM);
   }
 }
 return(Ok);
}
catch(const WyStr::Reply& e) {
  WY_RETURN(e);
};

// [Internal] algorithm#2:
//            Get the value of environment variable TZ and reset res
//            with the result
//
// [Ret] Ok
//       Wym_ESRCH       Search failed for local time zone name
//       Wym_ENOMEM      Not enough memory
//
static WyRet wy_local_tzname_2(WyStr& res)
try {
 bool tz_defined;
 WyStr tz_value;
 WyRet r( getenv_tz(tz_defined, tz_value) );
 if(r!=Ok) {
   WY_RETURN(Wym_ESRCH);
 }
 if(tz_defined==false) {
   WY_RETURN(Wym_ESRCH);
 }
 if(tz_value.is_default()) {
   WY_RETURN(Wym_ESRCH);
 }
 tz_value.swap(res);
 return(Ok);
}
catch(const WyStr::Reply& e) {
 WY_RETURN(e);
};

// [Internal] algorithm#3:
//            From ::tzname[2]/::timezone
//            Reset res with the result
//
// [Ret] Ok
//       Wym_ESRCH       Search failed for local time zone name
//       Wym_ENOMEM      Not enough memory
//
// Note: Different source reveals different algorithms
//
static WyRet wy_local_tzname_3(WyStr& res)
try {
 ::tzset();
 if((::tzname[0]==0)||(*::tzname[0]==0)) {
   WY_RETURN(Wym_ESRCH);
 }
 if((::tzname[1]==0)||(*::tzname[1]==0)) {
   WY_RETURN(Wym_ESRCH);
 }
 WyStr sbuf(::daylight? ::tzname[1]: ::tzname[0]);

 // append sbuf with OFFSET
 //
 {
  long toff(::timezone);
  int toff_sec,toff_min;

  if(toff<0) {
    sbuf+='-';
    toff=-toff;
  }
  toff_sec=toff%60;
  toff/=60;
  toff_min=toff%60;
  toff/=60;

  if(toff<10) {
    sbuf+=char(char('0')+char(toff));
  } else {
    sbuf+=char(char('0')+char(toff/10));
    sbuf+=char(char('0')+char(toff%10));
  }
  if((toff_min!=0)||(toff_sec!=0)) {
    sbuf+=':';
    if(toff_min<10) {
      sbuf+=char(char('0')+char(toff_min));
    } else {
      sbuf+=char(char('0')+char(toff_min/10));
      sbuf+=char(char('0')+char(toff_min%10));
    }
    if(toff_sec!=0) {
      sbuf+=':';
      if(toff_sec<10) {
        sbuf+=char(char('0')+char(toff_sec));
      } else {
        sbuf+=char(char('0')+char(toff_sec/10));
        sbuf+=char(char('0')+char(toff_sec%10));
      }
    }
  }
 }

 // not right for my machine (from other document read)
/*
 if(::daylight!=0) {
   if((::tzname[1]!=0)&&(*::tzname[1]!=0)) {
     sbuf+=::tzname[1];
   }
 }*/
 res.swap(sbuf);
 return(Ok);
}
catch(const WyStr::Reply& e) {
 WY_RETURN(e);
};

// Goal of this function is to get the identifying tzname of the system time
// in effect. The return time zone string should be valid TZ value that
// system time functions can recognize.
//
WyStr Wy::local_tzname()
try {
 WyRet r;
 WyStr str;
 if((r=wy_local_tzname_1(str))==Ok) {
   return(str);
 }
 if((r=wy_local_tzname_3(str))==Ok) {
   return(str);
 }
 if((r=wy_local_tzname_2(str))==Ok) {
   return(str);
 }
 if(r==Wym_ESRCH) {
   return WyStr(WyCSeg(WyDate::TIMEZONE_UTC,sizeof(WyDate::TIMEZONE_UTC)-1));
 }
 WY_THROW( WyDate::Reply(Wym_ENOMEM) );
}
catch(const WyStr::Reply&) {
 WY_THROW( WyDate::Reply(Wym_ENOMEM) );
}
catch(const WyRet& e) {
 WY_NDEBUG_MSG(e);
 WY_THROW( WyRet(e) );
};
