/* Copyright is licensed under GNU LGPL.                 by I.J.Wang 2003
*/

#define WYLIB_SOURCE
#include "WyRet.h"
#include <string.h>    // for ::strerror(int)
#include <cerrno>
#include <cstring>     // for strlen
#include <pthread.h>   // for mutex

//-------------------------------------------------
// Definition for Wy_RepInfo
//
// Note: Order of code is significant
//-------------------------------------------------
const Wy_RepInfo::Wy__RepMsgTID Wy_RepInfo::_Default_msgtid="";

Wy_RepInfo::Wy_RepInfo(Wy__RepCode ecode, const char* cstr) WY__NOTHROW__
 : _rcode(ecode), _rstr(cstr)
{
 if(cstr==NULL) {
   WY_TERMINATE("");
 }
};

Wy_RepInfo::Wy__RepCode Wy_RepInfo::repcode() const WY__NOTHROW__
{
 return _rcode; 
};

const char* Wy_RepInfo::repstr() const WY__NOTHROW__
{
 return _rstr;
};

size_t Wy_RepInfo::getmsg(Wy__RepMsgTID& mtid, void* buf, size_t blen) const WY__NOTHROW__
{
 mtid= _Default_msgtid;
 size_t slen( std::strlen(_rstr) );
 if(buf!=NULL) {
   if(blen<slen) {
     slen=blen;
   }
   std::memcpy(buf,_rstr,slen);
 }
 return slen; 
};

//-------------------------------------------------
// Definition for WyRet
//
// Note: Order of code is significant
//-------------------------------------------------
class Wy__ErrNo : public Wy_RepInfo {
    Wy__RepMsgTID _ap_mtid;  // application reply message type id
    const void* _ap_data;    // application reply message data
    size_t      _ap_dlen;    // length of the message pointed by _ap_data
    mutable ::pthread_mutex_t _mtx;

    // Hidden
    const Wy__ErrNo& operator =(const Wy__ErrNo&);
    bool operator ==(const Wy__ErrNo&);
    void* operator new(size_t);

  public:

    // [Syn] Construct object
    //
    //      repcode()= en
    //       repstr()= cstr
    //       getmsg()= copy out cstr
    //
    Wy__ErrNo(Wy__ErrNoType en, const char* cstr) WY__NOTHROW__
       : Wy_RepInfo(en,cstr),
          _ap_mtid(Wy_RepInfo::_Default_msgtid),
          _ap_data(cstr), _ap_dlen(std::strlen(cstr))
         {
           ::pthread_mutex_init(&_mtx,NULL);
         };

    Wy__ErrNo(Wy__ErrNoType en) WY__NOTHROW__
       : Wy_RepInfo(en,std::strerror(en)),
          _ap_mtid(Wy_RepInfo::_Default_msgtid)
         {
           _ap_data= Wy_RepInfo::repstr();
           _ap_dlen= std::strlen( static_cast<const char*>(_ap_data) );
           ::pthread_mutex_init(&_mtx,NULL);
         };

    ~Wy__ErrNo() WY__NOTHROW__
         {
           ::pthread_mutex_destroy(&_mtx); 
         };

    // Note: .Copy constructor is not defined.
    //       .Compiler (g++) seems need to 'see' this declaration.
    //        This may be from Standard requirement.
    //       .Wy_RepInfo object is not supposed to be derived by copy
    //        constructor
    //
    Wy__ErrNo(const Wy__ErrNo&) WY__NOTHROW__;

    // [Derived]
    // int repgrp() const _throw()
    // int repcode() const _throw()
    // const char* repstr() const _throw()

    // [Override]
    //
    size_t getmsg(Wy__RepMsgTID& mtid, void* buf, size_t blen) const WY__NOTHROW__
         {
           if(::pthread_mutex_lock(&_mtx)!=0) {
             WY_TERMINATE("");
           }
           mtid=_ap_mtid;
           if(buf==NULL) {
             const size_t rv(_ap_dlen);
             if(::pthread_mutex_unlock(&_mtx)!=0) {
               WY_TERMINATE("");
             }
             return(rv);
           }
           const size_t slen( (blen<_ap_dlen)? blen:_ap_dlen);
           std::memcpy(buf,_ap_data,slen);
           if(::pthread_mutex_unlock(&_mtx)!=0) {
             WY_TERMINATE("");
           }
           return slen; 
         };

    // [Override]
    //
    void setmsg(Wy__RepMsgTID mtid, const void* data, size_t dlen) WY__NOTHROW__
         {
           if(::pthread_mutex_lock(&_mtx)!=0) {
             WY_TERMINATE("");
           }
           _ap_mtid=mtid;
           if(data==NULL) {
             _ap_data=this->repstr();
             _ap_dlen=::strlen( static_cast<const char*>(_ap_data) );
           } else {
             _ap_data=data;
             _ap_dlen=dlen;
           }
           if(::pthread_mutex_unlock(&_mtx)!=0) {
             WY_TERMINATE("");
           }
         };
};

// The object for default WyReply to refer to
static Wy__ErrNo wye_default_repinfo(Wy::Wy__Default_ErrNo,Wy::Wy__Default_RepStr);

//-------------------------------------------------
// Definition for WyRet
//
// Note: Order of code is significant
//-------------------------------------------------

Wy_RepInfo* const WyReply::_default_ri_ptr=&wye_default_repinfo;

// Wy__ErrNo object for errno
// 
static Wy__ErrNo wye_tab[Wy::_max_errno+1]= {
  Wy__ErrNo(  0), Wy__ErrNo(  1), Wy__ErrNo(  2), Wy__ErrNo(  3), Wy__ErrNo(  4),
  Wy__ErrNo(  5), Wy__ErrNo(  6), Wy__ErrNo(  7), Wy__ErrNo(  8), Wy__ErrNo(  9),
  Wy__ErrNo( 10), Wy__ErrNo( 11), Wy__ErrNo( 12), Wy__ErrNo( 13), Wy__ErrNo( 14),
  Wy__ErrNo( 15), Wy__ErrNo( 16), Wy__ErrNo( 17), Wy__ErrNo( 18), Wy__ErrNo( 19),
  Wy__ErrNo( 20), Wy__ErrNo( 21), Wy__ErrNo( 22), Wy__ErrNo( 23), Wy__ErrNo( 24),
  Wy__ErrNo( 25), Wy__ErrNo( 26), Wy__ErrNo( 27), Wy__ErrNo( 28), Wy__ErrNo( 29),
  Wy__ErrNo( 30), Wy__ErrNo( 31), Wy__ErrNo( 32), Wy__ErrNo( 33), Wy__ErrNo( 34),
  Wy__ErrNo( 35), Wy__ErrNo( 36), Wy__ErrNo( 37), Wy__ErrNo( 38), Wy__ErrNo( 39),
  Wy__ErrNo( 40), Wy__ErrNo( 41), Wy__ErrNo( 42), Wy__ErrNo( 43), Wy__ErrNo( 44),
  Wy__ErrNo( 45), Wy__ErrNo( 46), Wy__ErrNo( 47), Wy__ErrNo( 48), Wy__ErrNo( 49),
  Wy__ErrNo( 50), Wy__ErrNo( 51), Wy__ErrNo( 52), Wy__ErrNo( 53), Wy__ErrNo( 54),
  Wy__ErrNo( 55), Wy__ErrNo( 56), Wy__ErrNo( 57), Wy__ErrNo( 58), Wy__ErrNo( 59),
  Wy__ErrNo( 60), Wy__ErrNo( 61), Wy__ErrNo( 62), Wy__ErrNo( 63), Wy__ErrNo( 64),
  Wy__ErrNo( 65), Wy__ErrNo( 66), Wy__ErrNo( 67), Wy__ErrNo( 68), Wy__ErrNo( 69),
  Wy__ErrNo( 70), Wy__ErrNo( 71), Wy__ErrNo( 72), Wy__ErrNo( 73), Wy__ErrNo( 74),
  Wy__ErrNo( 75), Wy__ErrNo( 76), Wy__ErrNo( 77), Wy__ErrNo( 78), Wy__ErrNo( 79),
  Wy__ErrNo( 80), Wy__ErrNo( 81), Wy__ErrNo( 82), Wy__ErrNo( 83), Wy__ErrNo( 84),
  Wy__ErrNo( 85), Wy__ErrNo( 86), Wy__ErrNo( 87), Wy__ErrNo( 88), Wy__ErrNo( 89),
  Wy__ErrNo( 90), Wy__ErrNo( 91), Wy__ErrNo( 92), Wy__ErrNo( 93), Wy__ErrNo( 94),
  Wy__ErrNo( 95), Wy__ErrNo( 96), Wy__ErrNo( 97), Wy__ErrNo( 98), Wy__ErrNo( 99),
  Wy__ErrNo(100), Wy__ErrNo(101), Wy__ErrNo(102), Wy__ErrNo(103), Wy__ErrNo(104),
  Wy__ErrNo(105), Wy__ErrNo(106), Wy__ErrNo(107), Wy__ErrNo(108), Wy__ErrNo(109),
  Wy__ErrNo(110), Wy__ErrNo(111), Wy__ErrNo(112), Wy__ErrNo(113), Wy__ErrNo(114),
  Wy__ErrNo(115), Wy__ErrNo(116), Wy__ErrNo(117), Wy__ErrNo(118), Wy__ErrNo(119),
  Wy__ErrNo(120), Wy__ErrNo(121), Wy__ErrNo(122), Wy__ErrNo(123), Wy__ErrNo(124),
  Wy__ErrNo(125), Wy__ErrNo(126), Wy__ErrNo(127), Wy__ErrNo(128), Wy__ErrNo(129),
  Wy__ErrNo(130), Wy__ErrNo(131),
};

// libwy added errinfo objects 
//
static Wy__ErrNo wye_ETHRDCAN(Wy::Wy__Default_ErrNo,"Thread canceled");
static Wy__ErrNo wye_EDIVZERO(Wy::Wy__Default_ErrNo,"Divided by zero");
static Wy__ErrNo wye_EBADTYPE(Wy::Wy__Default_ErrNo,"Bad argument type");

//-------------------------------------------------------
// Note: Order of code is significant
//-------------------------------------------------------

// WyReply constants. Text is probably the same as for C-library errno
//
 const WyReply Ok( wye_tab[0] );
 const WyReply Wym_EPERM( wye_tab[1] );
 const WyReply Wym_ENOENT( wye_tab[2] );
 const WyReply Wym_ESRCH( wye_tab[3] );
 const WyReply Wym_EINTR( wye_tab[4] );
 const WyReply Wym_EIO	( wye_tab[5] );
 const WyReply Wym_ENXIO( wye_tab[6] );
 const WyReply Wym_E2BIG( wye_tab[7] );
 const WyReply Wym_ENOEXEC( wye_tab[8] );
 const WyReply Wym_EBADF( wye_tab[9] );
 const WyReply Wym_ECHILD( wye_tab[10] );
 const WyReply Wym_EAGAIN( wye_tab[11] );
 const WyReply Wym_ENOMEM( wye_tab[12] );
 const WyReply Wym_EACCES( wye_tab[13] );
 const WyReply Wym_EFAULT( wye_tab[14] );
 const WyReply Wym_ENOTBLK( wye_tab[15] );
 const WyReply Wym_EBUSY( wye_tab[16] );
 const WyReply Wym_EEXIST( wye_tab[17] );
 const WyReply Wym_EXDEV( wye_tab[18] );
 const WyReply Wym_ENODEV( wye_tab[19] );
 const WyReply Wym_ENOTDIR( wye_tab[20] );
 const WyReply Wym_EISDIR( wye_tab[21] );
 const WyReply Wym_EINVAL( wye_tab[22] );
 const WyReply Wym_ENFILE( wye_tab[23] );
 const WyReply Wym_EMFILE( wye_tab[24] );
 const WyReply Wym_ENOTTY( wye_tab[25] );
 const WyReply Wym_ETXTBSY( wye_tab[26] );
 const WyReply Wym_EFBIG( wye_tab[27] );
 const WyReply Wym_ENOSPC( wye_tab[28] );
 const WyReply Wym_ESPIPE( wye_tab[29] );
 const WyReply Wym_EROFS( wye_tab[30] );
 const WyReply Wym_EMLINK( wye_tab[31] );
 const WyReply Wym_EPIPE( wye_tab[32] );
 const WyReply Wym_EDOM( wye_tab[33] );
 const WyReply Wym_ERANGE( wye_tab[34] );
 const WyReply Wym_EDEADLK( wye_tab[35] );
 const WyReply Wym_ENAMETOOLONG( wye_tab[36] );
 const WyReply Wym_ENOLCK( wye_tab[37] );
 const WyReply Wym_ENOSYS( wye_tab[38] );
 const WyReply Wym_ENOTEMPTY( wye_tab[39] );
 const WyReply Wym_ELOOP( wye_tab[40] );
 const WyReply Wym_EWOULDBLOCK( wye_tab[11] );   // ==EAGAIN
 const WyReply Wym_E41( wye_tab[41] );
 const WyReply Wym_ENOMSG( wye_tab[42] );
 const WyReply Wym_EIDRM( wye_tab[43] );
 const WyReply Wym_ECHRNG( wye_tab[44] );
 const WyReply Wym_EL2NSYNC( wye_tab[45] );
 const WyReply Wym_EL3HLT( wye_tab[46] );
 const WyReply Wym_EL3RST( wye_tab[47] );
 const WyReply Wym_ELNRNG( wye_tab[48] );
 const WyReply Wym_EUNATCH( wye_tab[49] );
 const WyReply Wym_ENOCSI( wye_tab[50] );
 const WyReply Wym_EL2HLT( wye_tab[51] );
 const WyReply Wym_EBADE( wye_tab[52] );
 const WyReply Wym_EBADR( wye_tab[53] );
 const WyReply Wym_EXFULL( wye_tab[54] );
 const WyReply Wym_ENOANO( wye_tab[55] );
 const WyReply Wym_EBADRQC( wye_tab[56] );
 const WyReply Wym_EBADSLT( wye_tab[57] );
 const WyReply Wym_EDEADLOCK( wye_tab[35] );     // ==EDEADLK
 const WyReply Wym_E58( wye_tab[58] );
 const WyReply Wym_EBFONT( wye_tab[59] );
 const WyReply Wym_ENOSTR( wye_tab[60] );
 const WyReply Wym_ENODATA( wye_tab[61] );
 const WyReply Wym_ETIME( wye_tab[62] );
 const WyReply Wym_ENOSR( wye_tab[63] );
 const WyReply Wym_ENONET( wye_tab[64] );
 const WyReply Wym_ENOPKG( wye_tab[65] );
 const WyReply Wym_EREMOTE( wye_tab[66] );
 const WyReply Wym_ENOLINK( wye_tab[67] );
 const WyReply Wym_EADV( wye_tab[68] );
 const WyReply Wym_ESRMNT( wye_tab[69] );
 const WyReply Wym_ECOMM( wye_tab[70] );
 const WyReply Wym_EPROTO( wye_tab[71] );
 const WyReply Wym_EMULTIHOP( wye_tab[72] );
 const WyReply Wym_EDOTDOT( wye_tab[73] );
 const WyReply Wym_EBADMSG( wye_tab[74] );
 const WyReply Wym_EOVERFLOW( wye_tab[75] );
 const WyReply Wym_ENOTUNIQ( wye_tab[76] );
 const WyReply Wym_EBADFD( wye_tab[77] );
 const WyReply Wym_EREMCHG( wye_tab[78] );
 const WyReply Wym_ELIBACC( wye_tab[79] );
 const WyReply Wym_ELIBBAD( wye_tab[80] );
 const WyReply Wym_ELIBSCN( wye_tab[81] );
 const WyReply Wym_ELIBMAX( wye_tab[82] );
 const WyReply Wym_ELIBEXEC( wye_tab[83] );
 const WyReply Wym_EILSEQ( wye_tab[84] );
 const WyReply Wym_ERESTART( wye_tab[85] );
 const WyReply Wym_ESTRPIPE( wye_tab[86] );
 const WyReply Wym_EUSERS( wye_tab[87] );
 const WyReply Wym_ENOTSOCK( wye_tab[88] );
 const WyReply Wym_EDESTADDRREQ( wye_tab[89] );
 const WyReply Wym_EMSGSIZE( wye_tab[90] );
 const WyReply Wym_EPROTOTYPE( wye_tab[91] );
 const WyReply Wym_ENOPROTOOPT( wye_tab[92] );
 const WyReply Wym_EPROTONOSUPPORT( wye_tab[93] );
 const WyReply Wym_ESOCKTNOSUPPORT( wye_tab[94] );
 const WyReply Wym_EOPNOTSUPP( wye_tab[95] );
 const WyReply Wym_EPFNOSUPPORT( wye_tab[96] );
 const WyReply Wym_EAFNOSUPPORT( wye_tab[97] );
 const WyReply Wym_EADDRINUSE( wye_tab[98] );
 const WyReply Wym_EADDRNOTAVAIL( wye_tab[99] );
 const WyReply Wym_ENETDOWN( wye_tab[100] );
 const WyReply Wym_ENETUNREACH( wye_tab[101] );
 const WyReply Wym_ENETRESET( wye_tab[102] );
 const WyReply Wym_ECONNABORTED( wye_tab[103] );
 const WyReply Wym_ECONNRESET( wye_tab[104] );
 const WyReply Wym_ENOBUFS( wye_tab[105] );
 const WyReply Wym_EISCONN( wye_tab[106] );
 const WyReply Wym_ENOTCONN( wye_tab[107] );
 const WyReply Wym_ESHUTDOWN( wye_tab[108] );
 const WyReply Wym_ETOOMANYREFS( wye_tab[109] );
 const WyReply Wym_ETIMEDOUT( wye_tab[110] );
 const WyReply Wym_ECONNREFUSED( wye_tab[111] );
 const WyReply Wym_EHOSTDOWN( wye_tab[112] );
 const WyReply Wym_EHOSTUNREACH( wye_tab[113] );
 const WyReply Wym_EALREADY( wye_tab[114] );
 const WyReply Wym_EINPROGRESS( wye_tab[115] );
 const WyReply Wym_ESTALE( wye_tab[116] );
 const WyReply Wym_EUCLEAN( wye_tab[117] );
 const WyReply Wym_ENOTNAM( wye_tab[118] );
 const WyReply Wym_ENAVAIL( wye_tab[119] );
 const WyReply Wym_EISNAM( wye_tab[120] );
 const WyReply Wym_EREMOTEIO( wye_tab[121] );
 const WyReply Wym_EDQUOT( wye_tab[122] );
 const WyReply Wym_ENOMEDIUM( wye_tab[123] );
 const WyReply Wym_EMEDIUMTYPE( wye_tab[124] );
 const WyReply Wym_E125( wye_tab[125] );
 const WyReply Wym_E126( wye_tab[126] );
 const WyReply Wym_E127( wye_tab[127] );
 const WyReply Wym_E128( wye_tab[128] );
 const WyReply Wym_E129( wye_tab[129] );
 const WyReply Wym_E130( wye_tab[130] );
 const WyReply Wym_E131( wye_tab[131] );

// libwy added global reply constants
//
 const WyReply Wym_ETHRDCAN( wye_ETHRDCAN );
 const WyReply Wym_EDIVZERO( wye_EDIVZERO );
 const WyReply Wym_EBADTYPE( wye_EBADTYPE );

//-----------------------------------------------------------------------------
// Definition code for Wy_SrcLoc
//
// Note: Order of code is significant
//-----------------------------------------------------------------------------
// No code yet

//-----------------------------------------------------------------------------
// Definition code for WyReply
//
// Note: Order of code is significant
//-----------------------------------------------------------------------------
Wy_RepInfo* WyReply::_en2ptr(Wy__ErrNoType en) WY__NOTHROW__
{
 if((en<0)||(en>Wy::_max_errno)) {
   WY_ASSERT(_default_ri_ptr!=NULL);
   return(_default_ri_ptr);
 }
 WY_ASSERT(wye_tab[en].repcode()==en);
 return(&wye_tab[en]);
};

Wy__ErrNoType WyReply::reset(Wy__ErrNoType en) WY__NOTHROW__
{
 if((en<0)||(en>Wy::_max_errno)) {
   _ri_ptr=_default_ri_ptr;
   return(Wy::Wy__Default_ErrNo);
 }
 _ri_ptr=&wye_tab[en];
 WY_ASSERT(_ri_ptr->repcode()==en);
 return(en);
};

//-------------------------------------------------
// Definition for WyRet
//
// Note: Order of code is significant
//-------------------------------------------------
// No code yet
