/* Copyright is licensed under GNU LGPL.                 by I.J.Wang 2003

  WySockAddrINet4 deals with the IPv4 address (AF_INET)
*/

#ifndef WYSOCKADDRINET_H__
#define WYSOCKADDRINET_H__
#define WYSOCKADDRINET_VERSION 33

#include "WyRet.h"
#include "WyStr.h"
#include "Wy_SockAddr.h"
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>

// Class for IPv4 address (family=AF_INET)
//
class WySockAddrINet4 : public Wy_SockAddr {
  public:
    typedef ::in_port_t PortType;
    typedef ::in_addr_t InAddrType;

  private:
    static const InAddrType DefaultIP=INADDR_ANY;  // net format
    static const PortType DefaultPort=0;           // net format
    static const socklen_t DefaultSockLen=sizeof(::sockaddr_in);

  public:
    class IPAddr {
      public:
        typedef struct ::in_addr C_AddrType;
        static const char class_name[];
        WY_THROW_REPLY;

        IPAddr() WY__TSPC();
        IPAddr(const IPAddr&) WY__TSPC();
        IPAddr(const C_AddrType&) WY__TSPC();
        // The following two ctor's are exmplicit for one reason that g++ 4.1.1 interprets
        //   WySockAddrINet4(INADDR_ANY,port) to use IPAddr(const char*) version
        explicit
        IPAddr(InAddrType ip) WY__TSPC(); // ip is in host representation
        explicit
        IPAddr(const char* cstr);
        IPAddr(const WyStr& str);
        IPAddr(IPAddr&,Wy::ByMove_t) WY__TSPC();

        bool is_default() const WY__TSPC();
        C_AddrType c_addr() const WY__TSPC();
        WyRet notation(WyStr& str) const;
        void reset() WY__TSPC();
        void reset(const IPAddr&) WY__TSPC();
        void reset(const C_AddrType&) WY__TSPC();
        void reset(InAddrType ip) WY__TSPC();
        WyRet reset(const char* cstr);
        WyRet reset(const WyStr& str);
        void swap(IPAddr& addr) WY__TSPC();

        const IPAddr& operator =(const IPAddr&) WY__TSPC();

        bool operator ==(const IPAddr& rhs) const WY__TSPC();
        bool operator !=(const IPAddr& rhs) const WY__TSPC()
             { return !this->operator==(rhs); };
        bool operator ==(const C_AddrType& rhs) const WY__TSPC();
        bool operator !=(const C_AddrType& rhs) const WY__TSPC()
             { return !this->operator==(rhs); };
        bool operator ==(InAddrType rhs) const WY__TSPC();
        bool operator !=(InAddrType rhs) const WY__TSPC()
             { return !this->operator==(rhs); };
        bool operator ==(const WyStr& rhs) const;
        bool operator !=(const WyStr& rhs) const
             { return !this->operator==(rhs); };
        bool operator ==(const char* rhs) const;
        bool operator !=(const char* rhs) const
             { return !this->operator==(rhs); };

        const C_AddrType& wy_addr_ref() const WY__NOTHROW__
           { return _addr; };

      private:
        C_AddrType _addr;

        // [Ret] Ok
        //       Wym_EFAULT cstr inaccessible
        //       Wym_EADDRNOTAVAIL *cstr not recognizable
        //       Wym_EAFNOSUPPORT  cstr invalid
        //
        WyRet _init(const char* cstr);
    };

    static const char class_name[];
    inline static int family() WY__TSPC()
              { return AF_INET; };
    WY_THROW_REPLY;

    WySockAddrINet4() WY__TSPC();
    WySockAddrINet4(const WySockAddrINet4& src) WY__TSPC();
    WySockAddrINet4(const struct ::sockaddr_in&) WY__TSPC();
    WySockAddrINet4(const IPAddr& ip, PortType port) WY__TSPC();
    WySockAddrINet4(WySockAddrINet4& src, Wy::ByMove_t) WY__TSPC();

    bool is_default() const WY__TSPC();

    IPAddr ip() const WY__TSPC();
    PortType port() const WY__TSPC();

    // [Override]
    virtual WyRet notation(WyStr& str) const;

    virtual void reset() WY__TSPC();
    virtual void reset(const WySockAddrINet4& src) WY__TSPC();
    virtual void reset(const struct ::sockaddr_in&) WY__TSPC();
    virtual void reset(const IPAddr& ip, PortType port) WY__TSPC();

    virtual const WySockAddrINet4& operator=(const WySockAddrINet4 &rhs) WY__TSPC();

    virtual WyRet _swap(Wy_SockAddr& iaddr) WY__TSPC();
    virtual WySockAddrINet4* _alloc(WyRet&) const;

    bool operator==(const WySockAddrINet4 &rhs) const WY__TSPC();
    bool operator!=(const WySockAddrINet4 &rhs) const WY__TSPC()
              { return !operator==(rhs); };

  private:
    virtual struct ::sockaddr& c_sockaddr() WY__TSPC()
              { return (struct ::sockaddr&)(_saddr); };
    virtual const struct ::sockaddr& c_sockaddr() const WY__TSPC()
              { return (struct ::sockaddr&)(_saddr); };

    virtual size_t c_sockaddr_size() const WY__TSPC()
              { return sizeof(_saddr); };

    virtual  socklen_t& c_addrlen() WY__TSPC()
              { return _slen; };
    virtual const socklen_t& c_addrlen() const WY__TSPC()
              { return _slen; };

    // [Internal] Initialize default object
    //
    void _init() WY__TSPC();

    // [Internal] Initialize object by copy from src
    //
    void _init(const WySockAddrINet4 &src) WY__TSPC();
    void _init(const struct ::sockaddr_in&) WY__TSPC();

    // [Internal] Initialize object to IPv4 address (family=AF_INET)
    //
    //        port= IPv4 Port number
    //            = 0, no port (ephemeral port with bind())
    //        ip  = IP address.
    //
    void _init(const IPAddr& ip, PortType port) WY__TSPC();

  private:
    class Wy_Assert_Check;

    struct ::sockaddr_in _saddr;
    socklen_t _slen;
};

inline bool operator ==(const WySockAddrINet4::IPAddr::C_AddrType& ip1, const WySockAddrINet4::IPAddr& ip2) WY__TSPC()
{ return ip2==ip1; };
inline bool operator !=(const WySockAddrINet4::IPAddr::C_AddrType& ip1, const WySockAddrINet4::IPAddr& ip2) WY__TSPC()
{ return ip2!=ip1; };
inline bool operator ==(WySockAddrINet4::InAddrType ip1, const WySockAddrINet4::IPAddr& ip2) WY__TSPC()
{ return ip2==ip1; };
inline bool operator !=(WySockAddrINet4::InAddrType ip1, const WySockAddrINet4::IPAddr& ip2) WY__TSPC()
{ return ip2!=ip1; };
inline bool operator ==(const char* ip1, const WySockAddrINet4::IPAddr& ip2)
{ return ip2==ip1; };
inline bool operator !=(const char* ip1, const WySockAddrINet4::IPAddr& ip2)
{ return ip2!=ip1; };
inline bool operator ==(const WyStr& ip1, const WySockAddrINet4::IPAddr& ip2)
{ return ip2==ip1; };
inline bool operator !=(const WyStr& ip1, const WySockAddrINet4::IPAddr& ip2)
{ return ip2!=ip1; };

// Define equivalence of IPAddr output type C_AddrType.
inline bool operator==(const WySockAddrINet4::IPAddr::C_AddrType& lhs, const WySockAddrINet4::IPAddr::C_AddrType& rhs) WY__TSPC()
{ return lhs.s_addr==rhs.s_addr; };
inline bool operator!=(const WySockAddrINet4::IPAddr::C_AddrType& lhs, const WySockAddrINet4::IPAddr::C_AddrType& rhs) WY__TSPC()
{ return lhs.s_addr!=rhs.s_addr; };

namespace Wy {
 WyStr wrd(const WySockAddrINet4::IPAddr& addr);
 WyStr wrd(const WySockAddrINet4& addr);
};

#endif
