/* Copyright is licensed under GNU LGPL.                 by I.J.Wang 2005

  WySockAddrINet6 deals with the IPv6 address (AF_INET6)
*/

#ifndef WYSOCKADDRINET6_H__
#define WYSOCKADDRINET6_H__
#define WYSOCKADDRINET6_VERSION 33

#include "WyStr.h"
#include "Wy_SockAddr.h"
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>

// Class for IPv6 address (family=AF_INET6)
//
class WySockAddrINet6 : public Wy_SockAddr {
  public:
    typedef ::in_port_t PortType;

  private:
    class Wy_Assert_Check;
    static const PortType DefaultPort=0;          // net format
    static const uint32_t DefaultFlowInfo=0;      // net format
    static const uint32_t DefaultScopeId=0;       // net format
    static const socklen_t DefaultSockLen=sizeof(struct ::sockaddr_in6); 

    struct ::sockaddr_in6 _saddr;
    socklen_t _slen;

  public:
    class IPAddr {
      public:
        typedef struct ::in6_addr C_AddrType;
        static const char class_name[];
        WY_THROW_REPLY;

        IPAddr() WY__TSPC();
        IPAddr(const IPAddr&) WY__TSPC();
        IPAddr(const C_AddrType&) WY__TSPC();
        IPAddr(const char* cstr);
        IPAddr(const WyStr& str);
        IPAddr(IPAddr&,Wy::ByMove_t) WY__TSPC();
        bool is_default() const WY__TSPC();
        C_AddrType c_addr() const WY__TSPC();
        WyRet notation(WyStr& addr) const;
        void reset() WY__TSPC();
        void reset(const IPAddr&) WY__TSPC();
        void reset(const C_AddrType&) WY__TSPC();
        WyRet reset(const char* cstr);
        WyRet reset(const WyStr& str);
        void swap(IPAddr& addr) WY__TSPC();

        const IPAddr& operator =(const IPAddr&) WY__TSPC();
        const IPAddr& operator =(const C_AddrType&) WY__TSPC();

        bool operator ==(const IPAddr& rhs) const WY__TSPC();
        bool operator !=(const IPAddr& rhs) const WY__TSPC();
        bool operator ==(const C_AddrType& rhs) const WY__TSPC();
        bool operator !=(const C_AddrType& rhs) const WY__TSPC();
        bool operator ==(const WyStr& rhs) const;
        bool operator !=(const WyStr& rhs) const
             { return !this->operator==(rhs); };
        bool operator ==(const char* rhs) const;
        bool operator !=(const char* rhs) const
             { return !this->operator==(rhs); };

        const C_AddrType& wy_addr_ref() const WY__NOTHROW__
           { return _addr; };

      private:
        C_AddrType _addr;

        // [Ret] Ok
        //       Wym_EFAULT str inaccessible
        //       Wym_EADDRNOTAVAIL str not recognizable
        //       Wym_EAFNOSUPPORT  str invalid
        //
        WyRet _init(const char* str);
    };

    static const char class_name[];
    inline static int family() WY__TSPC()
              { return AF_INET6; };
    WY_THROW_REPLY;

    WySockAddrINet6() WY__TSPC();
    WySockAddrINet6(const WySockAddrINet6 &) WY__TSPC();
    WySockAddrINet6(const struct ::sockaddr_in6&) WY__TSPC();
    WySockAddrINet6(const IPAddr&, PortType) WY__TSPC();
    WySockAddrINet6(WySockAddrINet6& src, Wy::ByMove_t) WY__TSPC();

    bool is_default() const WY__TSPC();

    IPAddr ip() const WY__TSPC();
    PortType port() const WY__TSPC();
    uint32_t _flowinfo() const WY__TSPC();
    uint32_t _scope_id() const WY__TSPC();

    // [Override]
    virtual WyRet notation(WyStr& str) const;

    virtual void reset() WY__TSPC();
    virtual void reset(const struct ::sockaddr_in6& src) WY__TSPC();
    virtual void reset(const IPAddr& ip, PortType port) WY__TSPC();
    virtual void reset(const WySockAddrINet6 &src) WY__TSPC();

    virtual const WySockAddrINet6 &operator=(const WySockAddrINet6 &rhs) WY__TSPC();
    virtual const WySockAddrINet6 &operator=(const struct ::sockaddr_in6& rhs) WY__TSPC();

    virtual WyRet _swap(Wy_SockAddr& iaddr) WY__TSPC();
    virtual WySockAddrINet6* _alloc(WyRet&) const;

    bool operator==(const WySockAddrINet6 &rhs) const WY__TSPC();
    bool operator!=(const WySockAddrINet6 &rhs) const WY__TSPC()
              { return !operator==(rhs); };
    bool operator==(const struct ::sockaddr_in6& rhs) const WY__TSPC();
    bool operator!=(const struct ::sockaddr_in6& rhs) const WY__TSPC()
              { return !operator==(rhs); };

    const struct ::sockaddr_in6& wy_sockaddr_ref() const WY__TSPC()
           { return _saddr; };
  private:
    static const IPAddr::C_AddrType& DefaultIP;

    virtual struct ::sockaddr& c_sockaddr() WY__TSPC()
              { return (struct ::sockaddr&)(_saddr); };
    virtual const struct ::sockaddr& c_sockaddr() const WY__TSPC()
              { return (struct ::sockaddr&)(_saddr); };

    virtual size_t c_sockaddr_size() const WY__TSPC()
              { return sizeof(_saddr); };

    virtual  socklen_t& c_addrlen() WY__TSPC()
              { return _slen; };
    virtual const socklen_t& c_addrlen() const WY__TSPC()
              { return _slen; };

    // [Internal] Initialize object to the default state
    //
    //     family() = AF_INET6
    //         ip() = ::in6addr_any
    //       port() = 0
    //   _flowinfo()= 0
    //   _scope_id()= 0
    //
    void _init() WY__TSPC();

    // [Internal] Initialize object from ::sockaddr_in6
    //
    //     family() = AF_INET6
    //         ip() = from src.sin6_addr
    //       port() = from src.sin6_port
    //   _flowinfo()= from src.sin6_flowinfo
    //   _scope_id()= from src.sin6_scope_id
    //
    void _init(const struct ::sockaddr_in6& src) WY__TSPC();

    // [Internal] Initialize object to the state by the arguments
    //
    //     family() = AF_INET6
    //         ip() = ip.ip()
    //       port() = port
    //   _flowinfo()= 0
    //   _scope_id()= 0
    //
    void _init(const IPAddr& ip, PortType port) WY__TSPC();

    // [Internal] Reconstruct object by copy from src
    //
    //     family() = AF_INET6
    //         ip() = from src
    //       port() = ditto
    //   _flowinfo()= ditto
    //   _scope_id()= ditto
    //
    void _init(const WySockAddrINet6 &src) WY__TSPC();
};

inline bool operator ==(const WySockAddrINet6::IPAddr::C_AddrType& ip1, const WySockAddrINet6::IPAddr& ip2) WY__TSPC()
{ return ip2==WySockAddrINet6::IPAddr(ip1); };
inline bool operator !=(const WySockAddrINet6::IPAddr::C_AddrType& ip1, const WySockAddrINet6::IPAddr& ip2) WY__TSPC()
{ return ip2!=WySockAddrINet6::IPAddr(ip1); };
inline bool operator ==(const char* ip1, const WySockAddrINet6::IPAddr& ip2) 
{ return ip2==ip1; };
inline bool operator !=(const char* ip1, const WySockAddrINet6::IPAddr& ip2)
{ return ip2!=ip1; };
inline bool operator ==(const WyStr& ip1, const WySockAddrINet6::IPAddr& ip2)
{ return ip2==ip1; };
inline bool operator !=(const WyStr& ip1, const WySockAddrINet6::IPAddr& ip2)
{ return ip2!=ip1; };

inline bool operator==(const struct ::sockaddr_in6& a1, const WySockAddrINet6& a2) WY__TSPC()
              { return a2==a1; };
inline bool operator!=(const struct ::sockaddr_in6& a1, const WySockAddrINet6& a2) WY__TSPC()
              { return a2!=a1; };

// Define equivalence of IPAddr output type C_AddrType.
bool operator==(const WySockAddrINet6::IPAddr::C_AddrType& lhs, const WySockAddrINet6::IPAddr::C_AddrType& rhs) WY__TSPC();
bool operator!=(const WySockAddrINet6::IPAddr::C_AddrType& lhs, const WySockAddrINet6::IPAddr::C_AddrType& rhs) WY__TSPC();

namespace Wy {
 WyStr wrd(const WySockAddrINet6::IPAddr& addr);
 WyStr wrd(const WySockAddrINet6& addr);
};

#endif
