/* Copyright is licensed under GNU LGPL.                 by I.J.Wang 2003

  -- Note --
  Documents says differently(to me) with the data stored in sun_path[]
  Manual page [unix(7),Redhat v9.0]: sun_path contains the zero-terminated
             pathname.., and it indicates that the rest of bytes in sun_path
             is meaningful?
  Unix Network Programming [W. RICHARD STEVENS] v1,p374: The pathname stored
             in the sun_path array must be null terminated..
  Linux Application Development [Johnson,Troan] 2nd, p419: The sun_path does
             not need to be '\0' terminated.
  Other documents indicates either the trailing zero is counted or not.

  Implement takes what is meant in the header file un.h:
  # define SUN_LEN(ptr) ((size_t) (((struct sockaddr_un *) 0)->sun_path)\
	      + std::strlen ((ptr)->sun_path))
  i.e. zero-terminated, the trailing zero is not counted.

  -- Note --
   (7)unix (Fedora3) says "..Linux supports an abstract namespace which is
   independent of the file system.... sun_path contains the zero terminated 
   pathname of the socket in the file system. If sun_path starts with a zero
   byte it refers to the abstract namespace... The socket's address in this 
   namespace is given by the rest of the bytes in sun_path.."

  Implement Basic#1:
      If sun_path contains a system file pathname, the pathname has to 
      terminate with 0 (and may not contain 0), or SUN_LEN fails. 
      (at least for now). And,the trailing zero is not counted in _slen.
  Implement Basic#2:
      If sun_path contains the system supported(Fedora3) namespace, rare
      info. available. SUN_LEN may fail for the 'abstract-namespace'.
     
*/

#define WYLIB_SOURCE
#include "WySockAddrLocal.h"
#include <memory>
#include <typeinfo>
#include <cstring>

// Length of sockaddr_un _saddr containing no data in sun_path[]
//
static const socklen_t Wy__SockLenNoData= (socklen_t)(((struct ::sockaddr_un*)0)->sun_path);

// Length of default sockaddr_un
//
static const socklen_t Wy__DefaultSockLen=Wy__SockLenNoData;

// [Ret] Length of data in sun_path (from the effective size, _slen)
//
inline static socklen_t wy__sun_path_dlen(size_t slen) WY__NOTHROW__
{ return(slen-Wy__SockLenNoData); };

const char WySockAddrLocal::class_name[]="WySockAddrLocal";

#ifndef NDEBUG
class WySockAddrLocal::Wy__Assert_Check__ {
  public:
    Wy__Assert_Check__() {
        struct ::sockaddr_un addr;
        addr.sun_path[0]=0;
        if(Wy__SockLenNoData!=SUN_LEN(&addr)) {
          WY_TERMINATE("");
        }
        if(Wy__SockLenNoData!=Wy__DefaultSockLen) {
          WY_TERMINATE("");
        }
      };
} static const wy__do_not_use_me__;
#endif

WyRet WySockAddrLocal::_init_sockaddr(const char* pdata)
{
 // pdata points in _saddr ok
 if(pdata==NULL) {
   WY_RETURN(Wym_EFAULT);
 }
 const size_t plen( std::strlen(pdata) );
 if(plen>WySockAddrLocal::MaxAddrLen) {
   WY_RETURN(Wym_ENAMETOOLONG);
 }
 _saddr.sun_family=AF_LOCAL;

 std::memmove(_saddr.sun_path,pdata,plen+1);   // trailing zero is also copied
 _slen=Wy__SockLenNoData+plen;
 _padding=0;
 return(Ok);
};

WyRet WySockAddrLocal::_init_sockaddr(const char* pdata,size_t plen) WY__NOTHROW__
{
 // pdata points in _saddr ok
 if(pdata==NULL) {
   WY_RETURN(Wym_EFAULT);
 }
 if(plen>WySockAddrLocal::MaxAddrLen) {
   WY_RETURN(Wym_ENAMETOOLONG);
 }
 _saddr.sun_family=AF_LOCAL;
 std::memmove(_saddr.sun_path,pdata,plen);

 // note: trailing zero is appended for the SUN_LEN to work
 //
 _saddr.sun_path[plen]=0;
 _slen=Wy__SockLenNoData+plen;
 _padding=0;
 return(Ok);
};

WySockAddrLocal::WySockAddrLocal() WY__TSPC()
 : _slen(Wy__DefaultSockLen), _padding(0)
{
  _saddr.sun_family=AF_LOCAL;
  _saddr.sun_path[0]=0;
};

WySockAddrLocal::WySockAddrLocal(const WySockAddrLocal &src) WY__TSPC()
 : _slen(src._slen), _padding(0)
{
 std::memcpy(&_saddr,&src._saddr,src._slen+1);  // implement uses one more byte
};

WySockAddrLocal::WySockAddrLocal(const char* path)
{
 const WyRet r( this->_init_sockaddr(path) );
 if(r!=Ok) {
   WY_THROW( Reply(r) );
 }
};

WySockAddrLocal::WySockAddrLocal(const WyStr& path) WY__TSPC(WyRet)
{
 WyRet r= this->_init_sockaddr(path.begin(),path.size());
 if(r!=Ok) {
   WY_THROW( Reply(r) );
 }
};

WySockAddrLocal::WySockAddrLocal(const WyCSeg& path) WY__TSPC(WyRet)
{
 WyRet r= this->_init_sockaddr(path.begin(),path.size());
 if(r!=Ok) {
   WY_THROW( Reply(r) );
 }
};

WySockAddrLocal::WySockAddrLocal(WySockAddrLocal &src, Wy::ByMove_t) WY__TSPC()
{
 if(&this->_saddr==&src._saddr) {
   return;      // self move
 }
 std::memcpy(&_saddr,&src._saddr,src._slen+1);
 _slen=src._slen;
 _padding=src._padding;

#ifdef WY_DEBUG
 src._slen=0;   // for safety and error detection
#endif
};

bool WySockAddrLocal::is_default() const WY__TSPC()
{
 return _slen==Wy__DefaultSockLen; 
};

const char* WySockAddrLocal::addrname() const WY__TSPC()
{
 return _saddr.sun_path;
};

size_t WySockAddrLocal::addrsize() const WY__TSPC()
{
 return wy__sun_path_dlen(_slen);
};

WyRet WySockAddrLocal::notation(WyStr& str) const
{
 WY_RETURN( str.reset( WyCSeg(_saddr.sun_path,wy__sun_path_dlen(_slen))) );
};

void WySockAddrLocal::reset() WY__TSPC()
{
 _saddr.sun_family=AF_LOCAL;
 _saddr.sun_path[0]=0;
 _slen=Wy__DefaultSockLen; 
};

void WySockAddrLocal::reset(const WySockAddrLocal &src) WY__TSPC()
{
 if(&_saddr!=&src._saddr) {
   _slen=src._slen;
   std::memcpy(&_saddr,&src._saddr,src._slen+1);
 }
};

WyRet WySockAddrLocal::reset(const char* path)
{
 return this->_init_sockaddr(path);
};

WyRet WySockAddrLocal::reset(const WyStr& path) WY__TSPC()
{
 return this->_init_sockaddr(path.begin(),path.size());
};

WyRet WySockAddrLocal::reset(const WyCSeg& path) WY__TSPC()
{
 return this->_init_sockaddr(path.begin(),path.size());
};

WyRet WySockAddrLocal::_swap(Wy_SockAddr& laddr) WY__TSPC()
{
 if(WY__EQU_TYPEID(*this,laddr)==false) {
   WY_RETURN(Wym_EBADTYPE);
 }
 WySockAddrLocal* lptr( dynamic_cast<WySockAddrLocal*>(&laddr) );
 Wy__Base::memswp(&_saddr,&lptr->_saddr,sizeof(_saddr)+1);
 Wy__Base::valswp(_slen,lptr->_slen);
 return(Ok);
};

WySockAddrLocal* WySockAddrLocal::_alloc(WyRet& r) const
try {
 if(WY__EQU_TYPEID(*this,WySockAddrLocal)==false) {
   r=Wym_ENOSYS; WY_HERE(r);        // _alloc not overridden
   return(NULL);
 }
 WySockAddrLocal* p=new(std::nothrow) WySockAddrLocal();
 if(p==NULL) {
   r=Wym_ENOMEM; WY_HERE(r);
 } else {
   r=Ok;
 }
 return(p);
}
catch(const Reply& e) {
 r=e; WY_HERE(r);
 return(NULL);
};

const WySockAddrLocal& WySockAddrLocal::operator=(const WySockAddrLocal& rhs) WY__TSPC()
{
 if(&_saddr!=&rhs._saddr) {
   _slen=rhs._slen;
   std::memcpy(&_saddr,&rhs._saddr,rhs._slen+1);
 }
 return(*this);
};

const WySockAddrLocal& WySockAddrLocal::operator=(const char* rhs)
{
 const WyRet r( this->_init_sockaddr(rhs) );
 if(r!=Ok) {
   WY_THROW( Reply(r) );
 }
 return(*this);
};

const WySockAddrLocal& WySockAddrLocal::operator=(const WyStr& rhs) WY__TSPC(WyRet)
{
 const WyRet r( this->_init_sockaddr(rhs.begin(),rhs.size()) );
 if(r!=Ok) {
   WY_THROW( Reply(r) );
 }
 return(*this);
};

const WySockAddrLocal& WySockAddrLocal::operator=(const WyCSeg& rhs) WY__TSPC(WyRet)
{
 const WyRet r( this->_init_sockaddr(rhs.begin(), rhs.size()) );
 if(r!=Ok) {
   WY_THROW( Reply(r) );
 }
 return(*this);
};

bool WySockAddrLocal::operator==(const WySockAddrLocal &rhs) const WY__TSPC()
{
 if(_slen!=rhs._slen) {
   return(false);
 }
#ifdef WY_DEBUG
 if(_saddr.sun_family!=rhs._saddr.sun_family) {
   WY_TERMINATE("");
 }
 if((_slen<Wy__DefaultSockLen)||(_slen>sizeof(struct ::sockaddr_un))) {
   WY_TERMINATE("");
 }
#endif
 return std::memcmp(_saddr.sun_path,rhs._saddr.sun_path,wy__sun_path_dlen(_slen))==0;
};

bool WySockAddrLocal::operator!=(const WySockAddrLocal &rhs) const WY__TSPC()
{ return !operator==(rhs); };

// note: interface can remove the EFAULT document, for the rhs is invalid
//       but the definition is comparision with WySockAddr(rhs). Take the
//       latter because the former is stricter and should be harder to
//       maintain
//
bool WySockAddrLocal::operator==(const char* rhs) const
{
 if(rhs==NULL) {
   WY_THROW( Reply(Wym_EFAULT) );
 }
 const size_t rlen( std::strlen(rhs) );
 if(rlen!=wy__sun_path_dlen(_slen)) {
   if(rlen>WySockAddrLocal::MaxAddrLen) {
     WY_THROW( Reply(Wym_ENAMETOOLONG) );
   }
   return(false);
 }
 return(std::memcmp(_saddr.sun_path,rhs,rlen)==0);
};

bool WySockAddrLocal::operator!=(const char* rhs) const
{ return !operator==(rhs); };

bool WySockAddrLocal::operator==(const WyStr& rhs) const WY__TSPC(WyRet)
{
 const socklen_t tlen=wy__sun_path_dlen(_slen);
 if(tlen!=rhs.size()) {
   if(rhs.size()>WySockAddrLocal::MaxAddrLen) {
     WY_THROW( Reply(Wym_ENAMETOOLONG) );
   }
   return(false);
 }
 return(std::memcmp(_saddr.sun_path,rhs.begin(),tlen)==0);
};

bool WySockAddrLocal::operator!=(const WyStr& rhs) const WY__TSPC(WyRet)
{ return !operator==(rhs); };

bool WySockAddrLocal::operator==(const WyCSeg& rhs) const WY__TSPC(WyRet)
{
 const socklen_t tlen=wy__sun_path_dlen(_slen);
 if(tlen!=rhs.size()) {
   if(rhs.size()>WySockAddrLocal::MaxAddrLen) {
     WY_THROW( Reply(Wym_ENAMETOOLONG) );
   }
   return(false);
 }
 return(std::memcmp(_saddr.sun_path,rhs.begin(),tlen)==0);
};

bool WySockAddrLocal::operator!=(const WyCSeg& rhs) const WY__TSPC(WyRet)
{ return !operator==(rhs); };

WyStr Wy::wrd(const WySockAddrLocal& addr)
{
 return WyStr( WyCSeg(addr.addrname(),addr.addrsize()) );
};
