/* Copyright is licensed under GNU LGPL.                 by I.J.Wang 2003
*/

#define WYLIB_SOURCE
#include "WyTimeSpec.h"
#include <limits>
#include <ctime>
#include <cstdlib>
#include <new>

static const long long LL_MaxV= static_cast<long long>(
           std::numeric_limits<WyTimeSpec::SecondType>::max())*WY__CONST_GIGA
          +WY__CONST_GIGA-1;
static const long long LL_MinV= static_cast<long long>(
           std::numeric_limits<WyTimeSpec::SecondType>::min())*WY__CONST_GIGA
          -WY__CONST_GIGA+1;

// [Syn] Test overflow 
//
// [Ret] true: a+b will overflow
//      false: otherwise
//
template<typename T>
static bool wy__add_ovf(T a, T b)
{
 WY_ASSERT(std::numeric_limits<T>::is_integer);
 return (b>=0)? (a>std::numeric_limits<T>::max()-b)
              : (a<std::numeric_limits<T>::min()-b);
};

// [Syn] Test overflow 
//
// [Ret] true: a-b will overflow
//      false: otherwise
//
template<typename T>
static bool wy__sub_ovf(T a, T b)
{
 WY_ASSERT(std::numeric_limits<T>::is_integer);
 return (b>=0)? (a<std::numeric_limits<T>::min()+b)    // min-0
              : (a>std::numeric_limits<T>::max()+b);   // 0-min
};

const char WyTimeSpec::class_name[]="WyTimeSpec";

class WyTimeSpec::_Assertion {
  public:
    _Assertion() {

        // SecondType is time_t
        //
        if((std::numeric_limits<SecondType>::max()!=
            std::numeric_limits<time_t>::max())||
           (std::numeric_limits<SecondType>::min()!=
            std::numeric_limits<time_t>::min())) {
          WY_TERMINATE("");
        }

        // SecondType is representable by long
        //
        if((std::numeric_limits<SecondType>::max()>
            std::numeric_limits<long>::max())||
           (std::numeric_limits<SecondType>::min()<
            std::numeric_limits<long>::min())) {
          WY_TERMINATE("");
        }

        // NanoSecType is long
        //
        if((std::numeric_limits<NanoSecType>::max()!=
            std::numeric_limits<signed long>::max())||
           (std::numeric_limits<NanoSecType>::max()!=
            std::numeric_limits<signed long>::max())) {
          WY_TERMINATE("");
        }

        // Default value of WyTimeSpec is 0.0
        //
        WyTimeSpec tmp;
        if((tmp.second()!=0)||(tmp.nano()!=0)) {
          WY_TERMINATE("");
        }

        // type long long has enough digits (arithmetic ops use lot of long long type)
        if(2*std::numeric_limits<SecondType>::digits>std::numeric_limits<long long>::digits) {
          WY_TERMINATE("");
        }
    };
} static const wy__do_not_use_me__;

//
// [Ret] A=(lhs.sec*10^9+lhs.nsec)*(rhs.sec*10^9+rhs.nsec)
//       >LL_MaxV if over range (toward positive)
//       <LL_MinV if over range (toward negative)
//
inline long long WyTimeSpec::_mul(const WyTimeSpec& lhs, const WyTimeSpec& rhs) WY__NOTHROW__
{
 long long v1= static_cast<long long>(lhs._tspc.tv_sec)*rhs._tspc.tv_sec;
 if(v1>std::numeric_limits<SecondType>::max()) {
   return LL_MaxV+1;
 }
 if(v1<std::numeric_limits<SecondType>::min()) {
   return LL_MinV-1;
 }
 v1*=WY__CONST_GIGA;
 v1+= static_cast<long long>(lhs._tspc.tv_sec)*rhs._tspc.tv_nsec
     +static_cast<long long>(rhs._tspc.tv_sec)*lhs._tspc.tv_nsec
     +(static_cast<long long>(lhs._tspc.tv_nsec)*rhs._tspc.tv_nsec)/WY__CONST_GIGA;
 return(v1);
};

inline long long WyTimeSpec::_mul(const WyTimeSpec& lhs, SecondType rhs) WY__NOTHROW__
{
 long long v1= static_cast<long long>(lhs._tspc.tv_sec)*rhs;
 if(v1>std::numeric_limits<SecondType>::max()) {
   return LL_MaxV+1;
 }
 if(v1<std::numeric_limits<SecondType>::min()) {
   return LL_MinV-1;
 }
 return(v1*WY__CONST_GIGA + lhs._tspc.tv_nsec*static_cast<long long>(rhs));
};

WyRet WyTimeSpec::_normalize(long& sec, long& nsec) WY__TSPC()
{
 const std::ldiv_t fr=std::ldiv(nsec,Wy__Giga);
 const SecondType ts=sec+fr.quot;

 if(wy__add_ovf(sec,fr.quot)) {
   WY_RETURN(Wym_ERANGE);   // result not representable
 }

 // make sec/nano the same sign
 if(ts>0) {
   if(fr.rem<0) {
     sec=ts-1;
     nsec=fr.rem+Wy__Giga;
     return(Ok);
   }
   // FALL_THROUGH
 } else if(ts<0) {
   if(fr.rem>0) {
     sec=ts+1;
     nsec=fr.rem-Wy__Giga;
     return(Ok);
   }
   // FALL_THROUGH
 }
 sec=ts;
 nsec=fr.rem;
 return(Ok);
};

WyRet WyTimeSpec::_normalize(long long int& sec, long long int& nsec) WY__TSPC()
{
 const std::lldiv_t fr=std::lldiv(nsec,Wy__Giga);
 const long long ts=sec+fr.quot;

 if(wy__add_ovf(sec,fr.quot)) {
   WY_RETURN(Wym_ERANGE);   // result not representable
 }

 // make sec/nano the same sign
 if(ts>0) {
   if(fr.rem<0) {
     sec=ts-1;
     nsec=fr.rem+Wy__Giga;
     return(Ok);
   }
   // FALL_THROUGH
 } else if(ts<0) {
   if(fr.rem>0) {
     sec=ts+1;
     nsec=fr.rem-Wy__Giga;
     return(Ok);
   }
   // FALL_THROUGH
 }
 sec=ts;
 nsec=fr.rem;
 return(Ok);
};

// [Internal] Initialize object to contain the text indicated number
//            The text is blen characters pointed by buf in the form:
//            [+-]ddd.ddd[(eE)[+-]ddd], radix is 10
//
//            second()= converted from the text
//            nano()  = ditto
// [Ret] Ok
//       Wym_EFAULT
//       Wym_EBADMSG
//       Wym_ERANGE
//
inline WyRet WyTimeSpec::_init(const char* buf,size_t blen) WY__TSPC()
{
 WyTimeSpec tm;
 const WyRet r( Wy::_strnum(tm,WyCSeg(buf,blen)) );
 if(r!=Ok) {
   WY_RETURN(r);
 }
 _tspc=tm._tspc;
 return(Ok);
};

WyTimeSpec::WyTimeSpec() WY__TSPC()
 : _tspc() 
{
};

WyTimeSpec::WyTimeSpec(const WyTimeSpec& c) WY__TSPC()
 : _tspc(c._tspc)   // self-copy ok
{
};

WyTimeSpec::WyTimeSpec(SecondType sec, NanoSecType nsec) WY__TSPC(Reply)
{
 const WyRet r(_normalize(sec, nsec)); 
 if(r!=Ok) {
   WY_THROW( Reply(r) );
 }
 _tspc.tv_sec=sec;
 _tspc.tv_nsec=nsec;
};

WyTimeSpec::WyTimeSpec(SecondType sec) WY__TSPC()
{
 _tspc.tv_sec=sec;
 _tspc.tv_nsec=0;
};

WyTimeSpec::WyTimeSpec(const WyCSeg& cseg) WY__TSPC(Reply)
{
 const WyRet r( this->_init(cseg.begin(),cseg.size()) );
 if(r!=Ok) {
   WY_THROW( Reply(r) );
 }
};

WyTimeSpec::WyTimeSpec(WyTimeSpec& c, Wy::ByMove_t) WY__TSPC()
 : _tspc(c._tspc)   // self-copy ok
{
};

bool WyTimeSpec::is_default() const WY__TSPC()
{
 return (_tspc.tv_sec==SecondType())&&(_tspc.tv_nsec==NanoSecType()); 
};

WyTimeSpec::SecondType WyTimeSpec::second() const WY__TSPC()
{
 return _tspc.tv_sec; 
};

WyTimeSpec::NanoSecType WyTimeSpec::nano() const WY__TSPC()
{
 return _tspc.tv_nsec; 
};

void WyTimeSpec::reset() WY__TSPC()
{
 _tspc.tv_sec=SecondType(); _tspc.tv_nsec=NanoSecType(); 
};

void WyTimeSpec::reset(const WyTimeSpec &t) WY__TSPC()
{
 _tspc=t._tspc; 
};

WyRet WyTimeSpec::reset(SecondType sec, NanoSecType nano) WY__TSPC()
{
 const WyRet r( _normalize(sec, nano) ); 
 if(r!=Ok) {
   WY_RETURN(r);
 }
 _tspc.tv_sec=sec;
 _tspc.tv_nsec=nano;
 return(Ok);
};

void WyTimeSpec::reset(SecondType sec) WY__TSPC()
{
 _tspc.tv_sec=sec;
 _tspc.tv_nsec=0;
};

WyRet WyTimeSpec::reset(const WyCSeg& cseg) WY__TSPC()
{
  WY_RETURN( this->_init(cseg.begin(),cseg.size()) );
};

void WyTimeSpec::swap(WyTimeSpec &tm) WY__TSPC()
{
 Wy__Base::valswp(_tspc,tm._tspc); 
};

WyTimeSpec& WyTimeSpec::operator =(const WyTimeSpec &rhs) WY__TSPC()
{
 _tspc=rhs._tspc; return(*this); 
};

WyTimeSpec& WyTimeSpec::operator =(SecondType sec) WY__TSPC()
{
 _tspc.tv_sec=sec; 
 _tspc.tv_nsec=0; 
 return(*this); 
};

WyTimeSpec& WyTimeSpec::operator =(const WyCSeg& cseg) WY__TSPC(Reply)
{
  const WyRet r( this->_init(cseg.begin(),cseg.size()) );
  if(r!=Ok) {
    WY_THROW( Reply(r) );
  }
  return(*this);
};

const WyTimeSpec WyTimeSpec::operator -() const WY__TSPC(Reply)
{
 if(wy__sub_ovf(std::time_t(0),this->_tspc.tv_sec)) {
   WY_THROW( Reply(Wym_ERANGE) );
 }
 return WyTimeSpec(-_tspc.tv_sec,-_tspc.tv_nsec);
};

WyTimeSpec& WyTimeSpec::operator ++() WY__TSPC(Reply)
{
 if(_tspc.tv_sec>=std::numeric_limits<SecondType>::max()) {
   WY_THROW( Reply(Wym_ERANGE) );
 } 
 ++_tspc.tv_sec;
 return(*this);
};

WyTimeSpec& WyTimeSpec::operator --() WY__TSPC(Reply)
{
 if(_tspc.tv_sec<=std::numeric_limits<SecondType>::min()) {
   WY_THROW( Reply(Wym_ERANGE) );
 } 
 --_tspc.tv_sec;
 return(*this);
};

const WyTimeSpec WyTimeSpec::operator ++(int) WY__TSPC(Reply)
{
 if(_tspc.tv_sec>=std::numeric_limits<SecondType>::max()) {
   WY_THROW( Reply(Wym_ERANGE) );
 } 
 const WyTimeSpec tm(*this);
 _tspc.tv_sec++;
 return(tm);
};

const WyTimeSpec WyTimeSpec::operator --(int) WY__TSPC(Reply)
{
 if(_tspc.tv_sec<=std::numeric_limits<SecondType>::min()) {
   WY_THROW( Reply(Wym_ERANGE) );
 }
 const WyTimeSpec tm(*this);
 _tspc.tv_sec--;
 return(tm);
};

WyTimeSpec & WyTimeSpec::operator +=(const WyTimeSpec &rhs) WY__TSPC(Reply)
{
 if(wy__add_ovf(this->_tspc.tv_sec,rhs._tspc.tv_sec)) {
   WY_THROW( Reply(Wym_ERANGE) );   // result not representable
 }

 WyTimeSpec v(*this);

 v._tspc.tv_sec+=rhs._tspc.tv_sec;
 v._tspc.tv_nsec+=rhs._tspc.tv_nsec;

 const WyRet r( _normalize(v._tspc.tv_sec,v._tspc.tv_nsec) );
 if(r!=Ok) {
   WY_THROW( Reply(r) );
 }
 this->_tspc=v._tspc;
 return(*this);
};

WyTimeSpec & WyTimeSpec::operator +=(SecondType rhs) WY__TSPC(Reply)
{
 if(wy__add_ovf(this->_tspc.tv_sec,rhs)) {
   WY_THROW( Reply(Wym_ERANGE) );   // result not representable
 }
 WyTimeSpec v(*this);
 v._tspc.tv_sec+=rhs;
 const WyRet r( _normalize(v._tspc.tv_sec,v._tspc.tv_nsec) );
 if(r!=Ok) {
   WY_THROW( Reply(r) );
 }
 this->_tspc=v._tspc;
 return(*this);
};

WyTimeSpec & WyTimeSpec::operator -=(const WyTimeSpec &rhs) WY__TSPC(Reply)
{
 if(wy__sub_ovf(this->_tspc.tv_sec,rhs._tspc.tv_sec)) {
   WY_THROW( Reply(Wym_ERANGE) );   // result not representable
 }

 WyTimeSpec v(*this);

 v._tspc.tv_sec-=rhs._tspc.tv_sec;
 v._tspc.tv_nsec-=rhs._tspc.tv_nsec;

 const WyRet r( _normalize(v._tspc.tv_sec,v._tspc.tv_nsec) );
 if(r!=Ok) {
   WY_THROW( Reply(r) );
 }
 this->_tspc=v._tspc;
 return(*this);
};

WyTimeSpec & WyTimeSpec::operator -=(SecondType rhs) WY__TSPC(Reply)
{
 if(wy__sub_ovf(this->_tspc.tv_sec,rhs)) {
   WY_THROW( Reply(Wym_ERANGE) );   // result not representable
 }
 WyTimeSpec v(*this);

 v._tspc.tv_sec-=rhs;

 const WyRet r( _normalize(v._tspc.tv_sec,v._tspc.tv_nsec) );
 if(r!=Ok) {
   WY_THROW( Reply(r) );
 }
 this->_tspc=v._tspc;
 return(*this);
};

WyTimeSpec& WyTimeSpec::operator *=(const WyTimeSpec &rhs) WY__TSPC(Reply)
{
 const long long v1=WyTimeSpec::_mul(*this,rhs);
 if((v1>LL_MaxV)||(v1<LL_MinV)) {
   WY_THROW( Reply(Wym_ERANGE) );
 }
 std::lldiv_t dt=std::lldiv(v1, WY__CONST_GIGA);
 _tspc.tv_sec=dt.quot;
 _tspc.tv_nsec=dt.rem;
 return(*this);
};

WyTimeSpec& WyTimeSpec::operator *=(SecondType rhs) WY__TSPC(Reply)
{
 const long long v1=WyTimeSpec::_mul(*this,rhs);
 if((v1>LL_MaxV)||(v1<LL_MinV)) {
   WY_THROW( Reply(Wym_ERANGE) );
 }
 std::lldiv_t dt=std::lldiv(v1, WY__CONST_GIGA);
 _tspc.tv_sec=dt.quot;
 _tspc.tv_nsec=dt.rem;
 return(*this);
};

WyTimeSpec& WyTimeSpec::operator /=(const WyTimeSpec &rhs) WY__TSPC(Reply)
{
 typedef signed long long LLType;
 LLType b=static_cast<LLType>(rhs._tspc.tv_sec)*WY__CONST_GIGA+rhs._tspc.tv_nsec;
 if(b==0) {
   WY_THROW( Reply(Wym_EDIVZERO) );
 }
 LLType a=static_cast<LLType>(this->_tspc.tv_sec)*WY__CONST_GIGA+this->_tspc.tv_nsec;
 bool is_neg=false;
 if(a<0) {
   a=-a;
   if(b>0) {
     is_neg=true;
   } else {
     b=-b;
   }
 } else if(b<0) {
   b=-b;
   if(a>0) {
     is_neg=true;
   } else {
     a=-a;
   }
 } else {
 };
 const std::lldiv_t dt=std::lldiv(a,b);
 LLType c=dt.quot;
 if(is_neg) {
  if(c>-static_cast<LLType>(std::numeric_limits<SecondType>::min()) ) {
    WY_THROW( Reply(Wym_ERANGE) );
  }
 } else {
  if(c>static_cast<LLType>(std::numeric_limits<SecondType>::max())) {
    WY_THROW( Reply(Wym_ERANGE) );
  }
 };
 LLType q=0;
 a=dt.rem;
 static const int FrBits=31;
 for(size_t cnt=FrBits; cnt; --cnt) {
   a<<=1;
   q<<=1;
   if(a>=b) {
     q|=1;
     a-=b;
   }
 }
 ++q;                // round-up bias
 if(is_neg) {
   q*=-WY__CONST_GIGA;
   q/=(1LL<<FrBits);
   if(q<=-WY__CONST_GIGA) {
     q+=WY__CONST_GIGA;
     ++c;
     if(c>-static_cast<LLType>(std::numeric_limits<SecondType>::min()) ) {
       WY_THROW( Reply(Wym_ERANGE) );
     }
   }
   this->_tspc.tv_sec=-c;
   this->_tspc.tv_nsec=q;
 } else {
   q*=WY__CONST_GIGA;
   q/=(1LL<<FrBits);
   if(q>=WY__CONST_GIGA) {
     q-=WY__CONST_GIGA;
     ++c;
     if(c>static_cast<LLType>(std::numeric_limits<SecondType>::max())) {
       WY_THROW( Reply(Wym_ERANGE) );
     }
   }
   this->_tspc.tv_sec=c;
   this->_tspc.tv_nsec=q;
 }
 return *this;
};

WyTimeSpec& WyTimeSpec::operator /=(SecondType rhs) WY__TSPC(Reply)
{
 typedef signed long long LLType;
 if(rhs==0) {
   WY_THROW( Reply(Wym_EDIVZERO) );
 }
 const LLType b=(static_cast<LLType>(this->_tspc.tv_sec)*WY__CONST_GIGA+this->_tspc.tv_nsec)/rhs;
 const std::lldiv_t dt=std::lldiv(b,WY__CONST_GIGA);
 this->_tspc.tv_sec=dt.quot;
 this->_tspc.tv_nsec=dt.rem;
 return *this;
};

WyTimeSpec WyTimeSpec::operator +(const WyTimeSpec &rhs) const WY__TSPC(Reply)
{
 if(wy__add_ovf(this->_tspc.tv_sec,rhs._tspc.tv_sec)) {
   WY_THROW( Reply(Wym_ERANGE) );   // result not representable
 }

 WyTimeSpec v(*this);

 v._tspc.tv_sec+=rhs._tspc.tv_sec;
 v._tspc.tv_nsec+=rhs._tspc.tv_nsec;

 const WyRet r( _normalize(v._tspc.tv_sec,v._tspc.tv_nsec) );
 if(r!=Ok) {
   WY_THROW( Reply(r) );
 }
 return(v);
};

WyTimeSpec WyTimeSpec::operator +(SecondType rhs) const WY__TSPC(Reply)
{
 WyTimeSpec v(*this);
 return(v+=rhs);
};

WyTimeSpec operator+(WyTimeSpec::SecondType lhs, const WyTimeSpec& rhs) WY__TSPC(WyTimeSpec::Reply)
{
 return rhs+lhs;
};

WyTimeSpec WyTimeSpec::operator -(const WyTimeSpec &rhs) const WY__TSPC(Reply)
{
 if(wy__sub_ovf(this->_tspc.tv_sec,rhs._tspc.tv_sec)) {
   WY_THROW( Reply(Wym_ERANGE) );   // result not representable
 }

 WyTimeSpec v(*this);

 v._tspc.tv_sec-=rhs._tspc.tv_sec;
 v._tspc.tv_nsec-=rhs._tspc.tv_nsec;

 const WyRet r( _normalize(v._tspc.tv_sec,v._tspc.tv_nsec) );
 if(r!=Ok) {
   WY_THROW( Reply(r) );
 }
 return(v);
};

WyTimeSpec WyTimeSpec::operator -(SecondType rhs) const WY__TSPC(Reply)
{
 WyTimeSpec v(*this);
 return(v-=rhs);
};

WyTimeSpec operator-(WyTimeSpec::SecondType lhs, const WyTimeSpec& rhs) WY__TSPC(WyTimeSpec::Reply)
{
 return WyTimeSpec(lhs,0)-rhs;
};

WyTimeSpec WyTimeSpec::operator *(const WyTimeSpec &rhs) const WY__TSPC(Reply)
{
 WyTimeSpec tmp(*this);
 return tmp*=rhs;
};

WyTimeSpec WyTimeSpec::operator *(SecondType rhs) const WY__TSPC(Reply)
{
 const long long v1=WyTimeSpec::_mul(*this,rhs);
 if((v1>LL_MaxV)||(v1<LL_MinV)) {
   WY_THROW( Reply(Wym_ERANGE) );
 }
 WyTimeSpec rv;
 const std::lldiv_t dt=std::lldiv(v1,WY__CONST_GIGA);
 rv._tspc.tv_sec=dt.quot;
 rv._tspc.tv_nsec=dt.rem;
 return(rv);
};

WyTimeSpec operator*(WyTimeSpec::SecondType lhs, const WyTimeSpec& rhs) WY__TSPC(WyTimeSpec::Reply)
{
 return rhs*lhs;
};

WyTimeSpec WyTimeSpec::operator /(const WyTimeSpec &rhs) const WY__TSPC(Reply)
{
 WyTimeSpec tmp(*this);
 return tmp/=rhs;
};

WyTimeSpec WyTimeSpec::operator /(SecondType rhs) const WY__TSPC(Reply)
{
 typedef signed long long LLType;
 if(rhs==0) {
   WY_THROW( Reply(Wym_EDIVZERO) );
 }
 const LLType b=(static_cast<LLType>(this->_tspc.tv_sec)*WY__CONST_GIGA+this->_tspc.tv_nsec)/rhs;
 const std::lldiv_t dt=std::lldiv(b,WY__CONST_GIGA);
 WyTimeSpec rv;
 rv._tspc.tv_sec=dt.quot;
 rv._tspc.tv_nsec=dt.rem;
 return rv;
};

WyTimeSpec operator/(WyTimeSpec::SecondType lhs, const WyTimeSpec& rhs) WY__TSPC(WyTimeSpec::Reply)
{
 return WyTimeSpec(lhs,0)/rhs;
};

bool WyTimeSpec::operator ==(const WyTimeSpec &rhs) const WY__TSPC()
{
 return (_tspc.tv_nsec==rhs._tspc.tv_nsec)&&(_tspc.tv_sec==rhs._tspc.tv_sec); 
};

bool WyTimeSpec::operator !=(const WyTimeSpec &rhs) const WY__TSPC() 
{
 return (_tspc.tv_nsec!=rhs._tspc.tv_nsec)||(_tspc.tv_sec!=rhs._tspc.tv_sec); 
};

bool WyTimeSpec::operator >(const WyTimeSpec &rhs) const WY__TSPC()
{
 if(_tspc.tv_sec>rhs._tspc.tv_sec) {
   return true;
 } else if(_tspc.tv_sec<rhs._tspc.tv_sec) {
   return false;
 } else {}

 if(_tspc.tv_nsec>rhs._tspc.tv_nsec) {
   return true;
 } else {
   return false;
 }
};

bool WyTimeSpec::operator >=(const WyTimeSpec &rhs) const WY__TSPC()
{
 if(_tspc.tv_sec>rhs._tspc.tv_sec) {
   return true;
 } else if(_tspc.tv_sec<rhs._tspc.tv_sec) {
   return false;
 } else {}

 if(_tspc.tv_nsec>=rhs._tspc.tv_nsec) {
   return true;
 } else {
   return false;
 }
};

bool WyTimeSpec::operator <(const WyTimeSpec &rhs) const WY__TSPC()
{
 if(_tspc.tv_sec<rhs._tspc.tv_sec) {
   return true;
 } else if(_tspc.tv_sec>rhs._tspc.tv_sec) {
   return false;
 } else {}

 if(_tspc.tv_nsec<rhs._tspc.tv_nsec) {
   return true;
 } else {
   return false;
 }
};

bool WyTimeSpec::operator <=(const WyTimeSpec &rhs) const WY__TSPC()
{
 if(_tspc.tv_sec<rhs._tspc.tv_sec) {
   return true;
 } else if(_tspc.tv_sec>rhs._tspc.tv_sec) {
   return false;
 } else {}

 if(_tspc.tv_nsec<=rhs._tspc.tv_nsec) {
   return true;
 } else {
   return false;
 }
};

bool WyTimeSpec::operator ==(SecondType rhs) const WY__TSPC()
{
 return (this->_tspc.tv_sec==rhs)&&(this->_tspc.tv_nsec==0);
};

bool WyTimeSpec::operator !=(SecondType rhs) const WY__TSPC()
{
 return (this->_tspc.tv_sec!=rhs)||(this->_tspc.tv_nsec!=0);
};

bool WyTimeSpec::operator >(SecondType rhs) const WY__TSPC()
{
 if(this->_tspc.tv_sec>rhs) {
   return true;
 } else if(this->_tspc.tv_sec==rhs) {
   return this->_tspc.tv_nsec>0;
 } else {
   return false;
 }
};

bool WyTimeSpec::operator >=(SecondType rhs) const WY__TSPC()
{
 if(this->_tspc.tv_sec>rhs) {
   return true;
 } else if(this->_tspc.tv_sec==rhs) {
   return this->_tspc.tv_nsec>=0;
 } else {
   return false;
 }
};

bool WyTimeSpec::operator <(SecondType rhs) const WY__TSPC()
{
 if(this->_tspc.tv_sec<rhs) {
   return true;
 } else if(this->_tspc.tv_sec==rhs) {
   return this->_tspc.tv_nsec<0;
 } else {
   return false;
 }
};

bool WyTimeSpec::operator <=(SecondType rhs) const WY__TSPC()
{
 if(this->_tspc.tv_sec<rhs) {
   return true;
 } else if(this->_tspc.tv_sec==rhs) {
   return this->_tspc.tv_nsec<=0;
 } else {
   return false;
 }
};

WyRet WyTimeSpec::add(SecondType sec, NanoSecType nsec) WY__TSPC()
{
 WyRet r( _normalize(sec,nsec) );
 if(r!=Ok) {
   WY_RETURN(r);
 }
 if(wy__add_ovf(_tspc.tv_sec,sec)) {
   WY_RETURN(Wym_ERANGE);
 }

 WyTimeSpec v(*this);

 v._tspc.tv_sec+=sec;
 v._tspc.tv_nsec+=nsec;

 if((r=_normalize(v._tspc.tv_sec,v._tspc.tv_nsec))!=Ok) {
   WY_RETURN(r);
 }
 _tspc=v._tspc;
 return(Ok);
};

WyRet WyTimeSpec::add(const WyTimeSpec& tm) WY__TSPC()
{
 return this->add(tm._tspc.tv_sec,tm._tspc.tv_nsec); 
};

WyRet WyTimeSpec::add(SecondType n) WY__TSPC()
{
 return this->add(n,0); 
};

WyRet WyTimeSpec::sub(SecondType sec, NanoSecType nsec) WY__TSPC()
{ 
 WyRet r( _normalize(sec,nsec) );
 if(r!=Ok) {
   WY_RETURN(r);
 }
 if(wy__sub_ovf(_tspc.tv_sec,sec)) {
   WY_RETURN(Wym_ERANGE);
 }

 WyTimeSpec v(*this);

 v._tspc.tv_sec-=sec;
 v._tspc.tv_nsec-=nsec;

 if((r=_normalize(v._tspc.tv_sec,v._tspc.tv_nsec))!=Ok) {
   WY_RETURN(r);
 }
 _tspc=v._tspc;
 return(Ok);
};

WyRet WyTimeSpec::sub(const WyTimeSpec& tm) WY__TSPC()
{
 return this->sub(tm._tspc.tv_sec,tm._tspc.tv_nsec); 
};

WyRet WyTimeSpec::sub(SecondType n) WY__TSPC()
{
 return this->sub(n,0); 
};

WyRet WyTimeSpec::mul(SecondType sec, NanoSecType nsec) WY__TSPC()
{
 const WyRet r=WyTimeSpec::_normalize(sec, nsec);
 if(r!=Ok) {
   WY_RETURN(r);
 }
 WyTimeSpec n;
 n._tspc.tv_sec=sec;
 n._tspc.tv_nsec=nsec;
 const long long v1=WyTimeSpec::_mul(*this, n);
 if((v1>LL_MaxV)||(v1<LL_MinV)) {
   WY_RETURN( Wym_ERANGE );
 }
 std::lldiv_t dt=std::lldiv(v1, WY__CONST_GIGA);
 _tspc.tv_sec=dt.quot;
 _tspc.tv_nsec=dt.rem;
 return(Ok);
};

WyRet WyTimeSpec::mul(SecondType n) WY__TSPC()
{
 long long ll((long long)_tspc.tv_sec*n);
 
 if( (ll>(long long)std::numeric_limits<SecondType>::max())
   ||(ll<(long long)std::numeric_limits<SecondType>::min())) {
   WY_RETURN( Wym_ERANGE );
 }
 std::lldiv_t dt=std::lldiv((long long)_tspc.tv_nsec*n,Wy__Giga);
 ll+=dt.quot;
 if( (ll>(long long)std::numeric_limits<SecondType>::max())
   ||(ll<(long long)std::numeric_limits<SecondType>::min())) {
   WY_RETURN( Wym_ERANGE );
 }
 
 _tspc.tv_sec=static_cast<SecondType>(ll);
 _tspc.tv_nsec=static_cast<NanoSecType>(dt.rem);
 return(Ok);
};

WyRet WyTimeSpec::mul(const WyTimeSpec& n) WY__TSPC()
{
 const long long v1=WyTimeSpec::_mul(*this,n);
 if((v1>LL_MaxV)||(v1<LL_MinV)) {
   WY_RETURN( Wym_ERANGE );
 }
 std::lldiv_t dt=std::lldiv(v1, WY__CONST_GIGA);
 _tspc.tv_sec=dt.quot;
 _tspc.tv_nsec=dt.rem;
 return(Ok);
};

WyRet WyTimeSpec::div(SecondType sec, NanoSecType nsec) WY__TSPC()
{
 const WyRet r=WyTimeSpec::_normalize(sec, nsec);
 if(r!=Ok) {
   WY_RETURN(r);
 }
 WyTimeSpec n;
 n._tspc.tv_sec=sec;
 n._tspc.tv_nsec=nsec;
 return this->div(n);
};

WyRet WyTimeSpec::div(const WyTimeSpec& rhs) WY__TSPC()
{
 // a similar copy is in operator /=(const WyTimeSpec&)
 typedef signed long long LLType;
 LLType b=static_cast<LLType>(rhs._tspc.tv_sec)*WY__CONST_GIGA+rhs._tspc.tv_nsec;
 if(b==0) {
   WY_RETURN( Wym_EDIVZERO );
 }
 LLType a=static_cast<LLType>(this->_tspc.tv_sec)*WY__CONST_GIGA+this->_tspc.tv_nsec;
 bool is_neg=false;
 if(a<0) {
   a=-a;
   if(b>0) {
     is_neg=true;
   } else {
     b=-b;
   }
 } else if(b<0) {
   b=-b;
   if(a>0) {
     is_neg=true;
   } else {
     a=-a;
   }
 } else {
 };
 const std::lldiv_t dt=std::lldiv(a,b);
 LLType c=dt.quot;
 if(is_neg) {
  if(c>-static_cast<LLType>(std::numeric_limits<SecondType>::min()) ) {
    WY_RETURN( Wym_ERANGE );
  }
 } else {
  if(c>static_cast<LLType>(std::numeric_limits<SecondType>::max())) {
    WY_RETURN( Wym_ERANGE );
  }
 };
 LLType q=0;
 a=dt.rem;
 static const int FrBits=31;
 for(size_t cnt=FrBits; cnt; --cnt) {
   a<<=1;
   q<<=1;
   if(a>=b) {
     q|=1;
     a-=b;
   }
 }
 ++q;                // round-up bias
 if(is_neg) {
   q*=-WY__CONST_GIGA;
   q/=(1LL<<FrBits);
   if(q<=-WY__CONST_GIGA) {
     q+=WY__CONST_GIGA;
     ++c;
     if(c>-static_cast<LLType>(std::numeric_limits<SecondType>::min()) ) {
       WY_RETURN( Wym_ERANGE );
     }
   }
   this->_tspc.tv_sec=-c;
   this->_tspc.tv_nsec=q;
 } else {
   q*=WY__CONST_GIGA;
   q/=(1LL<<FrBits);
   if(q>=WY__CONST_GIGA) {
     q-=WY__CONST_GIGA;
     ++c;
     if(c>static_cast<LLType>(std::numeric_limits<SecondType>::max())) {
       WY_RETURN( Wym_ERANGE );
     }
   }
   this->_tspc.tv_sec=c;
   this->_tspc.tv_nsec=q;
 }
 return Ok;
};

WyRet WyTimeSpec::div(SecondType n) WY__TSPC()
{
 if(n==0) {
   WY_RETURN( Wym_EDIVZERO );
 }
 const std::ldiv_t dt=std::ldiv(_tspc.tv_sec,n);

 _tspc.tv_sec=dt.quot;
 _tspc.tv_nsec= (((long long)dt.rem*(long long)Wy__Giga)+_tspc.tv_nsec)/n;
 return(Ok);
};

namespace Wy {

WyTimeSpec abs(const WyTimeSpec& src) WY__TSPC(WyTimeSpec::Reply)
{
 WyTimeSpec v; 
 if(src.second()>0) {
   v.wy_timespec().tv_sec=src.second();
   v.wy_timespec().tv_nsec=src.nano();
 } else if(src.second()<0) {
   if(wy__sub_ovf(std::time_t(0),src.second())) {
     WY_THROW( WyTimeSpec::Reply(Wym_ERANGE) );
   }
   v.wy_timespec().tv_sec=-src.second();
   v.wy_timespec().tv_nsec=-src.nano();
 } else {
   // v.tv_sec=0;  already 0
   if(src.nano()>=0) {
     v.wy_timespec().tv_nsec=src.nano();
   } else {
     v.wy_timespec().tv_nsec=-src.nano();
   }
 }
 return(v);
};

};  // end namespece Wy
