/* Copyright is licensed under GNU LGPL.                 by I.J.Wang 2006

  This file is header only.

  Note: Members are cancellation point iff the refered element members are
        cancellation points.

 .Conversion of some functions in consideration might need an array (poll,
  scattered i/o,getaddrinfo,reading dirents.., or might not). 
 .Wy_Array reduces dependency of this library on std::vector (and different rule)
  ,probably applications. Just go on and see.
 .for 2D/3D use, there are fast/smaller/convenient App. classes, Wy_Array need have
  features to convert underlying C-functions
 .t(size_t n) as capacity semantics is lower level: C++ code has to use std::malloc otherwise

 .Function members basically throw Reply to report error to avoid both testing
  returned and thrown object.
 .enum element type is limited in use
 .element type can't be member function pointer and existing C struct's (they don't 
  have the move ctor)
 .2D/3D dimension use are not particularly made convenient
 .Problem with static member class_name
 .function members for argument type WySeg<const T> are not documented
*/
#ifndef WY_ARRAY_HH__
#define WY_ARRAY_HH__
#define WY_ARRAY_VERSION 33

#include "WySeg.hh"
#include <new>
#include <climits>    // for INT_MAX

#ifdef WY_DEBUG
 // Precondition check
 #define WY__CHK_VALIDOBJ(x) \
     if( ((x)._b_bgn==NULL)||\
         ((x)._b_eob<(x)._b_end)||\
         ((x)._b_end<(x)._b_bgn) ) {\
         WY_TERMINATE("");\
       }
#else
 #define WY__CHK_VALIDOBJ(x) (void)0
#endif

const size_t Wy_Array_MaxSizeRawMemory=(INT_MAX/2)-1;  // maximal array buffer byte size
const size_t Wy_Array_MinimumCapacity=4;               // default capacity
const size_t Wy_Array_DefaultSize=0;                   // must be 0

template<typename T>
class Wy_Array {
    typedef T ElemType;
    typedef size_t SizeType;
    static const SizeType MaxSizeRawMemory;
    static const SizeType MinimumCapacity;
    static const SizeType DefaultSize;

    ElemType* _b_bgn;           // array begin pointer
    ElemType* _b_end;           // end pointer of constructed objects
    const ElemType* _b_eob;     // array end storage ptr (end of buffer)

    // [Syn] Allocate n bytes of raw memory block
    //
    // [Ret] ElemType pointer to n bytes of raw memory block
    //       NULL= Not enough memory
    //
    inline static ElemType* _alloc_raw(size_t n) WY__TSPC() {
             return reinterpret_cast<ElemType*>(::operator new(n,std::nothrow));
           };
 
    // [Syn] Free the memory block allocated by _alloc_raw(size_t)
    //
    inline static void _free_raw(ElemType* v) WY__TSPC() {
             ::operator delete(v);
           };

    // [Syn] Destruct elements in range [p1,p2).
    //       Destruction order is from the last element to the first.
    //
    // Postcondition: All elements of which the destructors have been called
    //
    // [Throw] from ~T()
    //
    //  Note: If destructing an element throws, the destruction proceeds for
    //        the rest and then resume the throw, or a second throw encountered and
    //        ??? std::terminate() is called ???.
    //
    inline static void _s_destruct(ElemType* p1, ElemType* p2)
           {
             for(; p1!=p2;) try {
               --p2;
               p2->~ElemType();
             }
             catch(...) {
               try {
                 while(p2!=p1) {
                   --p2;
                   p2->~ElemType();
                 }
               }
               catch(...) {
                 std::terminate();
               };
               throw;
             };
           };

    // [Syn] Destruct one element pointed by p
    //
    // [Throw] from ~T()
    //
    inline static void _s_destruct(ElemType* p)
           { p->~ElemType(); };

    // [Syn] Allocate an array of capacity cap, and copy construct the front
    //       part of the array from [buf, buf+n_elem)
    //
    //  Note: assert(cap>=n_elem) to call
    //
    // [Throw] from T(T)
    //
    // Note: ~T() throw leads to std::terminate()
    //
    // [Ret] ElemType pointer to an allocated array of which elements are copy
    //         constructed from buf
    //       NULL= Not enough memory
    //
    // [Depend] T(T), ~T()
    //
    static ElemType* _dup_array(const ElemType* buf, SizeType n_elem, SizeType cap)
       {
         // allocate another array of cap capacity
         ElemType *narr=_alloc_raw(cap*sizeof(ElemType));
         if(narr==NULL) {
           return(NULL);
         }

         // copy construct element in the raw array
         ElemType *des_ptr(narr);
         try {
           for( ; n_elem; ++buf,++des_ptr,--n_elem) {
             new(des_ptr) ElemType(*buf);
           };
         }
         catch(...) {
           try {
             _s_destruct(narr,des_ptr);
           }
           catch(...) {
             std::terminate();
           };
           _free_raw(narr);
           throw;
         };
         return(narr);
       };

     // [Syn] Move n objects pointed by src to raw memory pointed
     //       by dest
     //
     static void _objmove(ElemType* dest, ElemType* src, SizeType n) WY__NOTHROW__
       {
        if(dest<src) {
          for( ; n; ++src,++dest,--n) {
            new(dest) ElemType(*src,Wy::ByMove);
          }
        } else {
          if(n==0) {
            return;
          }
          --n;
          src+=n;
          dest+=n;
          for(;;) {
            new(dest) ElemType(*src,Wy::ByMove);
            if(n==0) {
              break;
            }
            --src;
            --dest;
            --n;
          };
        }
       };

  public:
    WY_THROW_REPLY;

    typedef ElemType value_type;
    typedef SizeType size_type;

    static  SizeType max_capacity() WY__TSPC()
       { return MaxSizeRawMemory/sizeof(ElemType); };
    static SizeType min_capacity() WY__TSPC()
       { return(MinimumCapacity); };

    Wy_Array() WY__TSPC(Reply)
      : _b_bgn(_alloc_raw(MinimumCapacity*sizeof(ElemType))),
        _b_end(_b_bgn),
        _b_eob(_b_bgn+MinimumCapacity)
    {
      if(_b_bgn==NULL) {
        WY_THROW( Reply(Wym_ENOMEM) );
      }
      WY__CHK_VALIDOBJ(*this);
    };

    // Note: Initialization problem: Wy_Array< Wy_Array<int> > tab(???)
    //       .or such use is not considered
    // Note: (conversion ctor) reset(SizeType) is not defined
    //       that ctor(SizeType n) means to reserve n capacity (== new T[n])
    //       has class responsibility explanation problem
    //
    //explicit Wy_Array(SizeType n_elem)

    Wy_Array(SizeType n_elem, const ElemType& elem)
      : _b_bgn(NULL), _b_end(NULL), _b_eob(NULL)
    {
      const SizeType cap( n_elem<MinimumCapacity? MinimumCapacity : n_elem );
      if(cap>this->max_capacity()) {
        WY_THROW( Reply(Wym_EFBIG) );
      }
      _b_end=_b_bgn=_alloc_raw(cap*sizeof(ElemType));
      if(_b_bgn==NULL) {
        WY_THROW( Reply(Wym_ENOMEM) );
      }
      _b_eob=_b_bgn+n_elem;
      try {
        for(; _b_end!=_b_eob; ++_b_end) {
          new(_b_end) ElemType(elem);
        }
      }
      catch(...) {
        try {
          _s_destruct(_b_bgn,_b_end);
        }
        catch(...) {
          std::terminate();
        };
        _free_raw(_b_bgn);
        _b_bgn=_b_end=NULL; _b_eob=NULL;
        throw;
      };
      _b_eob=_b_bgn+cap;
      WY__CHK_VALIDOBJ(*this);
    };

    Wy_Array(const WySeg<ElemType>& src)
      : _b_bgn(NULL), _b_end(NULL), _b_eob(NULL)
    {
      const SizeType src_size( src.size() );
      const SizeType cap( src_size<MinimumCapacity? MinimumCapacity : src_size);
      _b_end=_b_bgn=_alloc_raw(cap*sizeof(ElemType));
      if(_b_bgn==NULL) {
        WY_THROW( Reply(Wym_ENOMEM) );
      }
      _b_eob=_b_bgn+src_size;
      try {
        ElemType *ep( src.begin() );
        for(; _b_end!=_b_eob; ++_b_end, ++ep) {
          new(_b_end) ElemType(*ep);
        }
      }
      catch(...) {
        try {
          _s_destruct(_b_bgn,_b_end);
        }
        catch(...) {
          std::terminate();
        };
        _free_raw(_b_bgn);
        _b_bgn=_b_end=NULL; _b_eob=NULL;
        throw;
      };
      _b_eob=_b_bgn+cap;
      WY__CHK_VALIDOBJ(*this);
    };

    // note: This member might be considered an invisible overload. Manual does
    //       not list the versions of const element type, for looked inconsistent
    //
    Wy_Array(const WySeg<const ElemType>& src)
      : _b_bgn(NULL), _b_end(NULL), _b_eob(NULL)
    {
      const SizeType src_size( src.size() );
      const SizeType cap( src_size<MinimumCapacity? MinimumCapacity : src_size);
      _b_end=_b_bgn=_alloc_raw(cap*sizeof(ElemType));
      if(_b_bgn==NULL) {
        WY_THROW( Reply(Wym_ENOMEM) );
      }
      _b_eob=_b_bgn+src_size;
      try {
        const ElemType *ep( src.begin() );
        for(; _b_end!=_b_eob; ++_b_end, ++ep) {
          new(_b_end) ElemType(*ep);
        }
      }
      catch(...) {
        try {
          _s_destruct(_b_bgn,_b_end);
        }
        catch(...) {
          std::terminate();
        };
        _free_raw(_b_bgn);
        _b_bgn=_b_end=NULL; _b_eob=NULL;
        throw;
      };
      _b_eob=_b_bgn+cap;
      WY__CHK_VALIDOBJ(*this);
    };

    Wy_Array(const Wy_Array& src)
      : _b_bgn(NULL),  _b_end(NULL), _b_eob(NULL)
    {
      WY__CHK_VALIDOBJ(src);
      const SizeType cap( src.size()> Wy_Array::min_capacity()?
                          src.size(): Wy_Array::min_capacity() );
      _b_bgn=_dup_array(src._b_bgn,src.size(),cap);
      if(_b_bgn==NULL) {
        WY_THROW( Reply(Wym_ENOMEM) );
      }
      _b_eob=_b_bgn+cap;
      _b_end=_b_bgn+src.size();
      WY__CHK_VALIDOBJ(*this);
    };

    Wy_Array(Wy_Array& p, Wy::ByMove_t) WY__TSPC()
      : _b_bgn(p._b_bgn), _b_end(p._b_end), _b_eob(p._b_eob) // move semantics of p to this
       {
        /* Cleaning-up semantics of p can cause t.swap(t) fail
           Because impl. of swap can encounter 'invalid' p
         #ifdef WY_DEBUG
         // cleanup semantics of p
         p._b_bgn=p._b_end=NULL; p._b_eob=NULL;
         #endif
        */
         WY__CHK_VALIDOBJ(*this);
       };

    ~Wy_Array() 
    {
      WY__CHK_VALIDOBJ(*this);
      try {
        _s_destruct(_b_bgn,_b_end);
      }
      catch(...) {
        _free_raw(_b_bgn);
        throw;
      };
      _free_raw(_b_bgn);
      _b_bgn=_b_end=NULL; _b_eob=NULL;
    };

    bool is_default() const WY__TSPC()
      { WY__CHK_VALIDOBJ(*this);
        return _b_end==_b_bgn; 
      };

    SizeType size() const WY__TSPC()
      { WY__CHK_VALIDOBJ(*this);
        return _b_end-_b_bgn; 
      };

    // note: These members expose the address of the internal dynamic buffer
    //
    ElemType* begin() WY__TSPC()
      { WY__CHK_VALIDOBJ(*this);
        return _b_bgn; };
    const ElemType* begin() const WY__TSPC()
      { WY__CHK_VALIDOBJ(*this);
        return _b_bgn; };

    const ElemType* end() const WY__TSPC()
      { WY__CHK_VALIDOBJ(*this);
        return _b_end; };
    ElemType* end() WY__TSPC()
      { WY__CHK_VALIDOBJ(*this);
        return _b_end; };

    ElemType& front() WY__TSPC(Reply)
      {
        WY__CHK_VALIDOBJ(*this);
        if(_b_end==_b_bgn) {
          WY_THROW( Reply(Wym_ENOENT) );
        }
        return *_b_bgn; 
      };
    const ElemType& front() const WY__TSPC(Reply)
      {
        WY__CHK_VALIDOBJ(*this);
        if(_b_end==_b_bgn) {
          WY_THROW( Reply(Wym_ENOENT) );
        }
        return *_b_bgn; 
      };

    ElemType& back() WY__TSPC(Reply)
      {
        WY__CHK_VALIDOBJ(*this);
        if(_b_end==_b_bgn) {
          WY_THROW( Reply(Wym_ENOENT) );
        }
        return *(_b_end-1); 
      };
    const ElemType& back() const WY__TSPC(Reply)
      {
        WY__CHK_VALIDOBJ(*this);
        if(_b_end==_b_bgn) {
          WY_THROW( Reply(Wym_ENOENT) );
        }
        return *(_b_end-1); 
      };
    
    WySeg<ElemType> subseg() const WY__TSPC()
      {
        WY__CHK_VALIDOBJ(*this);
        return WySeg<ElemType>(_b_bgn,SizeType(_b_end-_b_bgn)); 
      };
    WySeg<ElemType> subseg(SizeType idx) const WY__TSPC(Reply)
      {
        WY__CHK_VALIDOBJ(*this);
        const SizeType csize(_b_end-_b_bgn);
        if(idx>csize) {
          WY_THROW( Reply(Wym_EINVAL) );
        }
        return WySeg<ElemType>(_b_bgn+idx,csize-idx);
      };
    WySeg<ElemType> subseg(SizeType idx, SizeType num) const WY__TSPC(Reply)
      {
        WY__CHK_VALIDOBJ(*this);
        SizeType rs(_b_end-_b_bgn);
        if(idx>rs) {
          WY_THROW( Reply(Wym_EINVAL) );
        }
        rs-=idx;
        return WySeg<ElemType>(_b_bgn+idx,rs>num? num : rs);
      };

    const ElemType& operator[](SizeType n) const WY__TSPC(Reply)
      {
        WY__CHK_VALIDOBJ(*this);
        if(n>=SizeType(_b_end-_b_bgn)) {
          WY_THROW( Reply(Wym_EINVAL) );
        }
        return _b_bgn[n];
      };
    ElemType& operator[](SizeType n) WY__TSPC(Reply)
      {
        WY__CHK_VALIDOBJ(*this);
        if(n>=SizeType(_b_end-_b_bgn)) {
          WY_THROW( Reply(Wym_EINVAL) );
        }
        return _b_bgn[n];
      };

    // Note: capacity is intact
    //
    void reset()
      {
        WY__CHK_VALIDOBJ(*this);
        try {
          _s_destruct(_b_bgn,_b_end);
        }
        catch(...) {
          _b_end=_b_bgn;
          throw;
        };
        _b_end=_b_bgn;
        WY__CHK_VALIDOBJ(*this);
      };

    void reset(const WySeg<const ElemType>& src)
      {
        WY__CHK_VALIDOBJ(*this);
        const SizeType src_size(src.size());
        if(src_size<=this->_capacity()) {
          if(src.is_overlap( WySeg<const ElemType>(_b_bgn,SizeType(_b_eob-_b_bgn)) )) {
            WY_THROW( Reply(Wym_ELOOP) );
          }

          // This algorithm reset *this first and reuse current array,
          // i.e. whatever copy constructor of ElemType throws will
          // be rethrown, *this is default.
          //
          try {
            _s_destruct(_b_bgn,_b_end);
          }
          catch(...) {        // postcondion, for stack unwind from ~T() 
            _b_end=_b_bgn;    // is default for this member
            throw;
          };
          _b_end=_b_bgn;

          // copy construct elements
          try {
            const ElemType *src_ptr(src.begin());
            for( ; src_ptr!=src.end(); ++src_ptr,++_b_end) {
              new(_b_end) ElemType(*src_ptr);
            };
          }
          catch(...) {
            try {
              _s_destruct(_b_bgn,_b_end);
            }
            catch(...) {
              std::terminate();
            };
            _b_end=_b_bgn;
            throw;
          };
        } else {

          // This algorithm duplicate src first
          //
          const SizeType cap( src_size>=Wy_Array::min_capacity()?
                              src_size : Wy_Array::min_capacity() );
          ElemType *ptmp=_dup_array(src.begin(),src_size,cap);
          if(ptmp==NULL) {
            WY_THROW( Reply(Wym_ENOMEM) );
          }
  
          // reset() *this
          try {
            _s_destruct(_b_bgn,_b_end);
          }
          catch(...) {        // postcondion, for stack unwind from ~T() 
            _b_end=_b_bgn;    // is default for this member
            try {
              _s_destruct(ptmp,ptmp+src_size);
            }
            catch(...) {
              std::terminate();
            };
            _free_raw(ptmp);
            throw;
          };
          _free_raw(_b_bgn);
          _b_bgn=ptmp;
          _b_end=ptmp+src_size;
          _b_eob=ptmp+cap;
        }
        WY__CHK_VALIDOBJ(*this);
      };

    void reset(const WySeg<ElemType>& src)
      {
        WY__CHK_VALIDOBJ(*this);
        const SizeType src_size(src.size());
        if(src_size<=this->_capacity()) {
          if(src.is_overlap( WySeg<ElemType>(_b_bgn,SizeType(_b_eob-_b_bgn)) )) {
            WY_THROW( Reply(Wym_ELOOP) );
          }

          // This algorithm reset *this first and reuse current array,
          // i.e. whatever copy constructor of ElemType throws will
          // be rethrown, *this is default.
          //
          try {
            _s_destruct(_b_bgn,_b_end);
          }
          catch(...) {        // postcondion, for stack unwind from ~T() 
            _b_end=_b_bgn;    // is default for this member
            throw;
          };
          _b_end=_b_bgn;

          // copy construct elements
          try {
            const ElemType *src_ptr(src.begin());
            for( ; src_ptr!=src.end(); ++src_ptr,++_b_end) {
              new(_b_end) ElemType(*src_ptr);
            };
          }
          catch(...) {
            try {
              _s_destruct(_b_bgn,_b_end);
            }
            catch(...) {
              std::terminate();
            };
            _b_end=_b_bgn;
            throw;
          };
        } else {

          // This algorithm duplicate src first
          //
          const SizeType cap( src_size>=Wy_Array::min_capacity()?
                              src_size : Wy_Array::min_capacity() );
          ElemType *ptmp=_dup_array(src.begin(),src_size,cap);
          if(ptmp==NULL) {
            WY_THROW( Reply(Wym_ENOMEM) );
          }
  
          // reset() *this
          try {
            _s_destruct(_b_bgn,_b_end);
          }
          catch(...) {        // postcondion, for stack unwind from ~T() 
            _b_end=_b_bgn;    // is default for this member
            try {
              _s_destruct(ptmp,ptmp+src_size);
            }
            catch(...) {
              std::terminate();
            };
            _free_raw(ptmp);
            throw;
          };
          _free_raw(_b_bgn);
          _b_bgn=ptmp;
          _b_end=ptmp+src_size;
          _b_eob=ptmp+cap;
        }
        WY__CHK_VALIDOBJ(*this);
      };

    // Note: capacity is set to needed minimum if reallocation occurred,
    //       otherwise capacity is intact
    //
    void reset(const Wy_Array& src)
      {
        WY__CHK_VALIDOBJ(*this);
        const SizeType src_size(src.size());
        if(src_size<=this->_capacity()) {
          if(_b_bgn==src._b_bgn) {
            return;        // v.reset(v)
          }

          // This algorithm reset *this first and reuse current array,
          // i.e. whatever copy constructor of ElemType throws will
          // be rethrown, *this is default.
          //
          try {
            _s_destruct(_b_bgn,_b_end);
          }
          catch(...) {        // postcondion, for stack unwind from ~T() 
            _b_end=_b_bgn;    // is default for this member
            throw;
          };
          _b_end=_b_bgn;

          // copy construct elements
          try {
            ElemType *src_ptr(src._b_bgn);
            for( ; src_ptr!=src._b_end; ++src_ptr,++_b_end) {
              new(_b_end) ElemType(*src_ptr);
            };
          }
          catch(...) {
            try {
              _s_destruct(_b_bgn,_b_end);
            }
            catch(...) {
              std::terminate();
            };
            _b_end=_b_bgn;
            throw;
          };
        } else {

          // This algorithm duplicate src first
          //
          const SizeType cap( src_size>=Wy_Array::min_capacity()?
                              src_size : Wy_Array::min_capacity() );
          ElemType *ptmp=_dup_array(src._b_bgn,src_size,cap);
          if(ptmp==NULL) {
            WY_THROW( Reply(Wym_ENOMEM) );
          }
  
          // reset() *this
          try {
            _s_destruct(_b_bgn,_b_end);
          }
          catch(...) {        // postcondion, for stack unwind from ~T() 
            _b_end=_b_bgn;    // is default for this member
            try {
              _s_destruct(ptmp,ptmp+src_size);
            }
            catch(...) {
              std::terminate();
            };
            _free_raw(ptmp);
            throw;
          };
          _free_raw(_b_bgn);
          _b_bgn=ptmp;
          _b_end=ptmp+src_size;
          _b_eob=ptmp+cap;
        }
        WY__CHK_VALIDOBJ(*this);
      };

    void reset(SizeType n_elem, const ElemType& elem)
    {
      WY__CHK_VALIDOBJ(*this);
      const SizeType cap( n_elem<MinimumCapacity? MinimumCapacity : n_elem );
      if(cap>this->max_capacity()) {
        WY_THROW( Reply(Wym_EFBIG) );
      }
      if(cap<=this->_capacity()) {
        const ElemType InsElem((elem));   // in case elem is in the buffer
        this->reset();
        try {
          WY_ASSERT(_b_bgn==_b_end);
          for(SizeType i=0; i<n_elem; ++i,++_b_end) {
            new(_b_end) ElemType(InsElem);
          }
        }
        catch(...) {
          try {
            _s_destruct(_b_bgn,_b_end);
          }
          catch(...) {
            std::terminate();
          };
          _b_end=_b_bgn;
          throw;
        };
      } else {
        ElemType *ptmp=_alloc_raw(cap*sizeof(ElemType));
        if(ptmp==NULL) {
          WY_THROW( Reply(Wym_ENOMEM) );
        }
        ElemType *p1(ptmp);
        ElemType *p2(ptmp+n_elem);
        try {
          for(; p1<p2; ++p1) {
            new(p1) ElemType(elem);
          }
        }
        catch(...) {
          try {
            _s_destruct(ptmp,p1);
          }
          catch(...) {
            std::terminate();
          };
          _free_raw(ptmp);
          throw;
        };
        p1= _b_bgn;
        p2= _b_end;
        _b_bgn=ptmp;
        _b_end=ptmp+n_elem;
        _b_eob=ptmp+cap;

        try {
          _s_destruct(p1,p2);
        }
        catch(...) {
          _free_raw(ptmp);
          throw;
        };
        _free_raw(p1);
      }
      WY__CHK_VALIDOBJ(*this);
    };

    // note: property relies on ElemType::operator=, otherwise definition is quite
    //       identical to reset(const Wy_Array&)
    //
    const Wy_Array& operator =(const Wy_Array& rhs)
      {
        WY__CHK_VALIDOBJ(*this);
        const SizeType src_size(rhs.size());
        if(src_size<=this->_capacity()) {
          if(_b_bgn==rhs._b_bgn) {
            return(*this);     // v.reset(v)
          }

          // This algorithm deliberately uses ElemType::operator= to require
          // it being defined, for object equivalence consideration.
          //
          // note: Whatever object thrown from ElemType is rethrown,
          //       *this is default. (for convenience)
          //
          const ElemType *src_ptr(rhs._b_bgn);
          ElemType *des_ptr(this->_b_bgn);
          try {
            if(this->size()<=src_size) {
              for( ; des_ptr!=_b_end; ++des_ptr,++src_ptr) {
                *des_ptr=*src_ptr;
              }
              for( ; src_ptr!=rhs._b_end; ++src_ptr,++_b_end) {
                new(_b_end) ElemType(*src_ptr);
              };
            } else {
              for( ; src_ptr!=rhs._b_end; ++des_ptr,++src_ptr) {
                *des_ptr=*src_ptr;
              }
              ElemType *ptmp(_b_end);
              _b_end=des_ptr;
              _s_destruct(_b_end,ptmp);
            }
          }
          catch(...) {
            try {
              _s_destruct(_b_bgn,_b_end);
            }
            catch(...) {
              std::terminate();
            };
            _b_end=_b_bgn;
            throw;
          };
        } else {

          // This algorithm duplicate src on different memory block
          //
          const SizeType cap( src_size>=Wy_Array::min_capacity()?
                              src_size : Wy_Array::min_capacity() );
          ElemType *ptmp=_dup_array(rhs._b_bgn,src_size,cap);
          if(ptmp==NULL) {
            WY_THROW( Reply(Wym_ENOMEM) );
          }
  
          // reset() *this
          try {
            _s_destruct(_b_bgn,_b_end);
          }
          catch(...) {        // postcondion, for stack unwind from ~T() 
            _b_end=_b_bgn;    // is default for this member
            try {
              _s_destruct(ptmp,ptmp+src_size);
            }
            catch(...) {
              std::terminate();
            };
            _free_raw(ptmp);
            throw;
          };
          _free_raw(_b_bgn);
          _b_bgn=ptmp;
          _b_end=ptmp+src_size;
          _b_eob=ptmp+cap;
        }
        WY__CHK_VALIDOBJ(*this);
        return(*this);
      };

    const Wy_Array& operator=(const WySeg<const ElemType>& src)
      {
        WY__CHK_VALIDOBJ(*this);
        const SizeType src_size(src.size());
        if(src_size<=this->_capacity()) {
          if(src.is_overlap( WySeg<const ElemType>(_b_bgn,SizeType(_b_eob-_b_bgn)) )) {
            WY_THROW( Reply(Wym_ELOOP) );
          }

          // This algorithm reset *this first and reuse current array,
          // i.e. whatever copy constructor of ElemType throws will
          // be rethrown, *this is default.
          //
          try {
            _s_destruct(_b_bgn,_b_end);
          }
          catch(...) {        // postcondion, for stack unwind from ~T() 
            _b_end=_b_bgn;    // is default for this member
            throw;
          };
          _b_end=_b_bgn;

          // copy construct elements
          try {
            const ElemType *src_ptr(src.begin());
            for( ; src_ptr!=src.end(); ++src_ptr,++_b_end) {
              new(_b_end) ElemType(*src_ptr);
            };
          }
          catch(...) {
            try {
              _s_destruct(_b_bgn,_b_end);
            }
            catch(...) {
              std::terminate();
            };
            _b_end=_b_bgn;
            throw;
          };
        } else {

          // This algorithm duplicate src first
          //
          const SizeType cap( src_size>=Wy_Array::min_capacity()?
                              src_size : Wy_Array::min_capacity() );
          ElemType *ptmp=_dup_array(src.begin(),src_size,cap);
          if(ptmp==NULL) {
            WY_THROW( Reply(Wym_ENOMEM) );
          }
  
          // reset() *this
          try {
            _s_destruct(_b_bgn,_b_end);
          }
          catch(...) {        // postcondion, for stack unwind from ~T() 
            _b_end=_b_bgn;    // is default for this member
            try {
              _s_destruct(ptmp,ptmp+src_size);
            }
            catch(...) {
              std::terminate();
            };
            _free_raw(ptmp);
            throw;
          };
          _free_raw(_b_bgn);
          _b_bgn=ptmp;
          _b_end=ptmp+src_size;
          _b_eob=ptmp+cap;
        }
        WY__CHK_VALIDOBJ(*this);
        return(*this);
      };

    const Wy_Array& operator=(const WySeg<ElemType>& src)
      {
        WY__CHK_VALIDOBJ(*this);
        const SizeType src_size(src.size());
        if(src_size<=this->_capacity()) {
          if(src.is_overlap( WySeg<ElemType>(_b_bgn,SizeType(_b_eob-_b_bgn)) )) {
            WY_THROW( Reply(Wym_ELOOP) );
          }

          // This algorithm reset *this first and reuse current array,
          // i.e. whatever copy constructor of ElemType throws will
          // be rethrown, *this is default.
          //
          try {
            _s_destruct(_b_bgn,_b_end);
          }
          catch(...) {        // postcondion, for stack unwind from ~T() 
            _b_end=_b_bgn;    // is default for this member
            throw;
          };
          _b_end=_b_bgn;

          // copy construct elements
          try {
            const ElemType *src_ptr(src.begin());
            for( ; src_ptr!=src.end(); ++src_ptr,++_b_end) {
              new(_b_end) ElemType(*src_ptr);
            };
          }
          catch(...) {
            try {
              _s_destruct(_b_bgn,_b_end);
            }
            catch(...) {
              std::terminate();
            };
            _b_end=_b_bgn;
            throw;
          };
        } else {

          // This algorithm duplicate src first
          //
          const SizeType cap( src_size>=Wy_Array::min_capacity()?
                              src_size : Wy_Array::min_capacity() );
          ElemType *ptmp=_dup_array(src.begin(),src_size,cap);
          if(ptmp==NULL) {
            WY_THROW( Reply(Wym_ENOMEM) );
          }
  
          // reset() *this
          try {
            _s_destruct(_b_bgn,_b_end);
          }
          catch(...) {        // postcondion, for stack unwind from ~T() 
            _b_end=_b_bgn;    // is default for this member
            try {
              _s_destruct(ptmp,ptmp+src_size);
            }
            catch(...) {
              std::terminate();
            };
            _free_raw(ptmp);
            throw;
          };
          _free_raw(_b_bgn);
          _b_bgn=ptmp;
          _b_end=ptmp+src_size;
          _b_eob=ptmp+cap;
        }
        WY__CHK_VALIDOBJ(*this);
        return(*this);
      };

    // std: no swap() function invalidates any references,pointers,.. ???
    //
    void swap(Wy_Array& v2) WY__TSPC()
      {
       WY__CHK_VALIDOBJ(*this);
       Wy__Base::TypeStore< Wy_Array > buf;
       new(&buf) Wy_Array(v2,Wy::ByMove);
       new(&v2) Wy_Array(*this,Wy::ByMove);
       new(this) Wy_Array(buf,Wy::ByMove);
       WY__CHK_VALIDOBJ(*this);
      };

    void push_back(const ElemType& elem)
      {
        WY__CHK_VALIDOBJ(*this);

        if(_b_end<_b_eob) {
          // elem in *this ok
          new(_b_end) ElemType(elem);
          ++_b_end;
        } else {
          // allocate another array twice as large
          const SizeType cap( (_b_eob-_b_bgn)<<1 );
          if(cap>this->max_capacity()) {
            WY_THROW( Reply(Wym_EFBIG) );
          }
          ElemType *ptmp=_alloc_raw(cap*sizeof(ElemType));
          if(ptmp==NULL) {
            WY_THROW( Reply(Wym_ENOMEM) );
          }
 
          ElemType *des_ptr(ptmp+this->size());
 
          // copy construct the argument object
          // (ok, if elem refers to the object inside *this)
          try {
            new(des_ptr) ElemType(elem);
          }
          catch(...) {
            _free_raw(ptmp);
            throw;
          };
 
          // move all elements of *this to ptmp
          ElemType *src_ptr(_b_end-1);    // _b_bgn!=_b_end
          --des_ptr;
          for(;;) {
            new(des_ptr) T(*src_ptr,Wy::ByMove);
            if(des_ptr==ptmp) {
              break;
            }
            --des_ptr;
            --src_ptr;
          }
          #ifdef WY_DEBUG
          if((des_ptr!=ptmp)||(src_ptr!=_b_bgn)) {
            WY_TERMINATE("");
          }
          #endif
 
          // update _b_bgn,_b_end,_b_eob to the new array(ptmp)
          des_ptr=_b_end;
          _b_end=ptmp+(this->size()+1);
          _b_bgn=ptmp;
          _b_eob=ptmp+cap;
 
          _free_raw(src_ptr);
        }
        WY__CHK_VALIDOBJ(*this);
      };

    void pop_back()
      {
        WY__CHK_VALIDOBJ(*this);
        if(_b_end==_b_bgn) {
          return;
        }
        --_b_end;
        _s_destruct(_b_end);
        WY__CHK_VALIDOBJ(*this);
      };

    void erase(SizeType idx, SizeType n)
      {
        WY__CHK_VALIDOBJ(*this);
        ElemType *p1(_b_bgn+idx);
        if(p1>=_b_end) {
          if(p1==_b_end) {
            WY__CHK_VALIDOBJ(*this);
            return;
          }
          WY_THROW( Reply(Wym_EINVAL) );
        }
        // assert( p1<_b_end );

        ElemType *p2=_b_end;
        if(SizeType(p2-p1)>=n) {
          if(n<=0) {
            WY__CHK_VALIDOBJ(*this);
            return;
          }
          p2=p1+n;
        }
        // destruct elements [p1,p2)
        try {
          _s_destruct(p1,p2);
        }
        catch(...) {
          for( ; p2!=_b_end; ++p1,++p2) {
            new(p1) T(*p2,Wy::ByMove);
          }
          _b_end=p1;
          throw;
        };

        for( ; p2!=_b_end; ++p1,++p2) {
          new(p1) T(*p2,Wy::ByMove);
        }
        _b_end=p1;
        WY__CHK_VALIDOBJ(*this);
      };

    // This member may be redundent from resize(len)
    // void erase(SizeType idx)

    void insert(SizeType idx, const ElemType& elem)
      {
       WY__CHK_VALIDOBJ(*this);
       const SizeType OrgSize( this->size() );
       if(idx>OrgSize) {
         WY_THROW( Reply(Wym_EINVAL) );
       }
       if(OrgSize<this->_capacity()) {
         Wy__Base::TypeStore<ElemType> tspace;  // might not work (see TypeStore, wy__base.h)
         new(&tspace) ElemType(elem);   // in case elem is in *this
         ElemType *p1(_b_bgn+idx);      // insert point
         ElemType *p2(_b_end);
         for(; p2>p1; ) {
           new(p2--) ElemType(*p2,Wy::ByMove);
         }
         new(p2) ElemType(tspace,Wy::ByMove); 
         ++_b_end; 
       } else {
         // allocate another array twice as large
         const SizeType cap( (_b_eob-_b_bgn)<<1 );
         if(cap>this->max_capacity()) {
           WY_THROW( Reply(Wym_EFBIG) );
         }
         ElemType *ptmp=_alloc_raw(cap*sizeof(ElemType));
         if(ptmp==NULL) {
           WY_THROW( Reply(Wym_ENOMEM) );
         }

         ElemType *p1(ptmp+idx);   // insert point
         try {
           new(p1) ElemType(elem);
         }
         catch(...) {
           _free_raw(ptmp);
           throw;
         };

         ElemType *des_ptr(ptmp);
         ElemType *p2(_b_bgn);

         // move all elements in _b_bgn to ptmp
         for(; des_ptr!=p1; ++p2,++des_ptr) {
           new(des_ptr) ElemType(*p2,Wy::ByMove);
         }
         ++des_ptr;
         for(; p2!=_b_end; ++p2,++des_ptr) {
           new(des_ptr) ElemType(*p2,Wy::ByMove);
         }

         // update _b_bgn,_b_end,_b_eob to the new array(ptmp)
         p2=_b_bgn;
         _b_bgn=ptmp;
         _b_end=des_ptr;
         _b_eob=ptmp+cap;

         _free_raw(p2);
       }
       WY__CHK_VALIDOBJ(*this);
      };

    void insert(SizeType idx, SizeType n_elem, const ElemType& elem)
      {
       WY__CHK_VALIDOBJ(*this);
       if(idx>this->size()) {
         WY_THROW( Reply(Wym_EINVAL) );
       }
       if(this->size()+n_elem<=this->_capacity()) {
         if(n_elem==0) {
           return;
         }
         const ElemType InsElem((elem));     // in case elem is in the buffer
         ElemType * const p0(_b_bgn+idx); // insert point
         ElemType *p1(_b_end);       // point to the last+1
         ElemType *p2(p1+n_elem);    // move dest.+1
         for(; p1>p0; ) {
           --p1;
           --p2;
           new(p2) ElemType(*p1,Wy::ByMove);
         }
         try {
           p2=p0;
           for(SizeType i=0; i<n_elem; ++i,++p2) {
             new(p2) ElemType(InsElem);
           }
         }
         catch(...) {
           p1=p0;
           try {
             _s_destruct(p1,p2);
           }
           catch(...) {
             std::terminate();
           };
           for(p2=p1+n_elem; p1<_b_end; ++p1) {
             new(p1) ElemType(*p2,Wy::ByMove);
           }
           throw;
         };
         _b_end+=n_elem;
       } else {
         // allocate another array twice as large
         if(n_elem>this->max_capacity()) {
           WY_THROW( Reply(Wym_EFBIG) );
         }
         const SizeType cap( (n_elem+this->size())<<1 );
         if(cap>this->max_capacity()) {
           WY_THROW( Reply(Wym_EFBIG) );
         }
         ElemType *ptmp=_alloc_raw(cap*sizeof(ElemType));
         if(ptmp==NULL) {
           WY_THROW( Reply(Wym_ENOMEM) );
         }

         ElemType *p1(ptmp+idx);   // insert point (in ptmp)
         ElemType *p2(p1);
         try {
           for(SizeType i=0; i<n_elem; ++i,++p2) {
             new(p2) ElemType(elem);
           }
         }
         catch(...) {
           try {
             _s_destruct(p1,p2);
           }
           catch(...) {
             std::terminate();
           };
           _free_raw(ptmp);
           throw;
         };

         ElemType *des_ptr(ptmp);
         p2=_b_bgn;

         // move all elements in _b_bgn to ptmp
         for(; des_ptr!=p1; ++p2,++des_ptr) {
           new(des_ptr) ElemType(*p2,Wy::ByMove);
         }
         des_ptr+=n_elem;
         for(; p2!=_b_end; ++p2,++des_ptr) {
           new(des_ptr) ElemType(*p2,Wy::ByMove);
         }

         // update _b_bgn,_b_end,_b_eob to the new array(ptmp)
         p2=_b_bgn;
         _b_bgn=ptmp;
         _b_end=des_ptr;
         _b_eob=ptmp+cap;

         _free_raw(p2);
       }
       WY__CHK_VALIDOBJ(*this);
      };

    void insert(SizeType idx, const WySeg<const ElemType>& src)
      {
       WY__CHK_VALIDOBJ(*this);
       if(idx>this->size()) {
         WY_THROW( Reply(Wym_EINVAL) );
       }
       if(src.is_overlap( WySeg<const ElemType>(_b_bgn,(ElemType*)_b_eob) )) {
         WY_THROW( Reply(Wym_ELOOP) );
       }
       const SizeType src_size( src.size() );
       if(this->size()+src_size<=this->_capacity()) {
         if(src_size==0) {
           return;
         }
         ElemType * const p0(_b_bgn+idx); // insert point
         ElemType *p1(_b_end);       // point to the last+1
         ElemType *p2(p1+src_size);  // move dest.+1
         for(; p1>p0; ) {
           --p1;
           --p2;
           new(p2) ElemType(*p1,Wy::ByMove);
         }
         try {
           p2=p0;
           for(SizeType i=0; i<src_size; ++i,++p2) {
             new(p2) ElemType(src[i]);
           }
         }
         catch(...) {
           p1=p0;
           try {
             _s_destruct(p1,p2);
           }
           catch(...) {
             std::terminate();
           };
           for(p2=p1+src_size; p1<_b_end; ++p1) {
             new(p1) ElemType(*p2,Wy::ByMove);
           }
           throw;
         };
         _b_end+=src_size;
       } else {
         // allocate another array twice as large
         if(src_size>this->max_capacity()) {
           WY_THROW( Reply(Wym_EFBIG) );
         }
         const SizeType cap( (src_size+this->size())<<1 );
         if(cap>this->max_capacity()) {
           WY_THROW( Reply(Wym_EFBIG) );
         }
         ElemType *ptmp=_alloc_raw(cap*sizeof(ElemType));
         if(ptmp==NULL) {
           WY_THROW( Reply(Wym_ENOMEM) );
         }

         ElemType *p1(ptmp+idx);   // insert point (in ptmp)
         ElemType *p2(p1);
         try {
           for(SizeType i=0; i<src_size; ++i,++p2) {
             new(p2) ElemType(src[i]);
           }
         }
         catch(...) {
           try {
             _s_destruct(p1,p2);
           }
           catch(...) {
             std::terminate();
           };
           _free_raw(ptmp);
           throw;
         };

         ElemType *des_ptr(ptmp);
         p2=_b_bgn;

         // move all elements in _b_bgn to ptmp
         for(; des_ptr!=p1; ++p2,++des_ptr) {
           new(des_ptr) ElemType(*p2,Wy::ByMove);
         }
         des_ptr+=src_size;
         for(; p2!=_b_end; ++p2,++des_ptr) {
           new(des_ptr) ElemType(*p2,Wy::ByMove);
         }

         // update _b_bgn,_b_end,_b_eob to the new array(ptmp)
         p2=_b_bgn;
         _b_bgn=ptmp;
         _b_end=des_ptr;
         _b_eob=ptmp+cap;

         _free_raw(p2);
       }
       WY__CHK_VALIDOBJ(*this);
      };

    void insert(SizeType idx, const WySeg<ElemType>& src)
      {
       WY__CHK_VALIDOBJ(*this);
       if(idx>this->size()) {
         WY_THROW( Reply(Wym_EINVAL) );
       }
       if(src.is_overlap( WySeg<ElemType>(_b_bgn,(ElemType*)_b_eob) )) {
         WY_THROW( Reply(Wym_ELOOP) );
       }
       const SizeType src_size( src.size() );
       if(this->size()+src_size<=this->_capacity()) {
         if(src_size==0) {
           return;
         }
         ElemType * const p0(_b_bgn+idx); // insert point
         ElemType *p1(_b_end);       // point to the last+1
         ElemType *p2(p1+src_size);  // move dest.+1
         for(; p1>p0; ) {
           --p1;
           --p2;
           new(p2) ElemType(*p1,Wy::ByMove);
         }
         try {
           p2=p0;
           for(SizeType i=0; i<src_size; ++i,++p2) {
             new(p2) ElemType(src[i]);
           }
         }
         catch(...) {
           p1=p0;
           try {
             _s_destruct(p1,p2);
           }
           catch(...) {
             std::terminate();
           };
           for(p2=p1+src_size; p1<_b_end; ++p1) {
             new(p1) ElemType(*p2,Wy::ByMove);
           }
           throw;
         };
         _b_end+=src_size;
       } else {
         // allocate another array twice as large
         if(src_size>this->max_capacity()) {
           WY_THROW( Reply(Wym_EFBIG) );
         }
         const SizeType cap( (src_size+this->size())<<1 );
         if(cap>this->max_capacity()) {
           WY_THROW( Reply(Wym_EFBIG) );
         }
         ElemType *ptmp=_alloc_raw(cap*sizeof(ElemType));
         if(ptmp==NULL) {
           WY_THROW( Reply(Wym_ENOMEM) );
         }

         ElemType *p1(ptmp+idx);   // insert point (in ptmp)
         ElemType *p2(p1);
         try {
           for(SizeType i=0; i<src_size; ++i,++p2) {
             new(p2) ElemType(src[i]);
           }
         }
         catch(...) {
           try {
             _s_destruct(p1,p2);
           }
           catch(...) {
             std::terminate();
           };
           _free_raw(ptmp);
           throw;
         };

         ElemType *des_ptr(ptmp);
         p2=_b_bgn;

         // move all elements in _b_bgn to ptmp
         for(; des_ptr!=p1; ++p2,++des_ptr) {
           new(des_ptr) ElemType(*p2,Wy::ByMove);
         }
         des_ptr+=src_size;
         for(; p2!=_b_end; ++p2,++des_ptr) {
           new(des_ptr) ElemType(*p2,Wy::ByMove);
         }

         // update _b_bgn,_b_end,_b_eob to the new array(ptmp)
         p2=_b_bgn;
         _b_bgn=ptmp;
         _b_end=des_ptr;
         _b_eob=ptmp+cap;

         _free_raw(p2);
       }
       WY__CHK_VALIDOBJ(*this);
      };

    void insert(SizeType idx, const Wy_Array& src)
      { this->insert(idx, src.subseg()); };

    void replace(SizeType sidx, SizeType slen, SizeType n_elem, const ElemType& elem)
      {
       WY__CHK_VALIDOBJ(*this);
       const SizeType OrgSize( this->size() );
       if(sidx>OrgSize) {
         WY_THROW( Reply(Wym_EINVAL) );
       }

       // Clip slen to at most the existing number of elements [sidx,_b_end)
       SizeType rs= OrgSize-sidx;   
       if(slen>rs) {
         slen=rs;
       }

       rs=(OrgSize-slen)+n_elem;      // resultant size of this array
       if(rs>this->max_capacity()) {
         WY_THROW( Reply(Wym_EFBIG) );
       }
       if(rs<=this->_capacity()) {
         const ElemType InsElem(elem);  // in case elem in *this

         // Destruct elements indicated by (sidx:slen)
         ElemType *des_ptr=_b_bgn+sidx;      // insert point
         ElemType *src_ptr(des_ptr+slen);
         try {
           _s_destruct(des_ptr,src_ptr);
         }
         catch(...) {
           // Make *this default and rethrow
           try {
             _s_destruct(_b_bgn,des_ptr);
             _s_destruct(src_ptr,_b_end);
           }
           catch(...) {
             std::terminate();
           };
           _b_end=_b_bgn;
           throw;
         };

         #ifdef WY_DEBUG
           if(((des_ptr-_b_bgn)+n_elem+(_b_end-src_ptr))!=rs) {
             WY_TERMINATE("");
           }
         #endif
         _objmove(des_ptr+n_elem,src_ptr,_b_end-src_ptr);

         for(SizeType i=0 ; i<n_elem; ++des_ptr,++i) try {
           new(des_ptr) ElemType(InsElem);
         }
         catch(...) {
           for( ; i; --i) try {
             --des_ptr;
               _s_destruct(des_ptr);
             // Make *this default and rethrow
             _s_destruct(_b_bgn,des_ptr);
             _s_destruct(src_ptr,_b_end);
           }
           catch(...) {
             std::terminate();
           };
           _b_end=_b_bgn;
           throw;
         };
         _b_end=_b_bgn+rs;
       } else {
         // allocate memory for the new array
         const SizeType cap( rs );
         ElemType *ptmp=_alloc_raw(cap*sizeof(ElemType));
         if(ptmp==NULL) {
           WY_THROW( Reply(Wym_ENOMEM) );
         }

         ElemType *p1(ptmp+sidx);   // insert point
         for(SizeType i=0; i<n_elem; ++p1,++i) try {
           new(p1) ElemType(elem);
         }
         catch(...) {
            try {
              _s_destruct(ptmp+sidx,p1);
            }
            catch(...) {
              std::terminate();
            };
           _free_raw(ptmp);
           throw;
         };

         // Destruct elements indicated by (sidx:slen)
         ElemType *des_ptr=_b_bgn+sidx;
         ElemType *p2(des_ptr+slen);
         try {
           _s_destruct(des_ptr,p2);
         }
         catch(...) {
           // Make *this default and rethrow
            try {
             _s_destruct(_b_bgn,des_ptr);
             _s_destruct(p2,_b_end);
             _b_end=_b_bgn;

              p1=ptmp+sidx;
              _s_destruct(p1,p1+n_elem);
            }
            catch(...) {
              std::terminate();
            };
           _free_raw(ptmp);
           throw;
         };

         des_ptr=ptmp;
         p1=ptmp+sidx;
         p2=_b_bgn;
         // move all elements in _b_bgn to ptmp
         for(; des_ptr!=p1; ++p2,++des_ptr) {
           new(des_ptr) T(*p2,Wy::ByMove);
         }
         des_ptr+=n_elem;
         p2+=slen;
         for(; p2!=_b_end; ++p2,++des_ptr) {
           new(des_ptr) T(*p2,Wy::ByMove);
         }

         // update _b_bgn,_b_end,_b_eob to the new array(ptmp)
         p2=_b_bgn;
         _b_bgn=ptmp;
         _b_end=des_ptr;
         _b_eob=ptmp+cap;

         _free_raw(p2);
       };
       WY__CHK_VALIDOBJ(*this);
      };

    void resize(SizeType nsize)
      {
       WY__CHK_VALIDOBJ(*this);
       if(nsize<this->size()) {
         ElemType *p1(_b_bgn+nsize);
         ElemType *p2(_b_end);
         _b_end=p1;
         _s_destruct(p1,p2);
       } else if(nsize>this->size()) {
         // reserve enough capacity
         const WyRet r( this->_reserve(nsize) );
         if(r!=Ok) {
           WY_THROW( Reply(r) );
         }

         ElemType *p1(_b_end);
         ElemType *p2(_b_bgn+nsize);
         try {
           for(; p1<p2; ++p1) {
             new(p1) ElemType();
           }
         }
         catch(...) {
           try {
             _s_destruct(_b_end,p1);
           }
           catch(...) {
             std::terminate();
           };
           throw;
         };
         _b_end=p2;
       } else {
       };
       WY__CHK_VALIDOBJ(*this);
      };

    void resize(SizeType nsize, const ElemType& elem)
      {
       WY__CHK_VALIDOBJ(*this);
       if(nsize<this->size()) {
         ElemType *p1(_b_bgn+nsize);
         ElemType *p2(_b_end);
         _b_end=p1;
         _s_destruct(p1,p2);
       } else if(nsize>this->size()) {
         // reserve enough capacity
         const WyRet r( this->_reserve(nsize) );
         if(r!=Ok) {
           WY_THROW( Reply(r) );
         }

         ElemType *p1(_b_end);
         ElemType *p2(_b_bgn+nsize);
         try {
           for(; p1<p2; ++p1) {
             new(p1) ElemType(elem);
           }
         }
         catch(...) {
           try {
             _s_destruct(_b_end,p1);
           }
           catch(...) {
             std::terminate();
           };
           throw;
         };
         _b_end=p2;
       } else {
       };
       WY__CHK_VALIDOBJ(*this);
      };

    // [Low level member function]
    // Set array size to nsize
    // If nsize>capacity, behavior undefined
    // Valid object state:
    //    For all position i<t.size(), t[i] is a valid ElemType object
    //    For all position i>=t.size()&&x<t._capacity(), no valid ElemType object exists.
    inline void __setsize(SizeType nsize)
      {
        _b_end=_b_bgn+nsize;
      };

    // [Low level member function]
    // Destruct the element at position n
    // pre: idx<capacity
    // post: position idx contains no object (*this) is invalid
    //       position idx is a empty cell
    //
    // compare: 1. t.__destruct(1)
    //          2. t[1].~T()
    inline void __destruct(SizeType idx)
      {
        _b_bgn[idx].~ElemType();
      };

    // [Low level member function]
    // Move the element at position src to position dest
    // position dest_idx must be 'raw' (no object)
    //   post: *this may not be a 'valid' object
    // compare: 1. t.__move(1,2)
    //          2. new(&t[1]) T(t[2],Wy::ByMove)
    inline void __move(SizeType dest_idx, SizeType src_idx)
      {
        new(_b_bgn+dest_idx) ElemType(*(_b_bgn+src_idx),Wy::ByMove);
      };
    inline void __move(SizeType dest_idx, ElemType* src_ptr)
      {
        new(_b_bgn+dest_idx) ElemType(*src_ptr,Wy::ByMove);
      };

    bool operator ==(const Wy_Array& rhs) const
      {
        WY__CHK_VALIDOBJ(*this);
        if(this->size()!=rhs.size()) {
          return(false);
        }
        for(const ElemType *p1(_b_bgn), *p2(rhs._b_bgn); p1!=_b_end; ++p1,++p2) {
          if((p1->operator ==(*p2))==false) {
            return(false);
          }
        }
        return(true);
      };
    bool operator !=(const Wy_Array& rhs) const 
      {
        WY__CHK_VALIDOBJ(*this);
        if(this->size()!=rhs.size()) {
          return(true);
        }
        for(const ElemType *p1(_b_bgn), *p2(rhs._b_bgn); p1!=_b_end; ++p1,++p2) {
          if((p1->operator !=(*p2))==true) {
            return(true);
          }
        }
        return(false);
      };

    SizeType _capacity() const WY__TSPC()
      { WY__CHK_VALIDOBJ(*this);
        return _b_eob-_b_bgn; };

    WyRet _reserve(SizeType rcap) WY__TSPC()
      {
        WY__CHK_VALIDOBJ(*this);
        if(rcap<=this->_capacity()) {
          return(Ok);
        }
        if(rcap>this->max_capacity()) {
          WY_RETURN( Wym_EFBIG );
        }
        // allocate another array of rcap capacity
        ElemType *ptmp=_alloc_raw(rcap*sizeof(ElemType));
        if(ptmp==NULL) {
          WY_RETURN( Wym_ENOMEM );
        }

        ElemType *des_ptr(ptmp+this->size());

        // move all elements of *this to ptmp
        ElemType *src_ptr(_b_end);
        while(des_ptr!=ptmp) {
          --des_ptr;
          --src_ptr;
          new(des_ptr) T(*src_ptr,Wy::ByMove);
        }
        #ifdef WY_DEBUG
        if((des_ptr!=ptmp)||(src_ptr!=_b_bgn)) {
           WY_TERMINATE("");
        }
        #endif

        // update _b_bgn,_b_end,_b_eob to the new array(ptmp)
        des_ptr=_b_end;
        _b_end=ptmp+(this->size());
        _b_bgn=ptmp;
        _b_eob=ptmp+rcap;

        // free the original array
        _free_raw(src_ptr);
        WY__CHK_VALIDOBJ(*this);
        return(Ok);
      };
};

template<typename T>
const size_t Wy_Array<T>::MaxSizeRawMemory
            =Wy_Array_MaxSizeRawMemory;

template<typename T>
const size_t Wy_Array<T>::MinimumCapacity
            =Wy_Array_MinimumCapacity;

template<typename T>
const size_t Wy_Array<T>::DefaultSize
            =Wy_Array_DefaultSize;

#include "Wy__ArrayValue.hh"

#endif
