/* Copyright is licensed under GNU LGPL.                 by I.J.Wang 2007

  This file is header only.

  This file is for testing and evaluation.
*/
#ifndef WY_MF_HH__
#define WY_MF_HH__
#define WY_MF_VERSION 33

#include "WyRet.h"

/*
  [Internal] Class template Wy_MF??? makes class member function call to normal(global)
             function call (by containing both object and function pointers)

  Class template Wy_MF? is discardable, exists for enabling Wy_Array
  to use class function member as element type.
  
  Wy_MF0 is for no-argument version, Wy_MF1 is for one-argument version,..etc.
  Suffix '_const' is appended if the template parameter denotes a const member
  function.

  E.g.
      struct S {
          void f(int v) { Wy::cout << "S::f(" << Wy::wrd(v) << ")\n"; };
      } s,t;

      Wy_Array< Wy_MF1<S,void,int> > arr;
      arr.push_back( Wy_MF1<S,void,int>(s,&S::f) );
      arr.push_back( Wy_MF1<S,void,int>(t,&S::f) );

      for(size_t i=0; i<arr.size(); ++i) {
        arr[i](i);      // call s.f(0) and t.f(1)
      }
*/

// [Syn] Wy_MF0 makes non-const function member call (no argument)
//       to normal(global) function call
//
template<typename BaseClass, typename ResType>
class Wy_MF0 {
    typedef ResType(BaseClass::*BaseFuncType)();
    BaseClass* _optr;
    BaseFuncType _fptr;

  public:
    WY_THROW_REPLY;

    Wy_MF0() WY__TSPC() : _optr(NULL), _fptr(NULL) {};
    Wy_MF0(const Wy_MF0& src) WY__TSPC() : _optr(src._optr), _fptr(src._fptr) {};
    Wy_MF0(Wy_MF0& src, Wy::ByMove_t) WY__TSPC()
      : _optr(src._optr), _fptr(src._fptr) {};

    // [Syn] Construct object to call obj.f.
    //       FClass can be static cast to BaseClass
    //
    //       E.g. 
    //           S s;                   // S has a member function 'void f(int)'
    //           Wy_MF0<S,void,int> mf(s,&S::f);
    //
    template<typename FClass>
    Wy_MF0(FClass& obj, ResType(FClass::*f)()) WY__TSPC(Reply) : _optr(&obj), 
       _fptr( static_cast<BaseFuncType>(f) ) {
       if(_fptr==NULL) {
         WY_THROW( Reply(Wym_EFAULT) );
       }
    };

    bool is_default() const WY__TSPC() { return(_optr==NULL); };
    void reset() WY__TSPC() { _optr=NULL; _fptr=NULL; };
    void reset(const Wy_MF0& src) WY__TSPC() { _fptr=src._fptr; _optr=src._optr; };
    template<typename FClass>
    WyRet reset(FClass& obj, ResType(FClass::*f)()) WY__TSPC() {
       if(f==NULL) {
         WY_RETURN(Wym_EFAULT);
       }
       _optr=&obj;
       _fptr= static_cast<BaseFuncType>(f);
       return Ok;
    };

    // Invoke the member function call registered in construction
    //
    ResType operator()() const {
      if(_optr==NULL) {
        WY_THROW( Reply(Wym_EFAULT) );
      }
      return (_optr->*_fptr)();   // g++ 4.1 can deal with when ResType=void !
    };

     const Wy_MF0& operator =(const Wy_MF0& src) WY__TSPC()
      { _fptr=src._fptr; _optr=src._optr; return *this; };

    bool operator==(const Wy_MF0& rhs) const WY__TSPC()
      { return (this->_fptr==rhs._fptr)&&(this->_optr==rhs._optr); };
    bool operator!=(const Wy_MF0& rhs) const WY__TSPC()
      { return (this->_fptr!=rhs._fptr)||(this->_optr!=rhs._optr); };
};

// [Syn] Wy_MF0_const makes const function member call (no argument)
//       to normal(global) function call
//
template<typename BaseClass, typename ResType>
class Wy_MF0_const {
    typedef ResType(BaseClass::*BaseFuncType)() const;
    const BaseClass* _optr;
    BaseFuncType _fptr;

  public:
    WY_THROW_REPLY;

    Wy_MF0_const() WY__TSPC() : _optr(NULL), _fptr(NULL) {};
    Wy_MF0_const(const Wy_MF0_const& src) WY__TSPC() : _optr(src._optr), _fptr(src._fptr) {};
    Wy_MF0_const(Wy_MF0_const& src, Wy::ByMove_t) WY__TSPC()
      : _optr(src._optr), _fptr(src._fptr) {};

    template<typename FClass>
    Wy_MF0_const(const FClass& obj, ResType(FClass::*f)()const) WY__TSPC(Reply) 
      : _optr(&obj), _fptr( static_cast<BaseFuncType>(f) ) {
       if(_fptr==NULL) {
         WY_THROW( Reply(Wym_EFAULT) );
       }
    };

    bool is_default() const WY__TSPC() { return(_optr==NULL); };
    void reset() WY__TSPC() { _optr=NULL; _fptr=NULL; };
    void reset(const Wy_MF0_const& src) WY__TSPC() { _fptr=src._fptr; _optr=src._optr; };
    template<typename FClass>
    WyRet reset(const FClass& obj, ResType(FClass::*f)()const) WY__TSPC() {
       if(f==NULL) {
         WY_RETURN(Wym_EFAULT);
       }
       _optr=&obj;
       _fptr= static_cast<BaseFuncType>(f);
       return Ok;
    };

    ResType operator()() const {
      if(_optr==NULL) {
        WY_THROW( Reply(Wym_EFAULT) );
      }
      return (_optr->*_fptr)();   // g++ 4.1 can deal with when ResType=void !
    };

     const Wy_MF0_const& operator =(const Wy_MF0_const& src) WY__TSPC()
      { _fptr=src._fptr; _optr=src._optr; return *this; };

    bool operator==(const Wy_MF0_const& rhs) const WY__TSPC()
      { return (this->_fptr==rhs._fptr)&&(this->_optr==rhs._optr); };
    bool operator!=(const Wy_MF0_const& rhs) const WY__TSPC()
      { return (this->_fptr!=rhs._fptr)||(this->_optr!=rhs._optr); };
};

// [Syn] Wy_MF1 makes non-const function member call (one argument)
//       to normal(global) function call
//
template<typename BaseClass, typename ResType, typename Arg1>
class Wy_MF1 {
    typedef ResType(BaseClass::*BaseFuncType)(Arg1);
    BaseClass* _optr;
    BaseFuncType _fptr;

  public:
    WY_THROW_REPLY;

    Wy_MF1() WY__TSPC() : _optr(NULL), _fptr(NULL) {};
    Wy_MF1(const Wy_MF1& src) WY__TSPC() : _optr(src._optr), _fptr(src._fptr) {};
    Wy_MF1(Wy_MF1& src, Wy::ByMove_t) WY__TSPC()
      : _optr(src._optr), _fptr(src._fptr) {};

    template<typename FClass>
    Wy_MF1(FClass& obj, ResType(FClass::*f)(Arg1)) WY__TSPC(Reply) : _optr(&obj), 
       _fptr( static_cast<BaseFuncType>(f) ) {
       if(_fptr==NULL) {
         WY_THROW( Reply(Wym_EFAULT) );
       }
    };

    bool is_default() const WY__TSPC() { return(_optr==NULL); };
    void reset() WY__TSPC() { _optr=NULL; _fptr=NULL; };
    void reset(const Wy_MF1& src) WY__TSPC() { _fptr=src._fptr; _optr=src._optr; };
    template<typename FClass>
    WyRet reset(FClass& obj, ResType(FClass::*f)(Arg1)) WY__TSPC() {
       if(f==NULL) {
         WY_RETURN(Wym_EFAULT);
       }
       _optr=&obj;
       _fptr= static_cast<BaseFuncType>(f);
       return Ok;
    };

    ResType operator()(Arg1 arg1) const {
      if(_optr==NULL) {
        WY_THROW( Reply(Wym_EFAULT) );
      }
      return (_optr->*_fptr)(arg1);   // g++ 4.1 can deal with when ResType=void !
    };

     const Wy_MF1& operator =(const Wy_MF1& src) WY__TSPC()
      { _fptr=src._fptr; _optr=src._optr; return *this; };

    bool operator==(const Wy_MF1& rhs) const WY__TSPC()
      { return (this->_fptr==rhs._fptr)&&(this->_optr==rhs._optr); };
    bool operator!=(const Wy_MF1& rhs) const WY__TSPC()
      { return (this->_fptr!=rhs._fptr)||(this->_optr!=rhs._optr); };
};

// [Syn] Wy_MF1_const makes const function member call (one argument)
//       to normal(global) function call
//
template<typename BaseClass, typename ResType, typename Arg1>
class Wy_MF1_const {
    typedef ResType(BaseClass::*BaseFuncType)(Arg1) const;
    const BaseClass* _optr;
    BaseFuncType _fptr;

  public:
    WY_THROW_REPLY;

    Wy_MF1_const() WY__TSPC() : _optr(NULL), _fptr(NULL) {};
    Wy_MF1_const(const Wy_MF1_const& src) WY__TSPC() : _optr(src._optr), _fptr(src._fptr) {};
    Wy_MF1_const(Wy_MF1_const& src, Wy::ByMove_t) WY__TSPC()
      : _optr(src._optr), _fptr(src._fptr) {};

    template<typename FClass>
    Wy_MF1_const(const FClass& obj, ResType(FClass::*f)(Arg1)const) WY__TSPC(Reply) 
      : _optr(&obj), _fptr( static_cast<BaseFuncType>(f) ) {
       if(_fptr==NULL) {
         WY_THROW( Reply(Wym_EFAULT) );
       }
    };

    bool is_default() const WY__TSPC() { return(_optr==NULL); };
    void reset() WY__TSPC() { _fptr=NULL; _optr=NULL; };
    void reset(const Wy_MF1_const& src) WY__TSPC() { _fptr=src._fptr; _optr=src._optr; };
    template<typename FClass>
    WyRet reset(const FClass& obj, ResType(FClass::*f)(Arg1)const) WY__TSPC() {
       if(f==NULL) {
         WY_RETURN(Wym_EFAULT);
       }
       _optr=&obj;
       _fptr= static_cast<BaseFuncType>(f);
       return Ok;
    };

    ResType operator()(Arg1 arg1) const {
      if(_optr==NULL) {
        WY_THROW( Reply(Wym_EFAULT) );
      }
      return (_optr->*_fptr)(arg1);   // g++ 4.1 can deal with when ResType=void !
    };

    const Wy_MF1_const& operator =(const Wy_MF1_const& src) WY__TSPC()
      { _fptr=src._fptr; _optr=src._optr; return *this; };

    bool operator==(const Wy_MF1_const& rhs) const WY__TSPC()
      { return (this->_fptr==rhs._fptr)&&(this->_optr==rhs._optr); };
    bool operator!=(const Wy_MF1_const& rhs) const WY__TSPC()
      { return (this->_fptr!=rhs._fptr)||(this->_optr!=rhs._optr); }; 
};

#endif // end WY_MF_HH__
