/* Copyright is licensed under GNU LGPL.                 by I.J.Wang 2003
*/

#define WYLIB_SOURCE
#include "Wy_Thread.h"
#include <pthread.h>
#include <new>        // for bad_alloc

#ifdef WY_DEBUG
#include "WyByteFlow.h"
#endif

const char Wy_Thread::class_name[]="Wy_Thread";

//-----------------------------------------------------------------
static ::pthread_once_t wy__toltab_once_var=PTHREAD_ONCE_INIT;

// [Internal] Array item of Wy__TOLTab
//            Pair of Wy_Thread::_ThreadID and Wy_Thread object pointer for association
//            The association should be 1-1, or result unpredictable.
//
class Wy__TOLItem {
  public:
    // [Syn] construct default object
    //
    //  tol_ptr() = NULL
    //  tol_tid() = Wy_Thread::Wy__Default_ThreadID
    //
    inline Wy__TOLItem() WY__NOTHROW__ : _tid(Wy_Thread::Wy__Default_ThreadID),_optr(NULL) {};
    inline Wy__TOLItem(const Wy__TOLItem& src) WY__NOTHROW__ 
       : _tid(src._tid),_optr(src._optr) {};
    inline Wy__TOLItem(Wy_Thread::_ThreadID id, const Wy_Thread* optr) WY__NOTHROW__ 
       : _tid(id),_optr(optr) {};
    inline void reset() WY__NOTHROW__
         { _tid=Wy_Thread::Wy__Default_ThreadID; _optr=NULL; };
    inline void reset(const Wy__TOLItem& src) WY__NOTHROW__ 
         { _tid=src._tid; _optr=src._optr; }; // self-ops ok
    inline void reset(Wy_Thread::_ThreadID id, const Wy_Thread* optr) WY__NOTHROW__ 
         {  _tid=id; _optr=optr; };
    inline bool is_default() const WY__NOTHROW__
         { return _optr==NULL; };
    inline const Wy_Thread* tol_ptr() const WY__NOTHROW__
         { return _optr; };
    inline Wy_Thread::_ThreadID tol_tid() const WY__NOTHROW__
         { return _tid; };
  private:
    Wy_Thread::_ThreadID _tid;
    const Wy_Thread* _optr;
};

// [Internal] class of set of Wy__TOL 
//
//  Note: Exactly one instance(wy__toltab_ptr) of this class is declared
//
//  Note: All member access should acquire the lock first
//
class Wy__TOLTab {
    size_t _find_idx(Wy_Thread::_ThreadID tid) WY__NOTHROW__
         {
           size_t i=0;
           for( ; (i<_tabsize)&&(_tab[i].tol_tid()!=tid); ++i) { // value cmp
           };
           return(i);
         };
  public:
    Wy__TOLTab() WY__NOTHROW__
       : _tabsize(0) {};

    // [Syn] Get mutex of *this
    //
    // [Ret] reply
    //
    WyMutex& mutex() WY__NOTHROW__
         { return _mtx; };

    // [Syn] Get the number of entries
    //
    // [Ret] number of Wy__TOL entries *this is handling
    //
    size_t size() const WY__NOTHROW__
         { return _tabsize; };

    // [Syn] Add entry<Wy_Thread pointer,Wy_Thread::_ThreadID>
    //
    //       Note: 0. tid is used as the key
    //             1. tid must be equal to optr->thread_id()
    //                (this member does not dereference optr)
    //
    // [Ret] Ok          entry added
    //       Wym_EEXIST  tid already in *this
    //       Wym_FAULT   optr is NULL
    //       Wym_ENOSPC  maximal number of threads reached
    //
    WyRet add(const Wy_Thread* optr,Wy_Thread::_ThreadID tid)
         {
           if(optr==NULL) {
             return(Wym_EINVAL);
           }
           if(this->_find_idx(tid)<_tabsize) {
             WY_RETURN( Wym_EEXIST );
           }
           if(_tabsize>=Wy_Thread::Max_Threads) {
             WY_RETURN(Wym_ENOSPC);
           }
           _tab[_tabsize++].reset(tid,optr);
           return(Ok);
         };

    // [Syn] Remove the entry whose Wy_Thread::_ThreadID is identical to tid
    //       If no entry (tid,*) in *this, no effect.
    //
    bool remove(Wy_Thread::_ThreadID tid) WY__NOTHROW__
         {
           const size_t i( this->_find_idx(tid) );
           if(i>=_tabsize) {
             return(false);        // tid not in tab
           }
           _tabsize--;
           _tab[i].reset(_tab[_tabsize]);
           return(true);
         };

    // [Syn] Find entry whose Wy_Thread::_ThreadID is identical to tid
    //       and return the Wy_Thread* associates with it
    //
    // [Ret] The Wy_Thread* associates with tid
    //       NULL, if no item found
    //
    const Wy_Thread* find(Wy_Thread::_ThreadID tid) WY__NOTHROW__
         {
           const size_t i( this->_find_idx(tid) );
           if(i>=_tabsize) {
             return(NULL);        // tid not in tab
           }
           return _tab[i].tol_ptr();
         };
  private:
    Wy__TOLItem _tab[Wy_Thread::Max_Threads];
    size_t _tabsize;
    mutable WyMutex _mtx;

} static *wy__toltab_ptr=NULL;

extern "C" void wy__toltab_once_func(void)
{
 try {
   wy__toltab_ptr= new Wy__TOLTab();  // note: never free'd
 }
 catch(...) {
   WY_TERMINATE("");
 };
};

//------------------------------------------------------------------------------
#ifdef WY_DEBUG
 size_t wy_joined_cnt=0;
 size_t wy_created_cnt=0;
 size_t wy_act_max=0;
 WyMutex wy_act_mutex;
#endif

Wy_Thread::NoCancel::NoCancel() WY__TSPC()
{
 const int v(::pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &_oldstate));
 if(v!=0) {
   // manpage says only parameter error. which is not supposed to occur.
   // and this library does not allow this function to throw
   WY_TERMINATE("");
 }
};

Wy_Thread::NoCancel::~NoCancel() WY__TSPC()
{
 const int v(::pthread_setcancelstate(_oldstate, NULL));
 if(v!=0) {
   // manpage says only parameter error. which is not supposed to occur.
   // and this library does not allow this function to throw
   WY_TERMINATE("");
 }
};

bool Wy_Thread::NoCancel::was_cancelable() const WY__TSPC()
{
 return _oldstate==PTHREAD_CANCEL_ENABLE;
};

// tid of main (program startup thread)
//
const Wy_Thread::_ThreadID Wy_Thread::wy_main_thrd_id(::pthread_self());

static ::pthread_once_t wy__thread_once_var=PTHREAD_ONCE_INIT;

bool Wy_Thread::set_cancelable(bool en) WY__TSPC()
{
 int oldstate;
 const int v(::pthread_setcancelstate( en? PTHREAD_CANCEL_ENABLE: 
                                  PTHREAD_CANCEL_DISABLE , &oldstate));
 if(v!=0) {
   // manpage says only parameter error. which is not supposed to occur.
   // and this library does not allow this function to throw
   WY_TERMINATE("");
 }
 return (oldstate==PTHREAD_CANCEL_ENABLE);
};

Wy_Thread::CancelType Wy_Thread::set_canceltype(CancelType ct) WY__TSPC()
{
 int oldtype;
 const int v(::pthread_setcanceltype(ct,&oldtype));
 if(v!=0) {
   // manpage says only parameter error. which is not supposed to occur.
   // and this library does not allow this function to throw
   WY_TERMINATE("");
 }
 return(static_cast<CancelType>(oldtype));
};

extern "C" void wy__thread_once_func(void) WY__TSPC()
{
 // intended for feature test code, nothing yet.
};

inline void Wy_Thread::_non_cancelable() WY__TSPC()
{
 if(::pthread_setcancelstate(PTHREAD_CANCEL_DISABLE,NULL)!=0) {
   WY_TERMINATE("");
 }
};

void Wy_Thread::_cleanup_thread(WyLock& aa)
{
  WyRet r;
  // Cancel the possibly active thread, and join the thread
  //
  if(_thrd_stat==Active) {
    // Note: Wait forever if the threead is not cancelable.
    //
    do {
      int v;
      while((v=::pthread_cancel(_thrd_id))!=0) {
        // ref. note5
        if(v==EAGAIN) {
          Wy__Base::yield();  // probably ok
          continue;
        }
        // ESRCH or others not documented
        // Wy__Base::put_abnmsg( WyReply(v)->repstr() ,__FILE__,__LINE__);
        Wy__Base::terminate();
      };
    
      // cancel_point
      if((r=_thrd_stat_cond.wait(aa))!=Ok) {
        WY_THROW(r); // Wym_EINVAL
      }
    } while(_thrd_stat==Active);

  } else if(_thrd_stat==Stopped) {
  } else {
    // Null (no thread)
  };
};

// [Spc] Wy_Thread relies on pthread. Default thread attributes are:
//
//   detatchstate= PTHREAD_CREATE_DETATCHED
//   schedpolicy = SCHED_OTHER (regular, non-realtime)
//   schedparam  = 0
//   inheritsched= PTHREAD_EXPLICIT_SCHED
//   scope       = PTHREAD_SCOPE_SYSTEM
//
Wy_Thread::Wy_Thread()
try : _thrd_stat_cond(), _thrd_stat(Null),
      _thrd_id(Wy__Default_ThreadID), _exit_obj(),
      _in_destroy(false), _mtx()
{
 ::pthread_once(&wy__toltab_once_var, wy__toltab_once_func);
 ::pthread_once(&wy__thread_once_var, wy__thread_once_func);

 WyLock aa(_mtx);
 int v;

  if((v=::pthread_attr_init(&_thrd_attr))!=0) {
    //
    // manu says only ENOMEM, linux man just says errno.
    // 
    WY_THROW( Reply(v) );
  }
  if((v=::pthread_attr_setdetachstate(&_thrd_attr,PTHREAD_CREATE_DETACHED))!=0) {
    // man. indicates only EINVAL
    WY_THROW( WyRet(v) );
  }
}
catch(const Wy_Thread::Reply&) {
 throw;
}
catch(const WyMutex::Reply& e) {
  WY_THROW( Reply(e) );
}
catch(const WyLock::Reply& e) {
  WY_THROW( WyRet(e) );
}
catch(const WyRet& e) {
 WY_THROW( WyRet(e) );
};

Wy_Thread::~Wy_Thread() WY__TSPC()
try {
 //  Wy_Thread destructor is non-cancelable 
 //  Because the active thread need to access *this 
 //
 _non_cancelable();
 WyLock aa(_mtx);
 WyRet r;

 if(_in_destroy==true) {
   WY_TERMINATE("");    // destroy twice
 }
 _in_destroy=true;

 this->_cleanup_thread(aa);
 
 // Cleanup members that need cleanup
 //
 const int v=::pthread_attr_destroy(&_thrd_attr);
 if(v!=0) {
   // manu says only EINVAL could happen
   Wy__Base::put_abnmsg( WyReply(v)->repstr() ,__FILE__,__LINE__);
   Wy__Base::terminate();
 }
}
catch(const WyLock::Reply& e) {
 Wy__Base::put_abnmsg( e->repstr() ,__FILE__,__LINE__);
 Wy__Base::terminate();
}
catch(const WyRet& e) {
 Wy__Base::put_abnmsg( e->repstr() ,__FILE__,__LINE__);
 Wy__Base::terminate();
}
catch(const std::exception&) {
 WY_TERMINATE("");
}
catch(...) {
 WY_TERMINATE("");
};

bool Wy_Thread::is_default() const
try {
 
 // _thrd_stat==Null ensures the default state, including
 //   
 //   1. _thrd_id==Wy__Default_ThreadID
 //   2. _exit_obj.is_default()
 //   3. _in_destroy==false
 //   4. _thrd_attr is default
 //
 WyLock aa(_mtx);
 return(_thrd_stat==Null);
}
catch(const WyLock::Reply& e) {
#ifdef WY_DEBUG
 Wy::cerr << Wy::wrd(e) << '\n';
#endif
 WY_THROW( WyRet(e) );
}
catch(const WyRet& e) {
#ifdef WY_DEBUG
 Wy::cerr << Wy::wrd(e) << '\n';
#endif
 WY_THROW( WyRet(e) );
};

// [Non-cancellation Point]
Wy_Thread::_ThreadID Wy_Thread::thread_id() const
try {
 WyLock aa(_mtx);
 return _thrd_id; 
}
catch(const WyLock::Reply& e) {
#ifdef WY_DEBUG
 Wy::cerr << Wy::wrd(e) << '\n';
#endif
 WY_THROW( WyRet(e) );
}
catch(const WyRet& e) {
#ifdef WY_DEBUG
 Wy::cerr << Wy::wrd(e) << '\n';
#endif
 WY_THROW( WyRet(e) );
};

// [Non-cancellation Point]
Wy_Thread::ThreadState Wy_Thread::thread_state() const
try {
 WyLock aa(_mtx);
 return _thrd_stat;
}
catch(const Wy_Thread::Reply&) {
 throw;
}
catch(const WyLock::Reply& e) {
#ifdef WY_DEBUG
 Wy::cerr << Wy::wrd(e) << '\n';
#endif
 WY_THROW( WyRet(e) );
}
catch(const WyRet& e) {
#ifdef WY_DEBUG
 Wy::cerr << Wy::wrd(e) << '\n';
#endif
 WY_THROW( WyRet(e) );
};

// [Non-cancellation Point]
WyRet Wy_Thread::exitcode() const
try {
 WyLock aa(_mtx);
 return(_exit_obj);
}
catch(const WyLock::Reply& e) {
#ifdef WY_DEBUG
 Wy::cerr << Wy::wrd(e) << '\n';
#endif
 WY_THROW( WyRet(e) );
}
catch(const WyRet& e) {
#ifdef WY_DEBUG
 Wy::cerr << Wy::wrd(e) << '\n';
#endif
 WY_THROW( WyRet(e) );
};

// [Non-cancellation Point]
WyRet Wy_Thread::reset()
try {
  NoCancel noc;
  WyLock aa(_mtx);

  if(_in_destroy==true) {
    WY_THROW( WyRet() );
  }

  WyRet r;
  // If thread exists (active), cancel it.
  //
  if(_thrd_stat==Active) {
    if(::pthread_equal(_thrd_id,::pthread_self())) {
      // [Illegal case] thread (tmain) is not allowed to reset itself
      // rule of libwy says reset() always returns with default object
      //
      WY_THROW( WyRet() );
    }
    this->_cleanup_thread(aa);
  }

  int v;
  // Reset thread attr object (destoy and initialize again)
  //
  if((v=::pthread_attr_destroy(&_thrd_attr))!=0) {
    // manu says only EINVAL could happen
    WY_THROW( WyRet(v) );
  }
  if((v=::pthread_attr_init(&_thrd_attr))!=0) {
    //
    // manu says only ENOMEM, linux manpage just says errno.
    // 
    ::pthread_attr_destroy(&_thrd_attr); // make this member invalid (release)
    WY_THROW( WyRet(v) );
  }
  if((v=::pthread_attr_setdetachstate(&_thrd_attr,PTHREAD_CREATE_DETACHED))!=0) {
    // man. indicates only EINVAL
    WY_THROW( WyRet(v) );
  }

  this->_reset_avars();
  _thrd_stat_cond.broadcast();
  return(Ok);
}
catch(const WyRet& e) {
#ifdef WY_DEBUG
 Wy::cerr << Wy::wrd(e) << '\n';
#endif
 WY_THROW( WyRet(e) );
}
catch(const std::exception&) {
#ifdef WY_DEBUG
 Wy::cerr << "std::exception\n";
#endif
 WY_THROW( WyRet() );
};

WyRet Wy_Thread::tmain() 
{
 return(Ok); 
};

// [Non-cancellation Point]
void Wy_Thread::tmain_close() WY__TSPC()
{
 _non_cancelable();
 WyLock aa(_mtx);
 WyRet r;

 // Detect early (in case default _thrd_id is meaningful to pthread)
 //
 if(_thrd_id==Wy__Default_ThreadID) {
   return;
 }

 if(::pthread_equal(_thrd_id,::pthread_self())) {
   return; // thread can not cancel itself
 }
 
 this->_cleanup_thread(aa);
};

void* Wy_Thread::_tmain_finish(const WyRet& ecode) WY__TSPC()
try {
  // thread cancelability disabled,  no need to recover, as going to end.
  //
 _non_cancelable();
 WyLock aa_toltab( wy__toltab_ptr->mutex() );   // lock toltab first
 WyLock aa_this(_mtx);

 _exit_obj=ecode;

 // Set the thread state to Stopped and boradcast, ref. note8
 //  (destructor and reset() may wait on this)
 //
 _thrd_stat=Stopped;
 _thrd_stat_cond.broadcast();

 // Decrease the number of total active(running) threads
 //
 if(wy__toltab_ptr->remove(_thrd_id)==false) {
   WY_TERMINATE("");
 }

 #ifdef WY_DEBUG
 {
   WyLock atmp(wy_act_mutex);
   ++wy_joined_cnt;
 }
 #endif

 return(&_exit_obj);
}
catch(const WyLock::Reply& e) {
 Wy__Base::put_abnmsg( e->repstr() ,__FILE__,__LINE__);
 Wy__Base::terminate();
}
catch(const WyRet& e) {
 Wy__Base::put_abnmsg( e->repstr() ,__FILE__,__LINE__);
 Wy__Base::terminate();
}
catch(const std::bad_alloc&) {
  WY_TERMINATE(""); // so we can trace the line
}
catch(const std::exception&) {
  WY_TERMINATE(""); // so we can trace the line
}
catch(...) {
  WY_TERMINATE(""); // so we can trace the line
};

// [Cancel Point]
//
// [Internal] The created thread function. tmain() is invoked from inside.
//
// [Ret] Pointer to _exit_obj
//
void * Wy_Thread::wy_thread_begin_func(Wy_Thread *obj_ptr)
{
 // Syncronize with begin() by waiting for Active state
 // begin() may cancel this thread because it may fail in its latter
 // processing after the thread had been created
 //
 {
   WyLock aa(obj_ptr->_mtx);
   #ifndef NDEBUG
    if(obj_ptr->_thrd_id!=::pthread_self()) {  // assertion check
      WY_TERMINATE("");
    }
   #endif
 }

 try {
   ::pthread_testcancel();   // syncronizing purpose
   return obj_ptr->_tmain_finish( obj_ptr->tmain() );
 }
 catch(...) {
   // assert: stack unwound from obj_ptr->tmain()
   //  (_exit_obj is reset the same)
   //
   obj_ptr->_tmain_finish( obj_ptr->_exit_obj );
   throw;
 };
 WY_ASSERT("UNREACHABLE");
};

extern "C" inline void * wy__thread_begin_func_c__(void* obj_ptr)
{
  return Wy_Thread::wy_thread_begin_func(
                    reinterpret_cast<Wy_Thread *>(obj_ptr));
};

// [Non-cancellation Point]
//
WyRet Wy_Thread::begin()
try {
  WyRet r;

  WyLock aa_toltab( wy__toltab_ptr->mutex() );   // lock toltab first
  WyLock aa_this(_mtx);

  if(_in_destroy==true) {
    WY_THROW( WyRet() );
  }

  if(_thrd_stat!=Null) {
    WY_RETURN(Wym_EBADF);
  }
  if(wy__toltab_ptr->size()>=Wy_Thread::Max_Threads) {
    WY_RETURN(Wym_ENFILE);
  }

  // Preset return status
  // (_thrd_id will be set again owing from  pthread_create)
  _thrd_stat=Active;
  _exit_obj=Wym_ETHRDCAN; WY_HERE(_exit_obj);

  int v;
  try {
    // assert: _thrd_id==default
    if((v=::pthread_create(&_thrd_id,&_thrd_attr, 
                           wy__thread_begin_func_c__,
                           static_cast<void*>(this)))!=0) {
       // ENOMEM can return, which is not in the manpage
       //
       this->_reset_avars();
       _thrd_stat_cond.broadcast();
       WY_RETURN( WyRet(v) );
    }

    #ifdef WY_DEBUG
    {
      WyLock atmp(wy_act_mutex);
      ++wy_created_cnt;
      if(wy_created_cnt<wy_joined_cnt) {
        WY_TERMINATE("");
      }
      const size_t act=wy_created_cnt-wy_joined_cnt;
      if(act>wy_act_max) {
        wy_act_max=act;
      }
    }
    #endif

    // Check that _thrd_id is not the value identical to the default id
    // Refer to definition of Wy__Default_ThreadID.
    //
    if(_thrd_id==Wy__Default_ThreadID) {
      WY_THROW( WyRet() ); // assertion failure
    }
    WyRet r( wy__toltab_ptr->add(this,_thrd_id) );
    if(r!=Ok) {
      WY_THROW(r);    // no error with wy__toltab_ptr->add(..) is expected
    }
  }
  catch(...) {
    if(_thrd_id!=Wy__Default_ThreadID) {
      while((v=::pthread_cancel(_thrd_id))!=0) {
        // ref. note5
        if(v==EAGAIN) {
          continue;     // no yield();
        }
        WY_THROW( WyRet(v) );    // ESRCH (or others not documented)
      };
      if(v==0) {
        NoCancel noc;
        while(_thrd_stat==Active) {
          // cancel_point
          if((r=_thrd_stat_cond.wait(aa_this))!=Ok) {
            WY_THROW(r); // Wym_EINVAL
          }
        }
      }
    }

    this->_reset_avars();
    _thrd_stat_cond.broadcast();
    throw;
  };

  _thrd_stat_cond.broadcast();
  return(Ok);
}
catch(const WyLock::Reply& e) {
#ifdef WY_DEBUG
 Wy::cerr << Wy::wrd(e) << '\n';
#endif
  WY_THROW( WyRet(e) );
}
catch(const WyRet& e) {
#ifdef WY_DEBUG
 Wy::cerr << Wy::wrd(e) << '\n';
#endif
  WY_THROW( WyRet(e) );
};

// [Non-cancellation Point]
//
WyRet Wy_Thread::cancel()
try {
  WyLock aa(_mtx);

  if(_in_destroy==true) {
    WY_THROW( WyRet() );
  }

  // Detect earily (in case default _thrd_id is meaningful to pthread)
  //
  if(_thrd_id==Wy__Default_ThreadID) {
    WY_RETURN(Wym_ESRCH);
  }

  if(::pthread_equal(_thrd_id,::pthread_self())) {
    // [Illegal case] thread can not cancel itself
    WY_THROW( WyRet(Wym_EPERM) );  
  }
  
  int v;
  while((v=::pthread_cancel(_thrd_id))!=0) {
    // ref. note5
    if(v==EAGAIN) {
      Wy__Base::yield();
      continue;
    }
    if(v==ESRCH) {  // note9
      // may happen sometimes (chk_thread2): cancelled running thread
      // effects in the meanwhile *this is being destroyed
      // See also ~Wy_Thread()
      //
      WY_RETURN( WyRet(v) );
    }
    WY_THROW( WyRet(v) ); // others not documented
  }
  return(Ok);
}
catch(const WyLock::Reply& e) {
#ifdef WY_DEBUG
 Wy::cerr << Wy::wrd(e) << '\n';
#endif
 WY_THROW( WyRet(e) );
}
catch(const WyRet& e) {
#ifdef WY_DEBUG
 Wy::cerr << Wy::wrd(e) << '\n';
#endif
 WY_THROW( WyRet(e) );
};

// [Cancel Point]
//
WyRet Wy_Thread::wait_not(ThreadState state)
try {
 WyLock aa(_mtx);
 WyRet r;

  if(_in_destroy==true) {
    WY_THROW( WyRet() );
  }

 if(_thrd_id!=Wy__Default_ThreadID) {
   if(::pthread_equal(_thrd_id,::pthread_self())) {
     // [Illegal case] thread cannot wait for the state of itself
     WY_THROW( WyRet(Wym_EPERM) );
   }
 }

 switch(state) {
   case Active: {
               while(_thrd_stat==Active) {
                 // cancel_point
                 if((r=_thrd_stat_cond.wait(aa))!=Ok) {
                   WY_THROW( WyRet(r) ); // Wym_EINVAL
                 }
               } 
              }; break;
   case Null: {
               while(_thrd_stat==Null) {
                 // cancel_point
                 if((r=_thrd_stat_cond.wait(aa))!=Ok) {
                   WY_THROW( WyRet(r) ); // Wym_EINVAL
                 }
               }
              }; break;
   case Stopped: {
               while(_thrd_stat==Stopped) {
                 // cancel_point
                 if((r=_thrd_stat_cond.wait(aa))!=Ok) {
                   WY_THROW( WyRet(r) ); // Wym_EINVAL
                 }
               } 
              }; break;
   default:
        WY_THROW( WyRet() );
 };

 return(Ok);
}
catch(const WyLock::Reply& e) {
#ifdef WY_DEBUG
 Wy::cerr << Wy::wrd(e) << '\n';
#endif
 WY_THROW( WyRet(e) );
}
catch(const WyRet& e) {
#ifdef WY_DEBUG
 Wy::cerr << Wy::wrd(e) << '\n';
#endif
 WY_THROW( WyRet(e) );
};

// [Non-cancellation Point]
//
Wy_Thread::_ThreadID Wy_Thread::thread_self()
{
 //::pthread_once(&wy__toltab_once_var, wy__toltab_once_func);
 return ::pthread_self(); 
};

// [Cancel Point]
//
void Wy_Thread::cancel_point()
{
 //::pthread_once(&wy__toltab_once_var, wy__toltab_once_func);
 ::pthread_testcancel(); 
};

// [Non-cancellation Point]
//
WyRet Wy_Thread::yield()
{
 ::pthread_once(&wy__toltab_once_var, wy__toltab_once_func);
 const int v=Wy__Base::yield();
 if(v==0) {
   return(Ok);
 }
 WY_RETURN( WyRet(v) );
};

// [Non-cancellation Point]
//
// Note: speed remains an issue with this member
//
Wy_Thread* Wy_Thread::tbase_ptr()
try {
 ::pthread_once(&wy__toltab_once_var, wy__toltab_once_func);
 WyLock aa( wy__toltab_ptr->mutex() );
 const Wy_Thread* retv( wy__toltab_ptr->find(Wy_Thread::thread_self()) );
 return(const_cast<Wy_Thread*>(retv));
}
catch(const WyRet& e) {
#ifdef WY_DEBUG
 Wy::cerr << Wy::wrd(e) << '\n';
#endif
 WY_THROW( WyRet(e) );
};

// [Non-cancellation Point]
//
void Wy_Thread::exit(WyRet exit_ret)
try {
 ::pthread_once(&wy__toltab_once_var, wy__toltab_once_func);
 WyRet r;
 const _ThreadID ctid( Wy_Thread::thread_self() );
 WyLock aa_toltab(wy__toltab_ptr->mutex());
 const Wy_Thread* optr=wy__toltab_ptr->find(ctid);
 if(optr==NULL) {
   //
   // Note: If the calling thread is not created by Wy_Thread,
   //       ::pthread_exit(r.repcode()) is called.
   //
   ::pthread_exit((void*)(exit_ret->repcode()));
 }
 WyLock aa_optr(optr->_mtx);

 #ifndef NDEBUG
  if(optr->_thrd_stat!=Active) {
    // assertion: thread tmain() always sees Active state   
    WY_TERMINATE("");
  }
 #endif
 const_cast<Wy_Thread*>(optr)->_exit_obj.reset(exit_ret);
 ::pthread_exit( &(const_cast<Wy_Thread*>(optr)->_exit_obj) );
}
catch(const WyLock::Reply& e) {
#ifdef WY_DEBUG
 Wy::cerr << Wy::wrd(e) << '\n';
#endif
 WY_THROW( WyRet(e) );
}
catch(const WyRet& e) {
#ifdef WY_DEBUG
 Wy::cerr << Wy::wrd(e) << '\n';
#endif
 WY_THROW( WyRet(e) );
}
catch(const std::bad_alloc&) {
#ifdef WY_DEBUG
 Wy::cerr << "std::bad_alloc\n";
#endif
 WY_THROW( WyRet() );
}
catch(const std::exception&) {
#ifdef WY_DEBUG
 Wy::cerr << "std::exception\n";
#endif
 WY_THROW( WyRet() );
};

// [Non-cancellation Point]
//
size_t Wy_Thread::active_threads()
try {
 ::pthread_once(&wy__toltab_once_var, wy__toltab_once_func);
 WyRet r;
 if((r=wy__toltab_ptr->mutex().lock())!=Ok) {
   WY_THROW(r);
 }
 const size_t retv( wy__toltab_ptr->size() );
 if((r=wy__toltab_ptr->mutex().unlock())!=Ok) {
   WY_THROW(r);
 }
 return(retv);
}
catch(const WyRet& e) {
#ifdef WY_DEBUG
 Wy::cerr << Wy::wrd(e) << '\n';
#endif
 WY_THROW( WyRet(e) );
};

// note: function form to limit usecases (so far)
//
size_t Wy_Thread::max_threads() WY__TSPC()
{
 return Max_Threads;
};

//------------------------------------------------------
namespace Wy {

 bool _thread_equal(const Wy_Thread::_ThreadID& t1, const Wy_Thread::_ThreadID& t2) WY__TSPC()
 {
   return ::pthread_equal(t1,t2);
 }
};
//----------------------------------------------------------------------------

/*
 note2: Wy_Thread assert the function call never fail:
        :::pthread_setcancelstate( PTHREAD_CANCEL_ENABLE: PTHREAD_CANCEL_DISABLE
                         , &oldstate)
        it might be better let this function always succeed.
 note5:
       ::pthread_cancel found to return EAGAIN (on RH Linux9.0)
       while man. indicates only fail on ESRCH.
 note6: Wy_Thread created threads always returns &_exit_obj, which is
        already set with desired value except cancelled.
        (So, value of PTHREAD_CANCELED cannot be any valid address, it is -1
         as I saw the source)
 note8: _thrd_stat transition general rule:
          Null -> Active    .. by begin()
          Active -> Stopped .. by _tmain_finish()
          * -> Null         .. by reset()
 note9: In rare testing situation, ESRCH may occur. Dont know how, but defensive

               begin()+   _tmain_finish+          reset()+
                      | ^
                      v |

             Stopped                  +----         ----+
 _thrd_stat: Active   +---         ---+             ----+
             Null   --+               .                 +-----
                      ...             .

 _thrd_id:    valid   +---           --+            ----+
            invalid --+               .+--              +-----

 real thread          .+-+--         --+
                    ---+-+             +----
*/
