/* Copyright is licensed under GNU LGPL.                 by I.J.Wang 2003

   Wy_Thread is a base class to handle a created thread.

   Note: Add compile option -pthread to g++
         Add link option -lpthread to g++

   Warning: Applications are not allowed to modify thread detatchstate
*/

#ifndef WY_THREAD_H__
#define WY_THREAD_H__

#define WY_THREAD_VERSION 33

#include "WyMutex.h"
#include "WyCond.h"

#ifdef WY_DEBUG
 extern size_t wy_joined_cnt;    // count. pthread_join succeeded
 extern size_t wy_created_cnt;   // count, pthread_create succeeded
 extern size_t wy_act_max;       // maximal reached active threads
#endif

//   Wy_Thread is a base class that tmain(...) can be run in a created thread.
//
class Wy_Thread {
  public:
    static const char class_name[];
    WY_THROW_REPLY;   // For declaraion of Reply

    enum ThreadState{
         Null,        // Object is construtced in Null state, no thread.
         Active,      // Thread is created (begin() called) to run tmain()
         Stopped      // Thread terminated
    };

    enum CancelType {
         Deferred=PTHREAD_CANCEL_DEFERRED,
         Asynchronous=PTHREAD_CANCEL_ASYNCHRONOUS
    };

    typedef ::pthread_t _ThreadID;

    // Because no document found for the value of ::pthread_t for no thread.
    // pick 0 for it, so far so good (the value it is initialized does not work)
    //
    static const _ThreadID Wy__Default_ThreadID=0;

    //------------------------------------------------------------
    // Thread function tmain() should not use these public members  
    // (except static ones) to refer to itself
    //------------------------------------------------------------
    Wy_Thread();
    virtual ~Wy_Thread() WY__TSPC();
 
    bool is_default() const;
    _ThreadID thread_id() const;
    ThreadState thread_state() const;
    WyRet exitcode() const;

    virtual WyRet reset();

    virtual WyRet begin();

    virtual WyRet cancel();
	 
    // [Cancel Point]
    virtual WyRet wait_not(ThreadState);

    //----------------
    // Static members
    //----------------
    static _ThreadID thread_self();

    static WyRet yield();
	 
    static Wy_Thread* tbase_ptr();

    static bool set_cancelable(bool en) WY__TSPC();

    static CancelType set_canceltype(CancelType ct) WY__TSPC();

    // [Cancel Point]
    static void cancel_point();

    static void exit(WyRet) __attribute__ ((__noreturn__));

    static size_t active_threads();

    static size_t max_threads() WY__TSPC();

    // Class for disabling cancelability of the calling thread
    //
    class NoCancel {
      public:

        NoCancel() WY__TSPC();
        ~NoCancel() WY__TSPC();
        bool was_cancelable() const WY__TSPC();

      private:
        int _oldstate;

        // Hidden
        NoCancel(const NoCancel&);
        const NoCancel& operator =(const NoCancel&);
        bool operator ==(const NoCancel&);
    };

    // [Internal] for check purpose
    //
    static const _ThreadID wy_main_thrd_id;

    // [Internal] Thread begin function
    //
    static void * wy_thread_begin_func(Wy_Thread *obj_ptr);
  protected:

    virtual WyRet tmain();

    // Make sure there is no active thread, the thread is cancelled if necessary
    // cancelability of the calling thread is disabled
    //
    void tmain_close() WY__TSPC();
  private:

    static const size_t Max_Threads=
#ifdef PTHREAD_THREADS_MAX
    PTHREAD_THREADS_MAX<512?  PTHREAD_THREADS_MAX:512;
#else
    512;
#endif

    mutable WyCond _thrd_stat_cond;// Condition that _thrd_stat set value
    ThreadState _thrd_stat;
    ::pthread_attr_t _thrd_attr;
    ::pthread_t _thrd_id;
    WyRet _exit_obj;
    bool _in_destroy;              // Destructor set this flag true
    mutable WyMutex _mtx;          // Mutex for all the private data members

    // [Not Cancelable] Set the calling thread to non-cancelable state
    //
    static void _non_cancelable() WY__TSPC();

    inline void _reset_avars() WY__NOTHROW__
         {
           _thrd_id=Wy__Default_ThreadID;
           _thrd_stat=Null;
           _exit_obj.reset();
         };

    // [Not Cancelable]
    //
    // [Internal] The final function of the thread tmain()
    //       _exit_obj is reset from ecode, and return pointer to _exit_obj
    //       tmain() return/throw invokes this function
    //
    //  Note: This member locks wy__toltab_ptr->mutex() and this->_mtx
    //
    void* _tmain_finish(const WyRet&) WY__TSPC();

    // [Internal] Cancel an Active thread and wait for non-Active state
    //            Make sure no active thread, by cancel and/or join
    //
    // post: thread state is Null or Stopped (no active thread)
    //
    void _cleanup_thread(WyLock&);

    friend class Wy__TOLTab;  // uses Max_Threads

    // Hidden members
    //
    Wy_Thread(const Wy_Thread&); 
    const Wy_Thread& operator =(const Wy_Thread&);
    bool operator ==(const Wy_Thread&);
    bool operator !=(const Wy_Thread&);
    virtual Wy_Thread* _alloc(WyRet&) const { return(0); };

    //void* set_terminate(void*);    // not defined yet
    //void* set_unexpected(void*);   // not defined yet
};

namespace Wy {

 bool _thread_equal(const Wy_Thread::_ThreadID& t1, const Wy_Thread::_ThreadID& t2) WY__TSPC();

};  // end namespace Wy

#endif
