/* Copyright is licensed under GNU LGPL.                 by I.J.Wang 2006

  This file is part of Wy_Array.hh

  Consider to add: class function member as array element

*/
#include <cstring>    // for memcpy,memmove

// Element type T is for POD (memcpy and operator= work) such as arithmetic types,
// pointer types,.. for those built-in types do not have the move constructor.
//
template<typename T>
class Wy__ArrayValue {
    typedef T ElemType;
    typedef size_t SizeType;
    static const SizeType MaxSizeRawMemory;
    static const SizeType MinimumCapacity;
    static const SizeType DefaultSize;

    ElemType* _b_bgn;           // array begin pointer
    ElemType* _b_end;           // end pointer of constructed objects
    const ElemType* _b_eob;     // array end storage ptr

    // [Syn] Allocate n bytes of raw memory block
    //
    // [Ret] ElemType pointer to n bytes of raw memory block
    //       NULL= Not enough memory
    //
    inline static ElemType* _alloc_raw(size_t n) WY__TSPC() {
             return reinterpret_cast<ElemType*>(::operator new(n,std::nothrow));
           };
 
    // [Syn] Free the memory block allocated by _alloc_raw(size_t)
    //
    inline static void _free_raw(ElemType* v) WY__TSPC() {
             ::operator delete(v);
           };

    // [Syn] Allocate an array of capacity cap, and copy the front
    //       part of the array from [buf, buf+n_elem)
    //
    //  Note: assert(cap>=n_elem) to call
    //
    // [Ret] ElemType pointer to such array
    //       NULL= Not enough memory
    //
    // [Depend] T(T), ~T()
    //
    static ElemType* _dup_array(const ElemType* buf, SizeType n_elem, SizeType cap)
       {
         // allocate another array of cap capacity
         ElemType *narr=_alloc_raw(cap*sizeof(ElemType));
         if(narr==NULL) {
           return(NULL);
         }
         std::memcpy(narr,buf,n_elem*sizeof(ElemType));
         return(narr);
       };

     static void _objmove(ElemType* dest, ElemType* src, SizeType n) WY__NOTHROW__
       {
        if(dest<src) {
          for( ; n; ++src,++dest,--n) {
            *dest=*src;
          }
        } else {
          if(n==0) {
            return;
          }
          --n;
          src+=n;
          dest+=n;
          for(;;) {
            *dest=*src;
            if(n==0) {
              break;
            }
            --src;
            --dest;
            --n;
          };
        }
       };

  public:
    //WY_THROW_REPLY; throw Wy_Array::Reply, not exactly in this class's name

    typedef ElemType value_type;
    typedef SizeType size_type;

    static  SizeType max_capacity() WY__TSPC()
       { return MaxSizeRawMemory/sizeof(ElemType); };
    static SizeType min_capacity() WY__TSPC()
       { return(MinimumCapacity); };

    Wy__ArrayValue()
      : _b_bgn(_alloc_raw(MinimumCapacity*sizeof(ElemType))),
        _b_end(_b_bgn),
        _b_eob(_b_bgn+MinimumCapacity)
    {
      if(_b_bgn==NULL) {
        WY_THROW(typename Wy_Array<ElemType>::Reply(Wym_ENOMEM) );
      }
      WY__CHK_VALIDOBJ(*this);
    };

    Wy__ArrayValue(size_t n_elem, const ElemType& elem)
      : _b_bgn(NULL), _b_end(NULL), _b_eob(NULL)
    {
      const SizeType cap( n_elem<MinimumCapacity? MinimumCapacity : n_elem );
      if(cap>this->max_capacity()) {
        WY_THROW( typename Wy_Array<ElemType>::Reply(Wym_EFBIG) );
      }
      _b_end=_b_bgn=_alloc_raw(cap*sizeof(ElemType));
      if(_b_bgn==NULL) {
        WY_THROW( typename Wy_Array<ElemType>::Reply(Wym_ENOMEM) );
      }
      _b_eob=_b_bgn+n_elem;
      for(; _b_end!=_b_eob; ++_b_end) {
        *_b_end= elem;
      }
      _b_eob=_b_bgn+cap;
      WY__CHK_VALIDOBJ(*this);
    };

    Wy__ArrayValue(const WySeg<ElemType>& src)
      : _b_bgn(NULL), _b_end(NULL), _b_eob(NULL)
    {
      const SizeType src_size( src.size() );
      const SizeType cap( src_size<MinimumCapacity? MinimumCapacity : src_size );
      _b_end=_b_bgn=_alloc_raw(cap*sizeof(ElemType));
      if(_b_bgn==NULL) {
        WY_THROW( typename Wy_Array<ElemType>::Reply(Wym_ENOMEM) );
      }
      _b_eob=_b_bgn+src_size;
      for(ElemType *ep(src.begin()); _b_end!=_b_eob; ++_b_end,++ep) {
        *_b_end= ElemType(*ep);
      }
      _b_eob=_b_bgn+cap;
      WY__CHK_VALIDOBJ(*this);
    };

    Wy__ArrayValue(const WySeg<const ElemType>& src)
      : _b_bgn(NULL), _b_end(NULL), _b_eob(NULL)
    {
      const SizeType src_size( src.size() );
      const SizeType cap( src_size<MinimumCapacity? MinimumCapacity : src_size );
      _b_end=_b_bgn=_alloc_raw(cap*sizeof(ElemType));
      if(_b_bgn==NULL) {
        WY_THROW( typename Wy_Array<ElemType>::Reply(Wym_ENOMEM) );
      }
      _b_eob=_b_bgn+src_size;
      for(const ElemType *ep(src.begin()); _b_end!=_b_eob; ++_b_end,++ep) {
        *_b_end= ElemType(*ep);
      }
      _b_eob=_b_bgn+cap;
      WY__CHK_VALIDOBJ(*this);
    };

    Wy__ArrayValue(const Wy__ArrayValue& src)
      : _b_bgn(NULL),  _b_end(NULL), _b_eob(NULL)
    {
      WY__CHK_VALIDOBJ(src);
      const SizeType cap( src.size()> Wy_Array<ElemType>::min_capacity()?
                          src.size(): Wy_Array<ElemType>::min_capacity() );
      _b_bgn=_dup_array(src._b_bgn,src.size(),cap);
      if(_b_bgn==NULL) {
        WY_THROW( typename Wy_Array<ElemType>::Reply(Wym_ENOMEM) );
      }
      _b_eob=_b_bgn+cap;
      _b_end=_b_bgn+src.size();
      WY__CHK_VALIDOBJ(*this);
    };

    Wy__ArrayValue(Wy__ArrayValue& p, Wy::ByMove_t) WY__TSPC()
      : _b_bgn(p._b_bgn), _b_end(p._b_end), _b_eob(p._b_eob) // move semantics of p to this
      {
         #ifdef WY_DEBUG
         // cleanup semantics of p
         p._b_bgn=p._b_end=NULL; p._b_eob=NULL;
         #endif
      };

    ~Wy__ArrayValue() 
    {
      WY__CHK_VALIDOBJ(*this);
      _free_raw(_b_bgn);
      _b_bgn=_b_end=NULL; _b_eob=NULL;
    };

    bool is_default() const WY__TSPC()
      { WY__CHK_VALIDOBJ(*this);
        return _b_end==_b_bgn; 
      };
    SizeType size() const WY__TSPC()
      { WY__CHK_VALIDOBJ(*this);
        return _b_end-_b_bgn; 
      };

    // ** Note **
    // These members expose the address of the internal dynamic buffer

    ElemType* begin() WY__TSPC()
      { WY__CHK_VALIDOBJ(*this);
        return _b_bgn; };
    const ElemType* begin() const WY__TSPC()
      { WY__CHK_VALIDOBJ(*this);
        return _b_bgn; };

    const ElemType* end() const WY__TSPC()
      { WY__CHK_VALIDOBJ(*this);
        return _b_end; };
    ElemType* end() WY__TSPC()
      { WY__CHK_VALIDOBJ(*this);
        return _b_end; };

    ElemType& front()
      {
        WY__CHK_VALIDOBJ(*this);
        if(_b_end==_b_bgn) {
          WY_THROW( typename Wy_Array<ElemType>::Reply(Wym_ENOENT) );
        }
        return *_b_bgn; 
      };
    const ElemType& front() const
      {
        WY__CHK_VALIDOBJ(*this);
        if(_b_end==_b_bgn) {
          WY_THROW( typename Wy_Array<ElemType>::Reply(Wym_ENOENT) );
        }
        return *_b_bgn; 
      };

    ElemType& back()
      {
        WY__CHK_VALIDOBJ(*this);
        if(_b_end==_b_bgn) {
          WY_THROW( typename Wy_Array<ElemType>::Reply(Wym_ENOENT) );
        }
        return *(_b_end-1); 
      };
    const ElemType& back() const
      {
        WY__CHK_VALIDOBJ(*this);
        if(_b_end==_b_bgn) {
          WY_THROW( typename Wy_Array<ElemType>::Reply(Wym_ENOENT) );
        }
        return *(_b_end-1); 
      };

    WySeg<ElemType> subseg() const WY__TSPC()
      {
        WY__CHK_VALIDOBJ(*this);
        return WySeg<ElemType>(_b_bgn,SizeType(_b_end-_b_bgn));
      };
    WySeg<ElemType> subseg(SizeType idx) const
      {
        WY__CHK_VALIDOBJ(*this);
        const SizeType csize(_b_end-_b_bgn);
        if(idx>csize) {
          WY_THROW( typename Wy_Array<ElemType>::Reply(Wym_EINVAL) );
        }
        return WySeg<ElemType>(_b_bgn+idx,csize-idx);
      };
    WySeg<ElemType> subseg(SizeType idx, SizeType num) const
      {
      WY__CHK_VALIDOBJ(*this);
        SizeType rs(_b_end-_b_bgn);
        if(idx>rs) {
          WY_THROW( typename Wy_Array<ElemType>::Reply(Wym_EINVAL) );
        }
        rs-=idx;
        if(rs>num) {
          rs=num;
        }
        return WySeg<ElemType>(_b_bgn+idx,rs);
      };

    const ElemType& operator[](SizeType n) const
      {
        WY__CHK_VALIDOBJ(*this);
        if(n>=SizeType(_b_end-_b_bgn)) {
          // Argument n is not in range [0,this->size())
          WY_THROW( typename Wy_Array<ElemType>::Reply(Wym_EINVAL) );
        }
        return _b_bgn[n];
      };
    ElemType& operator[](SizeType n)
      {
        WY__CHK_VALIDOBJ(*this);
        if(n>=SizeType(_b_end-_b_bgn)) {
          // Argument n is not in range [0,this->size())
          WY_THROW( typename Wy_Array<ElemType>::Reply(Wym_EINVAL) );
        }
        return _b_bgn[n];
      };

    // Note: capacity is intact
    void reset()
      {
        WY__CHK_VALIDOBJ(*this);
        _b_end=_b_bgn;
      };

    void reset(const WySeg<const ElemType>& src)
      {
        WY__CHK_VALIDOBJ(*this);
        const SizeType src_size(src.size());
        if(src_size<=this->_capacity()) {
          if(src.is_overlap( WySeg<const ElemType>(_b_bgn,_b_eob-_b_bgn) )) {
            WY_THROW( typename Wy_Array<ElemType>::Reply(Wym_ELOOP) );
          }
          ::memcpy(_b_bgn,src.begin(),src_size*sizeof(ElemType));
          _b_end=_b_bgn+src_size;
        } else {
          // This algorithm duplicate src first
          //
          const SizeType cap( src_size>=Wy_Array<ElemType>::min_capacity()?
                              src_size : Wy_Array<ElemType>::min_capacity() );
          ElemType *ptmp=_dup_array(src.begin(),src_size,cap);
          if(ptmp==NULL) {
            WY_THROW( typename Wy_Array<ElemType>::Reply(Wym_ENOMEM) );
          }
  
          _free_raw(_b_bgn);
          _b_bgn=ptmp;
          _b_end=ptmp+src_size;
          _b_eob=ptmp+cap;
        }
        WY__CHK_VALIDOBJ(*this);
      };

    void reset(const WySeg<ElemType>& src)
      {
        WY__CHK_VALIDOBJ(*this);
        const SizeType src_size(src.size());
        if(src_size<=this->_capacity()) {
          if(src.is_overlap( WySeg<ElemType>(_b_bgn,_b_eob-_b_bgn) )) {
            WY_THROW( typename Wy_Array<ElemType>::Reply(Wym_ELOOP) );
          }
          ::memcpy(_b_bgn,src.begin(),src_size*sizeof(ElemType));
          _b_end=_b_bgn+src_size;
        } else {
          // This algorithm duplicate src first
          //
          const SizeType cap( src_size>=Wy_Array<ElemType>::min_capacity()?
                              src_size : Wy_Array<ElemType>::min_capacity() );
          ElemType *ptmp=_dup_array(src.begin(),src_size,cap);
          if(ptmp==NULL) {
            WY_THROW( typename Wy_Array<ElemType>::Reply(Wym_ENOMEM) );
          }
  
          _free_raw(_b_bgn);
          _b_bgn=ptmp;
          _b_end=ptmp+src_size;
          _b_eob=ptmp+cap;
        }
        WY__CHK_VALIDOBJ(*this);
      };

    // Note: capacity is set to needed minimum if reallocation occurred,
    //       otherwise capacity is intact
    void reset(const Wy__ArrayValue& src)
      {
        WY__CHK_VALIDOBJ(*this);
        const SizeType src_size(src.size());
        if(src_size<=this->_capacity()) {
          if(_b_bgn==src._b_bgn) {
            return;        // v.reset(v)
          }
          ::memcpy(_b_bgn,src._b_bgn,src_size*sizeof(ElemType));
          _b_end=_b_bgn+src_size;
        } else {
          // This algorithm duplicate src first
          //
          const SizeType cap( src_size>=Wy_Array<ElemType>::min_capacity()?
                              src_size : Wy_Array<ElemType>::min_capacity() );
          ElemType *ptmp=_dup_array(src._b_bgn,src_size,cap);
          if(ptmp==NULL) {
            WY_THROW( typename Wy_Array<ElemType>::Reply(Wym_ENOMEM) );
          }
  
          _free_raw(_b_bgn);
          _b_bgn=ptmp;
          _b_end=ptmp+src_size;
          _b_eob=ptmp+cap;
        }
        WY__CHK_VALIDOBJ(*this);
      };

    // note: elem in buffer ok
    //
    void reset(size_t n_elem, const ElemType& elem)
      {
        WY__CHK_VALIDOBJ(*this);
        const SizeType cap( n_elem<MinimumCapacity? MinimumCapacity : n_elem );
        if(cap>this->max_capacity()) {
          WY_THROW( typename Wy_Array<ElemType>::Reply(Wym_EFBIG) );
        }
        if(cap<=this->_capacity()) {
          _b_end=_b_bgn;
          for(SizeType i=0; i<n_elem; ++i,++_b_end) {
            *_b_end= elem;
          }
        } else {
          ElemType *ptmp=_alloc_raw(cap*sizeof(ElemType));
          if(ptmp==NULL) {
            WY_THROW( typename Wy_Array<ElemType>::Reply(Wym_ENOMEM) );
          }
          for(_b_end=ptmp, _b_eob=ptmp+n_elem; _b_end<_b_eob; ++_b_end) {
            *_b_end= elem;
          }
          _free_raw(_b_bgn);
          _b_bgn=ptmp;
          _b_eob=ptmp+cap;
        }
        WY__CHK_VALIDOBJ(*this);
      };

    const Wy__ArrayValue& operator =(const Wy__ArrayValue& rhs)
      {
        this->reset(rhs);
        return *this; 
      };

    const Wy__ArrayValue& operator=(const WySeg<const ElemType>& src)
      {
        WY__CHK_VALIDOBJ(*this);
        const SizeType src_size(src.size());
        if(src_size<=this->_capacity()) {
          if(src.is_overlap( WySeg<const ElemType>(_b_bgn,SizeType(_b_eob-_b_bgn)) )) {
            WY_THROW( typename Wy_Array<ElemType>::Reply(Wym_ELOOP) );
          }
          ::memcpy(_b_bgn,src.begin(),src_size*sizeof(ElemType));
          _b_end=_b_bgn+src_size;
        } else {
          // This algorithm duplicate src first
          //
          const SizeType cap( src_size>=Wy_Array<ElemType>::min_capacity()?
                              src_size : Wy_Array<ElemType>::min_capacity() );
          ElemType *ptmp=_dup_array(src.begin(),src_size,cap);
          if(ptmp==NULL) {
            WY_THROW( typename Wy_Array<ElemType>::Reply(Wym_ENOMEM) );
          }
  
          _free_raw(_b_bgn);
          _b_bgn=ptmp;
          _b_end=ptmp+src_size;
          _b_eob=ptmp+cap;
        }
        WY__CHK_VALIDOBJ(*this);
        return(*this);
      };

    const Wy__ArrayValue& operator=(const WySeg<ElemType>& src)
      {
        WY__CHK_VALIDOBJ(*this);
        const SizeType src_size(src.size());
        if(src_size<=this->_capacity()) {
          if(src.is_overlap( WySeg<ElemType>(_b_bgn,SizeType(_b_eob-_b_bgn)) )) {
            WY_THROW( typename Wy_Array<ElemType>::Reply(Wym_ELOOP) );
          }
          ::memcpy(_b_bgn,src.begin(),src_size*sizeof(ElemType));
          _b_end=_b_bgn+src_size;
        } else {
          // This algorithm duplicate src first
          //
          const SizeType cap( src_size>=Wy_Array<ElemType>::min_capacity()?
                              src_size : Wy_Array<ElemType>::min_capacity() );
          ElemType *ptmp=_dup_array(src.begin(),src_size,cap);
          if(ptmp==NULL) {
            WY_THROW( typename Wy_Array<ElemType>::Reply(Wym_ENOMEM) );
          }
  
          _free_raw(_b_bgn);
          _b_bgn=ptmp;
          _b_end=ptmp+src_size;
          _b_eob=ptmp+cap;
        }
        WY__CHK_VALIDOBJ(*this);
        return(*this);
      };

    void swap(Wy__ArrayValue& v2) WY__TSPC()
      {
       WY__CHK_VALIDOBJ(*this);
       // v.swap(v) ok
       {
        ElemType* tmp(_b_bgn);
        _b_bgn=v2._b_bgn;
        v2._b_bgn=tmp;
        tmp=_b_end;
        _b_end=v2._b_end;
        v2._b_end=tmp;
       }
       {
        const ElemType* tmp(_b_eob);
        _b_eob=v2._b_eob;
        v2._b_eob=tmp;
       }
        WY__CHK_VALIDOBJ(*this);
       // This definition may be better
       // char buf[sizeof(Wy__ArrayValue)];   // alignment problem!
       // new(buf) Wy__ArrayValue(v2,ByMove);
       // new(v2) Wy__ArrayValue(*this,ByMove);
       // new(this) Wy__ArrayValue(*((Wy__ArrayValue*)buf),ByMove);
       //
      };

    void push_back(ElemType elem)
      {
       // elem in *this ok
       WY__CHK_VALIDOBJ(*this);

       if(_b_end<_b_eob) {
         *_b_end=elem;
         ++_b_end;
       } else {
         const SizeType OrgSize( this->size() );

         #ifdef WY_DEBUG
           if(_b_eob!=_b_end) {
             WY_TERMINATE("");
           }
         #endif
         // allocate another array twice as large
         const SizeType cap( OrgSize<<1 );
         
         if(cap>this->max_capacity()) {
           WY_THROW( typename Wy_Array<ElemType>::Reply(Wym_EFBIG) );
         }
         ElemType *ptmp=_alloc_raw(cap*sizeof(ElemType));
         if(ptmp==NULL) {
           WY_THROW( typename Wy_Array<ElemType>::Reply(Wym_ENOMEM) );
         }
         std::memcpy(ptmp,_b_bgn,sizeof(ElemType)*OrgSize);

         ptmp[OrgSize]=elem;

         // update _b_bgn,_b_end,_b_eob to the new array(ptmp)
         _free_raw(_b_bgn);
         _b_bgn=ptmp;
         _b_end=ptmp+(OrgSize+1);
         _b_eob=ptmp+cap;
       }
       WY__CHK_VALIDOBJ(*this);
      };

    void pop_back()
      {
       WY__CHK_VALIDOBJ(*this);
       if(_b_end==_b_bgn) {
         return;
       }
       --_b_end;
       WY__CHK_VALIDOBJ(*this);
      };

    void erase(SizeType idx, SizeType n)
      {
        WY__CHK_VALIDOBJ(*this);
        ElemType *p1(_b_bgn+idx);   // p1 -> start of erase
        if(p1>=_b_end) {
          if(p1==_b_end) {
            WY__CHK_VALIDOBJ(*this);
            return;
          }
          WY_THROW( typename Wy_Array<ElemType>::Reply(Wym_EINVAL) );
        }
        // assert( p1<_b_end );

        ElemType *p2=_b_end;
        if(SizeType(p2-p1)>=n) {
          if(n<=0) {
            WY__CHK_VALIDOBJ(*this);
            return;
          }
          p2=p1+n;
        }
        const SizeType NumRem(_b_end-p2);
        std::memmove(p1,p2,NumRem*sizeof(ElemType));
        _b_end=p1+NumRem;
        WY__CHK_VALIDOBJ(*this);
      };

    // This member may be redundent from resize(len)
    //    void erase(SizeType idx)

    // Note: In case elem is in the array buffer
    //       argument elem is a real object (not a reference type)
    //
    void insert(SizeType idx, const ElemType elem)
      {
       // elem in *this ok
       WY__CHK_VALIDOBJ(*this);
       const SizeType OrgSize( this->size() );
       if(idx>OrgSize) {
         WY_THROW( typename Wy_Array<ElemType>::Reply(Wym_EINVAL) );
       }
       if(OrgSize<this->_capacity()) {
         ElemType *p1(_b_bgn+idx);   // insert point
         std::memmove(p1+1,p1,(_b_end-p1)*sizeof(ElemType));
         *p1=elem;
         ++_b_end;
       } else {
         #ifdef WY_DEBUG
            if(OrgSize!=this->_capacity()) {
              WY_TERMINATE("");
            }
         #endif
         // allocate another array twice as large
         const SizeType cap( OrgSize<<1 );
         if(cap>this->max_capacity()) {
           WY_THROW( typename Wy_Array<ElemType>::Reply(Wym_EFBIG) );
         }
         ElemType *ptmp=_alloc_raw(cap*sizeof(ElemType));
         if(ptmp==NULL) {
           WY_THROW( typename Wy_Array<ElemType>::Reply(Wym_ENOMEM) );
         }

         std::memcpy(ptmp,_b_bgn,idx*sizeof(ElemType));
         ptmp[idx]=elem;
         std::memcpy(ptmp+idx+1,_b_bgn+idx,(OrgSize-idx)*sizeof(ElemType));

         // update _b_bgn,_b_end,_b_eob to the new array(ptmp)
         ElemType *p2(_b_bgn);
         _b_bgn=ptmp;
         _b_end=ptmp+OrgSize+1;
         _b_eob=ptmp+cap;

         _free_raw(p2);
        }
        WY__CHK_VALIDOBJ(*this);
      };

    // Note: In case elem is in the array buffer
    //       argument elem is a real object (not a reference type)
    //
    //
    void insert(SizeType idx, SizeType n_elem, const ElemType elem)
      {
       WY__CHK_VALIDOBJ(*this);
       const SizeType OrgSize( this->size() );
       if(idx>OrgSize) {
         WY_THROW( typename Wy_Array<ElemType>::Reply(Wym_EINVAL) );
       }
       if( (OrgSize+n_elem)<=this->_capacity()) {
         ElemType *p1(_b_bgn+idx);   // insert point
         std::memmove(p1+n_elem,p1,(_b_end-p1)*sizeof(ElemType));
         for(SizeType i=0; i<n_elem; ++i,++p1) {
            *p1=elem;
         }
         _b_end+=n_elem;
       } else {
         // allocate another array twice as large
         if(n_elem>this->max_capacity()) {
           WY_THROW( typename Wy_Array<ElemType>::Reply(Wym_EFBIG) );
         }
         const SizeType cap( (n_elem+OrgSize)<<1 );
         if(cap>this->max_capacity()) {
           WY_THROW( typename Wy_Array<ElemType>::Reply(Wym_EFBIG) );
         }
         ElemType *ptmp=_alloc_raw(cap*sizeof(ElemType));
         if(ptmp==NULL) {
           WY_THROW( typename Wy_Array<ElemType>::Reply(Wym_ENOMEM) );
         }

         std::memcpy(ptmp,_b_bgn,idx*sizeof(ElemType));
         ElemType *p1(ptmp+idx);
         for(SizeType i=0; i<n_elem; ++i,++p1) {
            *p1=elem;
         }
         std::memcpy(ptmp+(idx+n_elem),_b_bgn+idx,(OrgSize-idx)*sizeof(ElemType));

         // update _b_bgn,_b_end,_b_eob to the new array(ptmp)
         p1=_b_bgn;
         _b_bgn=ptmp;
         _b_end=ptmp+OrgSize+n_elem;
         _b_eob=ptmp+cap;

         _free_raw(p1);
        }
        WY__CHK_VALIDOBJ(*this);
      };

    void insert(SizeType idx, const WySeg<const ElemType>& src)
      {
       WY__CHK_VALIDOBJ(*this);
       const SizeType OrgSize( this->size() );
       if(idx>OrgSize) {
         WY_THROW( typename Wy_Array<ElemType>::Reply(Wym_EINVAL) );
       }
       if(src.is_overlap( WySeg<const ElemType>(_b_bgn,(ElemType*)_b_eob) )) {
         WY_THROW( typename Wy_Array<ElemType>::Reply(Wym_ELOOP) );
       }
       const SizeType src_size( src.size() );
       if( (OrgSize+src_size)<=this->_capacity()) {
         ElemType *p1(_b_bgn+idx);   // insert point
         std::memmove(p1+src_size,p1,(_b_end-p1)*sizeof(ElemType));
         std::memcpy(p1,src.begin(),src_size*sizeof(ElemType));
         _b_end+=src_size;
       } else {
         // allocate another array twice as large
         if(src_size>this->max_capacity()) {
           WY_THROW( typename Wy_Array<ElemType>::Reply(Wym_EFBIG) );
         }
         const SizeType cap( (src_size+OrgSize)<<1 );
         if(cap>this->max_capacity()) {
           WY_THROW( typename Wy_Array<ElemType>::Reply(Wym_EFBIG) );
         }
         ElemType *ptmp=_alloc_raw(cap*sizeof(ElemType));
         if(ptmp==NULL) {
           WY_THROW( typename Wy_Array<ElemType>::Reply(Wym_ENOMEM) );
         }

         std::memcpy(ptmp,_b_bgn,idx*sizeof(ElemType));
         std::memcpy(ptmp+idx,src.begin(),src_size*sizeof(ElemType));
         std::memcpy(ptmp+(idx+src_size),_b_bgn+idx,(OrgSize-idx)*sizeof(ElemType));

         // update _b_bgn,_b_end,_b_eob to the new array(ptmp)
         ElemType *p2(_b_bgn);
         _b_bgn=ptmp;
         _b_end=ptmp+OrgSize+src_size;
         _b_eob=ptmp+cap;

         _free_raw(p2);
        }
        WY__CHK_VALIDOBJ(*this);
      };

    void insert(SizeType idx, const WySeg<ElemType>& src)
      {
       WY__CHK_VALIDOBJ(*this);
       const SizeType OrgSize( this->size() );
       if(idx>OrgSize) {
         WY_THROW( typename Wy_Array<ElemType>::Reply(Wym_EINVAL) );
       }
       if(src.is_overlap( WySeg<ElemType>(_b_bgn,(ElemType*)_b_eob) )) {
         WY_THROW( typename Wy_Array<ElemType>::Reply(Wym_ELOOP) );
       }
       const SizeType src_size( src.size() );
       if( (OrgSize+src_size)<=this->_capacity()) {
         ElemType *p1(_b_bgn+idx);   // insert point
         std::memmove(p1+src_size,p1,(_b_end-p1)*sizeof(ElemType));
         std::memcpy(p1,src.begin(),src_size*sizeof(ElemType));
         _b_end+=src_size;
       } else {
         // allocate another array twice as large
         if(src_size>this->max_capacity()) {
           WY_THROW( typename Wy_Array<ElemType>::Reply(Wym_EFBIG) );
         }
         const SizeType cap( (src_size+OrgSize)<<1 );
         if(cap>this->max_capacity()) {
           WY_THROW( typename Wy_Array<ElemType>::Reply(Wym_EFBIG) );
         }
         ElemType *ptmp=_alloc_raw(cap*sizeof(ElemType));
         if(ptmp==NULL) {
           WY_THROW( typename Wy_Array<ElemType>::Reply(Wym_ENOMEM) );
         }

         std::memcpy(ptmp,_b_bgn,idx*sizeof(ElemType));
         std::memcpy(ptmp+idx,src.begin(),src_size*sizeof(ElemType));
         std::memcpy(ptmp+(idx+src_size),_b_bgn+idx,(OrgSize-idx)*sizeof(ElemType));

         // update _b_bgn,_b_end,_b_eob to the new array(ptmp)
         ElemType *p2(_b_bgn);
         _b_bgn=ptmp;
         _b_end=ptmp+OrgSize+src_size;
         _b_eob=ptmp+cap;

         _free_raw(p2);
        }
        WY__CHK_VALIDOBJ(*this);
      };

    void insert(SizeType idx, const Wy_Array<ElemType>& src)
      { this->insert(idx, src.subseg()); };

    void replace(SizeType sidx, SizeType slen, SizeType n_elem, const ElemType& elem)
      {
       WY__CHK_VALIDOBJ(*this);
       const SizeType OrgSize( this->size() );
       if(sidx>OrgSize) {
         WY_THROW( typename Wy_Array<ElemType>::Reply(Wym_EINVAL) );
       }

       // Clamp slen to at most the existing number of elements [sidx,_b_end)
       SizeType rs= OrgSize-sidx;   
       if(slen>rs) {
         slen=rs;
       }

       rs=OrgSize-slen+n_elem;      // resultant size of this array
       if(rs>this->max_capacity()) {
         WY_THROW( typename Wy_Array<ElemType>::Reply(Wym_EFBIG) );
       }
       if(rs<=this->_capacity()) {
         const ElemType InsElem(elem);  // in case elem in *this

         ElemType *des_ptr=_b_bgn+sidx;      // insert point
         ElemType *src_ptr(des_ptr+slen);
         #ifdef WY_DEBUG
           if(((des_ptr-_b_bgn)+n_elem+(_b_end-src_ptr))!=rs) {
             WY_TERMINATE("");
           }
         #endif
         _objmove(des_ptr+n_elem,src_ptr,_b_end-src_ptr);

         for(SizeType i=0 ; i<n_elem; ++des_ptr,++i) {
           *des_ptr=InsElem;
         }
         _b_end=_b_bgn+rs;
       } else {
         // allocate memory for the new array
         const SizeType cap( rs );
         ElemType *ptmp=_alloc_raw(cap*sizeof(ElemType));
         if(ptmp==NULL) {
           WY_THROW( typename Wy_Array<ElemType>::Reply(Wym_ENOMEM) );
         }

         ElemType *p1(ptmp+sidx);   // insert point
         for(SizeType i=0; i<n_elem; ++p1,++i) {
           *p1=elem;
         }

         ElemType* des_ptr=ptmp;
         ElemType* p2=_b_bgn;
         p1=ptmp+sidx;
         // move all elements in _b_bgn to ptmp
         for(; des_ptr!=p1; ++p2,++des_ptr) {
           *des_ptr=*p2;
         }
         des_ptr+=n_elem;
         p2+=slen;
         for(; p2!=_b_end; ++p2,++des_ptr) {
           *des_ptr=*p2;
         }

         // update _b_bgn,_b_end,_b_eob to the new array(ptmp)
         p2=_b_bgn;
         _b_bgn=ptmp;
         _b_end=des_ptr;
         _b_eob=ptmp+cap;

         _free_raw(p2);
       };
        WY__CHK_VALIDOBJ(*this);
      };

    void resize(SizeType nsize)
      {
       WY__CHK_VALIDOBJ(*this);
       if(nsize<this->size()) {
         ElemType *p1(_b_bgn+nsize);
         _b_end=p1;
       } else if(nsize>this->size()) {
         // reserve enough capacity
         const WyRet r( this->_reserve(nsize) );
         if(r!=Ok) {
           WY_THROW( typename Wy_Array<ElemType>::Reply(r) );
         }

         ElemType *p2(_b_bgn+nsize);
         for(; _b_end<p2; ++_b_end) {
           *_b_end=ElemType();
         }
       } else {
       };
       WY__CHK_VALIDOBJ(*this);
      };

    void resize(SizeType nsize, const ElemType& elem)
      {
       WY__CHK_VALIDOBJ(*this);
       if(nsize<this->size()) {
         ElemType *p1(_b_bgn+nsize);
         _b_end=p1;
       } else if(nsize>this->size()) {
         // reserve enough capacity
         const WyRet r( this->_reserve(nsize) );
         if(r!=Ok) {
           WY_THROW( typename Wy_Array<ElemType>::Reply(r) );
         }

         ElemType *p2(_b_bgn+nsize);
         for(; _b_end<p2; ++_b_end) {
           *_b_end=elem;
         }
       } else {
       };
       WY__CHK_VALIDOBJ(*this);
      };

    void __setsize(SizeType nsize)
      {
        _b_end= _b_bgn+nsize;
      };
 
    void __destruct(SizeType idx)
      {
        // ElemType is discardable (dtor should do nothing)
      };

    void __move(SizeType dest_idx, SizeType src_idx)
      {
        *(_b_bgn+dest_idx)=*(_b_bgn+src_idx);
      };
    void __move(SizeType dest_idx, ElemType* src_ptr)
      {
        *(_b_bgn+dest_idx)=*src_ptr;
      };

    bool operator ==(const Wy__ArrayValue& rhs) const
      {
        WY__CHK_VALIDOBJ(*this);
        if(this->size()!=rhs.size()) {
          return(false);
        }
        for(const ElemType *p1(_b_bgn), *p2(rhs._b_bgn); p1!=_b_end; ++p1,++p2) {
          if(*p1==*p2) {
          } else {
            return(false);
          }
        }
        return(true);
      };
    bool operator !=(const Wy__ArrayValue& rhs) const 
      {
        WY__CHK_VALIDOBJ(*this);
        if(this->size()!=rhs.size()) {
          return(true);
        }
        for(const ElemType *p1(_b_bgn), *p2(rhs._b_bgn); p1!=_b_end; ++p1,++p2) {
          if(*p1!=*p2) {
            return(true);
          } else {
          }
        }
        return(false);
      };

    SizeType _capacity() const WY__TSPC()
      { WY__CHK_VALIDOBJ(*this);
        return _b_eob-_b_bgn; };

    WyRet _reserve(SizeType rcap) WY__TSPC()
      {
        WY__CHK_VALIDOBJ(*this);
        if(rcap<=this->_capacity()) {
          return(Ok);
        }
        if(rcap>this->max_capacity()) {
          WY_RETURN( Wym_EFBIG );
        }
        const SizeType OrgSize(this->size());
        ElemType *ptmp=_dup_array(_b_bgn,OrgSize,rcap);
        if(ptmp==NULL) {
          WY_RETURN( Wym_ENOMEM );
        }

        _free_raw(_b_bgn);
        _b_bgn=ptmp;
        _b_end=ptmp+OrgSize;
        _b_eob=ptmp+rcap;

        WY__CHK_VALIDOBJ(*this);
        return(Ok);
      };

};

template<typename T>
const size_t Wy__ArrayValue<T>::MaxSizeRawMemory
            =Wy_Array_MaxSizeRawMemory;

template<typename T>
const size_t Wy__ArrayValue<T>::MinimumCapacity
            =Wy_Array_MinimumCapacity;

template<typename T>
const size_t Wy__ArrayValue<T>::DefaultSize
            =Wy_Array_DefaultSize;

template<>
struct Wy_Array<char> : Wy__ArrayValue<char> {
  WY_THROW_REPLY;
  Wy_Array();
  Wy_Array(size_type s,const value_type& elem);
  Wy_Array(const WySeg<value_type>& src);
  Wy_Array(const WySeg<const value_type>& src);
  Wy_Array(const Wy_Array& s);
  Wy_Array(Wy_Array& p, Wy::ByMove_t t);

 void reset();
 void reset(const WySeg<const value_type>& src);
 void reset(const WySeg<value_type>& src);
 void reset(const Wy_Array&);
 void reset(size_type n_elem, const value_type& elem);
 void swap(Wy_Array&) WY__TSPC();
 void push_back(const value_type&);
 void pop_back();
 void erase(size_type idx, size_type n);
 void insert(size_type idx, const value_type elem);
 void insert(size_type idx, size_type n_elem, const value_type elem);
 void insert(size_type idx, const WySeg<const value_type>& src);
 void insert(size_type idx, const WySeg<value_type>& src);
 void insert(size_type idx, const Wy_Array<value_type>& src);
 void replace(size_type idx, size_type slen, size_type n_elem, const value_type& elem);
 void resize(size_type nsize);
 void resize(size_type nsize, const value_type& elem);

  const Wy_Array<value_type>& operator=(const WySeg<value_type>& rhs);
  const Wy_Array<value_type>& operator=(const WySeg<const value_type>& rhs);
  const Wy_Array<value_type>& operator=(const Wy_Array& rhs);

  bool operator==(const Wy_Array&) const;
  bool operator!=(const Wy_Array&) const;

  size_t _capacity() const WY__TSPC();
  WyRet _reserve(size_t rcap) WY__TSPC();

  // [Derive]
  //void _move(SizeType dest_idx, SizeType src_idx)
  //void _setsize(size_t) WY__TSPC();
};

template<>
struct Wy_Array<signed char> : Wy__ArrayValue<signed char> {
  WY_THROW_REPLY;
  Wy_Array();
  Wy_Array(size_type s,const value_type& elem);
  Wy_Array(const WySeg<value_type>& src);
  Wy_Array(const WySeg<const value_type>& src);
  Wy_Array(const Wy_Array& s);
  Wy_Array(Wy_Array& p, Wy::ByMove_t t);

 void reset();
 void reset(const WySeg<const value_type>& src);
 void reset(const WySeg<value_type>& src);
 void reset(const Wy_Array&);
 void reset(size_type n_elem, const value_type& elem);
 void swap(Wy_Array&) WY__TSPC();
 void push_back(const value_type&);
 void pop_back();
 void erase(size_type idx, size_type n);
 void insert(size_type idx, const value_type elem);
 void insert(size_type idx, size_type n_elem, const value_type elem);
 void insert(size_type idx, const WySeg<const value_type>& src);
 void insert(size_type idx, const WySeg<value_type>& src);
 void insert(size_type idx, const Wy_Array<value_type>& src);
 void replace(size_type idx, size_type slen, size_type n_elem, const value_type& elem);
 void resize(size_type nsize);
 void resize(size_type nsize, const value_type& elem);

  const Wy_Array<value_type>& operator=(const WySeg<value_type>& rhs);
  const Wy_Array<value_type>& operator=(const WySeg<const value_type>& rhs);
  const Wy_Array<value_type>& operator=(const Wy_Array& rhs);

  bool operator==(const Wy_Array&) const;
  bool operator!=(const Wy_Array&) const;

  size_t _capacity() const WY__TSPC();
  WyRet _reserve(size_t rcap) WY__TSPC();
};

template<>
struct Wy_Array<unsigned char> : Wy__ArrayValue<unsigned char> {
  WY_THROW_REPLY;
  Wy_Array();
  Wy_Array(size_type s,const value_type& elem);
  Wy_Array(const WySeg<value_type>& src);
  Wy_Array(const WySeg<const value_type>& src);
  Wy_Array(const Wy_Array& s);
  Wy_Array(Wy_Array& p, Wy::ByMove_t t);

 void reset();
 void reset(const WySeg<const value_type>& src);
 void reset(const WySeg<value_type>& src);
 void reset(const Wy_Array&);
 void reset(size_type n_elem, const value_type& elem);
 void swap(Wy_Array&) WY__TSPC();
 void push_back(const value_type&);
 void pop_back();
 void erase(size_type idx, size_type n);
 void insert(size_type idx, const value_type elem);
 void insert(size_type idx, size_type n_elem, const value_type elem);
 void insert(size_type idx, const WySeg<const value_type>& src);
 void insert(size_type idx, const WySeg<value_type>& src);
 void insert(size_type idx, const Wy_Array<value_type>& src);
 void replace(size_type idx, size_type slen, size_type n_elem, const value_type& elem);
 void resize(size_type nsize);
 void resize(size_type nsize, const value_type& elem);

  const Wy_Array<value_type>& operator=(const WySeg<value_type>& rhs);
  const Wy_Array<value_type>& operator=(const WySeg<const value_type>& rhs);
  const Wy_Array<value_type>& operator=(const Wy_Array& rhs);

  bool operator==(const Wy_Array&) const;
  bool operator!=(const Wy_Array&) const;

  size_t _capacity() const WY__TSPC();
  WyRet _reserve(size_t rcap) WY__TSPC();
};

template<>
struct Wy_Array<bool> : Wy__ArrayValue<bool> {
  WY_THROW_REPLY;
  Wy_Array();
  Wy_Array(size_type s,const value_type& elem);
  Wy_Array(const WySeg<value_type>& src);
  Wy_Array(const WySeg<const value_type>& src);
  Wy_Array(const Wy_Array& s);
  Wy_Array(Wy_Array& p, Wy::ByMove_t t);

 void reset();
 void reset(const WySeg<const value_type>& src);
 void reset(const WySeg<value_type>& src);
 void reset(const Wy_Array&);
 void reset(size_type n_elem, const value_type& elem);
 void swap(Wy_Array&) WY__TSPC();
 void push_back(const value_type&);
 void pop_back();
 void erase(size_type idx, size_type n);
 void insert(size_type idx, const value_type elem);
 void insert(size_type idx, size_type n_elem, const value_type elem);
 void insert(size_type idx, const WySeg<const value_type>& src);
 void insert(size_type idx, const WySeg<value_type>& src);
 void insert(size_type idx, const Wy_Array<value_type>& src);
 void replace(size_type idx, size_type slen, size_type n_elem, const value_type& elem);
 void resize(size_type nsize);
 void resize(size_type nsize, const value_type& elem);

  const Wy_Array<value_type>& operator=(const WySeg<value_type>& rhs);
  const Wy_Array<value_type>& operator=(const WySeg<const value_type>& rhs);
  const Wy_Array<value_type>& operator=(const Wy_Array& rhs);

  bool operator==(const Wy_Array&) const;
  bool operator!=(const Wy_Array&) const;

  size_t _capacity() const WY__TSPC();
  WyRet _reserve(size_t rcap) WY__TSPC();
};

template<>
struct Wy_Array<signed short> : Wy__ArrayValue<signed short> {
  WY_THROW_REPLY;
  Wy_Array();
  Wy_Array(size_type s,const value_type& elem);
  Wy_Array(const WySeg<value_type>& src);
  Wy_Array(const WySeg<const value_type>& src);
  Wy_Array(const Wy_Array& s);
  Wy_Array(Wy_Array& p, Wy::ByMove_t t);

 void reset();
 void reset(const WySeg<const value_type>& src);
 void reset(const WySeg<value_type>& src);
 void reset(const Wy_Array&);
 void reset(size_type n_elem, const value_type& elem);
 void swap(Wy_Array&) WY__TSPC();
 void push_back(const value_type&);
 void pop_back();
 void erase(size_type idx, size_type n);
 void insert(size_type idx, const value_type elem);
 void insert(size_type idx, size_type n_elem, const value_type elem);
 void insert(size_type idx, const WySeg<const value_type>& src);
 void insert(size_type idx, const WySeg<value_type>& src);
 void insert(size_type idx, const Wy_Array<value_type>& src);
 void replace(size_type idx, size_type slen, size_type n_elem, const value_type& elem);
 void resize(size_type nsize);
 void resize(size_type nsize, const value_type& elem);

  const Wy_Array<value_type>& operator=(const WySeg<value_type>& rhs);
  const Wy_Array<value_type>& operator=(const WySeg<const value_type>& rhs);
  const Wy_Array<value_type>& operator=(const Wy_Array& rhs);

  bool operator==(const Wy_Array&) const;
  bool operator!=(const Wy_Array&) const;

  size_t _capacity() const WY__TSPC();
  WyRet _reserve(size_t rcap) WY__TSPC();
};

template<>
struct Wy_Array<signed int> : Wy__ArrayValue<signed int> {
  WY_THROW_REPLY;
  Wy_Array();
  Wy_Array(size_type s,const value_type& elem);
  Wy_Array(const WySeg<value_type>& src);
  Wy_Array(const WySeg<const value_type>& src);
  Wy_Array(const Wy_Array& s);
  Wy_Array(Wy_Array& p, Wy::ByMove_t t);

 void reset();
 void reset(const WySeg<const value_type>& src);
 void reset(const WySeg<value_type>& src);
 void reset(const Wy_Array&);
 void reset(size_type n_elem, const value_type& elem);
 void swap(Wy_Array&) WY__TSPC();
 void push_back(const value_type&);
 void pop_back();
 void erase(size_type idx, size_type n);
 void insert(size_type idx, const value_type elem);
 void insert(size_type idx, size_type n_elem, const value_type elem);
 void insert(size_type idx, const WySeg<const value_type>& src);
 void insert(size_type idx, const WySeg<value_type>& src);
 void insert(size_type idx, const Wy_Array<value_type>& src);
 void replace(size_type idx, size_type slen, size_type n_elem, const value_type& elem);
 void resize(size_type nsize);
 void resize(size_type nsize, const value_type& elem);

  const Wy_Array<value_type>& operator=(const WySeg<value_type>& rhs);
  const Wy_Array<value_type>& operator=(const WySeg<const value_type>& rhs);
  const Wy_Array<value_type>& operator=(const Wy_Array& rhs);

  bool operator==(const Wy_Array&) const;
  bool operator!=(const Wy_Array&) const;

  size_t _capacity() const WY__TSPC();
  WyRet _reserve(size_t rcap) WY__TSPC();
};

template<>
struct Wy_Array<signed long> : Wy__ArrayValue<signed long> {
  WY_THROW_REPLY;
  Wy_Array();
  Wy_Array(size_type s,const value_type& elem);
  Wy_Array(const WySeg<value_type>& src);
  Wy_Array(const WySeg<const value_type>& src);
  Wy_Array(const Wy_Array& s);
  Wy_Array(Wy_Array& p, Wy::ByMove_t t);

 void reset();
 void reset(const WySeg<const value_type>& src);
 void reset(const WySeg<value_type>& src);
 void reset(const Wy_Array&);
 void reset(size_type n_elem, const value_type& elem);
 void swap(Wy_Array&) WY__TSPC();
 void push_back(const value_type&);
 void pop_back();
 void erase(size_type idx, size_type n);
 void insert(size_type idx, const value_type elem);
 void insert(size_type idx, size_type n_elem, const value_type elem);
 void insert(size_type idx, const WySeg<const value_type>& src);
 void insert(size_type idx, const WySeg<value_type>& src);
 void insert(size_type idx, const Wy_Array<value_type>& src);
 void replace(size_type idx, size_type slen, size_type n_elem, const value_type& elem);
 void resize(size_type nsize);
 void resize(size_type nsize, const value_type& elem);

  const Wy_Array<value_type>& operator=(const WySeg<value_type>& rhs);
  const Wy_Array<value_type>& operator=(const WySeg<const value_type>& rhs);
  const Wy_Array<value_type>& operator=(const Wy_Array& rhs);

  bool operator==(const Wy_Array&) const;
  bool operator!=(const Wy_Array&) const;

  size_t _capacity() const WY__TSPC();
  WyRet _reserve(size_t rcap) WY__TSPC();
};

template<>
struct Wy_Array<signed long long> : Wy__ArrayValue<signed long long> {
  WY_THROW_REPLY;
  Wy_Array();
  Wy_Array(size_type s,const value_type& elem);
  Wy_Array(const WySeg<value_type>& src);
  Wy_Array(const WySeg<const value_type>& src);
  Wy_Array(const Wy_Array& s);
  Wy_Array(Wy_Array& p, Wy::ByMove_t t);

 void reset();
 void reset(const WySeg<const value_type>& src);
 void reset(const WySeg<value_type>& src);
 void reset(const Wy_Array&);
 void reset(size_type n_elem, const value_type& elem);
 void swap(Wy_Array&) WY__TSPC();
 void push_back(const value_type&);
 void pop_back();
 void erase(size_type idx, size_type n);
 void insert(size_type idx, const value_type elem);
 void insert(size_type idx, size_type n_elem, const value_type elem);
 void insert(size_type idx, const WySeg<const value_type>& src);
 void insert(size_type idx, const WySeg<value_type>& src);
 void insert(size_type idx, const Wy_Array<value_type>& src);
 void replace(size_type idx, size_type slen, size_type n_elem, const value_type& elem);
 void resize(size_type nsize);
 void resize(size_type nsize, const value_type& elem);

  const Wy_Array<value_type>& operator=(const WySeg<value_type>& rhs);
  const Wy_Array<value_type>& operator=(const WySeg<const value_type>& rhs);
  const Wy_Array<value_type>& operator=(const Wy_Array& rhs);

  bool operator==(const Wy_Array&) const;
  bool operator!=(const Wy_Array&) const;

  size_t _capacity() const WY__TSPC();
  WyRet _reserve(size_t rcap) WY__TSPC();
};

template<>
struct Wy_Array<unsigned short> : Wy__ArrayValue<unsigned short> {
  WY_THROW_REPLY;
  Wy_Array();
  Wy_Array(size_type s,const value_type& elem);
  Wy_Array(const WySeg<value_type>& src);
  Wy_Array(const WySeg<const value_type>& src);
  Wy_Array(const Wy_Array& s);
  Wy_Array(Wy_Array& p, Wy::ByMove_t t);

 void reset();
 void reset(const WySeg<const value_type>& src);
 void reset(const WySeg<value_type>& src);
 void reset(const Wy_Array&);
 void reset(size_type n_elem, const value_type& elem);
 void swap(Wy_Array&) WY__TSPC();
 void push_back(const value_type&);
 void pop_back();
 void erase(size_type idx, size_type n);
 void insert(size_type idx, const value_type elem);
 void insert(size_type idx, size_type n_elem, const value_type elem);
 void insert(size_type idx, const WySeg<const value_type>& src);
 void insert(size_type idx, const WySeg<value_type>& src);
 void insert(size_type idx, const Wy_Array<value_type>& src);
 void replace(size_type idx, size_type slen, size_type n_elem, const value_type& elem);
 void resize(size_type nsize);
 void resize(size_type nsize, const value_type& elem);

  const Wy_Array<value_type>& operator=(const WySeg<value_type>& rhs);
  const Wy_Array<value_type>& operator=(const WySeg<const value_type>& rhs);
  const Wy_Array<value_type>& operator=(const Wy_Array& rhs);

  bool operator==(const Wy_Array&) const;
  bool operator!=(const Wy_Array&) const;

  size_t _capacity() const WY__TSPC();
  WyRet _reserve(size_t rcap) WY__TSPC();
};

template<>
struct Wy_Array<unsigned int> : Wy__ArrayValue<unsigned int> {
  WY_THROW_REPLY;
  Wy_Array();
  Wy_Array(size_type s,const value_type& elem);
  Wy_Array(const WySeg<value_type>& src);
  Wy_Array(const WySeg<const value_type>& src);
  Wy_Array(const Wy_Array& s);
  Wy_Array(Wy_Array& p, Wy::ByMove_t t);

 void reset();
 void reset(const WySeg<const value_type>& src);
 void reset(const WySeg<value_type>& src);
 void reset(const Wy_Array&);
 void reset(size_type n_elem, const value_type& elem);
 void swap(Wy_Array&) WY__TSPC();
 void push_back(const value_type&);
 void pop_back();
 void erase(size_type idx, size_type n);
 void insert(size_type idx, const value_type elem);
 void insert(size_type idx, size_type n_elem, const value_type elem);
 void insert(size_type idx, const WySeg<const value_type>& src);
 void insert(size_type idx, const WySeg<value_type>& src);
 void insert(size_type idx, const Wy_Array<value_type>& src);
 void replace(size_type idx, size_type slen, size_type n_elem, const value_type& elem);
 void resize(size_type nsize);
 void resize(size_type nsize, const value_type& elem);

  const Wy_Array<value_type>& operator=(const WySeg<value_type>& rhs);
  const Wy_Array<value_type>& operator=(const WySeg<const value_type>& rhs);
  const Wy_Array<value_type>& operator=(const Wy_Array& rhs);

  bool operator==(const Wy_Array&) const;
  bool operator!=(const Wy_Array&) const;

  size_t _capacity() const WY__TSPC();
  WyRet _reserve(size_t rcap) WY__TSPC();
};

template<>
struct Wy_Array<unsigned long> : Wy__ArrayValue<unsigned long> {
  WY_THROW_REPLY;
  Wy_Array();
  Wy_Array(size_type s,const value_type& elem);
  Wy_Array(const WySeg<value_type>& src);
  Wy_Array(const WySeg<const value_type>& src);
  Wy_Array(const Wy_Array& s);
  Wy_Array(Wy_Array& p, Wy::ByMove_t t);

 void reset();
 void reset(const WySeg<const value_type>& src);
 void reset(const WySeg<value_type>& src);
 void reset(const Wy_Array&);
 void reset(size_type n_elem, const value_type& elem);
 void swap(Wy_Array&) WY__TSPC();
 void push_back(const value_type&);
 void pop_back();
 void erase(size_type idx, size_type n);
 void insert(size_type idx, const value_type elem);
 void insert(size_type idx, size_type n_elem, const value_type elem);
 void insert(size_type idx, const WySeg<const value_type>& src);
 void insert(size_type idx, const WySeg<value_type>& src);
 void insert(size_type idx, const Wy_Array<value_type>& src);
 void replace(size_type idx, size_type slen, size_type n_elem, const value_type& elem);
 void resize(size_type nsize);
 void resize(size_type nsize, const value_type& elem);

  const Wy_Array<value_type>& operator=(const WySeg<value_type>& rhs);
  const Wy_Array<value_type>& operator=(const WySeg<const value_type>& rhs);
  const Wy_Array<value_type>& operator=(const Wy_Array& rhs);

  bool operator==(const Wy_Array&) const;
  bool operator!=(const Wy_Array&) const;

  size_t _capacity() const WY__TSPC();
  WyRet _reserve(size_t rcap) WY__TSPC();
};

template<>
struct Wy_Array<unsigned long long> : Wy__ArrayValue<unsigned long long> {
  WY_THROW_REPLY;
  Wy_Array();
  Wy_Array(size_type s,const value_type& elem);
  Wy_Array(const WySeg<value_type>& src);
  Wy_Array(const WySeg<const value_type>& src);
  Wy_Array(const Wy_Array& s);
  Wy_Array(Wy_Array& p, Wy::ByMove_t t);

 void reset();
 void reset(const WySeg<const value_type>& src);
 void reset(const WySeg<value_type>& src);
 void reset(const Wy_Array&);
 void reset(size_type n_elem, const value_type& elem);
 void swap(Wy_Array&) WY__TSPC();
 void push_back(const value_type&);
 void pop_back();
 void erase(size_type idx, size_type n);
 void insert(size_type idx, const value_type elem);
 void insert(size_type idx, size_type n_elem, const value_type elem);
 void insert(size_type idx, const WySeg<const value_type>& src);
 void insert(size_type idx, const WySeg<value_type>& src);
 void insert(size_type idx, const Wy_Array<value_type>& src);
 void replace(size_type idx, size_type slen, size_type n_elem, const value_type& elem);
 void resize(size_type nsize);
 void resize(size_type nsize, const value_type& elem);

  const Wy_Array<value_type>& operator=(const WySeg<value_type>& rhs);
  const Wy_Array<value_type>& operator=(const WySeg<const value_type>& rhs);
  const Wy_Array<value_type>& operator=(const Wy_Array& rhs);

  bool operator==(const Wy_Array&) const;
  bool operator!=(const Wy_Array&) const;

  size_t _capacity() const WY__TSPC();
  WyRet _reserve(size_t rcap) WY__TSPC();
};

template<>
struct Wy_Array<float> : Wy__ArrayValue<float> {
  WY_THROW_REPLY;
  Wy_Array();
  Wy_Array(size_type s,const value_type& elem);
  Wy_Array(const WySeg<value_type>& src);
  Wy_Array(const WySeg<const value_type>& src);
  Wy_Array(const Wy_Array& s);
  Wy_Array(Wy_Array& p, Wy::ByMove_t t);

 void reset();
 void reset(const WySeg<const value_type>& src);
 void reset(const WySeg<value_type>& src);
 void reset(const Wy_Array&);
 void reset(size_type n_elem, const value_type& elem);
 void swap(Wy_Array&) WY__TSPC();
 void push_back(const value_type&);
 void pop_back();
 void erase(size_type idx, size_type n);
 void insert(size_type idx, const value_type elem);
 void insert(size_type idx, size_type n_elem, const value_type elem);
 void insert(size_type idx, const WySeg<const value_type>& src);
 void insert(size_type idx, const WySeg<value_type>& src);
 void insert(size_type idx, const Wy_Array<value_type>& src);
 void replace(size_type idx, size_type slen, size_type n_elem, const value_type& elem);
 void resize(size_type nsize);
 void resize(size_type nsize, const value_type& elem);

  const Wy_Array<value_type>& operator=(const WySeg<value_type>& rhs);
  const Wy_Array<value_type>& operator=(const WySeg<const value_type>& rhs);
  const Wy_Array<value_type>& operator=(const Wy_Array& rhs);

  bool operator==(const Wy_Array&) const;
  bool operator!=(const Wy_Array&) const;

  size_t _capacity() const WY__TSPC();
  WyRet _reserve(size_t rcap) WY__TSPC();
};

template<>
struct Wy_Array<double> : Wy__ArrayValue<double> {
  WY_THROW_REPLY;
  Wy_Array();
  Wy_Array(size_type s,const value_type& elem);
  Wy_Array(const WySeg<value_type>& src);
  Wy_Array(const WySeg<const value_type>& src);
  Wy_Array(const Wy_Array& s);
  Wy_Array(Wy_Array& p, Wy::ByMove_t t);

 void reset();
 void reset(const WySeg<const value_type>& src);
 void reset(const WySeg<value_type>& src);
 void reset(const Wy_Array&);
 void reset(size_type n_elem, const value_type& elem);
 void swap(Wy_Array&) WY__TSPC();
 void push_back(const value_type&);
 void pop_back();
 void erase(size_type idx, size_type n);
 void insert(size_type idx, const value_type elem);
 void insert(size_type idx, size_type n_elem, const value_type elem);
 void insert(size_type idx, const WySeg<const value_type>& src);
 void insert(size_type idx, const WySeg<value_type>& src);
 void insert(size_type idx, const Wy_Array<value_type>& src);
 void replace(size_type idx, size_type slen, size_type n_elem, const value_type& elem);
 void resize(size_type nsize);
 void resize(size_type nsize, const value_type& elem);

  const Wy_Array<value_type>& operator=(const WySeg<value_type>& rhs);
  const Wy_Array<value_type>& operator=(const WySeg<const value_type>& rhs);
  const Wy_Array<value_type>& operator=(const Wy_Array& rhs);

  bool operator==(const Wy_Array&) const;
  bool operator!=(const Wy_Array&) const;

  size_t _capacity() const WY__TSPC();
  WyRet _reserve(size_t rcap) WY__TSPC();
};

template<>
struct Wy_Array<long double> : Wy__ArrayValue<long double> {
  WY_THROW_REPLY;
  Wy_Array();
  Wy_Array(size_type s,const value_type& elem);
  Wy_Array(const WySeg<value_type>& src);
  Wy_Array(const WySeg<const value_type>& src);
  Wy_Array(const Wy_Array& s);
  Wy_Array(Wy_Array& p, Wy::ByMove_t t);

 void reset();
 void reset(const WySeg<const value_type>& src);
 void reset(const WySeg<value_type>& src);
 void reset(const Wy_Array&);
 void reset(size_type n_elem, const value_type& elem);
 void swap(Wy_Array&) WY__TSPC();
 void push_back(const value_type&);
 void pop_back();
 void erase(size_type idx, size_type n);
 void insert(size_type idx, const value_type elem);
 void insert(size_type idx, size_type n_elem, const value_type elem);
 void insert(size_type idx, const WySeg<const value_type>& src);
 void insert(size_type idx, const WySeg<value_type>& src);
 void insert(size_type idx, const Wy_Array<value_type>& src);
 void replace(size_type idx, size_type slen, size_type n_elem, const value_type& elem);
 void resize(size_type nsize);
 void resize(size_type nsize, const value_type& elem);

  const Wy_Array<value_type>& operator=(const WySeg<value_type>& rhs);
  const Wy_Array<value_type>& operator=(const WySeg<const value_type>& rhs);
  const Wy_Array<value_type>& operator=(const Wy_Array& rhs);

  bool operator==(const Wy_Array&) const;
  bool operator!=(const Wy_Array&) const;

  size_t _capacity() const WY__TSPC();
  WyRet _reserve(size_t rcap) WY__TSPC();
};

template<typename T>
struct Wy_Array<T*> : Wy__ArrayValue<T*> {
  WY_THROW_REPLY;
  Wy_Array() : Wy__ArrayValue<T*>() {};
  Wy_Array(size_t s,T* elem) : Wy__ArrayValue<T*>(s,elem) {};
  Wy_Array(const WySeg<T*>& src) : Wy__ArrayValue<T*>(src) {};
  Wy_Array(const WySeg<T* const>& src) : Wy__ArrayValue<T*>(src) {};
  Wy_Array(const Wy_Array& s) : Wy__ArrayValue<T*>(s) {};
  Wy_Array(Wy_Array& p, Wy::ByMove_t t) : Wy__ArrayValue<T*>(p,t) {};
};
