/* Copyright is licensed under GNU LGPL.                 by I.J.Wang 2003

  Test WySockFile to communicate with PF_LOCAL, PF_INET, PF_INET6
*/

#include "WySockFile.h"
#include "WySockAddrLocal.h"
#include "WySockAddrINet4.h"
#include "WySockAddrINet6.h"
#include "Wy_AtDestroy.hh"
#include "WyFileStat.h"
#include <sys/types.h>
#include <unistd.h>
#include <sys/wait.h>
#include <new>
#include <iostream>

#if WYSOCKFILE_VERSION!=33
#error Test code is for WYSOCKFILE_VERSION 33
#endif

#define WY__TEST_SOCK_INET__   // we are going to test INET

// Socket address (pathname for AF_LOCAL file) for test
//
static const WyStr LocalSockFileName("tmp.chk_sockfile.local.tmp");

// Message to send to the socket for the tests
//
static const WyCSeg SvrMsg("Messages sent from the server");
static const WyCSeg CliMsg("Messages sent from the client");

#ifdef WY__TEST_SOCK_INET__
const char* TestINetIP4="127.0.0.1";
const char* TestINetIP6="::1";
const ushort TestINetPort=8080;
#endif

// Assert ff is default
//
static void assert_default(const WySockFile& ff)
{
 WyRet r;

 if(ff.is_default()==false) {
   WY_THROW( WyRet() );
 }
 if(ff.fh().is_default()==false) {
   WY_THROW( WyRet() );
 }

 // chk fh()
 if(ff.fh()!=WyFileHandle()) {
   WY_THROW( WyRet() );
 }

 // stat
 {
   WyFileStat stt;
   if((r=ff.stat(stt))!=Wym_EBADF) {
     WY_THROW(r);
   }
   if(stt.is_default()==false) {
     WY_THROW( WyRet() );
   }
 }
 
 if(ff.family()!=PF_UNSPEC) {
   WY_THROW( WyRet() );
 }
 if(static_cast<const WyByteFlow&>(ff).ftype()!=0) {
   WY_THROW( WyRet() );
 }
};

static void assert_not_default(const WySockFile& ff)
{
 WyRet r;

 if(ff.is_default()) {
   WY_THROW( WyRet() );
 }
 if(ff.fh().is_default()) {
   WY_THROW( WyRet() );
 }

 // stat
 {
   WyFileStat stt;
   if((r=ff.stat(stt))!=Ok) {
     WY_THROW(r);
   }
   if(stt.is_sock()==false) {
     WY_THROW( WyRet() );
   }
 }
 
 if(ff.family()==PF_UNSPEC) {
   WY_THROW( WyRet() );
 }
 if(static_cast<const WyByteFlow&>(ff).ftype()!=S_IFSOCK) {
   WY_THROW( WyRet() );
 }
};

// Assert f1,f2 the same systerm file
//
static void assert_same_sysf(const WySockFile& f1,const WySockFile& f2)
{
 if(f1.is_default()) {
   assert_default(f1);
 } else {
   assert_not_default(f1);
 }

 if(f2.is_default()) {
   assert_default(f2);
 } else {
   assert_not_default(f2);
 }

 if(f1.is_default()&&f2.is_default()) {
   return;
 }
 WyRet r;
 WyFileStat stt1,stt2;
 if((r=f1.stat(stt1))!=Ok) {
   WY_THROW( WyRet() );
 }
 if((r=f2.stat(stt2))!=Ok) {
   WY_THROW( WyRet() );
 }
 if(stt1.sf_ino()!=stt2.sf_ino()) {
   WY_THROW( WyRet() );
 }
 if(stt1.sf_dev()!=stt2.sf_dev()) {
   WY_THROW( WyRet() );
 }
};

#define CHK_DEFAULT(p)\
    try { assert_default(p); } catch(...) { WY_THROW( WyRet()); }
#define CHK_NOT_DEFAULT(p)\
    try { assert_not_default(p); } catch(...) { WY_THROW( WyRet()); }
#define CHK_SAME_SYSF(p1,p2)\
    try { assert_same_sysf(p1,p2); } catch(...) { WY_THROW( WyRet()); }

static void exaust_cpy_ctor(const WySysFile& sktf)
{
 WySockFile tmp(sktf);
 exaust_cpy_ctor(sktf);   // recursive until failure
};

static void exaust_cpy_ctor(const WySockFile& sktf)
{
 WySockFile tmp(sktf);
 exaust_cpy_ctor(sktf);   // recursive until failure
};

// Setup the test 
//
static void setup_test(void)
{
};

// Undo setup_test
//
static void close_test(void)
{
};

// Check basic operations on default object
//
static void t1(void)
{
 WyRet r;

 WySockFile LocalObj;     // non-default object for test as argument
 CHK_DEFAULT(LocalObj);
 if((r=LocalObj.reset(PF_LOCAL,SOCK_STREAM))!=Ok) {
   WY_THROW(r);
 }
 CHK_NOT_DEFAULT(LocalObj);
 if(LocalObj.family()!=PF_LOCAL) {
   WY_THROW( WyRet() );
 }

 // stat
 {
   WyFileStat stt;
   if((r=LocalObj.stat(stt))!=Ok) {
     WY_THROW(r);
   }
   if(stt.is_sock()==false) {
     WY_THROW( WyRet() );
   }
 }

 // WySockFile()/WySockFile(const WySockFile&)
 {
   WySockFile tmp0;
   WySockFile tmp1(tmp0);
   CHK_DEFAULT(tmp0);
   CHK_DEFAULT(tmp1);

   WySockFile tmp2(LocalObj);
   CHK_NOT_DEFAULT(tmp2);
   CHK_SAME_SYSF(tmp2,LocalObj);
   if(tmp2.fh()==LocalObj.fh()) {
     WY_THROW( WyRet() );
   }

   WySockFile tmp3(static_cast<WyByteFlow&>(tmp2));
   CHK_SAME_SYSF(tmp2,tmp3);
 }

 // WySockFile(int,int,int=0)
 {
   WySockFile tmp(PF_INET,SOCK_STREAM);
   CHK_NOT_DEFAULT(tmp);
   if(tmp.family()!=AF_INET) {
     WY_THROW( WyRet() );
   }
   WySockFile tmp2(tmp);
   CHK_SAME_SYSF(tmp2,tmp);
 }

 // WySockFile(WyFileHandle)
 {
   WyFileHandle h;
   WySockFile tmp(h);
   CHK_DEFAULT(tmp);
 }

 // reset()/reset(const WySockFile&)
 {
   WySockFile tmp;
   CHK_DEFAULT(tmp);
   if((r=tmp.reset())!=Ok) {
     WY_THROW(r);
   }
   CHK_DEFAULT(tmp);

   WySockFile tmp2(LocalObj);
   CHK_NOT_DEFAULT(tmp2);
   CHK_SAME_SYSF(tmp2,LocalObj);
   if((r=tmp2.reset())!=Ok) {
     WY_THROW(r);
   }
   CHK_DEFAULT(tmp2);

   WySockFile tmp3;
   CHK_DEFAULT(tmp3);
   if((r=tmp3.reset(static_cast<WyByteFlow&>(tmp2)))!=Ok) {
     WY_THROW(r);
   }
   CHK_SAME_SYSF(tmp2,tmp3);
 }

 // reset(int,int,int=0)
 {
   WySockFile tmp;
   CHK_DEFAULT(tmp);
   if((r=tmp.reset(PF_INET,SOCK_STREAM))!=Ok) {
     WY_THROW(r);
   }
   CHK_NOT_DEFAULT(tmp);
   if(tmp.family()!=AF_INET) {
     WY_THROW( WyRet() );
   }
 }

 // reset(WyFileHandle)
 {
  WyFileHandle h;
   WySockFile tmp;
   CHK_DEFAULT(tmp);

   if((r=tmp.reset(h))!=Ok) {
     WY_THROW(r);
   }
   CHK_DEFAULT(tmp);
   if((r=tmp.reset( LocalObj.fh() ))!=Ok) {
     WY_THROW(r);
   }
   CHK_SAME_SYSF(tmp,LocalObj);
 }

 // read,write,bind
 {
   WySockFile tmp;
   CHK_DEFAULT(tmp);
   size_t n_rd,n_wr;
   char buf[2];
   WyStr str;
   n_rd=4;
   if((r=tmp.read(buf,sizeof(buf),n_rd))!=Wym_EBADF) {
     WY_THROW(r);
   }
   if(n_rd!=0) {
     WY_THROW( WyRet() );
   }
   n_wr=2;
   if((r=tmp.write(buf,sizeof(buf),n_wr))!=Wym_EBADF) {
     WY_THROW(r);
   }
   if(n_wr!=0) {
     WY_THROW( WyRet() );
   }
   n_wr=8;
   if((r=tmp.write(WyStr(),n_wr))!=Wym_EBADF) {
     WY_THROW(r);
   }
   if(n_wr!=0) {
     WY_THROW( WyRet() );
   }
   //if((r=tmp.drain())!=Wym_EBADF) {
   //  WY_THROW(r);
   //}
   if(tmp.family()!=PF_UNSPEC) {
     WY_THROW( WyRet() );
   }
 }

 // _swap
 {
  WySockFile tmp0(LocalObj),tmp1,tmp2;
  CHK_NOT_DEFAULT(tmp0);
  CHK_DEFAULT(tmp1);
  CHK_DEFAULT(tmp2);
  CHK_NOT_DEFAULT(LocalObj);

  if((r=tmp1._swap(tmp2))!=Ok) {
    WY_THROW(r);
  }
  CHK_DEFAULT(tmp1);
  CHK_DEFAULT(tmp2);

  if((r=tmp1._swap(tmp0))!=Ok) {
    WY_THROW(r);
  }
  CHK_DEFAULT(tmp0);
  CHK_NOT_DEFAULT(tmp1);
  CHK_SAME_SYSF(tmp1,LocalObj);

  if((r=tmp0._swap(tmp1))!=Ok) {
    WY_THROW(r);
  }
  CHK_DEFAULT(tmp1);
  CHK_NOT_DEFAULT(tmp0);
  CHK_SAME_SYSF(tmp0,LocalObj);

  WyByteFlow tmp3;
  if((r=tmp2._swap(tmp3))!=Wym_EBADTYPE) {
    WY_THROW(r);
  }
  if((r=tmp3._swap(tmp2))!=Wym_EBADTYPE) {
    WY_THROW(r);
  }
 }

 // chk _alloc
 {
   WySockFile tmp;
   WySockFile* p=tmp._alloc(r);
   if(r!=Ok) {
     WY_THROW(r);
   }
   if(p==0) {
     WY_THROW( WyRet() );
   }
   Wy_AtDestroy<void,WySockFile*> rrid(Wy::_delete_one,p);
   CHK_DEFAULT(*p);
   if((r=tmp._swap(*p))!=Ok) {
     WY_THROW(r);
   }
    CHK_DEFAULT(*p);
    CHK_DEFAULT(tmp);
 }

 // shutdown
 //
 {
   WySockFile tmp;
   CHK_DEFAULT(tmp);
   if((r=tmp.shutdown(SHUT_RD))!=Wym_EBADF) {
     WY_THROW(r);
   }
 }

 // bind,listen,accept,connect,sockname,peername
 //
 {
   WySockFile tmp;
   WySockAddrLocal addr;
   CHK_DEFAULT(tmp);
   
   if((r=tmp.bind(addr))!=Wym_EBADF) {
     WY_THROW(r);
   }
   if((r=tmp.listen(0))!=Wym_EBADF) {
     WY_THROW(r);
   }

   WySockFile tmp1;
   CHK_DEFAULT(tmp1);
   if((r=tmp.accept(tmp1,addr))!=Wym_EBADF) {
     WY_THROW(r);
   }
   if((r=tmp.connect(addr))!=Wym_EBADF) {
     WY_THROW(r);
   }
   if((r=tmp.sockname(addr))!=Wym_EBADF) {
     WY_THROW(r);
   }
   if((r=tmp.peername(addr))!=Wym_EBADF) {
     WY_THROW(r);
   }

   CHK_DEFAULT(tmp);
   CHK_DEFAULT(tmp1);
 }

 // sendto(..), recvfrom(..)
 {
   WySockFile tmp;
   WySockAddrLocal addr;
   CHK_DEFAULT(tmp);
   char buf[4];
   size_t n_size=1;
   if((r=tmp.sendto(buf,sizeof(buf),0,&addr,n_size))!=Wym_EBADF) {
     WY_THROW(r);
   }
   if(n_size!=0) {
     WY_THROW( WyRet() );
   }
   n_size=1;
   if((r=tmp.recvfrom(buf,sizeof(buf),0,&addr,n_size))!=Wym_EBADF) {
     WY_THROW(r);
   }
   if(n_size!=0) {
     WY_THROW( WyRet() );
   }
 }

 // chk sockpair(..)
 {
   WySockFile s1,s2;
   if((r=WySockFile::sockpair(PF_LOCAL,SOCK_STREAM,0,s1,s2))!=Ok) {
     WY_THROW(r);
   }
   const WyStr Msg("helo3");
   size_t n_wr(0);
   WyStr str;
   if(s1.write(Msg,n_wr)!=Ok) {
     WY_THROW(r);
   }
   if(n_wr!=Msg.size()) {
     WY_THROW( WyRet() );
   }
   if((r=WySockFile::sockpair(PF_LOCAL,SOCK_STREAM,0,s1,s1))!=Wym_EINVAL) {
     WY_THROW(r);
   }
 }
};

// Check basic operations on non-default object
//
static void t2(void)
{
 WyRet r;
 
 // non-default object for the test
 const WySockFile LocalObj(PF_LOCAL,SOCK_STREAM);
 CHK_NOT_DEFAULT(LocalObj);
 
 // stat LocalObj
 {
   WyFileStat stt;
   if((r=LocalObj.stat(stt))!=Ok) {
     WY_THROW(r);
   }
   if(stt.is_sock()==false) {
     WY_THROW( WyRet() );
   }
 }

 // reset
 {
   WySockFile tmp(LocalObj);
   CHK_NOT_DEFAULT(tmp);
   if((r=tmp.reset())!=Ok) {
     WY_THROW(r);
   }
   CHK_DEFAULT(tmp);
 }

 // reset(WySockFile())
 //
 {
   WySockFile tmp(LocalObj);
   CHK_NOT_DEFAULT(tmp);
   if((r=tmp.reset(WySockFile()))!=Ok) {
     WY_THROW(r);
   }
   CHK_DEFAULT(tmp);

   if((r=tmp.reset())!=Ok) {
     WY_THROW(r);
   }
   CHK_DEFAULT(tmp);
   if((r=tmp.reset(static_cast<const WyByteFlow&>(LocalObj)))!=Ok) {
     WY_THROW(r);
   }
   CHK_SAME_SYSF(tmp,LocalObj);
 }

 // reset(int,int,int=0)
 {
   WySockFile tmp(LocalObj);
   CHK_SAME_SYSF(tmp,LocalObj);
   if((r=tmp.reset(PF_LOCAL,SOCK_STREAM))!=Ok) {
     WY_THROW(r);
   }
   CHK_NOT_DEFAULT(tmp);
 }

 // reset(WyFileHandle())
 //
 {
   WySockFile tmp(LocalObj);
   CHK_NOT_DEFAULT(tmp);
   if((r=tmp.reset(WyFileHandle()))!=Ok) {
     WY_THROW(r);
   }
   CHK_DEFAULT(tmp);

   WySockFile tmp2;
   CHK_DEFAULT(tmp2);
   if((r=tmp2.reset(LocalObj.fh()))!=Ok) {
     WY_THROW(r);
   }   
   CHK_NOT_DEFAULT(tmp2);
   if(tmp2.family()!=LocalObj.family()) {
     WY_THROW( WyRet() );
   }
 }

 // (zero byte) read,write
 {
   WySockFile tmp(LocalObj);
   char buf[2]={-1,-1};
   WyStr str;
   size_t n_rd=5;   // anything >1
   CHK_NOT_DEFAULT(tmp);
   if((r=tmp.read(buf,0,n_rd))!=Ok) {
     WY_THROW(r);
   }
   if(n_rd!=0) {
     WY_THROW( WyRet() );
   }

   /* Write test disabled
      On Fedora core 8, Wym_ENOTCONN is returned while previous versions return Ok
   */
   /*
   if((r=tmp.write(buf,0,n_rd))!=Ok) {
     WY_THROW(r);
   }
   if(n_rd!=0) {
     WY_THROW( WyRet() );
   }
   if((r=tmp.write(WyStr(),n_rd))!=Ok) {
     WY_THROW(r);
   }
   if(n_rd!=0) {
     WY_THROW( WyRet() );
   }
   */
   //if((r=tmp.drain())!=Wym_EINVAL) {
   //  WY_THROW(r); // ref fsync(2)
   //}
   CHK_NOT_DEFAULT(tmp);
 }

 // shutdown,bind,listen,accept,connect,sockname,peername
 //
 {
   WySockFile tmp(LocalObj);
   const WySockAddrLocal addr;
   WySockAddrLocal addr2;
   CHK_NOT_DEFAULT(tmp);
   if(addr2!=addr) {
     WY_THROW( WyRet() );
   }
     
   // I am not sure what shoud be returned for the following functions
   // Coding takes whatever tested
   //
   if((r=tmp.shutdown(SHUT_RD))!=Ok) {  // test showed errno==0
     WY_THROW(r);
   }

   if((r=tmp.bind(addr))!=Ok) {         // test showed errno==0
     WY_THROW(r);
   }
   std::cout << "WySockFile (PF_LOCAL) bind a default address\n";

   if((r=tmp.listen(0))!=Ok) {         // test showed errno==0
     WY_THROW(r);
   }

   WySockFile tmp1;
   CHK_DEFAULT(tmp1);
   if((r=tmp.accept(tmp1,addr2))!=Wym_EINVAL) { // test showed errno=EINVAL
     WY_THROW(r);                      // and addr2 is modified
   }
   CHK_DEFAULT(tmp1);
   if(addr2.is_default()==false) {
     WY_THROW( WyRet() );
   }
   if(addr2.family()!=AF_LOCAL) {
     WY_THROW( WyRet() );
   }
   std::cout << "WySockFile (PF_LOCAL) accepted a default address\n";

   if((r=tmp.connect(addr))!=Wym_EINVAL) { // test showed errno==EINVAL
     WY_THROW(r);
   }

   if((r=tmp.sockname(addr2))!=Ok) {
     WY_THROW(r);
   }
   //if(addr!=addr2) {
   //  WY_THROW(r);
   //}
   if((r=tmp.peername(addr2))!=Wym_ENOTCONN) { // test showed errno==ENOTCONN
     WY_THROW(r);
   }
   //if(addr!=addr2) {
   //  WY_THROW(r);
   //}
   CHK_NOT_DEFAULT(tmp);
 }

 // chk move ctor
 {
  char raw_buf[2*sizeof(WySockFile)];
  ::memset(raw_buf,0,sizeof(raw_buf));
  WySockFile* const p1( reinterpret_cast<WySockFile*>(raw_buf) );
  WySockFile* const p2(p1+1);

  new(p1) WySockFile(PF_LOCAL,SOCK_STREAM);
  if(p1->family()!=PF_LOCAL) {
    p1->~WySockFile();
    WY_THROW( WyRet() );
  }

  // move *p1 to p2
  WyFileHandle ofh( p1->fh() );
  int ofamily( p1->family() );
  try {
    new(p2) WySockFile(*p1,Wy::ByMove);
  }
  catch(...) {
    p1->~WySockFile();
    throw;
  };
  WyFileHandle nfh( p2->fh() );
  int nfamily( p2->family() );
  p2->~WySockFile();

  if(nfh!=ofh) {
    WY_THROW( WyRet() );
  }
  if(nfamily!=ofamily) {
    WY_THROW( WyRet() );
  }
 }
};

static void t_exception(void)
{
 //
 // WySockFile(const WySockFile&)
 //
 try {
   const WySockFile tmp(PF_LOCAL,SOCK_STREAM);
   exaust_cpy_ctor(tmp);
   WY_THROW( WyRet() );  // not expected
 }
 catch(const WySockFile::Reply& e) {
   if((e!=Wym_EMFILE)&&(e!=Wym_ENOMEM)) {
     WyRet r=e;
     WY_THROW(r);
   }
   // FALL_THROUGH 
 };

 // WySockFile(int,int,int)
 try {
   WySockFile(PF_LOCAL,SOCK_STREAM,3);
   WY_THROW( WyRet() );
 }
 catch(const WySockFile::Reply& e) {
   if(e!=Wym_EPROTONOSUPPORT) {
     WY_THROW( WyRet(e) );
   }
   // FALL_THROUGH 
 };

 // WySockFile(WyFileHandle)
 try {
   WySockFile tmp(WyFileHandle(-2));
   WY_THROW( WyRet() );
 }
 catch(const WySockFile::Reply& e) {
   if(e!=Wym_EBADF) {
     WY_THROW( WyRet(e) );
   }
   // FALL_THROUGH 
 };

 // Test destructor exception
 //
 try {
   WySockFile tmp;
   std::memset(&tmp,-2,sizeof(tmp));
   // destructor of tmp is now not to cause exception for this setting
 }
 catch(...) {
   throw;
 };

};

static void t_error(void)
{
 WyRet r;
 WySockFile LocalObj(PF_LOCAL,SOCK_STREAM);

 // stat
 {
  WySockFile tmp;
  WyFileStat stt;
  if((r=tmp.stat(stt))!=Wym_EBADF) {
    WY_THROW(r);
  }
 }

 // reset(const WySockFile&)
 {
   WySockFile tmp(LocalObj);
   CHK_SAME_SYSF(tmp,LocalObj);
   if((r=tmp.reset(LocalObj))!=Ok) {
     WY_THROW(r);
   }
   CHK_SAME_SYSF(tmp,LocalObj);

   const WyFileHandle TmpFh( tmp.fh() );
   if((r=tmp.reset(tmp))!=Ok) {
     WY_THROW(r);
   }
   CHK_SAME_SYSF(tmp,LocalObj);
   if(tmp.fh()!=TmpFh) {
     // ok, equ or not
   }
 }

 // reset(int,int,int)
 {
   WySockFile tmp(LocalObj);
   CHK_SAME_SYSF(tmp,LocalObj);

   if((r=tmp.reset(PF_LOCAL,SOCK_STREAM,3))!=Wym_EPROTONOSUPPORT) {
     WY_THROW(r);
   }
   CHK_SAME_SYSF(tmp,LocalObj);

   if((r=tmp.reset(PF_LOCAL,SOCK_STREAM,0))!=Ok) {
     WY_THROW(r);
   }
   CHK_NOT_DEFAULT(tmp);
 }

 // reset(WyFileHandle)
 {
   WySockFile tmp(LocalObj);
   if((r=tmp.reset(LocalObj.fh()))!=Ok) {
     WY_THROW(r);
   }
   CHK_SAME_SYSF(tmp,LocalObj);
 }

 // read/write
 {
   WySockFile tmp;
   char buf[2];
   size_t n_rd,n_wr;
   WyStr str;
   n_rd=4;
   if((tmp.read(buf,sizeof(buf),n_rd))!=Wym_EBADF) {
     WY_THROW(r);
   }
   CHK_DEFAULT(tmp);
   if(n_rd!=0) {
     WY_THROW( WyRet() );
   }
   if((tmp.read(str,sizeof(buf),n_rd))!=Wym_EBADF) {
     WY_THROW(r);
   }
   if(n_rd!=0) {
     WY_THROW( WyRet() );
   }
   CHK_DEFAULT(tmp);
   n_wr=6;
   if((tmp.write(buf,sizeof(buf),n_wr))!=Wym_EBADF) {
     WY_THROW(r);
   }
   CHK_DEFAULT(tmp);
   if(n_wr!=0) {
     WY_THROW( WyRet() );
   }
   n_wr=6;
   if((tmp.write(str,n_wr))!=Wym_EBADF) {
     WY_THROW(r);
   }
   CHK_DEFAULT(tmp);
   if(n_wr!=0) {
     WY_THROW( WyRet() );
   }

   //if((tmp.drain())!=Wym_EBADF) {
   //  WY_THROW(r);
   //}
   CHK_DEFAULT(tmp);
 }

 // shutdown/
 {
   WySockFile tmp,tmp2;
   WySockAddrLocal addr;
   if((r=tmp.shutdown(SHUT_RD))!=Wym_EBADF) {
     WY_THROW(r);
   }
   if((r=tmp.bind(addr))!=Wym_EBADF) {
     WY_THROW(r);
   }
   if((r=tmp.listen(0))!=Wym_EBADF) {
     WY_THROW(r);
   }
   if((r=tmp.accept(tmp2,addr))!=Wym_EBADF) {
     WY_THROW(r);
   }
   if((r=tmp.connect(addr))!=Wym_EBADF) {
     WY_THROW(r);
   }
   if((r=tmp.sockname(addr))!=Wym_EBADF) {
     WY_THROW(r);
   }
   if((r=tmp.peername(addr))!=Wym_EBADF) {
     WY_THROW(r);
   }
   CHK_DEFAULT(tmp);
   if(addr.is_default()==false) {
     WY_THROW( WyRet() );
   }
 }

 // EFBIG
 {
  WySockFile tmp(LocalObj);
  WyStr buf;
  size_t n_rd(7);
  CHK_NOT_DEFAULT(tmp);
  if((r=tmp.read(buf,WyStr::max_capacity()+1,n_rd))!=Wym_EFBIG) {
    WY_THROW(r);
  }
  if(n_rd!=0) {
    WY_THROW( WyRet() );
  }
  n_rd=88;
  if((r=tmp.read(buf,std::numeric_limits<size_t>::max(),n_rd))!=Wym_EFBIG) {
    WY_THROW(r);
  }
  if(n_rd!=0) {
    WY_THROW( WyRet() );
  }
 }

 // chk _alloc not override
 {
   class T : public WySockFile {
   } tmp;
   WyByteFlow* p=tmp._alloc(r);
   if(r!=Wym_ENOSYS) {
     if(r==Ok) {
       delete p;
     }
     WY_THROW(r);
   }
 }
};

static void t_local_mis(void)
try {
 // chk sockname(..) result consistent
 //
 const WySockAddrLocal Addr(LocalSockFileName);
 WySockFile sock(PF_LOCAL,SOCK_STREAM,0);
 WyRet r;

 Wy::unlink(LocalSockFileName);

 if((r=sock.bind(Addr))!=Ok) {
   WY_THROW(r);
 }

 WySockAddrLocal sname;
 if((r=sock.sockname(sname))!=Ok) {
   WY_THROW(r);
 }

 if(sname.addrname()!=LocalSockFileName) {
   WY_THROW( WyRet() );
 }

 // Note: WySockFile is derived from WyByteFlow, and does not override
 //       reset(const char*,int). chk access from the base get ENXIO
 //
 {
   WySockFile tmp;
   WyByteFlow *bfp(&tmp);
   if((r=bfp->reset(LocalSockFileName,O_RDONLY))!=Wym_ENXIO) {
     WY_THROW( WyRet(r) );
   }
   if((r=bfp->reset(LocalSockFileName.c_str(),O_RDONLY))!=Wym_ENXIO) {
     WY_THROW( WyRet(r) );
   }
 }
 Wy::unlink(LocalSockFileName);
}
catch(...) {
 Wy::unlink(LocalSockFileName);
 throw;
};

static void t_local_mis2(void)
{
 // chk MaxAddrLen
 //
 const WyStr MLName(WySockAddrLocal::MaxAddrLen,'a');
 const WySockAddrLocal Addr(MLName);
 WySockFile sock(PF_LOCAL,SOCK_STREAM,0);
 WyRet r;

 Wy::unlink(MLName);

 if((r=sock.bind(Addr))!=Ok) {
   WY_THROW(r);
 }

 try {
   WySockAddrLocal sname;
   if((r=sock.sockname(sname))!=Ok) {
     WY_THROW(r);
   }

   if(sname.addrname()!=MLName) {
     WY_THROW( WyRet() );
   }
 }
 catch(...) {
   Wy::unlink(MLName);
   throw;
 };

 Wy::unlink(MLName);
};

static void t_inet4_mis(void)
{
 // chk sockname(..) result consistent
 //
 const uint32_t LBAddr=0x7F000001;   // 127.0.0.1
 const WySockAddrINet4::IPAddr IpAddr(LBAddr);
 const WySockAddrINet4 InAddr(IpAddr,TestINetPort);
 WySockFile sock(PF_INET,SOCK_STREAM,0);
 WyRet r;

 if((r=sock.bind(InAddr))!=Ok) {
   WY_THROW(r);
 }

 WySockAddrINet4 sname;
 if((r=sock.sockname(sname))!=Ok) {
   WY_THROW(r);
 }
 if(sname.ip()!=IpAddr) {
   WY_THROW( WyRet() );
 }

 {
  WyStr str;
  if((r=IpAddr.notation(str))!=Ok) {
    WY_THROW( WyRet() );
  }
  if(str!=TestINetIP4) {
    std::cerr << str.c_str() << '\n';
    WY_THROW( WyRet() );
  }
 }
};

// Test WySockFile communication (AF_LOCAL)
//
// Connect to LocalSockFileName and receieve SvrMsg, then send CliMsg 
// and exit(0).
//
static int process_client_local(void) throw()
try {
 WyRet r;
 WySockFile cli;
 const WySockAddrLocal Saddr(LocalSockFileName);
 
 // Open socket (PF_LOCAL) and connect to the server
 //
 if((r=cli.reset(PF_LOCAL,SOCK_STREAM))!=Ok) {
   WY_THROW(r);
 }
 if((r=cli.connect(Saddr))!=Ok) {
   WY_THROW(r);
 }

 // stat
 {
   WyFileStat stt;
   if((r=cli.stat(stt))!=Ok) {
     WY_THROW(r);
   }
   if(stt.is_sock()==false) {
     WY_THROW( WyRet() );
   }
 }

 // Dump addr
 {
  WyStr str(WyCSeg(Saddr.addrname(),Saddr.addrsize()));
  std::cout << "process_client_local connected to: " << str.c_str()
           << std::endl;
 }

 // chk peername
 {
  WySockAddrLocal taddr;
  if((r=cli.peername(taddr))!=Ok) {
    WY_THROW(r);
  }
  if(taddr!=WyCSeg(Saddr.addrname(),Saddr.addrsize())) {
    if(taddr.addrname()!=Saddr) {
      WY_THROW(r);
    }
    std::cout << "peername() has one more char(0) added\n";
  }
 }

 // Receive message from the server (SvrMsg expected)
 //
 size_t n_rd;
 char rbuf[64];
 if(sizeof(rbuf)<SvrMsg.size()+1) {
   WY_THROW( WyRet() );   // rbuf not long enough, reedit rbuf size.
 }
 if((r=cli.read(rbuf,sizeof(rbuf),n_rd))!=Ok) {
   WY_THROW(r);
 }
 if(n_rd!=SvrMsg.size()) {
   WY_THROW( WyRet() ); 
 }
 if(Wy::_strcmp(SvrMsg,WyCSeg(rbuf,n_rd))!=0) {
   WY_THROW( WyRet() );   // not expected server message
 }
 
 // Send message CliMsg to the server
 //
 size_t n_wr;
 if((r=cli.write(CliMsg.begin(),CliMsg.size(),n_wr))!=Ok) {
   WY_THROW(r);
 }
 if(n_wr!=CliMsg.size()) {
   WY_THROW( WyRet() ); 
 }

 return(0);
}
catch(const WyRet& e) {
 std::cerr << "process_client_local: "  << Wy::wrd(e).c_str() << std::endl;
 return(-1);
}
catch(...) {
 std::cerr << "process_client_local : caught throw" << std::endl;
 return(-1);
};

// Test WySockFile communication (AF_LOCAL)
// t_comm_local fork process_client_local and 
//
static void t_comm_local(void)
try {
 WyRet r;
 pid_t child_pid;
 int child_status;

 // Create server socket (PF_LOCAL) LocalSockFileName
 //
 ::unlink(LocalSockFileName.c_str());
 WySockAddrLocal saddr(LocalSockFileName);
 WySockFile svr;

 if((r=svr.reset(PF_LOCAL,SOCK_STREAM))!=Ok) {
   WY_THROW(r);
 }
 if((r=svr.bind(saddr))!=Ok) {
   WY_THROW(r);
 }
 if((r=svr.listen(1))!=Ok) {
   WY_THROW(r);
 }

 // stat
 {
   WyFileStat stt;
   if((r=svr.stat(stt))!=Ok) {
     WY_THROW(r);
   }
   if(stt.is_sock()==false) {
     WY_THROW( WyRet() );
   }
   std::cout << "socket(Local), _blksize= " << stt.sf_mode() << '\n';
 }

 // Fork child process
 //
 child_pid=::fork();
 if(child_pid<0) {
   WY_THROW( WyRet(WyReply(errno)) );
 }
 if(child_pid==0) {
   ::exit(process_client_local());
 }

 // Server accepts connection
 //
 WySockFile psock;
 WySockAddrLocal paddr;
 if((r=svr.accept(psock,paddr))!=Ok) {
   WY_THROW(r);
 } 

 // test showed default address returned (probable right for PF_LOCAL?)
 //
 if(paddr.is_default()==false) {
   // Dump the accept addr
   {
    WyStr str( WyCSeg(paddr.addrname(),paddr.addrsize()) );
    std::cout << "server(PF_LOCAL) accept() address: " << str.c_str() << std::endl;
   }
   WY_THROW( WyRet() );
 }

 // Send SvrMsg to the client
 //
 size_t n_wr;
 if((r=psock.write(SvrMsg.begin(),SvrMsg.size(),n_wr))!=Ok) {
   WY_THROW(r);
 }
 if(n_wr!=SvrMsg.size()) {
   WY_THROW( WyRet() ); 
 }

 // Receive message from the client (CliMsg expected)
 //
 size_t n_rd;
 char rbuf[64];
 if(sizeof(rbuf)<CliMsg.size()+1) {
   WY_THROW( WyRet() );   // rbuf not long enough
 }

 if((r=psock.read(rbuf,sizeof(rbuf),n_rd))!=Ok) {
   WY_THROW(r);
 }
 if(n_rd!=CliMsg.size()) {
   WY_THROW( WyRet() ); 
 }
 if(Wy::_strcmp(CliMsg,WyCSeg(rbuf,n_rd))!=0) {
   WY_THROW( WyRet() ); 
 }
 
 if(::wait(&child_status)!=child_pid) {
   WY_THROW( WyRet(WyReply(errno)) );
 }
 if(WIFEXITED(child_status)==0) {
   WY_THROW( WyRet() );  // child process was not terminated voluntarily
 }
 if(WEXITSTATUS(child_status)!=0) {
   WY_THROW( WyRet() );  // child process not exist(0)
 }

 // EOF test (Ok/zero byte)
 n_rd=99;
 if((r=psock.read(rbuf,sizeof(rbuf),n_rd))!=Ok) {
   WY_THROW(r);
 }
 if(n_rd!=0) {
   WY_THROW( WyRet() ); 
 }
 n_rd=99;
 if((r=psock.read(rbuf,sizeof(rbuf),n_rd))!=Ok) {
   WY_THROW(r);
 }
 if(n_rd!=0) {
   WY_THROW( WyRet() ); 
 }

 ::unlink(LocalSockFileName.c_str());
}
catch(...) {
 ::unlink(LocalSockFileName.c_str());
 throw;
};

#ifdef WY__TEST_SOCK_INET__

// Test WySockFile communication (AF_INET)
//
// Connect to TestINetIP4 and receieve SvrMsg, then send CliMsg 
// and exit(0).
//
static int process_client_inet4(void) throw()
try {
 WyRet r;
 WySockFile cli;
 WySockAddrINet4 saddr( WySockAddrINet4::IPAddr(TestINetIP4),TestINetPort);

 // Open socket (PF_INET) and connect to the server
 //
 if((r=cli.reset(PF_INET,SOCK_STREAM))!=Ok) {
   WY_THROW(r);
 }
 if((r=cli.connect(saddr))!=Ok) {
   WY_THROW(r);
 }

 // stat
 {
   WyFileStat stt;
   if((r=cli.stat(stt))!=Ok) {
     WY_THROW(r);
   }
   if(stt.is_sock()==false) {
     WY_THROW( WyRet() );
   }
 }

 // Dump addr
 {
  WyStr str;
  WyRet r=saddr.ip().notation(str);
  if(r!=Ok) {
    WY_THROW(r);
  }
  if((r=saddr.ip().notation(str))!=Ok) {
    WY_THROW( WyRet() );
  }
  std::cout << "process_client_inet4 connected to: " << str.c_str() << std::endl;
 }

 // Receive message from the server (SvrMsg expected)
 //
 size_t n_rd;
 char rbuf[64];
 if(sizeof(rbuf)<SvrMsg.size()+1) {
   WY_THROW( WyRet() );   // rbuf not long enough, reedit rbuf size.
 }
 if((r=cli.read(rbuf,sizeof(rbuf),n_rd))!=Ok) {
   WY_THROW(r);
 }
 if(n_rd!=SvrMsg.size()) {
   WY_THROW( WyRet() ); 
 }
 if(Wy::_strcmp(SvrMsg,WyCSeg(rbuf,n_rd))!=0) {
   WY_THROW( WyRet() ); 
 }
 
 // Send message CliMsg to the server
 //
 size_t n_wr;
 if((r=cli.write(CliMsg.begin(),CliMsg.size(),n_wr))!=Ok) {
   WY_THROW(r);
 }
 if(n_wr!=CliMsg.size()) {
   WY_THROW( WyRet() ); 
 }

 return(0);
}
catch(const WyRet& e) {
 std::cerr << "process_client_inet4: "  << Wy::wrd(e).c_str() << std::endl;
 return(-1);
}
catch(...) {
 std::cerr << "process_client_inet4 : caught throw" << std::endl;
 return(-1);
};

// Test WySockFile communication (AF_INET)
// t_comm_inet4 fork process_client_inet4 and runs as the server
// process_client_inet4 will run as the client
//
static void t_comm_inet4(void)
{
 WyRet r;
 pid_t child_pid;
 int child_status;

 // Create server socket
 //
 WySockFile svr;
 WySockAddrINet4 saddr( WySockAddrINet4::IPAddr(TestINetIP4),TestINetPort);

 if((r=svr.reset(PF_INET,SOCK_STREAM))!=Ok) {
   WY_THROW(r);
 }
 if((r=svr.bind(saddr))!=Ok) {
   WY_THROW(r);
 }
 if((r=svr.listen(1))!=Ok) {
   WY_THROW(r);
 }

 // stat
 {
   WyFileStat stt;
   if((r=svr.stat(stt))!=Ok) {
     WY_THROW(r);
   }
   if(stt.is_sock()==false) {
     WY_THROW( WyRet() );
   }
   std::cout << "stream socket(ipv4), _blksize= " << stt.sf_mode() << '\n';
 }

 // Fork child process
 //
 child_pid=::fork();
 if(child_pid<0) {
   WY_THROW( WyRet(WyReply(errno)) );
 }
 if(child_pid==0) {
   ::exit(process_client_inet4());
 }

 // Server accepts connection
 //
 WySockFile psock;
 WySockAddrINet4 paddr;
 if((r=svr.accept(psock,paddr))!=Ok) {
   WY_THROW(r);
 }
 if(paddr.is_default()) {
   WY_THROW( WyRet() );
 }

 // Dump the accept addr
 {
   WyStr str;
   if((r=paddr.ip().notation(str))!=Ok) {
     WY_THROW(r);
   }
  if((r=paddr.ip().notation(str))!=Ok) {
    WY_THROW(r);
  }
   std::cout << "server(PF_INET) accept() address: " << str.c_str() << std::endl;

   if((r=psock.sockname(paddr))!=Ok) {
     WY_THROW(r);
   }
   if((r=paddr.ip().notation(str))!=Ok) {
     WY_THROW(r);
   }
  if( Wy::wrd(paddr.ip())!=str) {
    WY_THROW( WyRet() );
  }
   std::cout << " accepted..sockname()= " << str.c_str() << std::endl;

   if((r=psock.peername(paddr))!=Ok) {
     WY_THROW(r);
   }
   if((r=paddr.ip().notation(str))!=Ok) {
     WY_THROW(r);
   }
  if((r=paddr.ip().notation(str))!=Ok) {
    WY_THROW(r);
  }
   std::cout << " accepted..peername()= " << str.c_str() << std::endl;
 }

 // Send SvrMsg to the client
 //
 size_t n_wr;
 if((r=psock.write(SvrMsg,n_wr))!=Ok) {
   WY_THROW(r);
 }
 if(n_wr!=SvrMsg.size()) {
   WY_THROW( WyRet() ); 
 }

 // Receive message from the client (CliMsg expected)
 //
 size_t n_rd;
 char rbuf[64];
 if(sizeof(rbuf)<CliMsg.size()+1) {
   WY_THROW( WyRet() );   // rbuf not long enough
 }
 if((r=psock.read(rbuf,sizeof(rbuf),n_rd))!=Ok) {
   WY_THROW(r);
 }
 if(n_rd!=CliMsg.size()) {
   WY_THROW( WyRet() ); 
 }
 if(Wy::_strcmp(CliMsg,WyCSeg(rbuf,n_rd))!=0) {
   WY_THROW( WyRet() ); 
 }
 
 if(::wait(&child_status)!=child_pid) {
   WY_THROW( WyRet(WyReply(errno)) );
 }
 if(WIFEXITED(child_status)==0) {
   WY_THROW( WyRet() );  // child process was not terminated voluntarily
 }
 if(WEXITSTATUS(child_status)!=0) {
   WY_THROW( WyRet() );  // child process not exist(0)
 }

 // EOF test (Ok/zero byte)
 n_rd=99;
 if((r=psock.read(rbuf,sizeof(rbuf),n_rd))!=Ok) {
   WY_THROW(r);
 }
 if(n_rd!=0) {
   WY_THROW( WyRet() ); 
 }
 n_rd=99;
 if((r=psock.read(rbuf,sizeof(rbuf),n_rd))!=Ok) {
   WY_THROW(r);
 }
 if(n_rd!=0) {
   WY_THROW( WyRet() ); 
 }

};

//-----------
//
// Test WySockFile communication (AF_INET6)
//
// Connect to TestINetIP6 and receieve SvrMsg, then send CliMsg 
// and exit(0).
//
static int process_client_inet6(void) throw()
try {
 WyRet r;
 WySockFile cli;
 WySockAddrINet6 saddr(TestINetIP6,TestINetPort);

 // Open socket (PF_INET6) and connect to the server
 //
 if((r=cli.reset(PF_INET6,SOCK_STREAM))!=Ok) {
   WY_THROW(r);
 }
 if((r=cli.connect(saddr))!=Ok) {
   WY_THROW(r);
 }

 // stat
 {
   WyFileStat stt;
   if((r=cli.stat(stt))!=Ok) {
     WY_THROW(r);
   }
   if(stt.is_sock()==false) {
     WY_THROW( WyRet() );
   }
 }

 // Dump addr
 {
  WyStr str;
  WyRet r=saddr.ip().notation(str);
  if(r!=Ok) {
    WY_THROW(r);
  }
  if( Wy::wrd(saddr.ip())!=str) {
    WY_THROW( WyRet() );
  }
  std::cout << "process_client_inet6 connected to: " << str.c_str() << std::endl;
 }

 // Receive message from the server (SvrMsg expected)
 //
 size_t n_rd;
 char rbuf[64];
 if(sizeof(rbuf)<SvrMsg.size()+1) {
   WY_THROW( WyRet() );   // rbuf not long enough, reedit rbuf size.
 }
 if((r=cli.read(rbuf,sizeof(rbuf),n_rd))!=Ok) {
   WY_THROW(r);
 }
 if(n_rd!=SvrMsg.size()) {
   WY_THROW( WyRet() ); 
 }
 if(Wy::_strcmp(SvrMsg,WyCSeg(rbuf,n_rd))!=0) {
   WY_THROW( WyRet() ); 
 }
 
 // Send message CliMsg to the server
 //
 size_t n_wr;
 if((r=cli.write(CliMsg.begin(),CliMsg.size(),n_wr))!=Ok) {
   WY_THROW(r);
 }
 if(n_wr!=CliMsg.size()) {
   WY_THROW( WyRet() ); 
 }

 return(0);
}
catch(const WyRet& e) {
 std::cerr << "process_client_inet6: "  << Wy::wrd(e).c_str() << std::endl;
 return(-1);
}
catch(...) {
 std::cerr << "process_client_inet6 : caught throw" << std::endl;
 return(-1);
};

// Test WySockFile communication (AF_INET6)
// t_comm_inet6 fork process_client_inet6 and runs as the server
// process_client_inet6 will run as the client
//
static void t_comm_inet6(void)
{
 WyRet r;
 pid_t child_pid;
 int child_status;

 // Create server socket
 //
 WySockFile svr;
 WySockAddrINet6 saddr(TestINetIP6,TestINetPort);

 if((r=svr.reset(PF_INET6,SOCK_STREAM))!=Ok) {
   WY_THROW(r);
 }
 if((r=svr.bind(saddr))!=Ok) {
   WY_THROW(r);
 }
 if((r=svr.listen(1))!=Ok) {
   WY_THROW(r);
 }

 // stat
 {
   WyFileStat stt;
   if((r=svr.stat(stt))!=Ok) {
     WY_THROW(r);
   }
   if(stt.is_sock()==false) {
     WY_THROW( WyRet() );
   }
   std::cout << "stream socket(ipv6), _blksize= " << stt.sf_mode() << '\n';
 }

 // Fork child process
 //
 child_pid=::fork();
 if(child_pid<0) {
   WY_THROW( WyRet(WyReply(errno)) );
 }
 if(child_pid==0) {
   ::exit(process_client_inet6());
 }

 // Server accepts connection
 //
 WySockFile psock;
 WySockAddrINet6 paddr;
 if((r=svr.accept(psock,paddr))!=Ok) {
   WY_THROW(r);
 }
 if(paddr.is_default()) {
   WY_THROW( WyRet() );
 }

 // Dump the accept addr
 {
   WyStr str;
   if((r=paddr.ip().notation(str))!=Ok) {
     WY_THROW(r);
   }
  if( Wy::wrd(paddr.ip())!=str) {
    WY_THROW( WyRet() );
  }
   std::cout << "server(PF_INET6) accept() address: " << str.c_str() << std::endl;

   if((r=psock.sockname(paddr))!=Ok) {
     WY_THROW(r);
   }
   if((r=paddr.ip().notation(str))!=Ok) {
     WY_THROW(r);
   }
  if( Wy::wrd(paddr.ip())!=str) {
    WY_THROW( WyRet() );
  }
   std::cout << " accepted..sockname()= " << str.c_str() << std::endl;

   if((r=psock.peername(paddr))!=Ok) {
     WY_THROW(r);
   }
   if((r=paddr.ip().notation(str))!=Ok) {
     WY_THROW(r);
   }
  if(Wy::wrd(paddr.ip())!=str) {
    WY_THROW( WyRet() );
  }
   std::cout << " accepted..peername()= " << str.c_str() << std::endl;
 }

 // Send SvrMsg to the client
 //
 size_t n_wr;
 if((r=psock.write(SvrMsg,n_wr))!=Ok) {
   WY_THROW(r);
 }
 if(n_wr!=SvrMsg.size()) {
   WY_THROW( WyRet() ); 
 }

 // Receive message from the client (CliMsg expected)
 //
 size_t n_rd;
 char rbuf[64];
 if(sizeof(rbuf)<CliMsg.size()+1) {
   WY_THROW( WyRet() );   // rbuf not long enough
 }
 if((r=psock.read(rbuf,sizeof(rbuf),n_rd))!=Ok) {
   WY_THROW(r);
 }
 if(n_rd!=CliMsg.size()) {
   WY_THROW( WyRet() ); 
 }
 if(Wy::_strcmp(CliMsg,WyCSeg(rbuf,n_rd))!=0) {
   WY_THROW( WyRet() ); 
 }
 
 if(::wait(&child_status)!=child_pid) {
   WY_THROW( WyRet(WyReply(errno)) );
 }
 if(WIFEXITED(child_status)==0) {
   WY_THROW( WyRet() );  // child process was not terminated voluntarily
 }
 if(WEXITSTATUS(child_status)!=0) {
   WY_THROW( WyRet() );  // child process not exist(0)
 }
}
#endif

static WyStr chdr("+---------------+\n"
                  "| main() caught:|\n"
                  "+---------------+\n");
int main(void) throw()
try {
 std::cout << "Checking WySockFile.h ...\n";
 if(Wy::_strcmp(WySockFile::class_name,"WySockFile")!=0) {
   WY_THROW( WyRet() );
 }
 try {
   setup_test();
   t1();
   t2();
   t_exception();
   t_error();
   t_local_mis();
   t_local_mis2();
   t_inet4_mis();
   t_comm_local();
#ifdef WY__TEST_SOCK_INET__
   try {
     t_comm_inet4(); 
   }
   catch(const WyRet& e) {
     std::cerr << "\at_comm_inet4() failed: (Beep: IPv4 may have not configured)\n"
               << Wy::wrd(e).c_str() << std::endl;
   };
   try {
     t_comm_inet6(); 
   }
   catch(const WyRet& e) {
     std::cerr << "\at_comm_inet6() failed: (Beep: IPv6 may have not configured)\n"
               << Wy::wrd(e).c_str() << std::endl;
   };
#endif
   close_test();
 }
 catch(...) {
   close_test();
   throw;
 }
 std::cout << "Checked Ok\n";
 return(0);
}
catch(const WyRet& e) {
 std::cerr << chdr.c_str() << Wy::wrd(e).c_str() << std::endl;
 return(-1);
}
catch(const std::exception& e) {
 std::cerr << chdr.c_str() << "std::exception" << std::endl;
 return(-1);
}
catch(...) {
 std::cerr << chdr.c_str() << "unknown unwind" << std::endl;
 return(-1);
};
