/* Copyright is licensed under GNU LGPL.                 by I.J.Wang 2007
*/
#include "WyTimeSpec.h"
#include "WyStr.h"
#include <iostream>
#include <sstream>
#include <string>
#include <limits>
#include <cmath>
//#include <sys/types.h>

#if WYTIMESPEC_VERSION!=33
#error Test code is for WYTIMESPEC_VERSION 33
#endif

static const double Nan=::sqrt(-1);
static const double Inf=::pow(10,1e999999);
static const int RandSmps=10000;

struct ChkItm1 {        // check WyTimeSpec(double)
  const double SVal;

  // Expected results
  const WyReply res_reply;
  const WyTimeSpec::SecondType res_sfloor;
  const WyTimeSpec::SecondType res_sceil;
  const WyTimeSpec::NanoSecType res_nfloor;
  const WyTimeSpec::NanoSecType res_nceil;

  void verify(void) const
  try {

    // chk WyTimeSpec(double)
    try {
      WyTimeSpec Tobj(SVal);
      if(res_reply!=Ok) {
        WY_THROW( WyRet() );
      }
      if((Tobj.second()<res_sfloor)||(Tobj.second()>res_sceil)) {
        std::cerr << "Tob.second()=" << Tobj.second() << '\n';
        WY_THROW( WyRet() );
      }
      if((Tobj.nano()<res_nfloor)||(Tobj.nano()>res_nceil)) {
        std::cerr << "Tob.nano()=" << Tobj.nano() << '\n';
        WY_THROW( WyRet() );
      }
    }
    catch(const WyTimeSpec::Reply& e) {
      if(e!=res_reply) {
        WY_THROW( WyRet(e) );
      }
    };

    // chk reset(double)
    {
     WyTimeSpec tobj;
     WyRet r=tobj.reset(SVal);
     if(r!=res_reply) {
       WY_THROW( WyRet(r) );
     }
     if(r==Ok) {
       if((tobj.second()<res_sfloor)||(tobj.second()>res_sceil)) {
         std::cerr << "Tob.second()=" << tobj.second() << '\n';
         WY_THROW( WyRet() );
       }
       if((tobj.nano()<res_nfloor)||(tobj.nano()>res_nceil)) {
         std::cerr << "Tob.nano()=" << tobj.nano() << '\n';
         WY_THROW( WyRet() );
       }
     } else {
       if(tobj.is_default()==false) {
         WY_THROW( WyRet() );
       }
     }
    }

    // chk operator=(double)
    try {
      WyTimeSpec tobj;
      tobj=SVal;
      if(res_reply!=Ok) {
        WY_THROW( WyRet() );
      }
      if((tobj.second()<res_sfloor)||(tobj.second()>res_sceil)) {
        std::cerr << "Tob.second()=" << tobj.second() << '\n';
        WY_THROW( WyRet() );
      }
      if((tobj.nano()<res_nfloor)||(tobj.nano()>res_nceil)) {
        std::cerr << "Tob.nano()=" << tobj.nano() << '\n';
        WY_THROW( WyRet() );
      }
    }
    catch(const WyTimeSpec::Reply& e) {
      if(e!=res_reply) {
        WY_THROW( WyRet(e) );
      }
    };

  } catch(...) {
    std::cerr << "Item={" << SVal << ", " << Wy::wrd(res_reply).c_str() << ", "
              << res_sfloor << ", " << res_sceil << ", " << res_nfloor << ", " << res_nceil
              << "}\n";
    throw;
  };
};

static void ck_double(void)
{
 const ChkItm1 chk_tab[]={
   // sval reply sfloor sceil nfloor nceil
   { 0.0, Ok, 0,0, 0,0},
   { 1.23, Ok, 1,1, 229000000,231000000},
   {-1.23, Ok, -1,-1,-231000000,-229000000},
   { 123e-3, Ok, 0,0, 122900000,123100000},
   { 1234e-6, Ok, 0,0, 1233900,1234100},
   { 4321e-9, Ok, 0,0, 4320,4322},
   { 1.23e-3, Ok, 0,0, 1229000,1231000},
   { 12.34e-6, Ok, 0,0, 12339,12341},
   { 432.1e-9, Ok, 0,0, 431,433},
   { 432.1e-19, Ok, 0,0, 0,1},
   { 432.1e+19, Wym_ERANGE, 0,0, 0,1},
   { 1e+90, Wym_ERANGE, 0,0, 0,1},
   { -1e+90, Wym_ERANGE, 0,0, 0,1},
   { 2147483647.0, Ok, 2147483647,2147483647, 0,1},
   { 0.2147483647, Ok, 0,0, 214748364,214748365},
   { 2147483649.0, Wym_ERANGE, 2147483647,2147483647, 0,1},
   { 9999999999e-5, Ok, 99999,100000, 999980000, 999991000},
   { 2147483648e-10, Ok, 0,0, 214748364,214748365},   // this is implement limits
   { -2147483647.0, Ok, -2147483647,-2147483647, 0,1},
   { Nan, Wym_ERANGE, 0,0, 0,0},
   { Inf, Wym_ERANGE, 0,0, 0,0},
 };
 
 for(size_t i=0; i<sizeof(chk_tab)/sizeof(ChkItm1); ++i) {
   chk_tab[i].verify();
 }

 // Random number check
 typedef double ResType;
 const long long RandExp=64;

 for(int i=0; i<RandSmps; ++i) {
   //  res= x*2^y  ,x in [-1.0,+1.0], y in [-RandExp, +RandExp]
   //
   const ResType Res= std::ldexp( ::rand()*(2.0/(double)RAND_MAX)-1.0
              , (::rand()*(long long)2*RandExp)/(long long)RAND_MAX-RandExp 
             );
   WyRet r;
   if(Res>Wy::to_double(WyTimeSpec::max())||(Res<Wy::to_double(WyTimeSpec::min()))) {
     ChkItm1 itm={ Res, Wym_ERANGE, 0,0, 0,0};
     itm.verify();
     if((Res<std::numeric_limits<WyTimeSpec::SecondType>::max()+1.0)&&
        (Res>std::numeric_limits<WyTimeSpec::SecondType>::min()-1.0)) {
       std::cout << "Error: Res=" << Res << '\n';
     }
   } else {
     ResType it,fr=std::modf(Res,&it);
     fr*=1e9;
     ChkItm1 itm={ Res, Ok, static_cast<WyTimeSpec::SecondType>(it-1),
                            static_cast<WyTimeSpec::SecondType>(it+1),
                            static_cast<WyTimeSpec::NanoSecType>(fr-1),
                            static_cast<WyTimeSpec::NanoSecType>(fr+1) };
     itm.verify();
   } 
 }
 
};

struct ChkItm_add_ts {        // check WyTimeSpec additions
 typedef double DType;
  const WyTimeSpec SVal_1;
  const WyTimeSpec SVal_2;

  // Expected results
  const WyReply res_reply;
  const WyTimeSpec res_floor;
  const WyTimeSpec res_ceil;

  void verify(void) const
  try {

    // chk operator +=(const WyTimeSpec&)
    try {
      WyTimeSpec tobj(SVal_1);
      tobj+= SVal_2;
      if(res_reply!=Ok) {
        WY_THROW( WyRet() );
      }
      if((tobj<res_floor)||(tobj>res_ceil)) {
        std::cerr << "Res=" << Wy::wrd(tobj).c_str() << '\n';
        WY_THROW( WyRet() );
      }
    }
    catch(const WyTimeSpec::Reply& e) {
      if(e!=res_reply) {
        WY_THROW( WyRet(e) );
      }
    };

    // chk operator+(const WyTimeSpec&)
    try {
      WyTimeSpec tobj= SVal_1+SVal_2;
      if(res_reply!=Ok) {
        WY_THROW( WyRet() );
      }
      if((tobj<res_floor)||(tobj>res_ceil)) {
        std::cerr << "Res=" << Wy::wrd(tobj).c_str() << '\n';
        WY_THROW( WyRet() );
      }
    }
    catch(const WyTimeSpec::Reply& e) {
      if(e!=res_reply) {
        WY_THROW( WyRet(e) );
      }
    };

    // chk add(const WyTimeSpec&)
    {
      WyRet r;
      WyTimeSpec tobj(SVal_1);
      r=tobj.add(SVal_2);
      if(r!=res_reply) {
        WY_THROW( r );
      }
      if(r==Ok) {
        if((tobj<res_floor)||(tobj>res_ceil)) {
          std::cerr << "Res=" << Wy::wrd(tobj).c_str() << '\n';
          WY_THROW( WyRet() );
        }
      }
    }

    // chk add(sec,nsec)
    {
      WyRet r;
      WyTimeSpec tobj(SVal_1);
      r=tobj.add(SVal_2.second(),SVal_2.nano());
      if(r!=res_reply) {
        WY_THROW( r );
      }
      if(r==Ok) {
        if((tobj<res_floor)||(tobj>res_ceil)) {
          std::cerr << "Res=" << Wy::wrd(tobj).c_str() << '\n';
          WY_THROW( WyRet() );
        }
      }
    }
  } catch(...) {
    std::cerr << "ChkItm_add_ts Item={" << Wy::wrd(SVal_1).c_str() << ", " 
                                        << Wy::wrd(SVal_2).c_str() << ", " 
              << Wy::wrd(res_reply).c_str() << ", "
              << Wy::wrd(res_floor).c_str() << ", " << Wy::wrd(res_ceil).c_str() << "}\n";
    throw;
  };
};

static void ckop_add_ts(void)
{
 const ChkItm_add_ts chk_tab[]={
   // sval reply sfloor sceil nfloor nceil
   { WyTimeSpec(0,0), WyTimeSpec(0,0), Ok, WyTimeSpec(), WyTimeSpec()},
   { WyTimeSpec(0,999999999), WyTimeSpec(0,2), Ok, WyTimeSpec(1,1), WyTimeSpec(1,1)},
   { WyTimeSpec(0,-999999999), WyTimeSpec(0,-2), Ok, WyTimeSpec(-1,-1), WyTimeSpec(-1,-1)},
   { WyTimeSpec(999999999,999999998), WyTimeSpec(0,1), Ok,
               WyTimeSpec(999999999,999999999), WyTimeSpec(999999999,999999999)},
   { WyTimeSpec(std::numeric_limits<WyTimeSpec::SecondType>::max(),999999998), WyTimeSpec(0,1), Ok,
               WyTimeSpec(std::numeric_limits<WyTimeSpec::SecondType>::max(),999999999),
               WyTimeSpec(std::numeric_limits<WyTimeSpec::SecondType>::max(),999999999)},
   { WyTimeSpec(std::numeric_limits<WyTimeSpec::SecondType>::min(),-999999998), WyTimeSpec(0,-1), Ok,
               WyTimeSpec(std::numeric_limits<WyTimeSpec::SecondType>::min(),-999999999),
               WyTimeSpec(std::numeric_limits<WyTimeSpec::SecondType>::min(),-999999999)},
   { WyTimeSpec(std::numeric_limits<WyTimeSpec::SecondType>::max(),999999999), WyTimeSpec(0,1),
               Wym_ERANGE, WyTimeSpec(), WyTimeSpec() },
   { WyTimeSpec(std::numeric_limits<WyTimeSpec::SecondType>::min(),-999999999), WyTimeSpec(0,-1),
               Wym_ERANGE, WyTimeSpec(), WyTimeSpec() },
 };
 
 for(size_t i=0; i<sizeof(chk_tab)/sizeof(ChkItm_add_ts); ++i) {
   chk_tab[i].verify();
 }

 // Random number check
 typedef double DType;
 const long long RandExp=std::numeric_limits<WyTimeSpec::SecondType>::digits+1;

 for(int i=0; i<RandSmps; ++i) {
   //  res= x*2^y  ,x in [-1.0,+1.0], y in [-RandExp, +RandExp]
   //
   DType num1= std::ldexp( ::rand()*(2.0/(double)RAND_MAX)-1.0
              , (::rand()*(long long)2*RandExp)/(long long)RAND_MAX-RandExp 
             );
   if(num1>Wy::to_double(WyTimeSpec::max())||(num1<Wy::to_double(WyTimeSpec::min()))) {
     continue;
   }
   if(std::abs(num1)<1e-9) {
     num1=0.0;
   }
   DType num2= std::ldexp( ::rand()*(2.0/(double)RAND_MAX)-1.0
              , (::rand()*(long long)2*RandExp)/(long long)RAND_MAX-RandExp 
             );
   if(num2>Wy::to_double(WyTimeSpec::max())||(num2<Wy::to_double(WyTimeSpec::min()))) {
     continue;
   }
   if(std::abs(num2)<1e-9) {
     num2=0.0;
   }
   DType Res=Wy::to_double(WyTimeSpec::max())+1.0;
   try { Res=Wy::to_double(WyTimeSpec(num1)+WyTimeSpec(num2)); }
   catch(const WyTimeSpec::Reply& e) {
     if(e!=Wym_ERANGE) {
       throw;
     }
   };

   WyRet r;
   if(Res>Wy::to_double(WyTimeSpec::max())||(Res<Wy::to_double(WyTimeSpec::min()))) {
     ChkItm_add_ts itm={ num1, num2, Wym_ERANGE, WyTimeSpec(), WyTimeSpec()};
     itm.verify();
     if((Res<std::numeric_limits<WyTimeSpec::SecondType>::max()+1.0)&&
        (Res>std::numeric_limits<WyTimeSpec::SecondType>::min()-1.0)) {
       std::cout << "Error: Res=" << Res << '\n';
       WY_THROW( WyRet() );
     }
   } else {
     if(Res>=0) {
       if(Res*1.000001+1e-9<=Wy::to_double(WyTimeSpec::max())) {
         ChkItm_add_ts itm={ WyTimeSpec(num1), WyTimeSpec(num2), Ok,
                             WyTimeSpec(Res*0.999999-1e-9), WyTimeSpec(Res*1.000001+1e-9)};
         itm.verify();
       }
     } else {
       if(Res*1.000001-1e-9>=Wy::to_double(WyTimeSpec::min())) {
         ChkItm_add_ts itm={ WyTimeSpec(num1), WyTimeSpec(num2), Ok,
                             WyTimeSpec(Res*1.000001-1e-9), WyTimeSpec(Res*0.999999+1e-9)};
         itm.verify();
       }
     }
   } 
 }
 
 (WyTimeSpec()+WyTimeSpec())++;   // chk return non-const
};

struct ChkItm_add_n {        // check WyTimeSpec addition
  const WyTimeSpec SVal_1;
  const WyTimeSpec::SecondType SVal_2;

  // Expected results
  const WyReply res_reply;
  const WyTimeSpec res_n;

  void verify(void) const
  try {

    // chk operator +=(SecondType)
    try {
      WyTimeSpec tobj(SVal_1);
      tobj+= SVal_2;
      if(res_reply!=Ok) {
        WY_THROW( WyRet() );
      }
      if(tobj!=res_n) {
        std::cerr << "Res=" << Wy::wrd(tobj).c_str() << '\n';
        WY_THROW( WyRet() );
      }
    }
    catch(const WyTimeSpec::Reply& e) {
      if(e!=res_reply) {
        WY_THROW( WyRet(e) );
      }
    };

    // chk operator+(SecondType)
    try {
      WyTimeSpec tobj= SVal_1+SVal_2;
      if(res_reply!=Ok) {
        WY_THROW( WyRet() );
      }
      if(tobj!=res_n) {
        std::cerr << "Res=" << Wy::wrd(tobj).c_str() << '\n';
        WY_THROW( WyRet() );
      }
    }
    catch(const WyTimeSpec::Reply& e) {
      if(e!=res_reply) {
        WY_THROW( WyRet(e) );
      }
    };

    // chk operator+(SecondType,WyTimeSpec)
    try {
      WyTimeSpec tobj= SVal_2+SVal_1;
      if(res_reply!=Ok) {
        WY_THROW( WyRet() );
      }
      if(tobj!=res_n) {
        std::cerr << "Res=" << Wy::wrd(tobj).c_str() << '\n';
        WY_THROW( WyRet() );
      }
    }
    catch(const WyTimeSpec::Reply& e) {
      if(e!=res_reply) {
        WY_THROW( WyRet(e) );
      }
    };

    // chk add(SecondType)
    {
      WyRet r;
      WyTimeSpec tobj(SVal_1);
      r=tobj.add(SVal_2);
      if(r!=res_reply) {
        WY_THROW( r );
      }
      if(r==Ok) {
        if(tobj!=res_n) {
          std::cerr << "Res=" << Wy::wrd(tobj).c_str() << '\n';
          WY_THROW( WyRet() );
        }
      }
    }
  } catch(...) {
    std::cerr << "ChkItm_add_n Item={" << Wy::wrd(SVal_1).c_str() << ", " 
                                        << Wy::wrd(SVal_2).c_str() << ", " 
              << Wy::wrd(res_reply).c_str() << ", "
              << Wy::wrd(res_n).c_str() << "}\n";
    throw;
  };
};

static void ckop_add_n(void)
{
 const ChkItm_add_n chk_tab[]={
   // sval1 sval2 reply n
   { WyTimeSpec(0,0), 0, Ok, WyTimeSpec() },
   { WyTimeSpec(0,5), 1, Ok, WyTimeSpec(1,5) },
   { WyTimeSpec(0,5),-1, Ok, WyTimeSpec(0,-999999995) },
   { WyTimeSpec(std::numeric_limits<WyTimeSpec>::max()),1, Wym_ERANGE, WyTimeSpec() },
 };
 
 for(size_t i=0; i<sizeof(chk_tab)/sizeof(ChkItm_add_n); ++i) {
   chk_tab[i].verify();
 }

 (WyTimeSpec()+WyTimeSpec::SecondType(0))++;   // chk return non-const
 (WyTimeSpec::SecondType(0)+WyTimeSpec())++;   // chk return non-const
};

struct ChkItm_sub_n {        // check WyTimeSpec subtraction
  const WyTimeSpec SVal_1;
  const WyTimeSpec::SecondType SVal_2;

  // Expected results
  const WyReply res_reply;
  const WyTimeSpec res_n;

  void verify(void) const
  try {

    // chk operator -=(SecondType)
    try {
      WyTimeSpec tobj(SVal_1);
      tobj-= SVal_2;
      if(res_reply!=Ok) {
        WY_THROW( WyRet() );
      }
      if(tobj!=res_n) {
        std::cerr << "Res=" << Wy::wrd(tobj).c_str() << '\n';
        WY_THROW( WyRet() );
      }
    }
    catch(const WyTimeSpec::Reply& e) {
      if(e!=res_reply) {
        WY_THROW( WyRet(e) );
      }
    };

    // chk operator-(SecondType)
    try {
      WyTimeSpec tobj= SVal_1-SVal_2;
      if(res_reply!=Ok) {
        WY_THROW( WyRet() );
      }
      if(tobj!=res_n) {
        std::cerr << "Res=" << Wy::wrd(tobj).c_str() << '\n';
        WY_THROW( WyRet() );
      }
    }
    catch(const WyTimeSpec::Reply& e) {
      if(e!=res_reply) {
        WY_THROW( WyRet(e) );
      }
    };

    // chk sub(SecondType)
    {
      WyRet r;
      WyTimeSpec tobj(SVal_1);
      r=tobj.sub(SVal_2);
      if(r!=res_reply) {
        WY_THROW( r );
      }
      if(r==Ok) {
        if(tobj!=res_n) {
          std::cerr << "Res=" << Wy::wrd(tobj).c_str() << '\n';
          WY_THROW( WyRet() );
        }
      }
    }
  } catch(...) {
    std::cerr << "ChkItm_sub_n Item={" << Wy::wrd(SVal_1).c_str() << ", " 
                                        << Wy::wrd(SVal_2).c_str() << ", " 
              << Wy::wrd(res_reply).c_str() << ", "
              << Wy::wrd(res_n).c_str() << "}\n";
    throw;
  };
};

static void ckop_sub_n(void)
{
 const ChkItm_sub_n chk_tab[]={
   // sval1 sval2 reply n
   { WyTimeSpec(0,0), 0, Ok, WyTimeSpec() },
   { WyTimeSpec(0,5), 1, Ok, WyTimeSpec(0,-999999995) },
   { WyTimeSpec(0,5),-1, Ok, WyTimeSpec(1,5) },
   { WyTimeSpec(std::numeric_limits<WyTimeSpec>::min()),1, Wym_ERANGE, WyTimeSpec() },
 };
 
 for(size_t i=0; i<sizeof(chk_tab)/sizeof(ChkItm_sub_n); ++i) {
   chk_tab[i].verify();
 }

 // just verify this function exists
 if(WyTimeSpec::SecondType(0)-WyTimeSpec(1,1)!=WyTimeSpec(-1,-1)) {
   WY_THROW( WyRet() );
 }

 (WyTimeSpec()+WyTimeSpec::SecondType(0))++;   // chk return non-const
 (WyTimeSpec::SecondType(0)+WyTimeSpec())++;   // chk return non-const
};

struct ChkItm_mul_n {        // check SecondType multiplication
  const WyTimeSpec SVal_1;
  const WyTimeSpec::SecondType SVal_2;

  // Expected results
  const WyReply res_reply;
  const WyTimeSpec res_n;

  void verify(void) const
  try {

    // chk operator *=(SecondType)
    try {
      WyTimeSpec tobj(SVal_1);
      tobj*= SVal_2;
      if(res_reply!=Ok) {
        WY_THROW( WyRet() );
      }
      if(tobj!=res_n) {
        std::cerr << "Res=" << Wy::wrd(tobj).c_str() << '\n';
        WY_THROW( WyRet() );
      }
    }
    catch(const WyTimeSpec::Reply& e) {
      if(e!=res_reply) {
        WY_THROW( WyRet(e) );
      }
    };

    // chk operator*(SecondType)
    try {
      WyTimeSpec tobj= SVal_1*SVal_2;
      if(res_reply!=Ok) {
        WY_THROW( WyRet() );
      }
      if(tobj!=res_n) {
        std::cerr << "Res=" << Wy::wrd(tobj).c_str() << '\n';
        WY_THROW( WyRet() );
      }
    }
    catch(const WyTimeSpec::Reply& e) {
      if(e!=res_reply) {
        WY_THROW( WyRet(e) );
      }
    };

    // chk operator*(SecondType,WyTimeSpec)
    try {
      WyTimeSpec tobj= SVal_2*SVal_1;
      if(res_reply!=Ok) {
        WY_THROW( WyRet() );
      }
      if(tobj!=res_n) {
        std::cerr << "Res=" << Wy::wrd(tobj).c_str() << '\n';
        WY_THROW( WyRet() );
      }
    }
    catch(const WyTimeSpec::Reply& e) {
      if(e!=res_reply) {
        WY_THROW( WyRet(e) );
      }
    };

    // chk mul(SecondType)
    {
      WyRet r;
      WyTimeSpec tobj(SVal_1);
      r=tobj.mul(SVal_2);
      if(r!=res_reply) {
        WY_THROW( r );
      }
      if(r==Ok) {
        if(tobj!=res_n) {
          std::cerr << "Res=" << Wy::wrd(tobj).c_str() << '\n';
          WY_THROW( WyRet() );
        }
      }
    }
  } catch(...) {
    std::cerr << "ChkItm_mul_n Item={" << Wy::wrd(SVal_1).c_str() << ", " 
                                        << Wy::wrd(SVal_2).c_str() << ", " 
              << Wy::wrd(res_reply).c_str() << ", "
              << Wy::wrd(res_n).c_str() << "}\n";
    throw;
  };
};

static void ckop_mul_n(void)
{
 const ChkItm_mul_n chk_tab[]={
   // sval1 sval2 reply n
   { WyTimeSpec(0,0), 0, Ok, WyTimeSpec() },
   { WyTimeSpec(0,5), 1, Ok, WyTimeSpec(0,5) },
   { WyTimeSpec(0,5),-1, Ok, WyTimeSpec(0,-5) },
   { WyTimeSpec(1,600000000),-2, Ok, WyTimeSpec(-3,-200000000) },
   { WyTimeSpec(std::numeric_limits<WyTimeSpec>::max()),2, Wym_ERANGE, WyTimeSpec() },
 };
 
 for(size_t i=0; i<sizeof(chk_tab)/sizeof(ChkItm_mul_n); ++i) {
   chk_tab[i].verify();
 }

 (WyTimeSpec()+WyTimeSpec::SecondType(0))++;   // chk return non-const
 (WyTimeSpec::SecondType(0)+WyTimeSpec())++;   // chk return non-const
};

struct ChkItm_div_n {        // check SecondType division
  const WyTimeSpec SVal_1;
  const WyTimeSpec::SecondType SVal_2;

  // Expected results
  const WyReply res_reply;
  const WyTimeSpec res_n;

  void verify(void) const
  try {

    // chk operator /=(SecondType)
    try {
      WyTimeSpec tobj(SVal_1);
      tobj/= SVal_2;
      if(res_reply!=Ok) {
        WY_THROW( WyRet() );
      }
      if(tobj!=res_n) {
        std::cerr << "Res=" << Wy::wrd(tobj).c_str() << '\n';
        WY_THROW( WyRet() );
      }
    }
    catch(const WyTimeSpec::Reply& e) {
      if(e!=res_reply) {
        WY_THROW( WyRet(e) );
      }
    };

    // chk operator/(SecondType)
    try {
      WyTimeSpec tobj= SVal_1/SVal_2;
      if(res_reply!=Ok) {
        WY_THROW( WyRet() );
      }
      if(tobj!=res_n) {
        std::cerr << "Res=" << Wy::wrd(tobj).c_str() << '\n';
        WY_THROW( WyRet() );
      }
    }
    catch(const WyTimeSpec::Reply& e) {
      if(e!=res_reply) {
        WY_THROW( WyRet(e) );
      }
    };

    // chk div(SecondType)
    {
      WyRet r;
      WyTimeSpec tobj(SVal_1);
      r=tobj.div(SVal_2);
      if(r!=res_reply) {
        WY_THROW( r );
      }
      if(r==Ok) {
        if(tobj!=res_n) {
          std::cerr << "Res=" << Wy::wrd(tobj).c_str() << '\n';
          WY_THROW( WyRet() );
        }
      }
    }
  } catch(...) {
    std::cerr << "ChkItm_div_n Item={" << Wy::wrd(SVal_1).c_str() << ", " 
                                        << Wy::wrd(SVal_2).c_str() << ", " 
              << Wy::wrd(res_reply).c_str() << ", "
              << Wy::wrd(res_n).c_str() << "}\n";
    throw;
  };
};

static void ckop_div_n(void)
{
 const ChkItm_div_n chk_tab[]={
   // sval1 sval2 reply n
   { WyTimeSpec(0,0), 0, Wym_EDIVZERO, WyTimeSpec() },
   { WyTimeSpec(0,5), 1, Ok, WyTimeSpec(0,5) },
   { WyTimeSpec(0,5),-1, Ok, WyTimeSpec(0,-5) },
   { WyTimeSpec(1,100000000),-2, Ok, WyTimeSpec(0,-550000000) },
 };
 
 for(size_t i=0; i<sizeof(chk_tab)/sizeof(ChkItm_div_n); ++i) {
   chk_tab[i].verify();
 }

 // just verify SecondType/WyTimeSpec exists
 if( WyTimeSpec::SecondType(2)/WyTimeSpec(1,500000000)!=WyTimeSpec(1,333333333)) {
   WY_THROW( WyRet() );
 }

 (WyTimeSpec()+WyTimeSpec::SecondType(0))++;   // chk return non-const
 (WyTimeSpec::SecondType(0)+WyTimeSpec())++;   // chk return non-const
};

struct ChkItm_add_double {        // check WyTimeSpec + double
 typedef double DType;
  const WyTimeSpec SVal_1;
  const DType SVal_2;

  // Expected results
  const WyReply res_reply;
  const WyTimeSpec res_floor;
  const WyTimeSpec res_ceil;

  void verify(void) const
  try {

    // chk operator +=(double)
    try {
      WyTimeSpec tobj(SVal_1);
      tobj+= SVal_2;
      if(res_reply!=Ok) {
        WY_THROW( WyRet() );
      }
      if((tobj<res_floor)||(tobj>res_ceil)) {
        std::cerr << "Res=" << Wy::wrd(tobj).c_str() << '\n';
        WY_THROW( WyRet() );
      }
    }
    catch(const WyTimeSpec::Reply& e) {
      if(e!=res_reply) {
        WY_THROW( WyRet(e) );
      }
    };

    // chk operator+(double)
    try {
      WyTimeSpec tobj= SVal_1+SVal_2;
      if(res_reply!=Ok) {
        WY_THROW( WyRet() );
      }
      if((tobj<res_floor)||(tobj>res_ceil)) {
        std::cerr << "Res=" << Wy::wrd(tobj).c_str() << '\n';
        WY_THROW( WyRet() );
      }
    }
    catch(const WyTimeSpec::Reply& e) {
      if(e!=res_reply) {
        WY_THROW( WyRet(e) );
      }
    };
    try {
      WyTimeSpec tobj= SVal_2+SVal_1;
      if(res_reply!=Ok) {
        WY_THROW( WyRet() );
      }
      if((tobj<res_floor)||(tobj>res_ceil)) {
        std::cerr << "Res=" << Wy::wrd(tobj).c_str() << '\n';
        WY_THROW( WyRet() );
      }
    }
    catch(const WyTimeSpec::Reply& e) {
      if(e!=res_reply) {
        WY_THROW( WyRet(e) );
      }
    };

  } catch(...) {
    std::cerr << "ChkItm_add_double Item={" << Wy::wrd(SVal_1).c_str() << ", " 
                                        << SVal_2 << ", " 
              << Wy::wrd(res_reply).c_str() << ", "
              << Wy::wrd(res_floor).c_str() << ", " << Wy::wrd(res_ceil).c_str() << "}\n";
    throw;
  };
};

static void ckop_add_double(void)
{
 const ChkItm_add_double chk_tab[]={
   // sval reply sfloor sceil nfloor nceil
   { WyTimeSpec(0,0), 0.0, Ok, WyTimeSpec(), WyTimeSpec()},
   { WyTimeSpec(0,999999999), 2e-9, Ok, WyTimeSpec(1,0), WyTimeSpec(1,2)},
   { WyTimeSpec(0,-999999999), -2e-9, Ok, WyTimeSpec(-1,-2), WyTimeSpec(-1,0)},
   { WyTimeSpec(999999999,999999998), 1e-9, Ok,
               WyTimeSpec(999999999,999999999), WyTimeSpec(999999999,999999999)},
   { WyTimeSpec(std::numeric_limits<WyTimeSpec::SecondType>::max(),999999998), 1e-9, Ok,
               WyTimeSpec(std::numeric_limits<WyTimeSpec::SecondType>::max(),999999999),
               WyTimeSpec(std::numeric_limits<WyTimeSpec::SecondType>::max(),999999999)},
   { WyTimeSpec(std::numeric_limits<WyTimeSpec::SecondType>::min(),-999999998), -1e-9, Ok,
               WyTimeSpec(std::numeric_limits<WyTimeSpec::SecondType>::min(),-999999999),
               WyTimeSpec(std::numeric_limits<WyTimeSpec::SecondType>::min(),-999999999)},
   { WyTimeSpec(std::numeric_limits<WyTimeSpec::SecondType>::max(),999999999), 1e-9,
               Wym_ERANGE, WyTimeSpec(), WyTimeSpec() },
   { WyTimeSpec(std::numeric_limits<WyTimeSpec::SecondType>::min(),-999999999), -1e-9,
               Wym_ERANGE, WyTimeSpec(), WyTimeSpec() },
   { WyTimeSpec(0,0), Nan, Wym_ERANGE, WyTimeSpec(), WyTimeSpec()},
   { WyTimeSpec(0,0), Inf, Wym_ERANGE, WyTimeSpec(), WyTimeSpec()},
 };
 
 for(size_t i=0; i<sizeof(chk_tab)/sizeof(ChkItm_add_double); ++i) {
   chk_tab[i].verify();
 }

 // Random number check
 typedef double DType;
 const long long RandExp=std::numeric_limits<WyTimeSpec::SecondType>::digits+1;

 for(int i=0; i<RandSmps; ++i) {
   //  res= x*2^y  ,x in [-1.0,+1.0], y in [-RandExp, +RandExp]
   //
   DType num1= std::ldexp( ::rand()*(2.0/(double)RAND_MAX)-1.0
              , (::rand()*(long long)2*RandExp)/(long long)RAND_MAX-RandExp 
             );
   if(num1>Wy::to_double(WyTimeSpec::max())||(num1<Wy::to_double(WyTimeSpec::min()))) {
     continue;
   }
   if(std::abs(num1)<1e-9) {
     num1=0.0;
   }
   DType num2= std::ldexp( ::rand()*(2.0/(double)RAND_MAX)-1.0
              , (::rand()*(long long)2*RandExp)/(long long)RAND_MAX-RandExp 
             );
   if(num2>Wy::to_double(WyTimeSpec::max())||(num2<Wy::to_double(WyTimeSpec::min()))) {
     continue;
   }
   if(std::abs(num2)<1e-9) {
     num2=0.0;
   }
   DType Res=Wy::to_double(WyTimeSpec::max())+1.0;
   try { Res=Wy::to_double(WyTimeSpec(num1)+num2); }
   catch(const WyTimeSpec::Reply& e) {
     if(e!=Wym_ERANGE) {
       throw;
     }
   };

   WyRet r;
   if(Res>Wy::to_double(WyTimeSpec::max())||(Res<Wy::to_double(WyTimeSpec::min()))) {
     ChkItm_add_double itm={ WyTimeSpec(num1), num2, Wym_ERANGE, WyTimeSpec(), WyTimeSpec()};
     itm.verify();
     if((Res<std::numeric_limits<WyTimeSpec::SecondType>::max()+1.0)&&
        (Res>std::numeric_limits<WyTimeSpec::SecondType>::min()-1.0)) {
       std::cout << "Error: Res=" << Res << '\n';
       WY_THROW( WyRet() );
     }
   } else {
     if(Res>=0) {
       if(Res*1.000001+1e-9<=Wy::to_double(WyTimeSpec::max())) {
         ChkItm_add_double itm={ WyTimeSpec(num1), num2, Ok,
                             WyTimeSpec(Res*0.999999-1e-9), WyTimeSpec(Res*1.000001+1e-9)};
         itm.verify();
       }
     } else {
       if(Res*1.000001-1e-9>=Wy::to_double(WyTimeSpec::min())) {
         ChkItm_add_double itm={ WyTimeSpec(num1), num2, Ok,
                             WyTimeSpec(Res*1.000001-1e-9), WyTimeSpec(Res*0.999999+1e-9)};
         itm.verify();
       }
     }
   } 
 }
 
 (WyTimeSpec()+double(1.2))++;   // chk return non-const
};

struct ChkItm_double_add {        // check double+WyTimeSpec
 typedef double DType;
  const DType SVal_1;
  const WyTimeSpec SVal_2;

  // Expected results
  const WyReply res_reply;
  const WyTimeSpec res_floor;
  const WyTimeSpec res_ceil;

  void verify(void) const
  try {

    // chk operator+(double)
    try {
      WyTimeSpec tobj= SVal_1+SVal_2;
      if(res_reply!=Ok) {
        WY_THROW( WyRet() );
      }
      if((tobj<res_floor)||(tobj>res_ceil)) {
        std::cerr << "Res=" << Wy::wrd(tobj).c_str() << '\n';
        WY_THROW( WyRet() );
      }
    }
    catch(const WyTimeSpec::Reply& e) {
      if(e!=res_reply) {
        WY_THROW( WyRet(e) );
      }
    };
    try {
      WyTimeSpec tobj= SVal_2+SVal_1;
      if(res_reply!=Ok) {
        WY_THROW( WyRet() );
      }
      if((tobj<res_floor)||(tobj>res_ceil)) {
        std::cerr << "Res=" << Wy::wrd(tobj).c_str() << '\n';
        WY_THROW( WyRet() );
      }
    }
    catch(const WyTimeSpec::Reply& e) {
      if(e!=res_reply) {
        WY_THROW( WyRet(e) );
      }
    };

  } catch(...) {
    std::cerr << "ChkItm_double_add Item={" << SVal_1 << ", " 
                                        << Wy::wrd(SVal_2).c_str() << ", " 
              << Wy::wrd(res_reply).c_str() << ", "
              << Wy::wrd(res_floor).c_str() << ", " << Wy::wrd(res_ceil).c_str() << "}\n";
    throw;
  };
};

static void ckop_double_add(void)
{
 const ChkItm_double_add chk_tab[]={
   // sval reply sfloor sceil nfloor nceil
   { 0.0, WyTimeSpec(0,0), Ok, WyTimeSpec(), WyTimeSpec()},
   { 2e-9, WyTimeSpec(0,999999999), Ok, WyTimeSpec(1,0), WyTimeSpec(1,2)},
   { -2e-9, WyTimeSpec(0,-999999999), Ok, WyTimeSpec(-1,-2), WyTimeSpec(-1,0)},
   { 1e-9, WyTimeSpec(999999999,999999998), Ok,
               WyTimeSpec(999999999,999999999), WyTimeSpec(999999999,999999999)},
   { 1e-9, WyTimeSpec(std::numeric_limits<WyTimeSpec::SecondType>::max(),999999998), Ok,
               WyTimeSpec(std::numeric_limits<WyTimeSpec::SecondType>::max(),999999999),
               WyTimeSpec(std::numeric_limits<WyTimeSpec::SecondType>::max(),999999999)},
   { -1e-9, WyTimeSpec(std::numeric_limits<WyTimeSpec::SecondType>::min(),-999999998), Ok,
               WyTimeSpec(std::numeric_limits<WyTimeSpec::SecondType>::min(),-999999999),
               WyTimeSpec(std::numeric_limits<WyTimeSpec::SecondType>::min(),-999999999)},
   { 1e-9, WyTimeSpec(std::numeric_limits<WyTimeSpec::SecondType>::max(),999999999),
               Wym_ERANGE, WyTimeSpec(), WyTimeSpec() },
   { -1e-9, WyTimeSpec(std::numeric_limits<WyTimeSpec::SecondType>::min(),-999999999),
               Wym_ERANGE, WyTimeSpec(), WyTimeSpec() },
   { Nan, WyTimeSpec(0,0), Wym_ERANGE, WyTimeSpec(), WyTimeSpec()},
   { Inf, WyTimeSpec(0,0), Wym_ERANGE, WyTimeSpec(), WyTimeSpec()},
 };
 
 for(size_t i=0; i<sizeof(chk_tab)/sizeof(ChkItm_double_add); ++i) {
   chk_tab[i].verify();
 }

 // Random number check
 typedef double DType;
 const long long RandExp=std::numeric_limits<WyTimeSpec::SecondType>::digits+1;

 for(int i=0; i<RandSmps; ++i) {
   //  res= x*2^y  ,x in [-1.0,+1.0], y in [-RandExp, +RandExp]
   //
   DType num1= std::ldexp( ::rand()*(2.0/(double)RAND_MAX)-1.0
              , (::rand()*(long long)2*RandExp)/(long long)RAND_MAX-RandExp 
             );
   if(num1>Wy::to_double(WyTimeSpec::max())||(num1<Wy::to_double(WyTimeSpec::min()))) {
     continue;
   }
   if(std::abs(num1)<1e-9) {
     num1=0.0;
   }
   DType num2= std::ldexp( ::rand()*(2.0/(double)RAND_MAX)-1.0
              , (::rand()*(long long)2*RandExp)/(long long)RAND_MAX-RandExp 
             );
   if(num2>Wy::to_double(WyTimeSpec::max())||(num2<Wy::to_double(WyTimeSpec::min()))) {
     continue;
   }
   if(std::abs(num2)<1e-9) {
     num2=0.0;
   }
   DType Res=Wy::to_double(WyTimeSpec::max())+1.0;
   try { Res=Wy::to_double(num1+WyTimeSpec(num2)); }
   catch(const WyTimeSpec::Reply& e) {
     if(e!=Wym_ERANGE) {
       throw;
     }
   };

   WyRet r;
   if(Res>Wy::to_double(WyTimeSpec::max())||(Res<Wy::to_double(WyTimeSpec::min()))) {
     ChkItm_double_add itm={ num1, WyTimeSpec(num2), Wym_ERANGE, WyTimeSpec(), WyTimeSpec()};
     itm.verify();
     if((Res<std::numeric_limits<WyTimeSpec::SecondType>::max()+1.0)&&
        (Res>std::numeric_limits<WyTimeSpec::SecondType>::min()-1.0)) {
       std::cout << "Error: Res=" << Res << '\n';
       WY_THROW( WyRet() );
     }
   } else {
     if(Res>=0) {
       if(Res*1.000001+1e-9<=Wy::to_double(WyTimeSpec::max())) {
         ChkItm_double_add itm={ num1, WyTimeSpec(num2), Ok,
                             WyTimeSpec(Res*0.999999-1e-9), WyTimeSpec(Res*1.000001+1e-9)};
         itm.verify();
       }
     } else {
       if(Res*1.000001-1e-9>=Wy::to_double(WyTimeSpec::min())) {
         ChkItm_double_add itm={ num1, WyTimeSpec(num2), Ok,
                             WyTimeSpec(Res*1.000001-1e-9), WyTimeSpec(Res*0.999999+1e-9)};
         itm.verify();
       }
     }
   } 
 }
 
 (double(1.2)+WyTimeSpec())++;   // chk return non-const
};

struct ChkItm_sub_ts {        // check WyTimeSpec substract
 typedef double DType;
  const WyTimeSpec SVal_1;
  const WyTimeSpec SVal_2;

  // Expected results
  const WyReply res_reply;
  const WyTimeSpec res_floor;
  const WyTimeSpec res_ceil;

  void verify(void) const
  try {

    // chk operator -=(const WyTimeSpec&)
    try {
      WyTimeSpec tobj(SVal_1);
      tobj-= SVal_2;
      if(res_reply!=Ok) {
        WY_THROW( WyRet() );
      }
      if((tobj<res_floor)||(tobj>res_ceil)) {
        std::cerr << "Res=" << Wy::wrd(tobj).c_str() << '\n';
        WY_THROW( WyRet() );
      }
    }
    catch(const WyTimeSpec::Reply& e) {
      if(e!=res_reply) {
        WY_THROW( WyRet(e) );
      }
    };

    // chk operator-(const WyTimeSpec)
    try {
      WyTimeSpec tobj= SVal_1-SVal_2;
      if(res_reply!=Ok) {
        WY_THROW( WyRet() );
      }
      if((tobj<res_floor)||(tobj>res_ceil)) {
        std::cerr << "Res=" << Wy::wrd(tobj).c_str() << '\n';
        WY_THROW( WyRet() );
      }
    }
    catch(const WyTimeSpec::Reply& e) {
      if(e!=res_reply) {
        WY_THROW( WyRet(e) );
      }
    };

    // chk sub(const WyTimeSpec&)
    {
      WyRet r;
      WyTimeSpec tobj(SVal_1);
      r=tobj.sub(SVal_2);
      if(r!=res_reply) {
        WY_THROW( r );
      }
      if(r==Ok) {
        if((tobj<res_floor)||(tobj>res_ceil)) {
          std::cerr << "Res=" << Wy::wrd(tobj).c_str() << '\n';
          WY_THROW( WyRet() );
        }
      }
    }

    // chk sub(sec,nsec)
    {
      WyRet r;
      WyTimeSpec tobj(SVal_1);
      r=tobj.sub(SVal_2.second(),SVal_2.nano());
      if(r!=res_reply) {
        WY_THROW( r );
      }
      if(r==Ok) {
        if((tobj<res_floor)||(tobj>res_ceil)) {
          std::cerr << "Res=" << Wy::wrd(tobj).c_str() << '\n';
          WY_THROW( WyRet() );
        }
      }
    }
  } catch(...) {
    std::cerr << "ChkItm_sub_ts Item={" << Wy::wrd(SVal_1).c_str() << ", " 
                                        << Wy::wrd(SVal_2).c_str() << ", " 
              << Wy::wrd(res_reply).c_str() << ", "
              << Wy::wrd(res_floor).c_str() << ", " << Wy::wrd(res_ceil).c_str() << "}\n";
    throw;
  };
};

static void ckop_sub_ts(void)
{
 const ChkItm_sub_ts chk_tab[]={
   // sval reply sfloor sceil nfloor nceil
   { WyTimeSpec(0,0), WyTimeSpec(0,0), Ok, WyTimeSpec(), WyTimeSpec()},
   { WyTimeSpec(1,0), WyTimeSpec(0,2), Ok, WyTimeSpec(0,999999998), WyTimeSpec(0,999999998)},
   { WyTimeSpec(0,-999999999), WyTimeSpec(0,2), Ok, WyTimeSpec(-1,-1), WyTimeSpec(-1,-1)},
   { WyTimeSpec(-999999999,-999999998), WyTimeSpec(0,1), Ok,
               WyTimeSpec(-999999999,-999999999), WyTimeSpec(-999999999,-999999999)},
   { WyTimeSpec(std::numeric_limits<WyTimeSpec::SecondType>::min(),-999999998), WyTimeSpec(0,1), Ok,
               WyTimeSpec(std::numeric_limits<WyTimeSpec::SecondType>::min(),-999999999),
               WyTimeSpec(std::numeric_limits<WyTimeSpec::SecondType>::min(),-999999999)},
   { WyTimeSpec(std::numeric_limits<WyTimeSpec::SecondType>::max(),999999998), WyTimeSpec(0,-1), Ok,
               WyTimeSpec(std::numeric_limits<WyTimeSpec::SecondType>::max(),999999999),
               WyTimeSpec(std::numeric_limits<WyTimeSpec::SecondType>::max(),999999999)},
   { WyTimeSpec(std::numeric_limits<WyTimeSpec::SecondType>::max(),999999999), WyTimeSpec(0,-1),
               Wym_ERANGE, WyTimeSpec(), WyTimeSpec() },
   { WyTimeSpec(std::numeric_limits<WyTimeSpec::SecondType>::min(),-999999999), WyTimeSpec(0,1),
               Wym_ERANGE, WyTimeSpec(), WyTimeSpec() },
 };
 
 for(size_t i=0; i<sizeof(chk_tab)/sizeof(ChkItm_sub_ts); ++i) {
   chk_tab[i].verify();
 }

 // Random number check
 typedef double DType;
 const long long RandExp=std::numeric_limits<WyTimeSpec::SecondType>::digits+1;

 for(int i=0; i<RandSmps; ++i) {
   //  res= x*2^y  ,x in [-1.0,+1.0], y in [-RandExp, +RandExp]
   //
   DType num1= std::ldexp( ::rand()*(2.0/(double)RAND_MAX)-1.0
              , (::rand()*(long long)2*RandExp)/(long long)RAND_MAX-RandExp 
             );
   if(num1>Wy::to_double(WyTimeSpec::max())||(num1<Wy::to_double(WyTimeSpec::min()))) {
     continue;
   }
   if(std::abs(num1)<1e-9) {
     num1=0.0;
   }
   DType num2= std::ldexp( ::rand()*(2.0/(double)RAND_MAX)-1.0
              , (::rand()*(long long)2*RandExp)/(long long)RAND_MAX-RandExp 
             );
   if(num2>Wy::to_double(WyTimeSpec::max())||(num2<Wy::to_double(WyTimeSpec::min()))) {
     continue;
   }
   if(std::abs(num2)<1e-9) {
     num2=0.0;
   }
   DType Res=Wy::to_double(WyTimeSpec::max())+1.0;
   try { Res=Wy::to_double(WyTimeSpec(num1)-WyTimeSpec(num2)); }
   catch(const WyTimeSpec::Reply& e) {
     if(e!=Wym_ERANGE) {
       throw;
     }
   };

   WyRet r;
   if(Res>Wy::to_double(WyTimeSpec::max())||(Res<Wy::to_double(WyTimeSpec::min()))) {
     ChkItm_sub_ts itm={ num1, num2, Wym_ERANGE, WyTimeSpec(), WyTimeSpec()};
     itm.verify();
     if((Res<std::numeric_limits<WyTimeSpec::SecondType>::max()+1.0)&&
        (Res>std::numeric_limits<WyTimeSpec::SecondType>::min()-1.0)) {
       std::cout << "Error: Res=" << Res << '\n';
       WY_THROW( WyRet() );
     }
   } else {
     if(Res>=0) {
       if(Res*1.000001+1e-9<=Wy::to_double(WyTimeSpec::max())) {
         ChkItm_sub_ts itm={ WyTimeSpec(num1), WyTimeSpec(num2), Ok,
                             WyTimeSpec(Res*0.999999-1e-9), WyTimeSpec(Res*1.000001+1e-9)};
         itm.verify();
       }
     } else {
       if(Res*1.000001-1e-9>=Wy::to_double(WyTimeSpec::min())) {
         ChkItm_sub_ts itm={ WyTimeSpec(num1), WyTimeSpec(num2), Ok,
                             WyTimeSpec(Res*1.000001-1e-9), WyTimeSpec(Res*0.999999+1e-9)};
         itm.verify();
       }
     }
   } 
 }
 
 (WyTimeSpec()-WyTimeSpec())++;   // chk return non-const
};

struct ChkItm_sub_double {        // check double substract
 typedef double DType;
  const WyTimeSpec SVal_1;
  const DType SVal_2;

  // Expected results
  const WyReply res_reply;
  const WyTimeSpec res_floor;
  const WyTimeSpec res_ceil;

  void verify(void) const
  try {

    // chk operator -=(const WyTimeSpec&)
    try {
      WyTimeSpec tobj(SVal_1);
      tobj-= SVal_2;
      if(res_reply!=Ok) {
        WY_THROW( WyRet() );
      }
      if((tobj<res_floor)||(tobj>res_ceil)) {
        std::cerr << "Res=" << Wy::wrd(tobj).c_str() << '\n';
        WY_THROW( WyRet() );
      }
    }
    catch(const WyTimeSpec::Reply& e) {
      if(e!=res_reply) {
        WY_THROW( WyRet(e) );
      }
    };

    // chk operator-(const WyTimeSpec)
    try {
      WyTimeSpec tobj= SVal_1-SVal_2;
      if(res_reply!=Ok) {
        WY_THROW( WyRet() );
      }
      if((tobj<res_floor)||(tobj>res_ceil)) {
        std::cerr << "Res=" << Wy::wrd(tobj).c_str() << '\n';
        WY_THROW( WyRet() );
      }
    }
    catch(const WyTimeSpec::Reply& e) {
      if(e!=res_reply) {
        WY_THROW( WyRet(e) );
      }
    };

  } catch(...) {
    std::cerr << "ChkItm_sub_double Item={" << Wy::wrd(SVal_1).c_str() << ", " 
                                        << SVal_2 << ", " 
              << Wy::wrd(res_reply).c_str() << ", "
              << Wy::wrd(res_floor).c_str() << ", " << Wy::wrd(res_ceil).c_str() << "}\n";
    throw;
  };
};

static void ckop_sub_double(void)
{
 const ChkItm_sub_double chk_tab[]={
   // sval reply sfloor sceil nfloor nceil
   { WyTimeSpec(0,0), 0.0, Ok, WyTimeSpec(), WyTimeSpec()},
   { WyTimeSpec(1,0), 2e-9, Ok, WyTimeSpec(0,999999998), WyTimeSpec(0,999999998)},
   { WyTimeSpec(0,-999999999), 2e-9, Ok, WyTimeSpec(-1,-1), WyTimeSpec(-1,-1)},
   { WyTimeSpec(-999999999,-999999998), 1e-9, Ok,
               WyTimeSpec(-999999999,-999999999), WyTimeSpec(-999999999,-999999999)},
   { WyTimeSpec(std::numeric_limits<WyTimeSpec::SecondType>::min(),-999999998), 1e-9, Ok,
               WyTimeSpec(std::numeric_limits<WyTimeSpec::SecondType>::min(),-999999999),
               WyTimeSpec(std::numeric_limits<WyTimeSpec::SecondType>::min(),-999999999)},
   { WyTimeSpec(std::numeric_limits<WyTimeSpec::SecondType>::max(),999999998), -1e-9, Ok,
               WyTimeSpec(std::numeric_limits<WyTimeSpec::SecondType>::max(),999999999),
               WyTimeSpec(std::numeric_limits<WyTimeSpec::SecondType>::max(),999999999)},
   { WyTimeSpec(std::numeric_limits<WyTimeSpec::SecondType>::max(),999999999), -1e-9,
               Wym_ERANGE, WyTimeSpec(), WyTimeSpec() },
   { WyTimeSpec(std::numeric_limits<WyTimeSpec::SecondType>::min(),-999999999), 1e-9,
               Wym_ERANGE, WyTimeSpec(), WyTimeSpec() },
   { WyTimeSpec(0,0), Nan, Wym_ERANGE, WyTimeSpec(), WyTimeSpec()},
   { WyTimeSpec(0,0), Inf, Wym_ERANGE, WyTimeSpec(), WyTimeSpec()},
 };
 
 for(size_t i=0; i<sizeof(chk_tab)/sizeof(ChkItm_sub_double); ++i) {
   chk_tab[i].verify();
 }

 // Random number check
 typedef double DType;
 const long long RandExp=std::numeric_limits<WyTimeSpec::SecondType>::digits+1;

 for(int i=0; i<RandSmps; ++i) {
   //  res= x*2^y  ,x in [-1.0,+1.0], y in [-RandExp, +RandExp]
   //
   DType num1= std::ldexp( ::rand()*(2.0/(double)RAND_MAX)-1.0
              , (::rand()*(long long)2*RandExp)/(long long)RAND_MAX-RandExp 
             );
   if(num1>Wy::to_double(WyTimeSpec::max())||(num1<Wy::to_double(WyTimeSpec::min()))) {
     continue;
   }
   if(std::abs(num1)<1e-9) {
     num1=0.0;
   }
   DType num2= std::ldexp( ::rand()*(2.0/(double)RAND_MAX)-1.0
              , (::rand()*(long long)2*RandExp)/(long long)RAND_MAX-RandExp 
             );
   if(num2>Wy::to_double(WyTimeSpec::max())||(num2<Wy::to_double(WyTimeSpec::min()))) {
     continue;
   }
   if(std::abs(num2)<1e-9) {
     num2=0.0;
   }
   DType Res=Wy::to_double(WyTimeSpec::max())+1.0;
   try { Res=Wy::to_double(WyTimeSpec(num1)-num2); }
   catch(const WyTimeSpec::Reply& e) {
     if(e!=Wym_ERANGE) {
       throw;
     }
   };

   WyRet r;
   if(Res>Wy::to_double(WyTimeSpec::max())||(Res<Wy::to_double(WyTimeSpec::min()))) {
     ChkItm_sub_double itm={ WyTimeSpec(num1), num2, Wym_ERANGE, WyTimeSpec(), WyTimeSpec()};
     itm.verify();
     if((Res<std::numeric_limits<WyTimeSpec::SecondType>::max()+1.0)&&
        (Res>std::numeric_limits<WyTimeSpec::SecondType>::min()-1.0)) {
       std::cout << "Error: Res=" << Res << '\n';
       WY_THROW( WyRet() );
     }
   } else {
     if(Res>=0) {
       if(Res*1.000001+1e-9<=Wy::to_double(WyTimeSpec::max())) {
         ChkItm_sub_double itm={ WyTimeSpec(num1), num2, Ok,
                             WyTimeSpec(Res*0.999999-1e-9), WyTimeSpec(Res*1.000001+1e-9)};
         itm.verify();
       }
     } else {
       if(Res*1.000001-1e-9>=Wy::to_double(WyTimeSpec::min())) {
         ChkItm_sub_double itm={ WyTimeSpec(num1), num2, Ok,
                             WyTimeSpec(Res*1.000001-1e-9), WyTimeSpec(Res*0.999999+1e-9)};
         itm.verify();
       }
     }
   } 
 }
 
 (WyTimeSpec()-double(1.2))++;   // chk return non-const
};

struct ChkItm_double_sub {        // check double substract
 typedef double DType;
  const DType SVal_1;
  const WyTimeSpec SVal_2;

  // Expected results
  const WyReply res_reply;
  const WyTimeSpec res_floor;
  const WyTimeSpec res_ceil;

  void verify(void) const
  try {

    // chk: double - WyTimeSpec
    try {
      WyTimeSpec tobj= SVal_1-SVal_2;
      if(res_reply!=Ok) {
        WY_THROW( WyRet() );
      }
      if((tobj<res_floor)||(tobj>res_ceil)) {
        std::cerr << "Res=" << Wy::wrd(tobj).c_str() << '\n';
        WY_THROW( WyRet() );
      }
    }
    catch(const WyTimeSpec::Reply& e) {
      if(e!=res_reply) {
        WY_THROW( WyRet(e) );
      }
    };

  } catch(...) {
    std::cerr << "ChkItm_double_sub Item={" << SVal_1 << ", " 
                                        << Wy::wrd(SVal_2).c_str() << ", " 
              << Wy::wrd(res_reply).c_str() << ", "
              << Wy::wrd(res_floor).c_str() << ", " << Wy::wrd(res_ceil).c_str() << "}\n";
    throw;
  };
};

static void ckop_double_sub(void)
{
 const ChkItm_double_sub chk_tab[]={
   // sval reply sfloor sceil nfloor nceil
   { 0.0, WyTimeSpec(0,0), Ok, WyTimeSpec(), WyTimeSpec()},
   { 1.0, WyTimeSpec(0,2), Ok, WyTimeSpec(0,999999998), WyTimeSpec(0,999999998)},
   { -0.999999999, WyTimeSpec(0,2), Ok, WyTimeSpec(-1,-1), WyTimeSpec(-1,-1)},
   { -999999999.999999998, WyTimeSpec(0,1), Ok,
               WyTimeSpec(-1000000000,-1), WyTimeSpec(-999999999,-999999997)},
   { -1e-9,  WyTimeSpec(std::numeric_limits<WyTimeSpec::SecondType>::max(),999999998), Ok,
             WyTimeSpec(-std::numeric_limits<WyTimeSpec::SecondType>::max(),-999999999),
             WyTimeSpec(-std::numeric_limits<WyTimeSpec::SecondType>::max(),-999999999)},
   { Nan, WyTimeSpec(0,0), Wym_ERANGE, WyTimeSpec(), WyTimeSpec()},
   { Inf, WyTimeSpec(0,0), Wym_ERANGE, WyTimeSpec(), WyTimeSpec()},
 };
 
 for(size_t i=0; i<sizeof(chk_tab)/sizeof(ChkItm_double_sub); ++i) {
   chk_tab[i].verify();
 }

 // Random number check
 typedef double DType;
 const long long RandExp=std::numeric_limits<WyTimeSpec::SecondType>::digits+1;

 for(int i=0; i<RandSmps; ++i) {
   //  res= x*2^y  ,x in [-1.0,+1.0], y in [-RandExp, +RandExp]
   //
   DType num1= std::ldexp( ::rand()*(2.0/(double)RAND_MAX)-1.0
              , (::rand()*(long long)2*RandExp)/(long long)RAND_MAX-RandExp 
             );
   if(num1>Wy::to_double(WyTimeSpec::max())||(num1<Wy::to_double(WyTimeSpec::min()))) {
     continue;
   }
   if(std::abs(num1)<1e-9) {
     num1=0.0;
   }
   DType num2= std::ldexp( ::rand()*(2.0/(double)RAND_MAX)-1.0
              , (::rand()*(long long)2*RandExp)/(long long)RAND_MAX-RandExp 
             );
   if(num2>Wy::to_double(WyTimeSpec::max())||(num2<Wy::to_double(WyTimeSpec::min()))) {
     continue;
   }
   if(std::abs(num2)<1e-9) {
     num2=0.0;
   }
   DType Res=Wy::to_double(WyTimeSpec::max())+1.0;
   try { Res=Wy::to_double(num1-WyTimeSpec(num2)); }
   catch(const WyTimeSpec::Reply& e) {
     if(e!=Wym_ERANGE) {
       throw;
     }
   };

   WyRet r;
   if(Res>Wy::to_double(WyTimeSpec::max())||(Res<Wy::to_double(WyTimeSpec::min()))) {
     ChkItm_double_sub itm={ num1, WyTimeSpec(num2), Wym_ERANGE, WyTimeSpec(), WyTimeSpec()};
     itm.verify();
     if((Res<std::numeric_limits<WyTimeSpec::SecondType>::max()+1.0)&&
        (Res>std::numeric_limits<WyTimeSpec::SecondType>::min()-1.0)) {
       std::cout << "Error: Res=" << Res << '\n';
       WY_THROW( WyRet() );
     }
   } else {
     if(Res>=0) {
       if(Res*1.000001+1e-9<=Wy::to_double(WyTimeSpec::max())) {
         ChkItm_double_sub itm={ num1, WyTimeSpec(num2), Ok,
                             WyTimeSpec(Res*0.999999-1e-9), WyTimeSpec(Res*1.000001+1e-9)};
         itm.verify();
       }
     } else {
       if(Res*1.000001-1e-9>=Wy::to_double(WyTimeSpec::min())) {
         ChkItm_double_sub itm={ num1, WyTimeSpec(num2), Ok,
                             WyTimeSpec(Res*1.000001-1e-9), WyTimeSpec(Res*0.999999+1e-9)};
         itm.verify();
       }
     }
   } 
 }
 
 (double(1.2)-WyTimeSpec())++;   // chk return non-const
};

struct ChkItm_mul_ts {        // check WyTimeSpec multiplication
 typedef double DType;
  const WyTimeSpec SVal_1;
  const WyTimeSpec SVal_2;

  // Expected results
  const WyReply res_reply;
  const WyTimeSpec res_floor;
  const WyTimeSpec res_ceil;

  void verify(void) const
  try {

    // chk operator *=(const WyTimeSpec&)
    try {
      WyTimeSpec tobj(SVal_1);
      tobj*= SVal_2;
      if(res_reply!=Ok) {
        WY_THROW( WyRet() );
      }
      if((tobj<res_floor)||(tobj>res_ceil)) {
        std::cerr << "Res=" << Wy::wrd(tobj).c_str() << '\n';
        WY_THROW( WyRet() );
      }
    }
    catch(const WyTimeSpec::Reply& e) {
      if(e!=res_reply) {
        WY_THROW( WyRet(e) );
      }
    };

    // chk operator*(const WyTimeSpec)
    try {
      WyTimeSpec tobj= SVal_1*SVal_2;
      if(res_reply!=Ok) {
        WY_THROW( WyRet() );
      }
      if((tobj<res_floor)||(tobj>res_ceil)) {
        std::cerr << "Res=" << Wy::wrd(tobj).c_str() << '\n';
        WY_THROW( WyRet() );
      }
    }
    catch(const WyTimeSpec::Reply& e) {
      if(e!=res_reply) {
        WY_THROW( WyRet(e) );
      }
    };

    // chk mul(const WyTimeSpec&)
    {
      WyTimeSpec tobj(SVal_1);
      WyRet r=tobj.mul(SVal_2);
      if(r!=res_reply) {
        WY_THROW( WyRet(r) );
      }
      if(r==Ok) {
        if((tobj<res_floor)||(tobj>res_ceil)) {
          std::cerr << "Res=" << Wy::wrd(tobj).c_str() << '\n';
          WY_THROW( WyRet() );
        }
      }
    }

    // chk mul(sec,nsec)
    {
      WyTimeSpec tobj(SVal_1);
      WyRet r=tobj.mul(SVal_2.second(),SVal_2.nano());
      if(r!=res_reply) {
        WY_THROW( WyRet(r) );
      }
      if(r==Ok) {
        if((tobj<res_floor)||(tobj>res_ceil)) {
          std::cerr << "Res=" << Wy::wrd(tobj).c_str() << '\n';
          WY_THROW( WyRet() );
        }
      }
    }
  } catch(...) {
    std::cerr << "ChkItm_mul_ts Item={" << Wy::wrd(SVal_1).c_str() << ", " 
                                        << Wy::wrd(SVal_2).c_str() << ", " 
              << Wy::wrd(res_reply).c_str() << ", "
              << Wy::wrd(res_floor).c_str() << ", " << Wy::wrd(res_ceil).c_str() << "}\n";
    throw;
  };
};

static void ckop_mul_ts(void)
{
 const ChkItm_mul_ts chk_tab[]={
   // sval reply sfloor sceil nfloor nceil
   { WyTimeSpec(0,0), WyTimeSpec(0,0), Ok, WyTimeSpec(), WyTimeSpec()},
   { WyTimeSpec(1,23), WyTimeSpec(2,34), Ok, WyTimeSpec(2,8), WyTimeSpec(2,80)},
   { WyTimeSpec(-1,-23), WyTimeSpec(2,34), Ok, WyTimeSpec(-2,-80), WyTimeSpec(-2,-80)},
   { WyTimeSpec(-1,-23), WyTimeSpec(-2,-34), Ok, WyTimeSpec(2,80), WyTimeSpec(2,80)},
   { WyTimeSpec(0,999999999), WyTimeSpec(2,0), Ok, WyTimeSpec(1,999999998), WyTimeSpec(1,999999998)},
   { WyTimeSpec(0,-999999999), WyTimeSpec(-2,0), Ok, WyTimeSpec(1,999999998), WyTimeSpec(1,999999998)},
   { WyTimeSpec(999999999,999999998), WyTimeSpec(1,0), Ok,
               WyTimeSpec(999999999,999999998), WyTimeSpec(999999999,999999998)},
   { WyTimeSpec(std::numeric_limits<WyTimeSpec::SecondType>::max(),0), WyTimeSpec(0,1), Ok,
               WyTimeSpec(0,std::numeric_limits<WyTimeSpec::SecondType>::max()),
               WyTimeSpec(0,std::numeric_limits<WyTimeSpec::SecondType>::max())},
   { WyTimeSpec(std::numeric_limits<WyTimeSpec::SecondType>::min(),0), WyTimeSpec(0,1), Ok,
               WyTimeSpec(0,std::numeric_limits<WyTimeSpec::SecondType>::min()),
               WyTimeSpec(0,std::numeric_limits<WyTimeSpec::SecondType>::min())},
   { WyTimeSpec(std::numeric_limits<WyTimeSpec::SecondType>::max()/2,499999999), WyTimeSpec(2,0),
               Ok,
               WyTimeSpec((std::numeric_limits<WyTimeSpec::SecondType>::max()/2)*2,999999998),
               WyTimeSpec((std::numeric_limits<WyTimeSpec::SecondType>::max()/2)*2,999999998)},
   { WyTimeSpec(std::numeric_limits<WyTimeSpec::SecondType>::max()/2,499999999), WyTimeSpec(2,1),
               Wym_ERANGE, WyTimeSpec(), WyTimeSpec()},
   { WyTimeSpec(std::numeric_limits<WyTimeSpec::SecondType>::min()/2,-499999999), WyTimeSpec(2,1),
               Wym_ERANGE, WyTimeSpec(), WyTimeSpec()},
   { WyTimeSpec(std::numeric_limits<WyTimeSpec::SecondType>::max()/2,499999999), WyTimeSpec(-2,-2),
               Wym_ERANGE, WyTimeSpec(), WyTimeSpec()},
   { WyTimeSpec(std::numeric_limits<WyTimeSpec::SecondType>::min()/2,-499999999), WyTimeSpec(-2,-1),
               Wym_ERANGE, WyTimeSpec(), WyTimeSpec()},
   { WyTimeSpec(std::numeric_limits<WyTimeSpec::SecondType>::min()/2,-499999999), WyTimeSpec(2,0),
               Ok,
               WyTimeSpec(2*(std::numeric_limits<WyTimeSpec::SecondType>::min()/2),-999999998),
               WyTimeSpec(2*(std::numeric_limits<WyTimeSpec::SecondType>::min()/2),-999999998) },
   { WyTimeSpec(std::numeric_limits<WyTimeSpec::SecondType>::min()/2,-500000000), WyTimeSpec(2,0),
               Wym_ERANGE, WyTimeSpec(), WyTimeSpec() },
 };
 
 for(size_t i=0; i<sizeof(chk_tab)/sizeof(ChkItm_mul_ts); ++i) {
   chk_tab[i].verify();
 }

 // Random number check
 typedef double DType;
 const long long RandExp=std::numeric_limits<WyTimeSpec::SecondType>::digits+1;
 WyTimeSpec num1,num2;
 for(int i=0; i<RandSmps; ++i) {
   
   //  res= x*2^y  ,x in [-1.0,+1.0], y in [-RandExp, +RandExp]
   //
   DType dtmp= std::ldexp( ::rand()*(2.0/(double)RAND_MAX)-1.0
              , (::rand()*(long long)2*RandExp)/(long long)RAND_MAX-RandExp 
             );
   if(num1.reset(dtmp)!=Ok) {
     continue;
   }
   dtmp= std::ldexp( ::rand()*(2.0/(double)RAND_MAX)-1.0
              , (::rand()*(long long)2*RandExp)/(long long)RAND_MAX-RandExp 
             );
   if(num2.reset(dtmp)!=Ok) {
     continue;
   }
   DType Res=Wy::to_double(WyTimeSpec::max())+1.0;
   try { Res=Wy::to_double(WyTimeSpec(num1)*WyTimeSpec(num2)); }
   catch(const WyTimeSpec::Reply& e) {
     if(e!=Wym_ERANGE) {
       throw;
     }
   };
  
   WyRet r;
   if(Res>Wy::to_double(WyTimeSpec::max())||(Res<Wy::to_double(WyTimeSpec::min()))) {
     ChkItm_mul_ts itm={ num1, num2, Wym_ERANGE, WyTimeSpec(), WyTimeSpec()};
     itm.verify();
     if((Res<std::numeric_limits<WyTimeSpec::SecondType>::max()+1.0)&&
        (Res>std::numeric_limits<WyTimeSpec::SecondType>::min()-1.0)) {
       std::cout << "Error: Res=" << Res << '\n';
       WY_THROW( WyRet() );
     }
   } else {
     if(Res>=0) {
       if(Res*1.000001+1e-9<=Wy::to_double(WyTimeSpec::max())) {
         ChkItm_mul_ts itm={ num1, num2, Ok,
                             WyTimeSpec(Res*0.999999-1e-9), WyTimeSpec(Res*1.000001+1e-9)};
         itm.verify();
       }
     } else {
       if(Res*1.000001-1e-9>=Wy::to_double(WyTimeSpec::min())) {
         ChkItm_mul_ts itm={ num1, num2, Ok,
                             WyTimeSpec(Res*1.000001-1e-9), WyTimeSpec(Res*0.999999+1e-9)};
         itm.verify();
       }
     }
   } 
 }
 
 (WyTimeSpec()*WyTimeSpec())++;   // chk return non-const
};

struct ChkItm_mul_double {        // check WyTimeSpec*double multiplication
 typedef double DType;
  const WyTimeSpec SVal_1;
  const DType SVal_2;

  // Expected results
  const WyReply res_reply;
  const WyTimeSpec res_floor;
  const WyTimeSpec res_ceil;

  void verify(void) const
  try {

    // chk operator *=(double)
    try {
      WyTimeSpec tobj(SVal_1);
      tobj*= SVal_2;
      if(res_reply!=Ok) {
        WY_THROW( WyRet() );
      }
      if((tobj<res_floor)||(tobj>res_ceil)) {
        std::cerr << "Res=" << Wy::wrd(tobj).c_str() << '\n';
        WY_THROW( WyRet() );
      }
    }
    catch(const WyTimeSpec::Reply& e) {
      if(e!=res_reply) {
        WY_THROW( WyRet(e) );
      }
    };

    // chk operator*(double)
    try {
      WyTimeSpec tobj= SVal_1*SVal_2;
      if(res_reply!=Ok) {
        WY_THROW( WyRet() );
      }
      if((tobj<res_floor)||(tobj>res_ceil)) {
        std::cerr << "Res=" << Wy::wrd(tobj).c_str() << '\n';
        WY_THROW( WyRet() );
      }
    }
    catch(const WyTimeSpec::Reply& e) {
      if(e!=res_reply) {
        WY_THROW( WyRet(e) );
      }
    };

  } catch(...) {
    std::cerr << "ChkItm_mul_double Item={" << Wy::wrd(SVal_1).c_str() << ", " 
                                        << SVal_2 << ", " 
              << Wy::wrd(res_reply).c_str() << ", "
              << Wy::wrd(res_floor).c_str() << ", " << Wy::wrd(res_ceil).c_str() << "}\n";
    throw;
  };
};

static void ckop_mul_double(void)
{
 const ChkItm_mul_double chk_tab[]={
   // sval reply sfloor sceil nfloor nceil
   { WyTimeSpec(0,0), 0.0, Ok, WyTimeSpec(), WyTimeSpec()},
   { WyTimeSpec(0,999999999), 2.0, Ok, WyTimeSpec(1,999999998), WyTimeSpec(1,999999998)},
   { WyTimeSpec(0,-999999999), -2.0, Ok, WyTimeSpec(1,999999998), WyTimeSpec(1,999999998)},
   { WyTimeSpec(999999999,999999998), 1.0, Ok,
               WyTimeSpec(999999999,999999998), WyTimeSpec(999999999,999999998)},
   { WyTimeSpec(std::numeric_limits<WyTimeSpec::SecondType>::max(),0), 1e-9, Ok,
               WyTimeSpec(0,std::numeric_limits<WyTimeSpec::SecondType>::max()),
               WyTimeSpec(0,std::numeric_limits<WyTimeSpec::SecondType>::max())},
   { WyTimeSpec(std::numeric_limits<WyTimeSpec::SecondType>::min(),0), 1e-9, Ok,
               WyTimeSpec(0,std::numeric_limits<WyTimeSpec::SecondType>::min()),
               WyTimeSpec(0,std::numeric_limits<WyTimeSpec::SecondType>::min())},
   { WyTimeSpec(std::numeric_limits<WyTimeSpec::SecondType>::max()/2,499999999), 2.0,
               Ok,
                 WyTimeSpec((std::numeric_limits<WyTimeSpec::SecondType>::max()/2)*2,999999998),
                 WyTimeSpec((std::numeric_limits<WyTimeSpec::SecondType>::max()/2)*2,999999998)},
   { WyTimeSpec(std::numeric_limits<WyTimeSpec::SecondType>::min()/2,-1), 2.0,
               Ok,
               WyTimeSpec(2*(std::numeric_limits<WyTimeSpec::SecondType>::min()/2),-2),
               WyTimeSpec(2*(std::numeric_limits<WyTimeSpec::SecondType>::min()/2),-2) },
   { WyTimeSpec(std::numeric_limits<WyTimeSpec::SecondType>::min()/2,-500000000), 2.0,
               Wym_ERANGE, WyTimeSpec(), WyTimeSpec() },
   { WyTimeSpec(0,0), Nan, Wym_ERANGE, WyTimeSpec(), WyTimeSpec()},
   { WyTimeSpec(0,0), Inf, Wym_ERANGE, WyTimeSpec(), WyTimeSpec()},
 };
 
 for(size_t i=0; i<sizeof(chk_tab)/sizeof(ChkItm_mul_double); ++i) {
   chk_tab[i].verify();
 }

 // Random number check
 typedef double DType;
 const long long RandExp=std::numeric_limits<WyTimeSpec::SecondType>::digits+1;
 DType num1,num2;
 for(int i=0; i<RandSmps; ++i) {
   
   //  res= x*2^y  ,x in [-1.0,+1.0], y in [-RandExp, +RandExp]
   //
   DType dtmp= std::ldexp( ::rand()*(2.0/(double)RAND_MAX)-1.0
              , (::rand()*(long long)2*RandExp)/(long long)RAND_MAX-RandExp 
             );
   num1=dtmp;

   dtmp= std::ldexp( ::rand()*(2.0/(double)RAND_MAX)-1.0
              , (::rand()*(long long)2*RandExp)/(long long)RAND_MAX-RandExp 
             );
   num2=dtmp;
   DType Res=Wy::to_double(WyTimeSpec::max())+1.0;
   try { Res=Wy::to_double(WyTimeSpec(num1)*num2); }
   catch(const WyTimeSpec::Reply& e) {
     if(e!=Wym_ERANGE) {
       throw;
     }
   };
  
   WyRet r;
   if(Res>Wy::to_double(WyTimeSpec::max())||(Res<Wy::to_double(WyTimeSpec::min()))) {
     ChkItm_mul_double itm={ WyTimeSpec(num1), num2, Wym_ERANGE, WyTimeSpec(), WyTimeSpec()};
     itm.verify();
     if((Res<std::numeric_limits<WyTimeSpec::SecondType>::max()+1.0)&&
        (Res>std::numeric_limits<WyTimeSpec::SecondType>::min()-1.0)) {
       std::cout << "Error: Res=" << Res << '\n';
       WY_THROW( WyRet() );
     }
   } else {
     if(Res>=0) {
       if(Res*1.000001+1e-9<=Wy::to_double(WyTimeSpec::max())) {
         ChkItm_mul_double itm={ WyTimeSpec(num1), num2, Ok,
                             WyTimeSpec(Res*0.999999-1e-9), WyTimeSpec(Res*1.000001+1e-9)};
         itm.verify();
       }
     } else {
       if(Res*1.000001-1e-9>=Wy::to_double(WyTimeSpec::min())) {
         ChkItm_mul_double itm={ WyTimeSpec(num1), num2, Ok,
                             WyTimeSpec(Res*1.000001-1e-9), WyTimeSpec(Res*0.999999+1e-9)};
         itm.verify();
       }
     }
   } 
 }
 
 (WyTimeSpec()*double(1.2))++;   // chk return non-const
};

struct ChkItm_double_mul {        // check WyTimeSpec*double multiplication
 typedef double DType;
  const DType SVal_1;
  const WyTimeSpec SVal_2;

  // Expected results
  const WyReply res_reply;
  const WyTimeSpec res_floor;
  const WyTimeSpec res_ceil;

  void verify(void) const
  try {

    // chk:  double * WyTimeSpec
    try {
      WyTimeSpec tobj= SVal_1*SVal_2;
      if(res_reply!=Ok) {
        WY_THROW( WyRet() );
      }
      if((tobj<res_floor)||(tobj>res_ceil)) {
        std::cerr << "Res=" << Wy::wrd(tobj).c_str() << '\n';
        WY_THROW( WyRet() );
      }
    }
    catch(const WyTimeSpec::Reply& e) {
      if(e!=res_reply) {
        WY_THROW( WyRet(e) );
      }
    };

  } catch(...) {
    std::cerr << "ChkItm_double_mul Item={" << SVal_1 << ", " 
                                        << Wy::wrd(SVal_2).c_str() << ", " 
              << Wy::wrd(res_reply).c_str() << ", "
              << Wy::wrd(res_floor).c_str() << ", " << Wy::wrd(res_ceil).c_str() << "}\n";
    throw;
  };
};

static void ckop_double_mul(void)
{
 const ChkItm_double_mul chk_tab[]={
   // sval reply sfloor sceil nfloor nceil
   { 0.0, WyTimeSpec(0,0), Ok, WyTimeSpec(), WyTimeSpec()},
   { 2.0, WyTimeSpec(0,999999999), Ok, WyTimeSpec(1,999999998), WyTimeSpec(1,999999998)},
   { -2.0, WyTimeSpec(0,-999999999), Ok, WyTimeSpec(1,999999998), WyTimeSpec(1,999999998)},
   { 1.0, WyTimeSpec(999999999,999999998), Ok,
               WyTimeSpec(999999999,999999998), WyTimeSpec(999999999,999999998)},
   { 1e-9, WyTimeSpec(std::numeric_limits<WyTimeSpec::SecondType>::max(),0), Ok,
               WyTimeSpec(0,std::numeric_limits<WyTimeSpec::SecondType>::max()),
               WyTimeSpec(0,std::numeric_limits<WyTimeSpec::SecondType>::max())},
   { 1e-9, WyTimeSpec(std::numeric_limits<WyTimeSpec::SecondType>::min(),0), Ok,
               WyTimeSpec(0,std::numeric_limits<WyTimeSpec::SecondType>::min()),
               WyTimeSpec(0,std::numeric_limits<WyTimeSpec::SecondType>::min())},
   { 2.0, WyTimeSpec(std::numeric_limits<WyTimeSpec::SecondType>::max()/2,499999999),
               Ok,
               WyTimeSpec((std::numeric_limits<WyTimeSpec::SecondType>::max()/2)*2,999999998),
               WyTimeSpec((std::numeric_limits<WyTimeSpec::SecondType>::max()/2)*2,999999998)},
   { 2.0, WyTimeSpec(std::numeric_limits<WyTimeSpec::SecondType>::min()/2,-1),
               Ok,
               WyTimeSpec(2*(std::numeric_limits<WyTimeSpec::SecondType>::min()/2),-2),
               WyTimeSpec(2*(std::numeric_limits<WyTimeSpec::SecondType>::min()/2),-2) },
   { 2.0, WyTimeSpec(std::numeric_limits<WyTimeSpec::SecondType>::min()/2,-500000000),
               Wym_ERANGE, WyTimeSpec(), WyTimeSpec() },
   { Nan, WyTimeSpec(0,0), Wym_ERANGE, WyTimeSpec(), WyTimeSpec()},
   { Inf, WyTimeSpec(0,0), Wym_ERANGE, WyTimeSpec(), WyTimeSpec()},
 };
 
 for(size_t i=0; i<sizeof(chk_tab)/sizeof(ChkItm_double_mul); ++i) {
   chk_tab[i].verify();
 }

 // Random number check
 typedef double DType;
 const long long RandExp=std::numeric_limits<WyTimeSpec::SecondType>::digits+1;
 DType num1, num2;
 for(int i=0; i<RandSmps; ++i) {
   
   //  res= x*2^y  ,x in [-1.0,+1.0], y in [-RandExp, +RandExp]
   //
   DType dtmp= std::ldexp( ::rand()*(2.0/(double)RAND_MAX)-1.0
              , (::rand()*(long long)2*RandExp)/(long long)RAND_MAX-RandExp 
             );
   num1=dtmp;

   dtmp= std::ldexp( ::rand()*(2.0/(double)RAND_MAX)-1.0
              , (::rand()*(long long)2*RandExp)/(long long)RAND_MAX-RandExp 
             );
   num2=dtmp;
   DType Res=Wy::to_double(WyTimeSpec::max())+1.0;
   try { Res=Wy::to_double( num1 * WyTimeSpec(num2)); }
   catch(const WyTimeSpec::Reply& e) {
     if(e!=Wym_ERANGE) {
       throw;
     }
   };
  
   WyRet r;
   if(Res>Wy::to_double(WyTimeSpec::max())||(Res<Wy::to_double(WyTimeSpec::min()))) {
     ChkItm_double_mul itm={ num1, WyTimeSpec(num2), Wym_ERANGE, WyTimeSpec(), WyTimeSpec()};
     itm.verify();
     if((Res<std::numeric_limits<WyTimeSpec::SecondType>::max()+1.0)&&
        (Res>std::numeric_limits<WyTimeSpec::SecondType>::min()-1.0)) {
       std::cout << "Error: Res=" << Res << '\n';
       WY_THROW( WyRet() );
     }
   } else {
     if(Res>=0) {
       if(Res*1.000001+1e-9<=Wy::to_double(WyTimeSpec::max())) {
         ChkItm_double_mul itm={ num1, WyTimeSpec(num2), Ok,
                             WyTimeSpec(Res*0.999999-1e-9), WyTimeSpec(Res*1.000001+1e-9)};
         itm.verify();
       }
     } else {
       if(Res*1.000001-1e-9>=Wy::to_double(WyTimeSpec::min())) {
         ChkItm_double_mul itm={ num1, WyTimeSpec(num2), Ok,
                             WyTimeSpec(Res*1.000001-1e-9), WyTimeSpec(Res*0.999999+1e-9)};
         itm.verify();
       }
     }
   } 
 }
 
 (double(1.2)*WyTimeSpec())++;   // chk return non-const
};

struct ChkItm_div_double {        // check WyTimeSpec/double multiplication
 typedef double DType;
  const WyTimeSpec SVal_1;
  const DType SVal_2;

  // Expected results
  const WyReply res_reply;
  const WyTimeSpec res_floor;
  const WyTimeSpec res_ceil;

  void verify(void) const
  try {

    // chk operator /=(double)
    try {
      WyTimeSpec tobj(SVal_1);
      tobj/= SVal_2;
      if(res_reply!=Ok) {
        WY_THROW( WyRet() );
      }
      if((tobj<res_floor)||(tobj>res_ceil)) {
        std::cerr << "Res=" << Wy::wrd(tobj).c_str() << '\n';
        WY_THROW( WyRet() );
      }
    }
    catch(const WyTimeSpec::Reply& e) {
      if(e!=res_reply) {
        WY_THROW( WyRet(e) );
      }
    };

    // chk operator/(double)
    try {
      WyTimeSpec tobj= SVal_1/SVal_2;
      if(res_reply!=Ok) {
        WY_THROW( WyRet() );
      }
      if((tobj<res_floor)||(tobj>res_ceil)) {
        std::cerr << "Res=" << Wy::wrd(tobj).c_str() << '\n';
        WY_THROW( WyRet() );
      }
    }
    catch(const WyTimeSpec::Reply& e) {
      if(e!=res_reply) {
        WY_THROW( WyRet(e) );
      }
    };

  } catch(...) {
    std::cerr << "ChkItm_div_double Item={" << Wy::wrd(SVal_1).c_str() << ", " 
                                        << SVal_2 << ", " 
              << Wy::wrd(res_reply).c_str() << ", "
              << Wy::wrd(res_floor).c_str() << ", " << Wy::wrd(res_ceil).c_str() << "}\n";
    throw;
  };
};

static void ckop_div_double(void)
{
 const ChkItm_div_double chk_tab[]={
   // sval reply sfloor sceil nfloor nceil
   { WyTimeSpec(0,0), 0.0, Wym_EDIVZERO, WyTimeSpec(), WyTimeSpec()},
   { WyTimeSpec(2,2), 2.0, Ok, WyTimeSpec(1,0), WyTimeSpec(1,3)},
   { WyTimeSpec(0,999999999), 2.0, Ok, WyTimeSpec(0,499999998), WyTimeSpec(0,500000000)},
   { WyTimeSpec(0,-999999999), 2.0, Ok, WyTimeSpec(-0,-500000000), WyTimeSpec(-0,-499999998)},
   { WyTimeSpec(999999999,999999998), 1.0, Ok,
               WyTimeSpec(999999999,999999997), WyTimeSpec(999999999,999999999)},
   { WyTimeSpec(1,0), 1e-9, Ok, WyTimeSpec(999999999,999999930), WyTimeSpec(1000000000,70) },
   { WyTimeSpec(0,9999), 2.0, Ok, WyTimeSpec(0,4999), WyTimeSpec(0,5001)},
   { WyTimeSpec(std::numeric_limits<WyTimeSpec::SecondType>::max(),0), -1.0, Ok,
                 WyTimeSpec(-std::numeric_limits<WyTimeSpec::SecondType>::max(),-100),
                 WyTimeSpec(-std::numeric_limits<WyTimeSpec::SecondType>::max()+10,0)},
   { WyTimeSpec(std::numeric_limits<WyTimeSpec::SecondType>::min()/2,-1), 0.5,
               Ok, WyTimeSpec(std::numeric_limits<WyTimeSpec::SecondType>::min(),-4),
                   WyTimeSpec(std::numeric_limits<WyTimeSpec::SecondType>::min()+1,0) },
   { WyTimeSpec(0,0), Nan, Wym_ERANGE, WyTimeSpec(), WyTimeSpec()},
   { WyTimeSpec(0,0), Inf, Wym_ERANGE, WyTimeSpec(), WyTimeSpec()},
 };
 
 for(size_t i=0; i<sizeof(chk_tab)/sizeof(ChkItm_div_double); ++i) {
   chk_tab[i].verify();
 }

 // Random number check
 typedef double DType;
 const long long RandExp=std::numeric_limits<WyTimeSpec::SecondType>::digits+1;
 DType num1,num2;
 for(int i=0; i<RandSmps; ++i) {
   
   //  res= x*2^y  ,x in [-1.0,+1.0], y in [-RandExp, +RandExp]
   //
   DType dtmp= std::ldexp( ::rand()*(2.0/(double)RAND_MAX)-1.0
              , (::rand()*(long long)2*RandExp)/(long long)RAND_MAX-RandExp 
             );
   num2=dtmp;
   if((num2>-11e-9)||(num2<1e-9)) {
     num2=1e-9;
   }

   dtmp= std::ldexp( ::rand()*(2.0/(double)RAND_MAX)-1.0
              , (::rand()*(long long)2*RandExp)/(long long)RAND_MAX-RandExp 
             );
   num1=dtmp;
   DType Res=Wy::to_double(WyTimeSpec::max())+1.0;
   try { Res=Wy::to_double(WyTimeSpec(num1)/num2); }
   catch(const WyTimeSpec::Reply& e) {
     if(e!=Wym_ERANGE) {
       throw;
     }
   };
  
   WyRet r;
   if(Res>Wy::to_double(WyTimeSpec::max())||(Res<Wy::to_double(WyTimeSpec::min()))) {
     ChkItm_div_double itm={ WyTimeSpec(num1), num2, Wym_ERANGE, WyTimeSpec(), WyTimeSpec()};
     itm.verify();
     if((Res<std::numeric_limits<WyTimeSpec::SecondType>::max()+1.0)&&
        (Res>std::numeric_limits<WyTimeSpec::SecondType>::min()-1.0)) {
       std::cout << "Error: Res=" << Res << '\n';
       WY_THROW( WyRet() );
     }
   } else {
     if(Res>=0) {
       if(Res*1.000001+1e-9<=Wy::to_double(WyTimeSpec::max())) {
         ChkItm_div_double itm={ WyTimeSpec(num1), num2, Ok,
                             WyTimeSpec(Res*0.999999-1e-9), WyTimeSpec(Res*1.000001+1e-9)};
         itm.verify();
       }
     } else {
       if(Res*1.000001-1e-9>=Wy::to_double(WyTimeSpec::min())) {
         ChkItm_div_double itm={ WyTimeSpec(num1), num2, Ok,
                             WyTimeSpec(Res*1.000001-1e-9), WyTimeSpec(Res*0.999999+1e-9)};
         itm.verify();
       }
     }
   } 
 }
 
 (WyTimeSpec()/double(1.2))++;   // chk return non-const
};

struct ChkItm_double_div {        // check double/WyTimeSpec
 typedef double DType;
  const DType SVal_1;
  const WyTimeSpec SVal_2;

  // Expected results
  const WyReply res_reply;
  const WyTimeSpec res_floor;
  const WyTimeSpec res_ceil;

  void verify(void) const
  try {

    // chk double / WyTimeSpec
    try {
      WyTimeSpec tobj= SVal_1/SVal_2;
      if(res_reply!=Ok) {
        WY_THROW( WyRet() );
      }
      if((tobj<res_floor)||(tobj>res_ceil)) {
        std::cerr << "Res=" << Wy::wrd(tobj).c_str() << '\n';
        WY_THROW( WyRet() );
      }
    }
    catch(const WyTimeSpec::Reply& e) {
      if(e!=res_reply) {
        WY_THROW( WyRet(e) );
      }
    };

  } catch(...) {
    std::cerr << "ChkItm_double_div Item={" << SVal_1 << ", " 
                                        << Wy::wrd(SVal_2).c_str() << ", " 
              << Wy::wrd(res_reply).c_str() << ", "
              << Wy::wrd(res_floor).c_str() << ", " << Wy::wrd(res_ceil).c_str() << "}\n";
    throw;
  };
};

static void ckop_double_div(void)
{
 const ChkItm_double_div chk_tab[]={
   // sval reply sfloor sceil nfloor nceil
   { 0.0, WyTimeSpec(0,0), Wym_EDIVZERO, WyTimeSpec(), WyTimeSpec()},
   { 2+2e-9, WyTimeSpec(2,0), Ok, WyTimeSpec(1,0), WyTimeSpec(1,3)},
   { 0.999999999, WyTimeSpec(2,0), Ok, WyTimeSpec(0,499999998), WyTimeSpec(0,500000000)},
   { -0.999999999, WyTimeSpec(2,0), Ok, WyTimeSpec(-0,-500000000), WyTimeSpec(-0,-499999998)},
   { 99.99, WyTimeSpec(1.1), Ok,
               WyTimeSpec(90,890000000), WyTimeSpec(90,999000000)},
   { 1.0, WyTimeSpec(0,1), Ok, WyTimeSpec(999999999,999999930), WyTimeSpec(1000000000,70) },
   { 0.9999, WyTimeSpec(2,0), Ok, WyTimeSpec(0,499900000), WyTimeSpec(0,500100000)},
   { std::numeric_limits<WyTimeSpec::SecondType>::min()/2.0, WyTimeSpec(0,500000000), Ok,
               WyTimeSpec(std::numeric_limits<WyTimeSpec::SecondType>::min(),-4),
               WyTimeSpec(std::numeric_limits<WyTimeSpec::SecondType>::min()+1,0) },
   { Nan, WyTimeSpec(0,0), Wym_ERANGE, WyTimeSpec(), WyTimeSpec()},
   { Inf, WyTimeSpec(0,0), Wym_ERANGE, WyTimeSpec(), WyTimeSpec()},
 };
 
 for(size_t i=0; i<sizeof(chk_tab)/sizeof(ChkItm_double_div); ++i) {
   chk_tab[i].verify();
 }

 // Random number check
 typedef double DType;
 const long long RandExp=std::numeric_limits<WyTimeSpec::SecondType>::digits+1;
 DType num1,num2;
 for(int i=0; i<RandSmps; ++i) {
   
   //  res= x*2^y  ,x in [-1.0,+1.0], y in [-RandExp, +RandExp]
   //
   DType dtmp= std::ldexp( ::rand()*(2.0/(double)RAND_MAX)-1.0
              , (::rand()*(long long)2*RandExp)/(long long)RAND_MAX-RandExp 
             );
   num2=dtmp;
   if((num2>-11e-9)||(num2<1e-9)) {
     num2=1e-9;
   }

   dtmp= std::ldexp( ::rand()*(2.0/(double)RAND_MAX)-1.0
              , (::rand()*(long long)2*RandExp)/(long long)RAND_MAX-RandExp 
             );
   num1=dtmp;
   DType Res=Wy::to_double(WyTimeSpec::max())+1.0;
   try { Res=Wy::to_double( num1/WyTimeSpec(num2)); }
   catch(const WyTimeSpec::Reply& e) {
     if(e!=Wym_ERANGE) {
       throw;
     }
   };
  
   WyRet r;
   if(Res>Wy::to_double(WyTimeSpec::max())||(Res<Wy::to_double(WyTimeSpec::min()))) {
     ChkItm_double_div itm={ num1, WyTimeSpec(num2), Wym_ERANGE, WyTimeSpec(), WyTimeSpec()};
     itm.verify();
     if((Res<std::numeric_limits<WyTimeSpec::SecondType>::max()+1.0)&&
        (Res>std::numeric_limits<WyTimeSpec::SecondType>::min()-1.0)) {
       std::cout << "Error: Res=" << Res << '\n';
       WY_THROW( WyRet() );
     }
   } else {
     if(Res>=0) {
       if(Res*1.000001+1e-9<=Wy::to_double(WyTimeSpec::max())) {
         ChkItm_double_div itm={ num1, WyTimeSpec(num2), Ok,
                             WyTimeSpec(Res*0.999999-1e-9), WyTimeSpec(Res*1.000001+1e-9)};
         itm.verify();
       }
     } else {
       if(Res*1.000001-1e-9>=Wy::to_double(WyTimeSpec::min())) {
         ChkItm_double_div itm={ num1, WyTimeSpec(num2), Ok,
                             WyTimeSpec(Res*1.000001-1e-9), WyTimeSpec(Res*0.999999+1e-9)};
         itm.verify();
       }
     }
   } 
 }
 
 (double(1.2)/WyTimeSpec(1,2))++;   // chk return non-const
};

struct ChkItm_div_ts {        // check WyTimeSpec division
 typedef double DType;
  const WyTimeSpec SVal_1;
  const WyTimeSpec SVal_2;

  // Expected results
  const WyReply res_reply;
  const WyTimeSpec res_floor;
  const WyTimeSpec res_ceil;

  void verify(void) const
  try {

    // chk operator /=(const WyTimeSpec&)
    try {
      WyTimeSpec tobj(SVal_1);
      tobj/= SVal_2;
      if(res_reply!=Ok) {
        WY_THROW( WyRet() );
      }
      if((tobj<res_floor)||(tobj>res_ceil)) {
        std::cerr << "Res=" << Wy::wrd(tobj).c_str() << '\n';
        WY_THROW( WyRet() );
      }
    }
    catch(const WyTimeSpec::Reply& e) {
      if(e!=res_reply) {
        WY_THROW( WyRet(e) );
      }
    };

    // chk operator/(const WyTimeSpec)
    try {
      WyTimeSpec tobj= SVal_1/SVal_2;
      if(res_reply!=Ok) {
        WY_THROW( WyRet() );
      }
      if((tobj<res_floor)||(tobj>res_ceil)) {
        std::cerr << "Res=" << Wy::wrd(tobj).c_str() << '\n';
        WY_THROW( WyRet() );
      }
    }
    catch(const WyTimeSpec::Reply& e) {
      if(e!=res_reply) {
        WY_THROW( WyRet(e) );
      }
    };

    // chk div(const WyTimeSpec&)
    {
      WyRet r;
      WyTimeSpec tobj(SVal_1);
      r=tobj.div(SVal_2);
      if(r!=res_reply) {
        WY_THROW( r );
      }
      if(r==Ok) {
        if((tobj<res_floor)||(tobj>res_ceil)) {
          std::cerr << "Res=" << Wy::wrd(tobj).c_str() << '\n';
          WY_THROW( WyRet() );
        }
      }
    }

    // chk div(sec,nsec)
    {
      WyRet r;
      WyTimeSpec tobj(SVal_1);
      r=tobj.div(SVal_2.second(),SVal_2.nano());
      if(r!=res_reply) {
        WY_THROW( r );
      }
      if(r==Ok) {
        if((tobj<res_floor)||(tobj>res_ceil)) {
          std::cerr << "Res=" << Wy::wrd(tobj).c_str() << '\n';
          WY_THROW( WyRet() );
        }
      }
    }

  } catch(...) {
    std::cerr << "ChkItm_div_ts Item={" << Wy::wrd(SVal_1).c_str() << ", " 
                                        << Wy::wrd(SVal_2).c_str() << ", " 
              << Wy::wrd(res_reply).c_str() << ", "
              << Wy::wrd(res_floor).c_str() << ", " << Wy::wrd(res_ceil).c_str() << "}\n";
    throw;
  };
};

static void ckop_div_ts(void)
{
 const ChkItm_div_ts chk_tab[]={
   // sval reply sfloor sceil nfloor nceil
   { WyTimeSpec(1,0), WyTimeSpec(1,0), Ok, WyTimeSpec(1,0), WyTimeSpec(1,0)},
   { WyTimeSpec(0,0), WyTimeSpec(0,0), Wym_EDIVZERO, WyTimeSpec(), WyTimeSpec()},
   { WyTimeSpec(0,999999999), WyTimeSpec(2,0), Ok, WyTimeSpec(0,499999999), WyTimeSpec(0,499999999)},
   { WyTimeSpec(0,-999999999), WyTimeSpec(-2,0), Ok, WyTimeSpec(0,499999999), WyTimeSpec(0,499999999)},
   { WyTimeSpec(999999999,999999998), WyTimeSpec(1,0), Ok,
               WyTimeSpec(999999999,999999998), WyTimeSpec(999999999,999999998)},
   { WyTimeSpec(999999999,999999998), WyTimeSpec(2,0), Ok,
               WyTimeSpec(499999999,999999999), WyTimeSpec(499999999,999999999)},
   { WyTimeSpec(0,9999), WyTimeSpec(2,0), Ok,
               WyTimeSpec(0,4999), WyTimeSpec(0,4999)},
   { WyTimeSpec(std::numeric_limits<WyTimeSpec::SecondType>::max(),999999999), WyTimeSpec(-1,0),
               Ok,
               WyTimeSpec(-std::numeric_limits<WyTimeSpec::SecondType>::max(),-999999999),
               WyTimeSpec(-std::numeric_limits<WyTimeSpec::SecondType>::max(),-999999999)},
   { WyTimeSpec(std::numeric_limits<WyTimeSpec::SecondType>::min(),-999999999), WyTimeSpec(1,0),
               Ok,
               WyTimeSpec(-std::numeric_limits<WyTimeSpec::SecondType>::min(),-999999999),
               WyTimeSpec(-std::numeric_limits<WyTimeSpec::SecondType>::min(),-999999999)},
   { WyTimeSpec(std::numeric_limits<WyTimeSpec::SecondType>::max()/2,499999999),
     WyTimeSpec(0,500000000), Ok, 
               WyTimeSpec(2*(std::numeric_limits<WyTimeSpec::SecondType>::max()/2),999999998),
               WyTimeSpec(2*(std::numeric_limits<WyTimeSpec::SecondType>::max()/2),999999998) },

   { WyTimeSpec(std::numeric_limits<WyTimeSpec::SecondType>::max(),999999999),
     WyTimeSpec(0,999999999), Wym_ERANGE, WyTimeSpec(), WyTimeSpec() },
   { WyTimeSpec(std::numeric_limits<WyTimeSpec::SecondType>::max(),0),
     WyTimeSpec(0,-999999999), Wym_ERANGE, WyTimeSpec(), WyTimeSpec() },
   { WyTimeSpec(std::numeric_limits<WyTimeSpec::SecondType>::min(),-999999999),
     WyTimeSpec(0,999999999), Wym_ERANGE, WyTimeSpec(), WyTimeSpec() },
   { WyTimeSpec(std::numeric_limits<WyTimeSpec::SecondType>::min(),-0),
     WyTimeSpec(0,-999999999), Wym_ERANGE, WyTimeSpec(), WyTimeSpec() },

   { WyTimeSpec(3*(std::numeric_limits<WyTimeSpec::SecondType>::max()/3),999999999),
     WyTimeSpec(3,0), Ok, 
               WyTimeSpec(std::numeric_limits<WyTimeSpec::SecondType>::max()/3,333333333),
               WyTimeSpec(std::numeric_limits<WyTimeSpec::SecondType>::max()/3,333333333) },
   { WyTimeSpec(3*(std::numeric_limits<WyTimeSpec::SecondType>::max()/3),999999999),
     WyTimeSpec(-3,0), Ok, 
               WyTimeSpec(-(std::numeric_limits<WyTimeSpec::SecondType>::max()/3),-333333333),
               WyTimeSpec(-(std::numeric_limits<WyTimeSpec::SecondType>::max()/3),-333333333) },

   { WyTimeSpec(std::numeric_limits<WyTimeSpec::SecondType>::min()/2,-499999999),
     WyTimeSpec(0,500000000), Ok, 
               WyTimeSpec(2*(std::numeric_limits<WyTimeSpec::SecondType>::min()/2),-999999998),
               WyTimeSpec(2*(std::numeric_limits<WyTimeSpec::SecondType>::min()/2),-999999998) },
   { WyTimeSpec(3*(std::numeric_limits<WyTimeSpec::SecondType>::min()/3),-999999999),
     WyTimeSpec(3,0), Ok, 
               WyTimeSpec(std::numeric_limits<WyTimeSpec::SecondType>::min()/3,-333333333),
               WyTimeSpec(std::numeric_limits<WyTimeSpec::SecondType>::min()/3,-333333333) },
   { WyTimeSpec(3*(std::numeric_limits<WyTimeSpec::SecondType>::min()/3),-999999999),
     WyTimeSpec(-3,0), Ok, 
               WyTimeSpec(-(std::numeric_limits<WyTimeSpec::SecondType>::min()/3),333333333),
               WyTimeSpec(-(std::numeric_limits<WyTimeSpec::SecondType>::min()/3),333333333) },

   { WyTimeSpec::min(), WyTimeSpec::max(), Ok, WyTimeSpec(-1,-1), WyTimeSpec(-1,0) },
   { WyTimeSpec::max(), WyTimeSpec::min(), Ok, WyTimeSpec(-1,0), WyTimeSpec(-1,0) },
   { WyTimeSpec::max(), WyTimeSpec::max(), Ok, WyTimeSpec(1,0), WyTimeSpec(1,0) },
   { WyTimeSpec::min(), WyTimeSpec::min(), Ok, WyTimeSpec(1,0), WyTimeSpec(1,0) },
 };
 
 for(size_t i=0; i<sizeof(chk_tab)/sizeof(ChkItm_div_ts); ++i) {
   chk_tab[i].verify();
 }

 // Random number check
 typedef double DType;
 const long long RandExp=std::numeric_limits<WyTimeSpec::SecondType>::digits+1;
 WyTimeSpec num1,num2;
 for(int i=0; i<RandSmps; ++i) {
   
   //  res= x*2^y  ,x in [-1.0,+1.0], y in [-RandExp, +RandExp]
   //
   DType dtmp= std::ldexp( ::rand()*(2.0/(double)RAND_MAX)-1.0
              , (::rand()*(long long)2*RandExp)/(long long)RAND_MAX-RandExp 
             );
   if(num1.reset(dtmp)!=Ok) {
     continue;
   }
   dtmp= std::ldexp( ::rand()*(2.0/(double)RAND_MAX)-1.0
              , (::rand()*(long long)2*RandExp)/(long long)RAND_MAX-RandExp 
             );
   if(num2.reset(dtmp)!=Ok) {
     continue;
   }
   if(num2.is_default()) {
     continue;
   }
   DType Res=Wy::to_double(WyTimeSpec::max())+1.0;
   try { Res=Wy::to_double(WyTimeSpec(num1)/WyTimeSpec(num2)); }
   catch(const WyTimeSpec::Reply& e) {
     if(e!=Wym_ERANGE) {
       throw;
     }
   };
  
   WyRet r;
   if(Res>Wy::to_double(WyTimeSpec::max())||(Res<Wy::to_double(WyTimeSpec::min()))) {
     ChkItm_div_ts itm={ num1, num2, Wym_ERANGE, WyTimeSpec(), WyTimeSpec()};
     itm.verify();
     if((Res<std::numeric_limits<WyTimeSpec::SecondType>::max()+1.0)&&
        (Res>std::numeric_limits<WyTimeSpec::SecondType>::min()-1.0)) {
       std::cout << "Error: Res=" << Res << '\n';
       WY_THROW( WyRet() );
     }
   } else {
     if(Res>=0) {
       if(Res*1.000001+1e-9<=Wy::to_double(WyTimeSpec::max())) {
         ChkItm_div_ts itm={ num1, num2, Ok,
                             WyTimeSpec(Res*0.999999-1e-9), WyTimeSpec(Res*1.00001+1e-9)};
         itm.verify();
       }
     } else {
       if(Res*1.000001-1e-9>=Wy::to_double(WyTimeSpec::min())) {
         ChkItm_div_ts itm={ num1, num2, Ok,
                             WyTimeSpec(Res*1.000001-1e-9), WyTimeSpec(Res*0.999999+1e-9)};
         itm.verify();
       }
     }
   } 
 }
 
 (WyTimeSpec()/WyTimeSpec(1,2))++;   // chk return non-const
};

void ck2_timespec(void)
{
 ck_double();
 ckop_add_ts();
 ckop_sub_ts();
 ckop_mul_ts();
 ckop_div_ts();
 ckop_add_n();
 ckop_sub_n();
 ckop_mul_n();
 ckop_div_n();

 ckop_add_double();
 ckop_double_add();
 ckop_sub_double();
 ckop_double_sub();
 ckop_mul_double();
 ckop_double_mul();
 ckop_div_double();
 ckop_double_div();
};
