/* Copyright is licensed under GNU LGPL.                 by I.J.Wang 2003

   Check WyRegFile public members

   Note: Test data files should exist in the working directory
         Files may be created in the working directory and unlinked
*/

#include "WyRegFile.h"
#include "WyFileStat.h"
#include "Wy_AtDestroy.hh"
#include "chk_data.h"
#include <stdlib.h>
#include <unistd.h>
#include <new>

#if WYREGFILE_VERSION!=33
#error Test code is for WYREGFILE_VERSION 33
#endif

void ck_regfile(bool,const char*);

#ifdef WY_DEBUG
extern int wydbg_get_fdcnt(void) throw();
extern int wydbg_get_fdcnt_max(void) throw();
static int initial_fdcnt=0;
#endif

static const WyStr Tmp_Name("tmp.chk_tobjile-open_tmpfile.tmp");
static bool exaustfd_flag=true;
static WyStr arg_rtfile;    // filename for read test file

// Assert regf is default
//
static void assert_default(const WyRegFile& regf)
{
 WyRet r;

 if(regf.is_default()==false) {
   WY_THROW( WyRet() );
 }
 if(regf.fh().is_default()==false) {
   WY_THROW( WyRet() );
 }

 // stat
 {
   WyFileStat stt;
   if((r=regf.stat(stt))!=Wym_EBADF) {
     WY_THROW(r);
   }
   if(stt.is_default()==false) {
     WY_THROW( WyRet() );
   }
 }

 // chk fh()
 if(regf.fh()!=WyFileHandle()) {
   WY_THROW( WyRet() );
 }

 // chk pos(0)
 if(regf.pos()!=0) {
   WY_THROW( WyRet() );
 }
 if(static_cast<const WyByteFlow&>(regf).ftype()!=0) {
   WY_THROW( WyRet() );
 }
};

static void assert_not_default(const WyRegFile& regf)
{
 WyRet r;

 if(regf.is_default()) {
   WY_THROW( WyRet() );
 }
 if(regf.fh().is_default()) {
   WY_THROW( WyRet() );
 }

 // stat
 {
   WyFileStat stt;
   if((r=regf.stat(stt))!=Ok) {
     WY_THROW(r);
   }
   if(stt.is_reg()==false) {
     WY_THROW( WyRet() );
   }
 }
 if(static_cast<const WyByteFlow&>(regf).ftype()!=S_IFREG) {
   WY_THROW( WyRet() );
 }
};

// Assert f1,f2 refers to the same systerm file
//
template<typename Dev1, typename Dev2>
static void assert_same_sysf(const Dev1& f1,const Dev2& f2)
{
 if(f1.is_default()) {
   assert_default(f1);
 } else {
   assert_not_default(f1);
 }

 if(f2.is_default()) {
   assert_default(f2);
 } else {
   assert_not_default(f2);
 }

 if(f1.is_default()&&f2.is_default()) {
   return;
 }
 WyRet r;
 WyFileStat stt1,stt2;
 if((r=f1.stat(stt1))!=Ok) {
   WY_THROW( WyRet() );
 }
 if((r=f2.stat(stt2))!=Ok) {
   WY_THROW( WyRet() );
 }
 if(stt1.sf_ino()!=stt2.sf_ino()) {
   WY_THROW( WyRet() );
 }
 if(stt1.sf_dev()!=stt2.sf_dev()) {
   WY_THROW( WyRet() );
 }
};

#define CHK_DEFAULT(p)\
    try { assert_default(p); } catch(...) { WY_THROW( WyRet()); }
#define CHK_NOT_DEFAULT(p)\
    try { assert_not_default(p); } catch(...) { WY_THROW( WyRet()); }
#define CHK_SAME_SYSF(p1,p2)\
    try { assert_same_sysf(p1,p2); } catch(...) { WY_THROW( WyRet()); }

static void exaust_cpy_ctor(const WySysFile& sysf)
{
 WyRegFile tmp(sysf);
 exaust_cpy_ctor(sysf);   // recursive until failure
};

// Check basic operations on default object
//
static void t1(void)
{
 WyRegFile EmptyFile(TestFile_Empty,O_RDONLY);

 WyRet r;

 // Prepare
 //
 CHK_NOT_DEFAULT(EmptyFile);

 // WyRegFile()/WyRegFile(const WyRegFile&)/(const WyByteFlow&)
 {
   WyRegFile tobj;
   WyRegFile tmp1(tobj);
   CHK_DEFAULT(tobj);
   CHK_DEFAULT(tmp1);

   WyRegFile tmp2(EmptyFile);
   CHK_NOT_DEFAULT(tmp2);
   if(tmp2.fh()==EmptyFile.fh()) {
     WY_THROW( WyRet() );
   }
   CHK_SAME_SYSF(tmp2,EmptyFile);
   if(tmp2.pos()!=EmptyFile.pos()) {
     WY_THROW( WyRet() );
   }

   WyRegFile tmp3(static_cast<WyByteFlow&>(EmptyFile));
   CHK_SAME_SYSF(tmp3,EmptyFile);

   WyRegFile tmp4(static_cast<WyByteFlow&>(tobj));
   CHK_SAME_SYSF(tmp4,tobj);
 }

 // WyRegFile(const char*,int)
 {
   WyRegFile tmp(TestFile_Empty.c_str(),O_RDONLY); // same as EmptyFile
   CHK_NOT_DEFAULT(tmp);
   CHK_SAME_SYSF(tmp,EmptyFile);
   if(tmp.pos()!=EmptyFile.pos()) {
     WY_THROW( WyRet() );
   }
 }

 // WyRegFile(const WyStr&,int)
 {
   WyRegFile tmp(TestFile_Empty,O_RDONLY); // same as EmptyFile
   CHK_NOT_DEFAULT(tmp);
   CHK_SAME_SYSF(tmp,EmptyFile);
   if(tmp.pos()!=EmptyFile.pos()) {
     WY_THROW( WyRet() );
   }
 }

 // WyRegFile(WyFileHandle)
 {
   WyFileHandle dfh;
   WyRegFile tmp1( dfh );
   CHK_DEFAULT(tmp1);

   WyRegFile tmp2( EmptyFile.fh() ); // same as EmptyFile
   CHK_NOT_DEFAULT(tmp2);
   CHK_SAME_SYSF(tmp2,EmptyFile);
   if(tmp2.pos()!=EmptyFile.pos()) {
     WY_THROW( WyRet() );
   }
 }

 // reset()
 {
   WyRegFile tmp;
   CHK_DEFAULT(tmp);
   if((r=tmp.reset())!=Ok) {
     WY_THROW(r);
   }
   CHK_DEFAULT(tmp);
 }

 // reset(const WyByteFlow&)
 {
   WyRegFile tmp;
   CHK_DEFAULT(tmp);
   if((r=tmp.reset(EmptyFile))!=Ok) {
     WY_THROW(r);
   }
   CHK_SAME_SYSF(tmp,EmptyFile);
   if(tmp.pos()!=EmptyFile.pos()) {
     WY_THROW( WyRet() );
   }
   if(tmp.reset()!=Ok) {
     WY_THROW(r);
   }
   if((r=tmp.reset(static_cast<const WyByteFlow&>(EmptyFile)))!=Ok) {
     WY_THROW(r);
   }
   CHK_SAME_SYSF(tmp,EmptyFile);
   if(tmp.pos()!=EmptyFile.pos()) {
     WY_THROW( WyRet() );
   }

   tmp.reset();
   CHK_DEFAULT(tmp);
   if((r=tmp.reset(WyRegFile()))!=Ok) {
     WY_THROW(r);
   }
   CHK_DEFAULT(tmp);
 }

 // reset(const char*,int)
 {
   WyRegFile tmp;
   CHK_DEFAULT(tmp);
   if((r=tmp.reset(TestFile_Empty.c_str(),O_RDONLY))!=Ok) {
     WY_THROW(r);
   }
   CHK_SAME_SYSF(tmp,EmptyFile);
   if(tmp.pos()!=EmptyFile.pos()) {
     WY_THROW( WyRet() );
   }
 }
 
 // reset(const WyStr&,int)
 {
   WyRegFile tmp;
   CHK_DEFAULT(tmp);
   if((r=tmp.reset(TestFile_Empty,O_RDONLY))!=Ok) {
     WY_THROW(r);
   }
   CHK_SAME_SYSF(tmp,EmptyFile);
   if(tmp.pos()!=EmptyFile.pos()) {
     WY_THROW( WyRet() );
   }
 }
 
 // reset(WyFileHandle)
 {
   WyRegFile tmp;
   CHK_DEFAULT(tmp);
   if((r=tmp.reset( EmptyFile.fh() ))!=Ok) {
     WY_THROW(r);
   }
   CHK_SAME_SYSF(tmp,EmptyFile);
   if(tmp.pos()!=EmptyFile.pos()) {
     WY_THROW( WyRet() );
   }

   tmp.reset();
   CHK_DEFAULT(tmp);
   if((r=tmp.reset( WyFileHandle() ))!=Ok) {
     WY_THROW(r);
   }
   CHK_DEFAULT(tmp);
 }

 // read
 {
   WyRegFile tobj;
   char buf[2]={-1,-1};
   size_t n_rd=5;   // anything not zero
   CHK_DEFAULT(tobj);
   if((r=tobj.read(buf,sizeof(buf),n_rd))!=Wym_EBADF) {
     WY_THROW(r);
   }
   if(n_rd!=0) {
     WY_THROW( WyRet() );
   }
   n_rd=5;
   if((r=EmptyFile.read(buf,sizeof(buf),n_rd))!=Ok) {
     WY_THROW(r);
   }
   if(n_rd!=0) {
     WY_THROW( WyRet() );
   }
   if(EmptyFile.pos()!=0) {
     WY_THROW( WyRet() );
   }
   CHK_DEFAULT(tobj);
 }

 // read WyStr
 {
   WyRegFile tobj;
   WyStr buf;
   size_t n_rd=5;   // anything not zero
   CHK_DEFAULT(tobj);
   if((r=tobj.read(buf,size_t(2),n_rd))!=Wym_EBADF) {
     WY_THROW(r);
   }
   if(n_rd!=0) {
     WY_THROW( WyRet() );
   }
   if(buf.is_default()==false) {
     WY_THROW( WyRet() );
   }
   n_rd=5;
   if((r=EmptyFile.read(buf,size_t(2),n_rd))!=Ok) {
     WY_THROW(r);
   }
   if(n_rd!=0) {
     WY_THROW( WyRet() );
   }
   if(EmptyFile.pos()!=0) {
     WY_THROW( WyRet() );
   }
   CHK_DEFAULT(tobj);
 }

 // write
 {
   WyRegFile tobj;
   char buf[2]={-1,-1};
   size_t n_wr=5;   // anything not zero
   CHK_DEFAULT(tobj);
   if((r=tobj.write(buf,sizeof(buf),n_wr))!=Wym_EBADF) {
     WY_THROW(r);
   }
   if(n_wr!=0) {
     WY_THROW( WyRet() );
   }
   n_wr=5;
   CHK_DEFAULT(tobj);
   if((r=tobj.write(buf,sizeof(buf),n_wr))!=Wym_EBADF) {
     WY_THROW(r);
   }
   if(n_wr!=0) {
     WY_THROW( WyRet() );
   }
   CHK_DEFAULT(tobj);
   if((r=tobj.write(WyStr( WyCSeg(buf,sizeof(buf)) ),n_wr))!=Wym_EBADF) {
     WY_THROW(r);
   }
   if(n_wr!=0) {
     WY_THROW( WyRet() );
   }
   CHK_DEFAULT(tobj);
 }

 // _fdatasync,_fsync
 {
   WyRegFile tobj;
   CHK_DEFAULT(tobj);
   if((r=tobj._fdatasync())!=Wym_EBADF) {
     WY_THROW(r);
   }
   if((r=tobj._fsync())!=Wym_EBADF) {
     WY_THROW(r);
   }
   CHK_DEFAULT(tobj);
 }

 // pos()/set_pos()
 {
   WyRegFile tobj;
   CHK_DEFAULT(tobj);
   if(tobj.pos()!=0) {
     WY_THROW( WyRet() );
   }
   CHK_DEFAULT(tobj);
   tobj.set_pos(100);
   if(tobj.pos()!=0) {
     WY_THROW( WyRet() );
   }
   CHK_DEFAULT(tobj);
 }

 // set_pos_end()
 {
   WyRegFile tobj;
   CHK_DEFAULT(tobj);
   if((r=tobj.seek_end())!=Wym_EBADF) {
     WY_THROW(r);
   }
   CHK_DEFAULT(tobj);
 }

 // _read_till(1)
 {
   WyRegFile tobj;
   char buf[2]={-1,-1};
   size_t n_rd=5;   // anything not zero 
   CHK_DEFAULT(tobj);
   if((r=tobj._read_till(buf,sizeof(buf),n_rd,'\n'))!=Wym_EBADF) {
     WY_THROW(r);
   }
   if(n_rd!=0) {
     WY_THROW( WyRet() );
   }
   n_rd=5;
   if((r=EmptyFile._read_till(buf,sizeof(buf),n_rd,'\n'))!=Ok) {
     WY_THROW(r);
   }
   if(n_rd!=0) {
     WY_THROW( WyRet() );
   }
   if(EmptyFile.pos()!=0) {
     WY_THROW( WyRet() );
   }
   CHK_DEFAULT(tobj);
 }

 // _read_till(2)
 {
   WyRegFile tobj;
   WyStr buf("a");
   size_t n_rd=5;   // anything not zero
   CHK_DEFAULT(tobj);
   if((r=tobj._read_till(buf,size_t(1),n_rd,'\n'))!=Wym_EBADF) {
     WY_THROW(r);
   }
   if(n_rd!=0) {
     WY_THROW(r);
   }
   if(buf!="a") {
     WY_THROW(r);
   }
   if((r=EmptyFile._read_till(buf,size_t(0),n_rd,'\n'))!=Ok) {
     WY_THROW(r);
   }
   if(n_rd!=0) {
     WY_THROW( WyRet() );
   }
   if(buf!="a") {
     WY_THROW(r);
   }

   n_rd=5;   // pre-set for test overwrite
   if((r=EmptyFile._read_till(buf,size_t(1),n_rd,'\n'))!=Ok) {
     WY_THROW(r);
   }
   if(n_rd!=0) {
     WY_THROW( WyRet() );
   }
   if(buf!="a") {
     WY_THROW(r);
   }
   if(EmptyFile.pos()!=0) {
     WY_THROW( WyRet() );
   }
   CHK_DEFAULT(tobj);
 }

 // resize
 {
   WyRegFile tobj;
   CHK_DEFAULT(tobj);
   if((r=tobj.resize(1))!=Wym_EBADF) {
     WY_THROW(r);
   }
   CHK_DEFAULT(tobj);
 }

 // reset(const char*,int)
 {
   WyRegFile tobj;
   CHK_DEFAULT(tobj);
   if((r=tobj.reset("tmp_use.del_me",0))!=Wym_ENOENT) {
     WY_THROW(r);
   }
   CHK_DEFAULT(tobj);
 }

 // reset(const WyStr&,int)
 {
   WyRegFile tobj;
   CHK_DEFAULT(tobj);
   if((r=tobj.reset(WyStr("tmp_use.del_me"),0))!=Wym_ENOENT) {
     WY_THROW(r);
   }
   CHK_DEFAULT(tobj);
 }

 // create_tmpfile
 //
 {
   WyRegFile tobj;
   WyStr tmp_name(Tmp_Name);
   CHK_DEFAULT(tobj);
   if((r=WyRegFile::create_tmpfile(tobj,tmp_name))!=Ok) {
     WY_THROW(r);
   }
   Wy_AtDestroy<WyRet,const WyStr&> rrid(Wy::unlink,tmp_name);
   CHK_NOT_DEFAULT(tobj);
   if(tmp_name.size()!=Tmp_Name.size()+6) {
     WY_THROW( WyRet() );
   }
   if(::access(tmp_name.c_str(),F_OK)!=0) {
     r=WyReply(errno); WY_THROW(r);
   }
 }

 // _swap
 {
  WyRegFile tmp0(EmptyFile),tmp1,tmp2;
  CHK_NOT_DEFAULT(tmp0);
  CHK_DEFAULT(tmp1);
  CHK_DEFAULT(tmp2);
  CHK_NOT_DEFAULT(EmptyFile);

  if((r=tmp1._swap(tmp2))!=Ok) {
    WY_THROW(r);
  }
  CHK_DEFAULT(tmp1);
  CHK_DEFAULT(tmp2);

  if((r=tmp1._swap(tmp0))!=Ok) {
    WY_THROW(r);
  }
  CHK_DEFAULT(tmp0);
  CHK_NOT_DEFAULT(tmp1);
  CHK_SAME_SYSF(tmp1,EmptyFile);

  if((r=tmp0._swap(tmp1))!=Ok) {
    WY_THROW(r);
  }
  CHK_DEFAULT(tmp1);
  CHK_NOT_DEFAULT(tmp0);
  CHK_SAME_SYSF(tmp0,EmptyFile);

  WySysFile abc;
  if((r=abc._swap(tmp1))!=Wym_EBADTYPE) {
    WY_THROW(r);
  }
 }

 // chk _alloc)
 {
    WyRegFile tmp;
    WyRegFile* p=tmp._alloc(r);
    if(r!=Ok) {
      WY_THROW(r);
    }
    if(p==0) {
      WY_THROW( WyRet() );
    }
    Wy_AtDestroy<void,WyRegFile*> rrid(Wy::_delete_one,p);
    CHK_DEFAULT(*p);
    if((r=tmp._swap(*p))!=Ok) {
      WY_THROW(r);
    }
 }
};

// Check basic operations on non-default object
//
static void t2(void)
{
 WyRegFile EmptyFile(TestFile_Empty,O_RDONLY);
 WyRegFile TFile1A(TestFile_1A,O_RDONLY);
 WyRegFile TFile2A(TestFile_2A,O_RDONLY);
 WyRet r;

 CHK_NOT_DEFAULT(EmptyFile);
 CHK_NOT_DEFAULT(TFile1A);
 CHK_NOT_DEFAULT(TFile2A);

 // chk EmptyFile
 //
 if(EmptyFile.pos()!=0) {
   WY_THROW( WyRet() );
 }
 if(EmptyFile.fh().is_default()) {
   WY_THROW( WyRet() );
 }

 // stat EmptyFile
 {
   WyFileStat stt;
   if((r=EmptyFile.stat(stt))!=Ok) {
     WY_THROW(r);
   }
   if(stt.is_default()) {
     WY_THROW( WyRet() );
   }
   if(stt.is_reg()==false) {
     WY_THROW( WyRet() );
   }
   if(stt.sf_size()!=0) {
     WY_THROW( WyRet() );
   }
 }

 // chk TFile1A
 //
 if(TFile1A.pos()!=0) {
   WY_THROW( WyRet() );
 }
 if(TFile1A.fh().is_default()) {
   WY_THROW( WyRet() );
 }

 // stat TFile1A
 {
   WyFileStat stt;
   if((r=TFile1A.stat(stt))!=Ok) {
     WY_THROW(r);
   }
   if(stt.is_default()) {
     WY_THROW( WyRet() );
   }
   if(stt.is_reg()==false) {
     WY_THROW( WyRet() );
   }
   if(stt.sf_size()!=1) {
     WY_THROW( WyRet() );
   }
 }

 // chk TFile2A
 //
 if(TFile2A.pos()!=0) {
   WY_THROW( WyRet() );
 }
 if(TFile2A.fh().is_default()) {
   WY_THROW( WyRet() );
 }

 // stat TFile2A
 {
   WyFileStat stt;
   if((r=TFile2A.stat(stt))!=Ok) {
     WY_THROW(r);
   }
   if(stt.is_default()) {
     WY_THROW( WyRet() );
   }
   if(stt.is_reg()==false) {
     WY_THROW( WyRet() );
   }
   if(stt.sf_size()!=2) {
     WY_THROW( WyRet() );
   }
 }

 // WyRegFile(const WyRegFile&)
 {
   WyRegFile tmp1(TFile1A);

   CHK_NOT_DEFAULT(tmp1);
   if(tmp1.fh()==TFile1A.fh()) {
     WY_THROW( WyRet() );
   }
   if(tmp1.pos()!=TFile1A.pos()) {
     WY_THROW( WyRet() );
   }
   tmp1.set_pos(5);

   WyRegFile tmp2(tmp1);
   CHK_NOT_DEFAULT(tmp2);
   if(tmp2.fh()==tmp1.fh()) {
     WY_THROW( WyRet() );
   }
   if(tmp2.pos()!=tmp1.pos()) {
     WY_THROW( WyRet() );
   }
 }

 // reset()
 {
   WyRegFile tmp(EmptyFile);
   CHK_NOT_DEFAULT(tmp);
   if((r=tmp.reset())!=Ok) {
     WY_THROW(r);
   }
   CHK_DEFAULT(tmp);
 }

 // read(void*,..) read(char&)
 {
   WyRegFile tmp(TFile1A);
   char buf[2]={-1,-1};
   size_t n_rd=5;   // anything >1
   CHK_NOT_DEFAULT(tmp);
   if((r=tmp.read(buf,size_t(0),n_rd))!=Ok) {
     WY_THROW(r);
   }
   if(n_rd!=0) {
     WY_THROW( WyRet() );
   }
   n_rd=99;
   if((r=tmp.read(buf,sizeof(buf),n_rd))!=Ok) {
     WY_THROW(r);
   }
   if(n_rd!=1) {
     WY_THROW( WyRet() );
   }
   if(buf[0]!='\n') {
     WY_THROW( WyRet() );
   }
   if(buf[1]!=char(-1)) {
     WY_THROW( WyRet() );
   }
   if(tmp.pos()!=1) {
     WY_THROW( WyRet() );
   }
   // eof
   if((r=tmp.read(buf,sizeof(buf),n_rd))!=Ok) {
     WY_THROW(r);
   }
   if(n_rd!=0) {
     WY_THROW( WyRet() );
   }
   if(buf[0]!='\n') {
     WY_THROW( WyRet() );
   }
   if(buf[1]!=char(-1)) {
     WY_THROW( WyRet() );
   }
   if(tmp.pos()!=1) {
     WY_THROW( WyRet() );
   }

   if((r=tmp.reset())!=Ok) {
     WY_THROW(r);
   }
   if((r=tmp.reset(TFile1A))!=Ok) {
     WY_THROW(r);
   }

   // chk reset default pos
   {
     WyRegFile tmp2(tmp);
     CHK_SAME_SYSF(tmp,tmp2);
     if((r=tmp2.reset())!=Ok) {
       WY_THROW(r);
     }
     CHK_DEFAULT(tmp2);

     if((r=tmp2.reset(tmp))!=Ok) {
       WY_THROW(r);
     }
     CHK_SAME_SYSF(tmp,tmp2);
     if((r=tmp2.reset(WyFileHandle()))!=Ok) {
       WY_THROW(r);
     }
     CHK_DEFAULT(tmp2);

     if((r=tmp2.reset(tmp))!=Ok) {
       WY_THROW(r);
     }
     CHK_SAME_SYSF(tmp,tmp2);
     if((r=tmp2.reset(WyRegFile()))!=Ok) {
       WY_THROW(r);
     }
     CHK_DEFAULT(tmp2);

     if((r=tmp2.reset(tmp))!=Ok) {
       WY_THROW(r);
     }
     CHK_SAME_SYSF(tmp,tmp2);
     if((r=tmp2.reset(WyByteFlow()))!=Ok) {
       WY_THROW(r);
     }
     CHK_DEFAULT(tmp2);
   }
 }

 // read WyStr
 {
   WyRegFile tmp(TFile1A);
   WyStr buf;
   size_t n_rd=5;   // anything >1
   CHK_NOT_DEFAULT(tmp);
   if((r=tmp.read(buf,size_t(0),n_rd))!=Ok) {
     WY_THROW(r);
   }
   if(n_rd!=0) {
     WY_THROW( WyRet() );
   }
   n_rd=5;    // pre-set to invalid
   if((r=tmp.read(buf,size_t(5),n_rd))!=Ok) {
     WY_THROW(r);
   }
   if(n_rd!=1) {
     WY_THROW( WyRet() );
   }
   if(buf.size()!=1) {
     WY_THROW( WyRet() );
   }
   if(buf[0]!='\n') {
     WY_THROW( WyRet() );
   }
   if(tmp.pos()!=1) {
     WY_THROW( WyRet() );
   }
   // eof
   if((r=tmp.read(buf,size_t(5),n_rd))!=Ok) {
     WY_THROW(r);
   }
   if(n_rd!=0) {
     WY_THROW( WyRet() );
   }
   if(buf.size()!=1) {
     WY_THROW( WyRet() );
   }
   if(buf[0]!='\n') {
     WY_THROW( WyRet() );
   }
   if(tmp.pos()!=1) {
     WY_THROW( WyRet() );
   }

   // read again from the begin
   tmp.set_pos(0);
   if((r=tmp.read(buf,size_t(5),n_rd))!=Ok) {
     WY_THROW(r);
   }
   if(n_rd!=1) {
     WY_THROW( WyRet() );
   }
   if(buf.size()!=2) {
     WY_THROW( WyRet() );
   }
   if(buf[0]!='\n') {
     WY_THROW( WyRet() );
   }
   if(buf[1]!='\n') {
     WY_THROW( WyRet() );
   }
   if(tmp.pos()!=1) {
     WY_THROW( WyRet() );
   }
 }

 // _read_till
 {
   WyRegFile tmp2(TFile2A);
   char buf[2]={-1,-1};
   size_t n_rd=5;   // anything >1
   CHK_NOT_DEFAULT(tmp2);
   buf[0]=99;
   buf[1]=99;
   n_rd=5;
   if((r=tmp2._read_till(buf,sizeof(buf),n_rd,'\n'))!=Ok) {
     WY_THROW(r);
   }
   if(n_rd!=1) {
     WY_THROW( WyRet() );
   }
   if(tmp2.pos()!=1) {
     WY_THROW( WyRet() );
   }
   if(buf[0]!='\n') {
     WY_THROW( WyRet() );
   }
   buf[0]=99;
   n_rd=5;
   if((r=tmp2._read_till(buf,sizeof(buf),n_rd,char(99)))!=Ok) {
     WY_THROW(r);
   }
   if(n_rd!=1) {
     WY_THROW( WyRet() );
   }
   if(tmp2.pos()!=2) {
     WY_THROW( WyRet() );
   }
   if(buf[0]!='\n') {
     WY_THROW( WyRet() );
   }

   tmp2.set_pos(0);
 }

 // _read_till(WyStr)
 {
   WyRegFile tmp(TFile1A);
   WyStr buf;
   size_t n_rd=5;   // anything >1
   WyRegFile tmp2(TFile2A);
   CHK_NOT_DEFAULT(tmp2);
   buf.reset();
   n_rd=9;
   if((r=tmp2._read_till(buf,0,n_rd,'\n'))!=Ok) {
     WY_THROW(r);
   }
   if(n_rd!=0) {
     WY_THROW( WyRet() );
   }
   n_rd=9;
   if((r=tmp2._read_till(buf,10,n_rd,'\n'))!=Ok) {
     WY_THROW(r);
   }
   if(n_rd!=1) {
     WY_THROW( WyRet() );
   }
   if(tmp2.pos()!=1) {
     WY_THROW( WyRet() );
   }
   if(buf.size()!=1) {
     WY_THROW( WyRet() );
   }
   if(buf[0]!='\n') {
     WY_THROW( WyRet() );
   }
   n_rd=9;
   if((r=tmp2._read_till(buf,10,n_rd,char(99)))!=Ok) {
     WY_THROW(r);
   }
   if(n_rd!=1) {
     WY_THROW( WyRet() );
   }
   if(tmp2.pos()!=2) {
     WY_THROW( WyRet() );
   }
   if(buf.size()!=2) {
     WY_THROW( WyRet() );
   }
   if(buf!="\n\n") {
     WY_THROW( WyRet() );
   }
 }

 // write
 {
   WyRegFile tmp;
   WyStr tnam(Tmp_Name);
   const char buf[2]={-1,-1};
   size_t n_wr=5;   // anything not zero
   CHK_DEFAULT(tmp);
   if((r=WyRegFile::create(tmp,tnam,O_WRONLY,0))!=Ok) {
     WY_THROW(r);
   }
   Wy_AtDestroy<WyRet,const WyStr&> rrid(Wy::unlink,tnam);
   CHK_NOT_DEFAULT(tmp);
   if((r=tmp.write(buf,sizeof(buf),n_wr))!=Ok) {
     WY_THROW(r);
   }
   if(n_wr!=sizeof(buf)) {
     WY_THROW( WyRet() );
   }
   if(tmp.pos()!=sizeof(buf)) {
     WY_THROW( WyRet() );
   }
   if((tmp._fdatasync())!=Ok) {
     WY_THROW( WyRet() );
   }
   if((tmp._fsync())!=Ok) {
     WY_THROW( WyRet() );
   }
   if((r=tmp.write(WyStr("x"),n_wr))!=Ok) {
     WY_THROW(r);
   }
   if(n_wr!=1) {
     WY_THROW( WyRet() );
   }
   if(tmp.pos()!=sizeof(buf)+1) {
     WY_THROW( WyRet() );
   }
   if((tmp._fdatasync())!=Ok) {
     WY_THROW( WyRet() );
   }
   CHK_NOT_DEFAULT(tmp);

   tmp.set_pos(1);
   if(tmp.pos()!=1) {
     WY_THROW( WyRet() );
   }
 }

 // seek_end
 {
   WyRegFile tmp(TFile1A);
   CHK_NOT_DEFAULT(tmp);
   if((r=tmp.seek_end())!=Ok) {
     WY_THROW(r);
   }
   if(tmp.pos()!=1) {
     WY_THROW( WyRet() );
   }
   if(TFile1A.pos()!=0) {
     WY_THROW( WyRet() ); // should not be modified
   }
   if((r=tmp.reset())!=Ok) {
     WY_THROW(r);
   }
   CHK_DEFAULT(tmp);
 }

 // _read_till(1), TFile1A
 {
   WyRegFile tmp(TFile1A);
   char buf[2]={-1,-1};
   size_t n_rd=5;   // anything not zero
   CHK_NOT_DEFAULT(tmp);
   if((r=tmp._read_till(buf,sizeof(buf),n_rd,'\n'))!=Ok) {
     WY_THROW(r);
   }
   if(n_rd!=1) {
     WY_THROW( WyRet() );
   }
   if(buf[0]!='\n') {
     WY_THROW( WyRet() );
   }
   if(tmp.pos()!=1) {
     WY_THROW( WyRet() );
   }

   CHK_NOT_DEFAULT(tmp);
   n_rd=5;   // pre-set for test overwrite
   if((r=tmp._read_till(buf,sizeof(buf),n_rd,'\n'))!=Ok) {
     WY_THROW(r);
   }
   if(n_rd!=0) {
     WY_THROW( WyRet() );
   }
   if((r=tmp.reset())!=Ok) {
     WY_THROW(r);
   }
   CHK_DEFAULT(tmp);
 }

 // _read_till(2), TFile1A
 {
   WyRegFile tmp(TFile1A);
   WyStr buf("b");
   size_t n_rd=5;   // anything not zero
   CHK_NOT_DEFAULT(tmp);
   if((r=tmp._read_till(buf,size_t(0),n_rd,'\n'))!=Ok) {
     WY_THROW(r);
   }
   if(n_rd!=0) {
     WY_THROW( WyRet() );
   }
   if(buf!="b") {
     WY_THROW( WyRet() );
   }
   if(tmp.pos()!=0) {
     WY_THROW( WyRet() );
   }

   n_rd=5;   // pre-set for test overwrite
   if((r=tmp._read_till(buf,size_t(2),n_rd,'\n'))!=Ok) {
     WY_THROW(r);
   }
   if(n_rd!=1) {
     WY_THROW( WyRet() );
   }
   if(buf!="b\n") {
     WY_THROW( WyRet() );
   }
   if(tmp.pos()!=1) {
     WY_THROW( WyRet() );
   }
   if((r=tmp._read_till(buf,size_t(5),n_rd,'\n'))!=Ok) {
     WY_THROW(r);
   }
   if(n_rd!=0) {
     WY_THROW( WyRet() );
   }
   if(buf!="b\n") {
     WY_THROW( WyRet() );
   }
   CHK_NOT_DEFAULT(tmp);

   // read again, see if it is appended
   tmp.set_pos(0);
   n_rd=5;   // pre-set for test overwrite
   if((r=tmp._read_till(buf,size_t(5),n_rd,'\n'))!=Ok) {
     WY_THROW(r);
   }
   if(n_rd!=1) {
     WY_THROW( WyRet() );
   }
   if(buf!="b\n\n") {
     WY_THROW( WyRet() );
   }
   if(tmp.pos()!=1) {
     WY_THROW( WyRet() );
   }
   CHK_NOT_DEFAULT(tmp);
 }

 // _read_till(1), TFile2A
 {
   WyRegFile tmp(TFile2A);
   char buf[3]={-1,-1,-1};
   size_t n_rd=5;   // anything not zero
   CHK_NOT_DEFAULT(tmp);

   if((r=tmp._read_till(buf,sizeof(buf),n_rd,'?'))!=Ok) {
     WY_THROW(r);
   }
   if(n_rd!=2) {    // TFile2A is two bytes long
     WY_THROW( WyRet() );
   }
   if((buf[0]!='\n')||(buf[1]!='\n')) {
     WY_THROW( WyRet() );
   }
   if(tmp.pos()!=2) {
     WY_THROW( WyRet() );
   }
   
   if((r=tmp.reset())!=Ok) {
     WY_THROW(r);
   } 
   if((r=tmp.reset(TFile2A))!=Ok) {
     WY_THROW(r);
   } 
   if((r=tmp._read_till(buf,sizeof(buf),n_rd,'\n'))!=Ok) {
     WY_THROW(r);
   }
   if(n_rd!=1) {
     WY_THROW( WyRet() );
   }
   if(buf[0]!='\n') {
     WY_THROW( WyRet() );
   }
   if(tmp.pos()!=1) {
     WY_THROW( WyRet() );
   }

   if((r=tmp._read_till(buf,sizeof(buf),n_rd,'\n'))!=Ok) {
     WY_THROW(r);
   }
   if(n_rd!=1) {
     WY_THROW( WyRet() );
   }
   if(buf[0]!='\n') {
     WY_THROW( WyRet() );
   }
   if(tmp.pos()!=2) {
     WY_THROW( WyRet() );
   }

   if((r=tmp._read_till(buf,sizeof(buf),n_rd,'\n'))!=Ok) {
     WY_THROW(r);
   }
   if(n_rd!=0) {
     WY_THROW( WyRet() );
   }
   if(tmp.pos()!=2) {
     WY_THROW( WyRet() );
   }
   if((r=tmp.reset())!=Ok) {
     WY_THROW(r);
   }
   CHK_DEFAULT(tmp);
 }

 // _read_till(2), TFile2A
 {
   WyRegFile tmp(TFile2A);
   WyStr buf;
   size_t n_rd=5;   // anything not zero
   CHK_NOT_DEFAULT(tmp);
   if((r=tmp._read_till(buf,size_t(5),n_rd,'\n'))!=Ok) {
     WY_THROW(r);
   }
   if(buf.size()!=1) {
     WY_THROW( WyRet() );
   }
   if(n_rd!=1) {
     WY_THROW( WyRet() );
   }
   if(buf[0]!='\n') {
     WY_THROW( WyRet() );
   }
   if(tmp.pos()!=1) {
     WY_THROW( WyRet() );
   }
   CHK_NOT_DEFAULT(tmp);

   if((r=tmp._read_till(buf,size_t(5),n_rd,'\n'))!=Ok) {
     WY_THROW(r);
   }
   if(buf.size()!=2) {
     WY_THROW( WyRet() );  // not appended
   }
   if(n_rd!=1) {
     WY_THROW( WyRet() );
   }
   if(buf[1]!='\n') {
     WY_THROW( WyRet() );
   }
   if(tmp.pos()!=2) {
     WY_THROW( WyRet() );
   }
   CHK_NOT_DEFAULT(tmp);

   if((r=tmp._read_till(buf,size_t(5),n_rd,'\n'))!=Ok) {
     WY_THROW(r);
   }
   if(n_rd!=0) {
     WY_THROW( WyRet() );
   }
   if(buf.size()!=2) {
     WY_THROW( WyRet() );  // modified
   }
   CHK_NOT_DEFAULT(tmp);
 }

 // _read_till, TFile2A
 {
   WyRegFile tmp("ck_regfile.cpp",O_RDONLY); // this file
   size_t fsize;      // size of this file
   WyStr buf;
   size_t n_rd=99999;
   CHK_NOT_DEFAULT(tmp);

   {
     WyFileStat stt;
     if((r=tmp.stat(stt))!=Ok) {
       WY_THROW(r);
     }
     fsize=stt.sf_size();
     if(fsize<1800) {        // known minimal
       WY_THROW( WyRet() );
     }
     if((r=buf.append(char(0)))!=Ok) {
       WY_THROW(r);
     }
   }

   // Find this unique character ` (exactly here in this cpp file)
   //
   const char UniqueChar='`';
   if((r=tmp._read_till(buf,fsize,n_rd,UniqueChar))!=Ok) {
     WY_THROW(r);
   }
   if((n_rd>=fsize)||(n_rd<1800)) { // n_rd range not right
     WY_THROW( WyRet() );
   }
   if(buf.size()!=n_rd+1) {
     WY_THROW( WyRet() );
   }
   if(buf[0]!=char(0)) {
     WY_THROW( WyRet() );  // pre-exit char overwritten
   }
   if(buf[n_rd]!=UniqueChar) {
     WY_THROW( WyRet() );
   }
   if(tmp.pos()!=off_t(n_rd)) {
     WY_THROW( WyRet() );
   }
 }

 // resize
 {
   WyRegFile tmp;
   const char buf[2]={-1,-1};
   size_t n_wr=5;   // anything not zero
   CHK_DEFAULT(tmp);
   if((r=WyRegFile::create(tmp,Tmp_Name.c_str(),O_WRONLY,0))!=Ok) {
     WY_THROW(r);
   }
   Wy_AtDestroy<WyRet,const WyStr&> rrid(Wy::unlink,Tmp_Name);
   CHK_NOT_DEFAULT(tmp);
   if((r=tmp.write(buf,sizeof(buf),n_wr))!=Ok) {
     WY_THROW(r);
   }
   if(n_wr!=sizeof(buf)) {
     WY_THROW( WyRet() );
   }
   if(tmp.pos()!=sizeof(buf)) {
     WY_THROW( WyRet() );
   }
   if((tmp._fdatasync())!=Ok) {
     WY_THROW( WyRet() );
   }

   // resize to 1
   if((r=tmp.resize(1))!=Ok) {
     WY_THROW(r);
   }
   if((r=tmp.seek_end())!=Ok) {
     WY_THROW(r);
   }
   if(tmp.pos()!=1) {
     WY_THROW( WyRet() );
   }

   // resize to 10
   if((r=tmp.resize(10))!=Ok) {
     WY_THROW(r);
   }
   if(tmp.pos()==10) {
     WY_THROW( WyRet() );
   }
   if((r=tmp.seek_end())!=Ok) {
     WY_THROW(r);
   }
   if(tmp.pos()!=10) {
     WY_THROW( WyRet() );
   }
 }

 // reset(WyStr..) Wym_ENOENT
 {
   WyRegFile tmp(TFile1A);
   CHK_NOT_DEFAULT(tmp);
   if((r=tmp.reset(WyStr(Tmp_Name),O_RDONLY))!=Wym_ENOENT) {
     WY_THROW(r);
   }
 }

 // reset(WyFileHandle())
 //
 {
   WyRegFile tmp(TFile1A);
   CHK_NOT_DEFAULT(tmp);
   if((r=tmp.reset(WyFileHandle()))!=Ok) {
     WY_THROW(r);
   }
   CHK_DEFAULT(tmp);
 }

 // reset(WyRegFile())
 //
 {
   WyRegFile tmp(TFile1A);
   CHK_NOT_DEFAULT(tmp);
   if((r=tmp.reset(WyRegFile()))!=Ok) {
     WY_THROW(r);
   }
   CHK_DEFAULT(tmp);
 }

 // reset() from base
 {
   WyRegFile tmp;
   CHK_DEFAULT(tmp);
   WyByteFlow *p=&tmp;
   if((r=p->reset(TFile1A))!=Ok) {
     WY_THROW(r);
   }
   CHK_SAME_SYSF(tmp,TFile1A);
 }

 // chk move ctor
 {
  char raw_buf[2*sizeof(WyRegFile)];
  ::memset(raw_buf,0,sizeof(raw_buf));
  WyRegFile * const p1( reinterpret_cast<WyRegFile*>(raw_buf) );
  WyRegFile * const p2(p1+1);

  new(p1) WyRegFile(__FILE__,O_RDONLY);
  {
    try {
      CHK_NOT_DEFAULT(*p1);
    }
    catch(...) {
      p1->~WyRegFile();
      throw;
    };
  }

  // move *p1 to p2
  WyFileHandle ofh( p1->fh() );
  try {
    new(p2) WyRegFile(*p1,Wy::ByMove);
  }
  catch(...) {
    p1->~WyRegFile();
    throw;
  };
  WyFileHandle nfh( p2->fh() );
  p2->~WyRegFile();

  if(nfh!=ofh) {
    WY_THROW( WyRet() );
  }
 }

};

// Test read operation (interleave read)
//
// Two objects are designated to read tfile alternatively.
// tfile points to any file that can be read.
//
static void interleaved_read(const WyStr& tfile)
{
 WyRet r;
 WyRegFile f0;   // tfile

 if((r=f0.reset(tfile,O_RDONLY))!=Ok) {
   WY_THROW(r);
 }
 WyRegFile f1(f0);   // f1 will read with the former half part of tfile
 WyRegFile f2(f0);   // f2 will read with the latter half part of tfile

 if((r=f0.seek_end())!=Ok) {
   WY_THROW(r);
 }

 //Wy::cout << tfile << "= " <<
 //             f0.pos() << " bytes, ";

 // file size of f0 and the start position of f2, also the defined file size of f1
 //
 const off_t FSize=f0.pos(); 
 const  off_t MidPos=FSize>>1; 
 f1.set_pos(0);                 // set f1 to begin from 0
 f2.set_pos(MidPos);            // set f2 to begin from MidPos

 // Set str0 with the whole file of f0
 //
 WyStr str0;
 {
   size_t n_rd;
   char* text0=(char*)::malloc(FSize);
   if(text0==0) {
     WY_THROW( WyRet(Wym_ENOMEM) );
   }
   Wy_AtDestroy<void,char*> rrid((void(*)(char*))::free,text0);
   
   f0.set_pos(0);
   if((r=f0.read(text0,FSize,n_rd))!=Ok) {
     WY_THROW(r);
   }
   if((off_t)n_rd!=FSize) {
     WY_THROW( WyRet() );
   }
   if(f0.pos()!=FSize) {
     WY_THROW( WyRet() ); // value of seek_end not identical to read EOF
   }
   str0.reset( WyCSeg(text0,FSize) ); // transfer text0 to str0
 }

 // Read f1 to str1 and f2 to str2 in a loop, each loop read one byte
 //
 WyStr str1,str2;
 for(;;) {
   char ch;
   size_t n_rd;
   bool all_eof=true;
   if(str1.size()<(size_t)MidPos) {  // f1 read up to MidPos characters
     if((r=f1.read(&ch,1,n_rd))!=Ok) {
       WY_THROW(r);
     }
     if(n_rd==1) {
       str1+=ch;         // append the read
       all_eof=false;
     } else {
       if(n_rd!=0) {
         WY_THROW( WyRet() );  // not expected value of n_rd
       }
       WY_THROW( WyRet() );  // not expected value of n_rd(0)
     }
   }

   if((r=f2.read(&ch,1,n_rd))!=Ok) {
     WY_THROW(r);
   }
   if(n_rd==1) {
     str2+=ch;           // append the read
     all_eof=false;
   } else {
     if(n_rd!=0) {
       WY_THROW( WyRet() );  // not expected value
     }
     if(f2.pos()!=FSize) {
       WY_THROW( WyRet() );  // not expected value
     }
     // FALL_THROUGH
   }

   if(all_eof==true) {
     break;
   }
 }

 if(f1.pos()!=MidPos) {
   WY_THROW( WyRet() );  // not expected value
 }
 if(f2.pos()!=FSize) {
   WY_THROW( WyRet() );  // not expected value
 }
 str1+=str2;   // append str2 to str1
 if(str0!=str1) {
   WY_THROW( WyRet() );  // assertion failed
 }

};

static void t_throw(void)
{

 // WyRegFile(const WySysFile&)
 //
 if(exaustfd_flag) {
   try {
     WyByteFlow tmp(__FILE__,O_RDONLY);
     exaust_cpy_ctor(tmp);
     WY_THROW( WyRet() );  // not expected
   }
   catch(const WyRegFile::Reply& e) {
     if((e!=Wym_EMFILE)&&(e!=Wym_ENOMEM)) {
       WY_THROW( WyRet(e) );
     }
     // FALL_THROUGH 
   };
 }

 // WyRegFile(const WyRegFile&)
 //
 if(exaustfd_flag) {
   try {
     WyRegFile tmp(__FILE__,O_RDONLY);
     exaust_cpy_ctor(tmp);
     WY_THROW( WyRet() );  // not expected
   }
   catch(const WyRegFile::Reply& e) {
     if((e!=Wym_EMFILE)&&(e!=Wym_ENOMEM)) {
       WY_THROW( WyRet(e) );
     }
     // FALL_THROUGH 
   };
 }

 // WyRegFile(const char*,int)
 //
 try {
   const char pn[]="nosuchfilename";
   WyRegFile tmp(pn,0);
   WY_THROW( WyRet() );
 }
 catch(const WyRegFile::Reply& e) {
   if(e!=Wym_ENOENT) {
     WY_THROW( WyRet(e) );
   }
   // FALL_THROUGH 
 };

 // WyRegFile(const WyStr&,int)
 //
 try {
   WyStr pn("nosuchfilename");
   WyRegFile tmp(pn,0);
   WY_THROW( WyRet() );
 }
 catch(const WyRegFile::Reply& e) {
   if(e!=Wym_ENOENT) {
     WY_THROW( WyRet(e) );
   }
   // FALL_THROUGH 
 };

 // WyRegFile(WyFileHandle)
 //
 try {
   WyRegFile tmp( WyFileHandle(STDERR_FILENO));
   WY_THROW( WyRet() );
 }
 catch(const WyRegFile::Reply& e) {
   if(e!=Wym_EBADF) {
     WY_THROW( WyRet(e) );
   }
   // FALL_THROUGH 
 };

 // Test destructor throw
 //
 {
   WyRegFile tmp;
   //std::memset(&tmp,-2,sizeof(tmp));
   // destructor of tmp is now not to thow
 }

};

static void t_error(void)
{
 WyRegFile EmptyFile(TestFile_Empty,O_RDONLY);
 WyRet r;

 CHK_NOT_DEFAULT(EmptyFile);

 // stat
 {
   WyRegFile tmp;
   WyFileStat stt;
   if((r=tmp.stat(stt))!=Wym_EBADF) {
     WY_THROW(r);
   }
   CHK_DEFAULT(tmp);
 }

 // reset(const WyRegFile&)
 {
   WyRegFile tmp(EmptyFile);
   const WyFileHandle TmpFh(tmp.fh());
   if((r=tmp.reset(tmp))!=Ok) {
     WY_THROW(r);
   }
   CHK_SAME_SYSF(tmp,EmptyFile);
   if(tmp.fh()!=TmpFh) {
     // ok either way
   }
 }

 // reset(const char*,int)
 {
   WyRegFile tmp;
   const char pn[]="nosuchfilename";
   if((r=tmp.reset(pn,O_RDONLY))!=Wym_ENOENT) {
     WY_THROW(r);
   }
   CHK_DEFAULT(tmp);
 }

 // reset(const WyStr&,int)
 {
   WyRegFile tmp;
   WyStr pn("nosuchfilename");
   if((r=tmp.reset(pn,O_RDONLY))!=Wym_ENOENT) {
     WY_THROW(r);
   }
   CHK_DEFAULT(tmp);
 }

 // reset(WyFileHandle)
 {
   WyRegFile tmp;
   WyFileHandle h(STDERR_FILENO);
   if((r=tmp.reset(h))!=Wym_EBADF) {
     WY_THROW(r);
   }
   CHK_DEFAULT(tmp);
 }

 // chk _alloc not override
 {
   class T : public WyRegFile {
   } tmp;
   WyByteFlow* p=tmp._alloc(r);
   if(r!=Wym_ENOSYS) {
     if(r==Ok) {
       delete p;
     }
     WY_THROW(r);
   }
 }
};

// verify _read_till(..) feature
//
static void t_read_line1(void)
{
 // check _read_till(..) does not append the buffer with a trailing zero
 //
 {
   WyRegFile tfile(__FILE__,O_RDONLY);
   char rbuf[256];   // enough for the first line of tfile +2
   WyRet r;
   size_t n_rd;
   ::memset(rbuf,-1,sizeof(rbuf));  // character to detect overwritten
   if((r=tfile._read_till(rbuf,4,n_rd,'\n'))!=Ok) {  // read 4 char of the first line
     WY_THROW(r);
   }
   if(n_rd!=4) {
     WY_THROW( WyRet() );
   }
   if(rbuf[4]==0) {
     WY_THROW( WyRet() );
   }
  
   tfile.set_pos(0);
   if((r=tfile._read_till(rbuf,sizeof(rbuf)-1,n_rd,'\n'))!=Ok) {
     WY_THROW(r);
   }
   if(n_rd>=sizeof(rbuf)-1) {
     WY_THROW( WyRet() );   // rbuf is small for the check
   }
   if(rbuf[n_rd]==0) {
     WY_THROW( WyRet() );   // chk not appended with zero
   }
 }
};

// Test read tfile (read_line against read)
//
static void t_read_line2(const WyStr& tfile)
{
 WyRet r;

 // Set str0 with the whole file of tfile
 //
 WyStr str0;
 {
   WyRegFile f0;
   size_t n_rd;

   if((r=f0.reset(tfile,O_RDONLY))!=Ok) {
     Wy::cerr << tfile << '\n';
     WY_THROW(r); // file open failure
   }
   if((r=f0.seek_end())!=Ok) {
     WY_THROW(r);
   }
   const size_t FSize=f0.pos();
   char* text0=(char*)::malloc(FSize);
   if(text0==0) {
     WY_THROW( WyRet(Wym_ENOMEM) );
   }
   Wy_AtDestroy<void,char*> rrid((void(*)(char*))::free,text0);
   f0.set_pos(0);

   if((r=f0.read(text0,FSize,n_rd))!=Ok) {
     WY_THROW(r);
   }
   if(n_rd!=FSize) {
     WY_THROW( WyRet() );
   }
   if(f0.pos()!=(off_t)FSize) {
     WY_THROW( WyRet() ); // value of seek_end not identical to read EOF
   }
   str0.reset( WyCSeg(text0,FSize) ); // transfer text0 to str0
 }

 // Read tfile line by line(1), and append to str1, check against str0
 //
 {
   WyRegFile f0;
   const size_t MaxLineLen=str0.size();
   size_t n_rd;
   char* buf=(char*)::malloc(MaxLineLen);
   if(buf==0) {
     WY_THROW( WyRet(Wym_ENOMEM) );
   }
   Wy_AtDestroy<void,char*> rrid((void(*)(char*))::free,buf);
   WyStr str1;

   if((r=f0.reset(tfile,O_RDONLY))!=Ok) {
     WY_THROW(r); // file open failure
   }
   for(;;) {
     if((r=f0._read_till(buf,MaxLineLen,n_rd,'\n'))!=Ok) {
       WY_THROW(r);
     }
     if(n_rd==0) {  // eof
       if(str1!=str0) {
         WY_THROW( WyRet() );
       }
       break;
     }
     str1.append( WyCSeg(buf,n_rd) );  // append the line read
   }
 }

 // Read tfile line by line(2), and append to str1, check against str0
 //
 {
   WyRegFile f0;
   size_t n_rd;
   WyStr str1;

   if((r=f0.reset(tfile,O_RDONLY))!=Ok) {
     WY_THROW(r); // file open failure
   }
   for(;;) {
     if((r=f0._read_till(str1,size_t(65535),n_rd,'\n'))!=Ok) {
       WY_THROW(r);
     }
     if(n_rd==0) {  // eof
       if(str1!=str0) {
         WY_THROW( WyRet() );
       }
       break;
     }
   }
 }
};

static void t_writr(void)
{
  WyRet r;
  WyRegFile tmp;
  const char buf[2]={-2,-2};
  size_t count=5;   // anything not zero
  CHK_DEFAULT(tmp);
  if((r=WyRegFile::create(tmp,WyStr(Tmp_Name),O_RDWR,0))!=Ok) {
    WY_THROW(r);
  }
  Wy_AtDestroy<WyRet,const WyStr&> rrid(Wy::unlink,Tmp_Name);
  CHK_NOT_DEFAULT(tmp);
  if((r=tmp.write(buf,0,count))!=Ok) {
    WY_THROW(r);
  }
  if(count!=0) {
    WY_THROW( WyRet() );
  }
  if(tmp.pos()!=0) {
    WY_THROW( WyRet() );
  }

  if((r=tmp.write(buf,1,count))!=Ok) {
    WY_THROW(r);
  }
  if(count!=1) {
    WY_THROW( WyRet() );
  }
  if(tmp.pos()!=1) {
    WY_THROW( WyRet() );
  }

  if((r=tmp.write(WyStr( WyCSeg(buf,sizeof(buf)) ),count))!=Ok) {
    WY_THROW(r);
  }
  if(count!=sizeof(buf)) {
    WY_THROW( WyRet() );
  }
  if(tmp.pos()!=sizeof(buf)+1) {
    WY_THROW( WyRet() );
  }

  if((tmp._fdatasync())!=Ok) {
    WY_THROW( WyRet() );
  }

  // read back
  tmp.set_pos(0);
  if(tmp.pos()!=0) {
    WY_THROW( WyRet() );
  }
  char rbuf[sizeof(buf)+2];
  if((r=tmp.read(rbuf,sizeof(rbuf),count))!=Ok) {
    WY_THROW(r);
  }
  if(count!=sizeof(buf)+1) {
    WY_THROW( WyRet() );
  }
  if(tmp.pos()!=sizeof(buf)+1) {
    WY_THROW( WyRet() );
  }
  if(rbuf[0]!=buf[0]) {
    WY_THROW( WyRet() );
  }
  if((rbuf[1]!=buf[0])||(rbuf[2]!=buf[1])) {
    WY_THROW( WyRet() );
  }
};

static void t3(void) 
{
 interleaved_read(TestFile_Empty);
 interleaved_read(TestFile_1A);
 interleaved_read(TestFile_2A);
 interleaved_read(WyStr("ck_regfile.cpp"));
 interleaved_read(WyStr("ck_regfile.o"));

 t_read_line1();
 t_read_line2(TestFile_1A);
 t_read_line2(TestFile_2A);
 //Wy::cout << "Check _read_till(" << arg_rtfile<< ") ...\n";
 t_read_line2(arg_rtfile);
 t_writr();
};

// exaustfd: exaust testing fd or not
// rtfile  : file name for _read_till test
//
void ck_regfile(bool exaustfd,const char* rtfile)
try {
 exaustfd_flag=exaustfd;
 arg_rtfile=rtfile;
#ifdef WY_DEBUG
 initial_fdcnt=wydbg_get_fdcnt();
#endif
 Wy::unlink(Tmp_Name);
 t1();
 t2();
 t_throw();
 t_error();
 t3();
#ifdef WY_DEBUG
  // fd may be lost. report is only valid in single thread test
 if(wydbg_get_fdcnt()!=initial_fdcnt) {
   WY_THROW( WyRet() );  
 }
 Wy::cout << "At most " << wydbg_get_fdcnt_max() << " fd's allocated\n";
#endif
 Wy::unlink(Tmp_Name);
}
catch(...) {
#ifdef WY_DEBUG
 // fd may be lost. report is only valid in single thread test
 if(wydbg_get_fdcnt()!=initial_fdcnt) {
   WY_THROW( WyRet() );
 }
#endif
 Wy::unlink(Tmp_Name);
 throw;
};
