/* Copyright is licensed under GNU LGPL.                 by I.J.Wang 2003

  This file is part of wydefs.h

  Wy__Base intends to provide the basic support for libwy functions.
*/
#ifndef WY__BASE_H__
#define WY__BASE_H__

#include <pthread.h>
#include <sched.h>     // for sched_yield

namespace Wy__Base {

    // [Syn] Mark cancellation point
    //
    inline void cancel_point()
        { ::pthread_testcancel(); };   // errno is assumed not modified

    // [Syn] Change cancellation state 
    //  
    //   state is one of the value:
    //          PTHREAD_CANCEL_ENABLE
    //          PTHREAD_CANCEL_DISABLE
    //
    // [Ret] 0  Ok
    //       errno otherwise
    //
    // [Refer] ::pthread_setcancelstate(int,int*)
    //
    inline int pthread_setcancelstate(int state, int* oldstate) WY__NOTHROW__
        {
          return ::pthread_setcancelstate(state,oldstate); 
        };

    // Convenient class to disable thread cancelablity
    //
    // Example:  NoCancel noc;
    //
    class NoCancel {
      public:

        // [Syn] Construct object and disable cancelablity of the calling 
        //       thread.
        // 
        // [Throw] int (errno)
        //
        NoCancel();
    
        // [Syn] Destruct *this and recover the previous cancelablity
        //
        // [Throw] int (errno)
        //
        ~NoCancel();
    
      private:
        int _oldstate;

        NoCancel(const NoCancel&);
        const NoCancel& operator =(const NoCancel&);
        bool operator ==(const NoCancel&);
    };

    // [Syn] Yield thread execution
    //
    // [Ret] 0= ok
    //       otherwise= failure, errno
    //
    inline int yield() WY__NOTHROW__
        {
          return (::sched_yield()==0)? 0: errno; 
        };
 
    // [Syn] Abort the program (NOT IN USE)
    //
    // Note: I had problem with this function inlined.
    //
    //void abort(void) throw() __attribute__ ((__noreturn__));

    // [Syn] Abort the program
    //
    // Note: I had problem with this function inlined.
    //       It seemed senseless to use throw() explicitly
    //
    void terminate() WY__NOTHROW__ __attribute__ ((__noreturn__));

    // [NoCancel]
    //
    // [Syn] Write 'abnormal message' to STDERR_FILENO
    //
    // [Ret] errno encountered
    //
    int put_abnmsg(const char* expr,const char *file_name, unsigned int lnum) WY__NOTHROW__;

    // [NoCancel] NOT IN USE
    //
    // [Syn] Write message to STDERR_FILENO
    //
    // [Ret] 0= Ok
    //       otherwise= errno
    //
    //int put_errmsg(const char* msg) WY__NOTHROW__;

    // [Syn] Swap n bytes of the two memory area pointed by p1 and p2
    //
    void memswp(void* p1,void* p2,size_t n) WY__NOTHROW__;

    // [Syn] Swap scalar type (value swap, copy constructor and assign
    //       operator work)
    //
    template<typename T>
    void valswp(T& t1,T& t2) WY__NOTHROW__
    {
      const T t0(t1);
      t1=t2;
      t2=t0;
    };

    // [Syn] Storage for type T
    //
    // Note: _long_double, _fptr might be not aligned. I.e. it's questionable if they
    //       are the first elements of T. Tests are OK (chk_ret.cpp).
    //
    template<typename T>
    union TypeStore {
        operator T&() { return reinterpret_cast<T&>(*_char_array); };
        operator const T&() const { return reinterpret_cast<const T&>(*this->_char_array); };
        char _char_array[sizeof(T)];
        char _char;
        signed char _schar;
        unsigned char _uchar;
        short _short;
        signed short _sshort;
        unsigned short _ushort;
        int _int;
        signed int _sint;
        unsigned int _uint;
        long _long;
        signed long _slong;
        unsigned long _ulong;
        long long  _llong;
        signed long long  _sllong;
        unsigned long long  _ullong;
        float _float;
        double _double;
        long double _long_double;
        bool  _bool;
        void *  _ptr;
        void (*_fptr)();
        struct _struct { char _char_content; } _dummy_struct;
        char _struct::* _struct_dptr;
        void (_struct::* _struct_fptr)();

      private:
        union _union { char _char_content; } _dummy_union;
        class _class { char _char_content; } _dummy_class;
        enum _enum { _enum_content } _dummy_enum;
    };

}; // End of Wy__Base

// This macro is for convenience
//
#define WY__EQU_TYPEID(a,b) (typeid(a)==typeid(b))

#endif
