/* Copyright is licensed by GNU LGPL.                         by I.J.Wang 2005

   This file contains function template wy__itoa_?(char*, size_t, NumT, int)
   for inclusion to instantiate integral type of functions,
*/
#ifndef WY_ITOA_HH__
#define WY_ITOA_HH__

#define WYLIB_SOURCE
#include "wydefs.h"
#include <cstdlib>   // for ldiv(..)
#include <limits>

// [Internal] Break num into radix notation and store into buf
//
// Note: 1. buf have to have enough space for the num/radix before call
//       2. buf can not be NULL, radix should be >=2
//       3. the stored 'digit' is saved in reversed order
//       4. if num is negative, the stored 'digit' is negative
//       5. the stored 'digit' is in range [0-radix) or (-radix,0]
// 
// [Ret] number of digits stored into buf
//
inline size_t wy__brknum(char* buf,int num, int radix)
{
 std::div_t dt={num,0};
 size_t n=0;
 do {
   dt=std::div(dt.quot,radix);
   buf[n++]=(char)(dt.rem);
 } while(dt.quot!=0);
 return(n);
};

inline size_t wy__brknum(char* buf,long num, long radix)
{
 std::ldiv_t dt={num,0};
 size_t n=0;
 do {
   dt=std::ldiv(dt.quot,radix);
   buf[n++]=(char)(dt.rem);
 } while(dt.quot!=0);
 return(n);
};

inline static size_t wy__brknum(char* buf,long long num, long long radix)
{
 std::lldiv_t dt={num,0};
 size_t n=0;
 do {
   dt=std::lldiv(dt.quot,radix);
   buf[n++]=(char)(dt.rem);
 } while(dt.quot!=0);
 return(n);
};

inline size_t wy__brknum(char* buf,unsigned int num, unsigned int radix)
{
 typedef unsigned int U_Type;
 typedef signed   int S_Type;
 if(num<=static_cast<U_Type>(std::numeric_limits<S_Type>::max())) {
   return( wy__brknum(buf,static_cast<S_Type>(num),static_cast<S_Type>(radix)) );
 }

 const U_Type q=num/radix;
 WY_ASSERT(q<=static_cast<U_Type>(std::numeric_limits<S_Type>::max()));
 buf[0]= char(num%radix);
 return( wy__brknum(buf+1,static_cast<S_Type>(q),static_cast<S_Type>(radix))+1 );
};

inline size_t wy__brknum(char* buf,unsigned long num, unsigned long radix)
{
 typedef unsigned long U_Type;
 typedef signed   long S_Type;
 if(num<=static_cast<U_Type>(std::numeric_limits<S_Type>::max())) {
   return( wy__brknum(buf,static_cast<S_Type>(num),static_cast<S_Type>(radix)) );
 }

 const U_Type q=num/radix;
 WY_ASSERT(q<=static_cast<U_Type>(std::numeric_limits<S_Type>::max()));
 buf[0]= char(num%radix);
 return( wy__brknum(buf+1,static_cast<S_Type>(q),static_cast<S_Type>(radix))+1 );
};

inline static size_t wy__brknum(char* buf,unsigned long long num, unsigned long long radix)
{
 typedef unsigned long long U_Type;
 typedef signed   long long S_Type;
 if(num<=static_cast<U_Type>(std::numeric_limits<S_Type>::max())) {
   return( wy__brknum(buf,static_cast<S_Type>(num),static_cast<S_Type>(radix)) );
 }

 const U_Type q=num/radix;
 WY_ASSERT(q<=static_cast<U_Type>(std::numeric_limits<S_Type>::max()));
 buf[0]= char(num%radix);
 return( wy__brknum(buf+1,static_cast<S_Type>(q),static_cast<S_Type>(radix))+1 );
};

// [Internal] Convert the broken down representation in ibuf/blen
//            to ascii and store into obuf
//
// Note: 1. for all x<blen, ibuf[x]>=0
//       2. ibuf/obuf can not be NULL, blen must be >=1
//
inline void wy__cvrrev_pos(char* obuf, const char* ibuf, int blen)
{
 const char OffV10='W';  // 'W' for lowercase. '7' for uppercase
 obuf+=blen-1;
 for(int i=blen; i>0; --i,++ibuf,--obuf) {
   const char ch=*ibuf;
   *obuf= (ch>9)? (ch+OffV10):(ch+'0');
 }
};

// [Internal] Convert the broken down representation in ibuf/blen
//            to ascii and store into obuf
//
// Note: 1. for all x<blen, ibuf[x]<=0
//       2. ibuf/obuf can not be NULL, blen must be >=1
//
// Note: obuf contaions a leading character ('-') after return
//
inline void wy__cvrrev_neg(char* obuf, const char* ibuf, int blen)
{
 const char OffV10='W';  // 'W' for lowercase. '7' for uppercase
 *obuf='-';
 obuf+=blen;
 for(int i=blen; i>0; --i,++ibuf,--obuf) {
   const char ch=-(*ibuf);
   *obuf= (ch>9)? (ch+OffV10):(ch+'0');
 }
};

// [Interna] Convert num(integer type) to string of format [-]<ddd> and store to
//           buf pointed array according to left or right alignment.
//
//           For wy__itoa_l(..), the result is stored from buf
//           For wy__itoa_r(..), the result is stored from buf+blen-v(return value)
//            e.g. App. would sometimes like buf leading with blanks:
//              v=wy__itoa_r(buf, blen, num, radix);   // blen>0
//              std::memset(buf,' ',blen-v);
//
//     Note: 1. buf must not be NULL
//           2. radix must be in range [2,36]
//
// [Ret] Number of characters converted and written into buf
//       0= buf space is not enought for the result
//       (no trailing zero is written)
//
template <typename NumT>
inline size_t wy__itoa_l(char* buf, size_t blen, NumT num, Wy::WyRadix radix)
{
 WY_ASSERT(std::numeric_limits<NumT>::is_integer);
 WY_ASSERT(std::numeric_limits<NumT>::radix==2); // so radix= bits

 char dbuf[std::numeric_limits<NumT>::digits+2];  // [-] digits+ one spare
 size_t n=wy__brknum(dbuf,num,static_cast<NumT>(radix));

 // Compiler's warning about the next line should be ok (g++ 4.1.1)
 // (don't know why only for unsigned char and unsigned short)
 //
 if((std::numeric_limits<NumT>::is_signed)&&(num<0)) {
   ++n;
   if(n>blen) {
     return(0);     // buf/blen not enough space to store result
   }
   wy__cvrrev_neg(buf,dbuf,n-1);
   return(n);
 } else {
   if(n>blen) {
     return(0);     // buf/blen not enough space to store result
   }
   wy__cvrrev_pos(buf,dbuf,n);
   return(n);
 }
};

template <typename NumT>
inline size_t wy__itoa_r(char* buf, size_t blen, NumT num, Wy::WyRadix radix)
{
 WY_ASSERT(std::numeric_limits<NumT>::is_integer);
 WY_ASSERT(std::numeric_limits<NumT>::radix==2); // so radix= bits

 char dbuf[std::numeric_limits<NumT>::digits+2];  // [-] digits+ one spare
 size_t n=wy__brknum(dbuf,num,static_cast<NumT>(radix));

 // Compiler's warning about the next line should be ok (g++ 4.1.1)
 // (don't know why only for unsigned char and unsigned short)
 //
 if((std::numeric_limits<NumT>::is_signed)&&(num<0)) {
   ++n;
   if(n>blen) {
     return(0);     // buf/blen not enough space to store result
   }
   wy__cvrrev_neg(buf+(blen-n),dbuf,n-1);
   return(n);
 } else {
   if(n>blen) {
     return(0);     // buf/blen not enough space to store result
   }
   wy__cvrrev_pos(buf+(blen-n),dbuf,n);
   return(n);
 }
};

#endif // end of WY_ITOA_HH__
