/* Copyright is licensed under GNU LGPL.                 by I.J.Wang 2003
*/

#define WYLIB_SOURCE
#include "WyStr.h"
#include "WyFileStat.h"
#include "wy__cfunc.h"
#include "Wy_AtDestroy.hh"
#include <unistd.h>      // access,chdir,stat,link...
#include <cstdlib>       // for std::malloc
#include <climits>       // for PATH_MAX

namespace Wy {

WyRet unlink(const WyStr& pathname)
{
 if(::unlink(pathname.c_str())!=0) {
   WY_RETURN(errno);
 }
 return(Ok);
};

WyRet unlink(const char* pathname)
{
 if(::unlink(pathname)!=0) {
   WY_RETURN(errno);
 }
 return(Ok);
};

WyRet access(const WyStr& pathname,int mode)
{
 if(::access(pathname.c_str(),mode)!=0) {
   WY_RETURN(errno);
 };
 return(Ok);
};

WyRet access(const char* pathname,int mode)
{
 if(::access(pathname,mode)!=0) {
   WY_RETURN(errno);
 };
 return(Ok);
};

WyRet chdir(const WyStr& path)
{
 if(::chdir(path.c_str())!=0) {
   WY_RETURN(errno);
 };
 return(Ok);
};

WyRet chdir(const char* path)
{
 if(::chdir(path)!=0) {
   WY_RETURN(errno);
 };
 return(Ok);
};

WyRet getcwd(WyStr& path)
try {
 for(size_t bsize=256; bsize<=PATH_MAX+256; bsize+=256) {
   char* buf=(char*)std::malloc(bsize);
   if(buf==0) {
     WY_RETURN(Wym_ENOMEM);
   }
   Wy_AtDestroy<void,void*> rrid(&std::free, static_cast<void*>(buf));

   if(::getcwd(buf,bsize)== buf) {
     const WyRet r=path.reset( WyCSeg(buf) );
     if(r!=Ok) {
       WY_RETURN(r);
     }
     return(Ok);
   }
   if(errno==ERANGE) {
     continue;
   }
   WY_RETURN(errno);
 }
 WY_RETURN(Wym_ENAMETOOLONG);
}
catch(const WyRet& e) {
  WY_THROW( WyRet(e) );
};

WyRet chown(const WyStr& path,uid_t owner,gid_t group)
{
 if(::chown(path.c_str(),owner,group)!=0) {
   WY_RETURN(errno);
 };
 return(Ok);
};

WyRet chown(const char* path,uid_t owner,gid_t group)
{
 if(::chown(path,owner,group)!=0) {
   WY_RETURN(errno);
 };
 return(Ok);
};

WyRet chmod(const WyStr& path,mode_t mode)
{
 if(::chmod(path.c_str(),mode)!=0) {
   WY_RETURN(errno);
 };
 return(Ok);
};

WyRet chmod(const char* path,mode_t mode)
{
 if(::chmod(path,mode)!=0) {
   WY_RETURN(errno);
 };
 return(Ok);
};

WyRet link(const WyStr& oldpath,const WyStr& newpath)
{
  if(::link(oldpath.c_str(),newpath.c_str())!=0) {
    WY_RETURN(errno);
  }
  return(Ok);
};

WyRet link(const char* oldpath,const char* newpath)
{
  if(::link(oldpath,newpath)!=0) {
    WY_RETURN(errno);
  }
  return(Ok);
};

WyRet link(const char* oldpath,const WyStr& newpath)
{
 return link(oldpath,newpath.c_str()); 
};

WyRet link(const WyStr& oldpath,const char* newpath)
{
 return link(oldpath.c_str(),newpath); 
};

WyRet gethostname(WyStr& buf)
try {
#ifndef HOST_NAME_MAX      // man. indicated this name but not found
#define HOST_NAME_MAX 255
#endif
 WyRet r;
 WyStr tmp;
 if((r=tmp._reserve(HOST_NAME_MAX+1))!=Ok) {
   WY_RETURN(r);
 }

 // pre-write trailing zero in case the name is HOST_NAME_MAX long
 *(tmp.begin()+HOST_NAME_MAX)=0;  

 const int v=::gethostname(tmp.begin(),HOST_NAME_MAX+1);
 if(v!=0) {
   WY_RETURN(errno);
 }
 if(*(tmp.begin()+HOST_NAME_MAX)!=0) {  // overwritten, there is no trailing zero written
   WY_THROW( WyRet(Wym_ENAMETOOLONG) ); // insufficient tmp buffer
 }
 tmp._setsize( std::strlen(tmp.begin()) );
 buf.swap(tmp);
 return(Ok);
}
catch(const WyStr::Reply& e) {
  return(e);
};

};  // End namespace Wy
