/* Copyright is licensed under GNU LGPL.                 by I.J.Wang 2007
*/

#define WYLIB_SOURCE
#include "WyNum.h"
#include <cmath>       // std::isnan, isinf
#include "WyMath.h"    // for scif(..)
#include <cassert>
#include "WyStr.h"

namespace Wy {

//  [Syn] Convert number num into radix string representation, store the result into buf
//        according to left (l) or right (r) alignmentment suffix
//        The number of characters for fraction is specified by frdig.
//
//        radix=10: [-]ddd[.<ggg>]               --> float number decimal representation 
//        radix= 0: [-]<ddd>[.<ggg>]e(+-)<kkk>   --> scientific representation
//        radix=16: [-]ddd[.<ggg>]               --> float number hexidecimal representation 
//
//    Implementation works for radix 10, 0 and 16
//
// [Ret] Number of characters converted and stored into buf
//       Unused area for the result is intact (no trailing zero is written)
//       If 0 is returned, it is one of the following three reasons:
//            1. length blen of buf space is not enough to store the result
//            2. buf is NULL
//            3. radix is not supported
//
// note: 1.'e+0' is ignored
//       2.0x<nnn>p<kkk> like format should be easy by using frexp, function protocol
//         is might change if designated for this specific format.
//
size_t _ntoa(char* buf, size_t blen, double num, Wy::WyRadix radix, size_t frdig) WY__TSPC()
{
 typedef double NumType;            // type of argument num
 typedef signed long long FI_Type;  // intetral type holding precision part of NumType
 const char DPChar='.';
 const char ExpChar='e';
 size_t wr_lft=blen;
 char* wr_ptr=buf;
 double wnum;   // whole num
 double frac;   // fraction

 if(buf==NULL) {
   return(0);
 }

 // It is convenient and probably OK to check out nan/inf before checking for accepted radix
 //
 if(std::isnan(num)) {
   if(blen<3) {
     return(0);
   } else {
     buf[0]='n'; buf[1]='a'; buf[2]='n';
     return(3);
   }
 }
 if(int v=std::isinf(num)) {
   if(v>0) {
     if(blen<3) {
       return(0);
     } else {
       buf[0]='i'; buf[1]='n'; buf[2]='f';
       return(3);
     }
   } else {
     if(blen<4) {
       return(0);
     } else {
       buf[0]='-'; buf[1]='i'; buf[2]='n'; buf[3]='f';
       return(4);
     }
   }
 }

 if(radix==10) {  // radix 10 notation (l)
   const size_t FI_Digits( std::numeric_limits<FI_Type>::digits10 );
   size_t wdigs;
   size_t stmp;
   frac=std::modf(num,&wnum);
   { // wdigs= digits of wnum (sign symbol not counted)
     const double tmp=std::abs(wnum);
     wdigs= (tmp<10.0)? 1 : (static_cast<size_t>(std::log10(tmp))+1);
   }

   // check for enough blen before buf is written
   // (big frdig is also sieved)
   {
     size_t tmp=blen;
     if(tmp==0) {
       return(0);
     }
     if(tmp<=frdig) {
       return(0);
     }
     if(frdig) {
       tmp-=frdig+1;   // one for '.'
     }
     if(num>=0.0) {
       if(tmp<wdigs) {
         return(0);
       }
     } else {
       if(tmp<wdigs+1) {  // one for '-'
         return(0);
       }
     }
   }

   if(wdigs>FI_Digits) {
     // output FI_Digits and append overprecision '0's
     const size_t Trail_Zeros=wdigs-FI_Digits;
     
     stmp=Wy::wy_itoa_l(wr_ptr,wr_lft, 
          static_cast<FI_Type>( wnum/std::pow(10.0,static_cast<double>(Trail_Zeros)) ),radix);
     if(stmp==0) {
       return(0);    // should not occur
     }
     wr_lft-=stmp;
     wr_ptr+=stmp;
     std::memset(wr_ptr,'0',Trail_Zeros);
     wr_ptr+=Trail_Zeros;
     wr_lft-=Trail_Zeros;
   } else {
     const FI_Type fi_tmp=static_cast<FI_Type>(wnum);
     if(fi_tmp==0) {  // if num=-0.001, wnum casts to 0, lose the sign.
       if(num<0) {
         *wr_ptr='-'; ++wr_ptr;
         --wr_lft;
       }
       *wr_ptr='0'; ++wr_ptr;
       --wr_lft;
     } else {
       stmp=Wy::wy_itoa_l(wr_ptr,wr_lft,fi_tmp,radix);
       if(stmp==0) {
         return(0);
       }
       wr_lft-=stmp;
       wr_ptr+=stmp;
     }
   }

   if(frdig==0) {
     return(wr_ptr-buf);
   }
   if(frdig+1>wr_lft) {
     return(0);
   }
   *wr_ptr=DPChar;
   ++wr_ptr;
   frac=std::abs(frac);
   if(frdig<=FI_Digits) {
     std::modf(frac*std::pow(10.0,static_cast<double>(frdig)),&frac);
     if((stmp=Wy::wy_itoa_r(wr_ptr,frdig,static_cast<FI_Type>(frac),radix))==0) {
       return(0);  // should not occur
     }
     std::memset(wr_ptr,'0',frdig-stmp);
     wr_ptr+=frdig;
   } else {
     std::modf(frac*std::pow(10.0,static_cast<double>(FI_Digits)),&frac);
     if((stmp=Wy::wy_itoa_r(wr_ptr,FI_Digits,static_cast<FI_Type>(frac),radix))==0) {
       return(0);  // should not occur
     }
     std::memset(wr_ptr,'0',FI_Digits-stmp);
     wr_ptr+=FI_Digits;
     stmp=frdig-FI_Digits;
     std::memset(wr_ptr,'0',stmp);
     wr_ptr+=stmp;
   }
   return(wr_ptr-buf);
 } else if(radix==0) {  // scientific radix 10 notation (l)
   char e_buf[std::numeric_limits<int>::digits10+2]; // take larger one for simplicity
   char* e_ptr=e_buf+sizeof(e_buf);
   size_t e_digs;
   size_t stmp;
   frac=WyMath::_scif(num,wnum);
   if(wnum!=0) {
     if((stmp=Wy::wy_itoa_r(e_buf,sizeof(e_buf),static_cast<int>(wnum),10))==0) {
       return(0);  // should not occur
     }
     e_ptr-=stmp;
     if(*e_ptr=='-') {
       --e_ptr;
       *e_ptr=ExpChar;
     } else {
       --e_ptr;
       *e_ptr='+';
       --e_ptr;
       *e_ptr=ExpChar;
     }
     e_digs=(e_buf+sizeof(e_buf))-e_ptr;
   } else {
     e_digs=0;
   }
   if(blen<=e_digs) {
     return(0);
   }
   if((stmp=Wy::_ntoa(wr_ptr,blen-e_digs,frac,10,frdig))==0) {
     return(0);   // should not occur
   }
   wr_ptr+=stmp;
   std::memcpy(wr_ptr,e_ptr,e_digs);
   wr_ptr+=e_digs;
   return(wr_ptr-buf);
 } else if(radix==16) {  // hexidecimal float notation (l)
   assert(std::numeric_limits<FI_Type>::radix==2);
   const size_t FI_Digits( std::numeric_limits<FI_Type>::digits>>2 );
   size_t wdigs;
   size_t stmp;
   frac=std::modf(num,&wnum);
   { // wdigs= digits of wnum (sign symbol not counted)
     const double tmp=std::abs(wnum);
     wdigs= (tmp<16.0)? 1 : ((static_cast<size_t>(std::log(tmp)/M_LN2)>>2)+1);
   }

   // check for enough blen before buf is written
   // (big frdig is also sieved)
   {
     stmp=blen;
     if(stmp==0) {
       return(0);
     }
     if(stmp<=frdig) {
       return(0);
     }
     if(frdig) {
       stmp-=frdig+1;   // one for '.'
     }
     if(num>=0.0) {
       if(stmp<wdigs) {
         return(0);
       }
     } else {
       if(stmp<wdigs+1) {  // one for '-'
         return(0);
       }
     }
   }

   if(wdigs>FI_Digits) {
     // output FI_Digits and append overprecision '0's
     const size_t Trail_Zeros=wdigs-FI_Digits;
     
     stmp=Wy::wy_itoa_l(wr_ptr,wr_lft, 
          static_cast<FI_Type>( std::ldexp(wnum,-(Trail_Zeros<<2)) ),radix);
     if(stmp==0) {
       return(0);
     }
     wr_lft-=stmp;
     wr_ptr+=stmp;
     std::memset(wr_ptr,'0',Trail_Zeros);
     wr_ptr+=Trail_Zeros;
     wr_lft-=Trail_Zeros;
   } else {
     const FI_Type fi_tmp=static_cast<FI_Type>(wnum);
     if(fi_tmp==0) {  // if num=-0.001, wnum casts to 0, lose the sign.
       if(num<0) {
         *wr_ptr='-'; ++wr_ptr;
         --wr_lft;
       }
       *wr_ptr='0'; ++wr_ptr;
       --wr_lft;
     } else {
       stmp=Wy::wy_itoa_l(wr_ptr,wr_lft,fi_tmp,radix);
       if(stmp==0) {
         return(0);
       }
       wr_lft-=stmp;
       wr_ptr+=stmp;
     }
   }

   if(frdig==0) {
     return(wr_ptr-buf);
   }
   if(frdig+1>wr_lft) {
     return(0);
   }
   *wr_ptr=DPChar;
   ++wr_ptr;
   frac=std::abs(frac);
   if(frdig<=FI_Digits) {
     std::modf(std::ldexp(frac,frdig<<2),&frac);
     if((stmp=Wy::wy_itoa_r(wr_ptr,frdig,static_cast<FI_Type>(frac),radix))==0) {
       return(0);  // should not occur
     }
     std::memset(wr_ptr,'0',frdig-stmp);
     wr_ptr+=frdig;
   } else {
     std::modf(std::ldexp(frac,FI_Digits<<2),&frac);
     if((stmp=Wy::wy_itoa_r(wr_ptr,FI_Digits,static_cast<FI_Type>(frac),radix))==0) {
       return(0);  // should not occur
     }
     std::memset(wr_ptr,'0',FI_Digits-stmp);
     wr_ptr+=FI_Digits;
     stmp=frdig-FI_Digits;
     std::memset(wr_ptr,'0',stmp);
     wr_ptr+=stmp;
   }
   return(wr_ptr-buf);
 } else {
   return(0);
 };
 WY_ASSERT("UNREACHABLE");
};

// If return value v>0, result is stored at [buf+blen-v, buf+blen)
//
size_t _ntoa_r(char* buf, size_t blen, double num, Wy::WyRadix radix, size_t frdig) WY__TSPC()
{
 typedef double NumType;            // type of argument num
 typedef signed long long FI_Type;  // intetral type holding precision part of NumType
 const char DPChar='.';
 const char ExpChar='e';
 char* wr_ptr=buf+blen;
 double wnum;   // whole num
 double frac;   // fraction

 if(buf==NULL) {
   return(0);
 }

 // It is convenient and probably OK to check out nan/inf before checking for accepted radix
 //
 if(std::isnan(num)) {
   if(blen<3) {
     return(0);
   } else {
     wr_ptr-=3;
     wr_ptr[0]='n'; wr_ptr[1]='a'; wr_ptr[2]='n';
     return(3);
   }
 }
 if(int v=std::isinf(num)) {
   if(v>0) {
     if(blen<3) {
       return(0);
     } else {
       wr_ptr-=3;
       wr_ptr[0]='i'; wr_ptr[1]='n'; wr_ptr[2]='f';
       return(3);
     }
   } else {
     if(blen<4) {
       return(0);
     } else {
       wr_ptr-=4;
       wr_ptr[0]='-'; wr_ptr[1]='i'; wr_ptr[2]='n'; wr_ptr[3]='f';
       return(4);
     }
   }
 }

 if(radix==10) {  // radix 10 notation (r)
   const size_t FI_Digits( std::numeric_limits<FI_Type>::digits10 );
   size_t wdigs;
   size_t stmp;
   frac=std::modf(num,&wnum);
   { // wdigs= digits of wnum (sign symbol not counted)
     const double tmp=std::abs(wnum);
     wdigs= (tmp<10.0)? 1 : (static_cast<size_t>(std::log10(tmp))+1);
   }

   // check for enough blen before buf is written
   // (big frdig is also sieved)
   {
     stmp=blen;
     if(stmp==0) {
       return(0);
     }
     if(stmp<=frdig) {
       return(0);
     }
     if(frdig) {
       stmp-=frdig+1;   // one for '.'
     }
     if(num>=0.0) {
       if(stmp<wdigs) {
         return(0);
       }
     } else {
       if(stmp<wdigs+1) {  // one for '-'
         return(0);
       }
     }
   }

   if(frdig!=0) {
     frac=std::abs(frac);
     if(frdig<=FI_Digits) {
       std::modf(frac*std::pow(10.0,static_cast<double>(frdig)),&frac);
       if((stmp=Wy::wy_itoa_r(buf,wr_ptr-buf,static_cast<FI_Type>(frac),radix))==0) {
         return(0);  // should not occur
       }
       wr_ptr-=stmp;
       stmp=frdig-stmp;
       wr_ptr-=stmp;
       std::memset(wr_ptr,'0',stmp);
     } else {
       stmp=frdig-FI_Digits;
       wr_ptr-=stmp;
       std::memset(wr_ptr,'0',stmp);
       wr_ptr-=FI_Digits;
       std::modf(frac*std::pow(10.0,static_cast<double>(FI_Digits)),&frac);
       if((stmp=Wy::wy_itoa_r(buf,wr_ptr-buf,static_cast<FI_Type>(frac),radix))==0) {
         return(0);  // should not occur
       }
       wr_ptr-=FI_Digits;
       stmp=FI_Digits-stmp;
       std::memset(wr_ptr,'0',stmp);
     }
     --wr_ptr;
     *wr_ptr=DPChar;
   }

   if(wdigs>FI_Digits) {
     // output FI_Digits and append overprecision '0's
     const size_t Trail_Zeros=wdigs-FI_Digits;
     wr_ptr-=Trail_Zeros;
     std::memset(wr_ptr,'0',Trail_Zeros);
     stmp=Wy::wy_itoa_r(buf,wr_ptr-buf,
          static_cast<FI_Type>( wnum/std::pow(10.0,static_cast<double>(Trail_Zeros)) ),radix);
     if(stmp==0) {
       return(0);    // should not occur
     }
     wr_ptr-=stmp;
   } else {
     const FI_Type fi_tmp=static_cast<FI_Type>(wnum);
     if(fi_tmp==0) {  // if num=-0.001, wnum casts to 0, lose the sign.
       --wr_ptr;
       *wr_ptr='0'; 
       if(num<0) {
         --wr_ptr;
         *wr_ptr='-';
       }
     } else {
       stmp=Wy::wy_itoa_r(buf,wr_ptr-buf,fi_tmp,radix);
       if(stmp==0) {
         return(0);
       }
       wr_ptr-=stmp;
     }
   }
   return (buf+blen)-wr_ptr;
 } else if(radix==0) {  // scientific radix 10 notation (r)
   char e_buf[std::numeric_limits<int>::digits10+2]; // take larger one for simplicity
   char* e_ptr=e_buf+sizeof(e_buf);
   size_t e_digs;
   size_t stmp;
   frac=WyMath::_scif(num,wnum); 
   if(wnum!=0) {
     if((stmp=Wy::wy_itoa_r(e_buf,sizeof(e_buf),static_cast<int>(wnum),10))==0) {
       return(0);  // should not occur
     }
     e_ptr-=stmp;
     if(*e_ptr=='-') {
       --e_ptr;
       *e_ptr=ExpChar;
     } else {
       --e_ptr;
       *e_ptr='+';
       --e_ptr;
       *e_ptr=ExpChar;
     }
     e_digs=(e_buf+sizeof(e_buf))-e_ptr;
     wr_ptr-=e_digs;
   } else {
     e_digs=0;
   }
   if(blen<=e_digs) {
     return(0);
   }
   if((stmp=Wy::_ntoa_r(buf,wr_ptr-buf,frac,10,frdig))==0) {
     return(0);
   }
   std::memcpy(wr_ptr,e_ptr,e_digs);
   return stmp+e_digs;
 } else if(radix==16) {  // hexidecimal float notation (r)
   assert(std::numeric_limits<FI_Type>::radix==2);
   const size_t FI_Digits( std::numeric_limits<FI_Type>::digits>>2 );
   size_t wdigs;
   size_t stmp;
   frac=std::modf(num,&wnum);
   { // wdigs= digits of wnum (sign symbol not counted)
     const double tmp=std::abs(wnum);
     wdigs= (tmp<16.0)? 1 : ((static_cast<size_t>(std::log(tmp)/M_LN2)>>2)+1);
   }

   // check for enough blen before buf is written
   // (big frdig is also sieved)
   {
     stmp=blen;
     if(stmp==0) {
       return(0);
     }
     if(stmp<=frdig) {
       return(0);
     }
     if(frdig) {
       stmp-=frdig+1;   // one for '.'
     }
     if(num>=0.0) {
       if(stmp<wdigs) {
         return(0);
       }
     } else {
       if(stmp<wdigs+1) {  // one for '-'
         return(0);
       }
     }
   }

   if(frdig!=0) {
     frac=std::abs(frac);
     if(frdig<=FI_Digits) {
       std::modf(std::ldexp(frac,frdig<<2),&frac);
       if((stmp=Wy::wy_itoa_r(buf,wr_ptr-buf,static_cast<FI_Type>(frac),radix))==0) {
         return(0);  // should not occur
       }
       wr_ptr-=stmp;
       stmp=frdig-stmp;
       wr_ptr-=stmp;
       std::memset(wr_ptr,'0',stmp);
     } else {
       stmp=frdig-FI_Digits;
       wr_ptr-=stmp;
       std::memset(wr_ptr,'0',stmp);
       wr_ptr-=FI_Digits;
       std::modf(std::ldexp(frac,FI_Digits<<2),&frac);
       if((stmp=Wy::wy_itoa_r(buf,wr_ptr-buf,static_cast<FI_Type>(frac),radix))==0) {
         return(0);  // should not occur
       }
       wr_ptr-=FI_Digits;
       stmp=FI_Digits-stmp;
       std::memset(wr_ptr,'0',stmp);
     }
     --wr_ptr;
     *wr_ptr=DPChar;
   }

   if(wdigs>FI_Digits) {
     // output FI_Digits and append overprecision '0's
     const size_t Trail_Zeros=wdigs-FI_Digits;
     wr_ptr-=Trail_Zeros;
     std::memset(wr_ptr,'0',Trail_Zeros);
     stmp=Wy::wy_itoa_r(buf,wr_ptr-buf, 
          static_cast<FI_Type>( std::ldexp(wnum,-(Trail_Zeros<<2)) ),radix);
     if(stmp==0) {
       return(0);
     }
     wr_ptr-=stmp;
   } else {
     const FI_Type fi_tmp=static_cast<FI_Type>(wnum);
     if(fi_tmp==0) {  // if num=-0.001, wnum casts to 0, lose the sign.
        --wr_ptr;
       *wr_ptr='0';
       if(num<0) {
         --wr_ptr;
         *wr_ptr='-';
       }
     } else {
       stmp=Wy::wy_itoa_r(buf,wr_ptr-buf,fi_tmp,radix);
       if(stmp==0) {
         return(0);
       }
       wr_ptr-=stmp;
     }
   }
   return (buf+blen)-wr_ptr;
 } else {
   return(0);
 };
 WY_ASSERT("UNREACHABLE");
};

size_t _ntoa(char* buf, size_t blen, float num, Wy::WyRadix radix, size_t frdig) WY__TSPC()
{
 return Wy::_ntoa(buf,blen,static_cast<double>(num),radix,frdig);
};

size_t _ntoa_r(char* buf, size_t blen, float num, Wy::WyRadix radix, size_t frdig) WY__TSPC()
{
 return Wy::_ntoa_r(buf,blen,static_cast<double>(num),radix,frdig);
};

// Note: For WyStr versions: If str has enough capacity, no reallocation, so no ENOMEM
//
WyRet _mkstr(WyStr& str, double num, Wy::WyRadix radix, size_t frdig) WY__TSPC()
{
 size_t v=Wy::_ntoa(str.begin(), str._capacity(), num, radix, frdig);
 if(v<=0) {
   switch(radix) {
     case 0:
     case 10:
     case 16: break;
     default: WY_RETURN( Wym_EINVAL );
   }
   WyRet r;
   if((r=str._reserve( std::numeric_limits<double>::digits+ frdig+ 7))!=Ok) {
     WY_RETURN(r);
   }
   v=Wy::_ntoa(str.begin(), str._capacity(), num, radix, frdig);
   if(v==0) {
     WY_ASSERT(v>0);
     WY_RETURN( Wym_ENOMEM );   // should not occur
   }
 }
 str._setsize(v);
 return(Ok);
};

WyRet _mkstr(WyStr& str, float num, Wy::WyRadix radix, size_t frdig) WY__TSPC()
{
 size_t v=Wy::_ntoa(str.begin(), str._capacity(), num, radix, frdig);
 if(v<=0) {
   switch(radix) {
     case 0:
     case 10:
     case 16: break;
     default: WY_RETURN( Wym_EINVAL );
   }
   WyRet r;
   if((r=str._reserve( std::numeric_limits<float>::digits+ frdig+ 7))!=Ok) {
     WY_RETURN(r);
   }
   v=Wy::_ntoa(str.begin(), str._capacity(), num, radix, frdig);
   if(v==0) {
     WY_ASSERT(v>0);
     WY_RETURN( Wym_ENOMEM );   // should not occur
   }
 }
 str._setsize(v);
 return(Ok);
};

WyStr wrd(double num, Wy::WyRadix radix, size_t frdig)
{
 WyStr str;
 size_t v=Wy::_ntoa(str.begin(), str._capacity(), num, radix, frdig);
 if(v<=0) {
   switch(radix) {
     case 0:
     case 10:
     case 16: break;
     default: WY_THROW( WyStr::Reply(Wym_EINVAL) );
   }
   WyRet r;
   if((r=str._reserve( std::numeric_limits<double>::digits+ frdig+ 7))!=Ok) {
     WY_THROW( WyStr::Reply(r) );
   }
   v=Wy::_ntoa(str.begin(), str._capacity(), num, radix, frdig);
   if(v==0) {
     WY_ASSERT(v>0);
     WY_THROW( WyStr::Reply(Wym_ENOMEM) );   // pretending ENOMEM
   }
 }
 str._setsize(v);
 return(str);
};

WyStr wrd(float num, Wy::WyRadix radix, size_t frdig)
{
 WyStr str;
 size_t v=Wy::_ntoa(str.begin(), str._capacity(), num, radix, frdig);
 if(v<=0) {
   switch(radix) {
     case 0:
     case 10:
     case 16: break;
     default: WY_THROW( WyStr::Reply(Wym_EINVAL) );
   }
   WyRet r;
   if((r=str._reserve( std::numeric_limits<float>::digits+ frdig+ 7))!=Ok) {
     WY_THROW( WyStr::Reply(r) );
   }
   v=Wy::_ntoa(str.begin(), str._capacity(), num, radix, frdig);
   if(v==0) {
     WY_ASSERT(v>0);
     WY_THROW( WyStr::Reply(Wym_ENOMEM) );   // pretending ENOMEM
   }
 }
 str._setsize(v);
 return(str);
};

};  // end namespace Wy
