/* Copyright is licensed under GNU LGPL.                 by I.J.Wang 2007
*/

#define WYLIB_SOURCE
#include "WyNum.h"
#include "WyTimeSpec.h"
#include <stdint.h>   // for uint64_t
#include <cassert>

namespace Wy {

// The following values are from statistics generated by g_ntoa_timespec.cpp
//
static const char Wy_DReq[]= {
   // Exact number of digits of radix r to hold value=999999999
   // E.g. Nine digits of radix 10 are required to hold the value
   //      Eight digits of radix 16 are required to hold the value, etc.
   //
   0,0, 30, 19, 15, 13, 12, 11, 10, 10, 9, 9, 9, 9, 8, 8, 8, 8, 8, 8, 7, 7,
    7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 6, 6, 6, 6, 6,
};

static const uint64_t Wy_Scale=670483147LL;
static const uint64_t Wy_Mi[]= {
   // Multiplier coefficient mi, so that n= tv_nsec*mi/scale, n/r^d==tv_nsec/Giga
   //        r=radix, d=Wy_DReq[r]
   //
   //   Note: 64 bits are required to hold 999999999*wy_mi[31]
   //         35 bits are required to hold the n value above
   //
    0,0,
    719925797ULL,   // radix=2
    779276726ULL,   // radix=3
    719925797ULL,   // radix=4
    818460872ULL,   // radix=5
   1459495870ULL,   // radix=6
   1325764257ULL,   // radix=7
    719925797ULL,   // radix=8
   2337830178ULL,   // radix=9
    670483147ULL,   // radix=10
   1580964188ULL,   // radix=11
   3459545768ULL,   // radix=12
   7110138111ULL,   // radix=13
    989491690ULL,   // radix=14
   1718374971ULL,   // radix=15
   2879703188ULL,   // radix=16
   4677127801ULL,   // radix=17
   7388697846ULL,   // radix=18
  11387192795ULL,   // radix=19
    858218428ULL,   // radix=20
   1207599512ULL,   // radix=21
   1672424926ULL,   // radix=22
   2282878080ULL,   // radix=23
   3075151793ULL,   // radix=24
   4092304364ULL,   // radix=25
   5385193362ULL,   // radix=26
   7013490534ULL,   // radix=27
   9046781170ULL,   // radix=28
  11565751353ULL,   // radix=29
  14663466424ULL,   // radix=30
  18446744091ULL,   // radix=31
    719925797ULL,   // radix=32
    865907508ULL,   // radix=33
   1035765326ULL,   // radix=34
   1232526121ULL,   // radix=35
   1459495870ULL,   // radix=36
};

// [Interna] Convert num to radix string representation which is written into
//           buf pointed array with trailing zero added. The result is like 
//           normal string representation without decimal point.
//             [-]<ddd><ggg>
//
//  Limits: 1.buf must be non-NULL
//          2.buf pointed array must have enough space
//            space for <ggg> is at least Wy::Wy_DReq[radix] bytes
//          3.radix is in range [2-36]
//
// [Ret] The number of characters written into buf substracting frsize which
//       is written with the number of digits for fraction <ggg>.
//       If the whole number part is zero (i.e. num.second()==0), <ddd> is a 
//       single '0' and return value is 1.
//       If conversion failed, zero is returned and frsize is intact.
//
size_t wy_ntoa_impl(char* buf, size_t& frsize, WyTimeSpec num, Wy::WyRadix radix)
{
 typedef WyTimeSpec NumType;     // type of argument num
 assert( std::numeric_limits<uint64_t>::digits>=64 );
 assert( std::numeric_limits<long long>::digits>=35 );
 WY_ASSERT(buf!=NULL);
 WY_ASSERT((radix>=2)&&(radix<=36));

 const size_t DigitsR= Wy::Wy_DReq[static_cast<int>(radix)];  // compiler may warn subscript type
 const size_t wdigs=Wy::wy_ndigits(num.second(),radix);
 char* wr_ptr=buf;
 size_t stmp;

 WyTimeSpec::SecondType wnum=num.second();
 WyTimeSpec::NanoSecType frac=num.nano();
 if(wnum<0) {
   frac=-frac;
 } else if (wnum>0) {
 } else {
   if(frac<0) {
     *wr_ptr='-';
     ++wr_ptr;
     frac=-frac;
   }
 }

 if((stmp=Wy::wy_itoa_l(wr_ptr,
                        std::numeric_limits<WyTimeSpec::SecondType>::digits+2,
                        wnum,radix))==0) {
   return(0);  // should not occur
 }
 WY_ASSERT(stmp<=std::numeric_limits<WyTimeSpec::SecondType>::digits+2);
 wr_ptr+=stmp;

 // Compiler may warn subscript type being char
 if((stmp=Wy::wy_itoa_r(wr_ptr,DigitsR,
                        static_cast<long long>((frac*Wy::Wy_Mi[static_cast<int>(radix)])/Wy::Wy_Scale)
                       ,radix))==0) {
   return(0);  // should not occur
 }
 WY_ASSERT(stmp<=DigitsR);

 std::memset(wr_ptr,'0',DigitsR-stmp);
 frsize=DigitsR;
 return( (buf[0]=='-')? wdigs+1: wdigs);
};

size_t _ntoa(char* buf, size_t blen, const WyTimeSpec& num, Wy::WyRadix radix, size_t frdig) WY__TSPC()
{
 typedef WyTimeSpec NumType;     // type of argument num
 const char DPChar='.';
 const char ExpChar='e';
 char* wr_ptr=buf;

 if(buf==NULL) {
   return(0);
 }

 if((radix>=Wy::Wy_MinRadix)&&(radix<=Wy::Wy_MaxRadix)) {  // radix [2,36] notation (l)
   char tmpbuf[ std::numeric_limits<WyTimeSpec::SecondType>::digits 
                +std::numeric_limits<WyTimeSpec::NanoSecType>::digits+2
              ];
   size_t fr_size;
   size_t stmp;
   if((stmp=wy_ntoa_impl(tmpbuf,fr_size,num,radix))==0) {
     return(0);
   }
   const char* fr_ptr=tmpbuf+stmp;
   if(blen<frdig) {
     return(0);
   }
   if(blen-frdig<stmp) {
     return(0);
   }
   const char* p=tmpbuf;
   for( ; p!=fr_ptr; ) {
     *wr_ptr=*p++;
     ++wr_ptr;
   }
   if(frdig==0) {
     return(wr_ptr-buf);
   }
   *wr_ptr=DPChar;
   ++wr_ptr;

   stmp= (frdig<=fr_size)? frdig : fr_size;
   for( ; stmp>0; --stmp) {
     *wr_ptr=*p++;
     ++wr_ptr;
   }
   return(wr_ptr-buf);
 } else if(radix==0) {  // scientific radix 10 notation (l)
   const int Eff_Radix=10;
   char tmpbuf[ std::numeric_limits<WyTimeSpec::SecondType>::digits10
                +std::numeric_limits<WyTimeSpec::NanoSecType>::digits10+2
              ];
   char e_buf[4]; // e(+-)<dd>
   char* hd_ptr;
   char* fr_ptr;
   size_t fr_size;
   char* e_ptr;     // Is initialized (g++ warning: using may be uninitialized)
   int e_num;       // Is initialized (g++ warning: using may be uninitialized)
   size_t e_digs;   // Is initialized (g++ warning: using may be uninitialized)
   size_t stmp;
   if(num.is_default()) {
     if(blen<1) {
       return(0);
     }
     buf[0]='0';
     return(1);
   }
   if((stmp=wy_ntoa_impl(tmpbuf,fr_size,num,Eff_Radix))==0) {
     return(0);
   }
   WY_ASSERT(stmp<sizeof(tmpbuf));
   fr_ptr=tmpbuf+stmp;
   if(blen<frdig) {
     return(0);
   }
   if(blen-frdig<stmp) {
     return(0);
   }

   if(tmpbuf[0]=='-') {
     --stmp;
   }
   hd_ptr=tmpbuf;
   if(stmp<=1) {
     // one digit whole number
     e_num=0;
     if(fr_ptr[-1]=='0') {
       for(;;) {
         if(*fr_ptr!='0') {  // fr_ptr will see non-'0', for num==0 won't get here
           break;
         }
         ++fr_ptr;
         --fr_size;
         --e_num;
       }
       ++fr_ptr;
       --fr_size;
       --e_num;
       if(tmpbuf[0]=='-') {
         hd_ptr=fr_ptr-2;
         *hd_ptr='-';
       } else {
         hd_ptr=fr_ptr-1;
       }
     }
   } else {
     e_num=stmp-1;
     fr_ptr-=e_num;
     fr_size+=e_num;
   }

    if(e_num!=0) {
      WY_ASSERT((e_num>=-99)&&(e_num<=99));
      e_ptr=e_buf+sizeof(e_buf);
     if((stmp=Wy::wy_itoa_r(e_buf,sizeof(e_buf),e_num,10))==0) {
       return(0);  // should not occur
     }
     WY_ASSERT(stmp<=sizeof(e_buf)-2);
     e_ptr-=stmp;
     if(*e_ptr=='-') {
       --e_ptr;
       *e_ptr=ExpChar;
     } else {
       --e_ptr;
       *e_ptr='+';
       --e_ptr;
       *e_ptr=ExpChar;
     }
     e_digs=(e_buf+sizeof(e_buf))-e_ptr;
   }

   stmp=fr_ptr-hd_ptr;
   if(blen<stmp) {
     return(0);
   }
   if(frdig!=0) {
     stmp+=frdig+1;
     if(blen<stmp) {
       return(0);
     }
   }
   if(e_num!=0) {
     stmp+=e_digs;
     if(blen<stmp) {
       return(0);
     }
   }

   stmp=fr_ptr-hd_ptr;
   std::memcpy(wr_ptr,hd_ptr,stmp);
   wr_ptr+=stmp;
   if(frdig!=0) {
     *wr_ptr=DPChar;
     ++wr_ptr;
     stmp=(frdig<=fr_size)? frdig: fr_size;
     std::memcpy(wr_ptr,fr_ptr,stmp);
     wr_ptr+=stmp;
   }
   if(e_num!=0) {
     std::memcpy(wr_ptr,e_ptr,e_digs);
     wr_ptr+=e_digs;
   }
   return(wr_ptr-buf);
 } else {
   return(0);
 };
 WY_ASSERT("UNREACHABLE");
};

WyRet _mkstr(WyStr& str, WyTimeSpec num, Wy::WyRadix radix, size_t frdig) WY__TSPC()
{
 size_t v=Wy::_ntoa(str.begin(), str._capacity(), num, radix, frdig);
 if(v<=0) {
   if((radix<Wy::Wy_MinRadix)||(radix>Wy::Wy_MaxRadix)) {
     if(radix!=0) {
       WY_RETURN( Wym_EINVAL );
     }
   }
   WyRet r;
   if((r=str._reserve(std::numeric_limits<WyTimeSpec::SecondType>::digits+frdig+7))!=Ok) {
     WY_RETURN(r);
   }
   v=Wy::_ntoa(str.begin(), str._capacity(), num, radix, frdig);
   if(v==0) {
     WY_ASSERT(v>0);
     WY_RETURN( Wym_ENOMEM );   // should not occur
   }
 }
 str._setsize(v);
 return(Ok);
};

WyStr wrd(const WyTimeSpec& tm, Wy::WyRadix radix, size_t frdig)
{
 WyStr str;
 WyRet r( Wy::_mkstr(str,tm,radix,frdig) );
 if(r!=Ok) {
   WY_THROW( WyStr::Reply(r) );
 }
 return(str);
};

};  // end namespace Wy

