/* Copyright is licensed under GNU LGPL.                 by I.J.Wang 2007

   Header only for inclusion
   Template for instantiate _strnum(..) of integer types
*/
#ifndef WY_PARSE_NSTR_H__
#define WY_PARSE_NSTR_H__

#include "WyRet.h"
#include <limits>

namespace Wy {

//--------------------------------------------------------------------------------------------
// [Internal] Parse the string pointed by nstr to integer
//
//          <S> ::= [+-][<p>]<ddd>
//          
//          <ddd>::= non-empty combination of digit string
//
// [Ret] Ok           endptr= nstr+ns_len,  the whole string is interpret
//       Wym_EINVAL   ns_radix not in range [0,2-36]. res=0 endptr==nstr
//       Wym_ENOENT   No number recognized. res=0 endptr==nstr
//       Wym_EBADMSG  endptr= the bad character [nstr,endptr)
//       Wym_ERANGE   endptr= the character that may cause overflow [nstr,endptr)
//
//       [nstr,endptr) = The prefix of nstr string accepted by S and converted
//       res = the converted integer
//       ns_radix = If ns_radix is zero, <p> is parsed as the radix, the interpreted radix
//               is then written back to ns_radix
//               <p>= 0           radix=8  (00, 00.? are radix eight. 0 and 0.? are radix ten)
//               <p>= 0[xX]       radix=16
//               <p>= otherwise   radix=10
//
//  Note: endptr must not be from wy_begin_ref()
//
template <typename ResType>
WyRet wy_parse_nstr(ResType& res,  const char*& endptr, WyCSeg nseg, Wy::WyRadix& ns_radix)
{
 if(std::numeric_limits<ResType>::is_integer==false) {
   WY_TERMINATE("");
 }

 res=ResType(0);
 endptr=nseg.begin();    // always written before return

 if(nseg.size()<=0) {
   WY_RETURN( Wym_ENOENT );
 }

 const char* const Zptr(nseg.end());
 const char* sptr(nseg.begin());
 bool is_neg(false);

 // Resolve the sign symbol
 //
 if(std::numeric_limits<ResType>::is_signed) {
   if(*sptr=='+') {
     ++sptr;
     if(sptr==Zptr) {
       WY_RETURN(Wym_ENOENT);
     }
   } else if(*sptr=='-') {
     ++sptr;
     if(sptr==Zptr) {
       WY_RETURN(Wym_ENOENT);
     }
     is_neg=true;
   } else {
     // no sign symbol recognized
   }
 } else {
   if(*sptr=='+') {
     ++sptr;
     if(sptr==Zptr) {
       WY_RETURN(Wym_ENOENT);
     }
   }
 }

 WY_ASSERT(sptr!=Zptr);

 // Resolve ns_radix
 //               V------------ sptr= first character after [+-]
 //  <S>::= [+-][<p>]...
 //
 if(ns_radix>Wy::Wy_MaxRadix) {
   WY_RETURN(Wym_EINVAL);
 } else if(ns_radix<Wy::Wy_MinRadix) {
   if(ns_radix!=0) {
     WY_RETURN(Wym_EINVAL);
   }

   // Determine effective ns_radix from the prefix
   //
   if(sptr[0]=='0') {
     endptr=++sptr;
     if(sptr==Zptr) {
       ns_radix=10;
       return Ok;        // one digit string [+-]0
     } else {
       if((*sptr=='x')||(*sptr=='X')) {
         // [+-]0x
         ++sptr;
         if(sptr==Zptr) {
           ns_radix=10;
           WY_RETURN(Wym_EBADMSG);  // [+-]0x
         }
         const int tmp( Wy::_charnum(*sptr) );
         if((tmp<0)||(tmp>=16)) {
           // [+-]0x<non-hex>  e.g. "0xx" "0xg"
           ns_radix=10;
           WY_RETURN(Wym_EBADMSG);
         }
         // FALLTHROUGH [+-]0x<hex>  ---sptr at first character of <hex>
         ns_radix=16;
       } else {
         // [+-]0<non-x>  e.g. "08" "0."
         const int tmp( Wy::_charnum(*sptr) );
         if((tmp<0)||(tmp>=8)) {
           ns_radix=10;
           WY_RETURN(Wym_EBADMSG); // [+-]0<non-oct>
         }
         ns_radix=8;
       };
     }
   } else {
     ns_radix=10;
   };
 } else {
   if(ns_radix==16) {
     if(Zptr-sptr>=2) {
       if(sptr[0]=='0') {
         endptr=++sptr;
         if((*sptr=='x')||(*sptr=='X')) {
           ++sptr;
           if(sptr==Zptr) {
             WY_RETURN(Wym_EBADMSG);  // [+-]0x
           }
           // FALL_THROUGH  [+-]0x?
         };
       };
     };
   } else {
   };
 };

 //                 V----------- sptr= first character after <p>
 //  S::= [+-][<p>]<ddd>
 //
 WY_ASSERT(sptr!=Zptr);
 if((std::numeric_limits<ResType>::is_signed)&&(is_neg)) {
   const ResType MultMin=std::numeric_limits<ResType>::min()/ns_radix;
   for(;;) {
     const int d0= Wy::_charnum(*sptr);
     if((d0<0)||(d0>=ns_radix)) {
       if(endptr==nseg.begin()) {
         WY_RETURN( Wym_ENOENT );
       }
       WY_RETURN( Wym_EBADMSG );
     }
     if(res<MultMin) {
       WY_RETURN( Wym_ERANGE );
     }
     const ResType tmp1=res*ns_radix;
     if(tmp1<std::numeric_limits<ResType>::min()+d0) {
       WY_RETURN( Wym_ERANGE );
     }
     res=tmp1-d0;
     if( (endptr=++sptr)==Zptr ) {
       return(Ok);
     }
   };
 } else {
   const ResType MultMax=std::numeric_limits<ResType>::max()/ns_radix;
   for(;;) {
     const int d0= Wy::_charnum(*sptr);
     if((d0<0)||(d0>=ns_radix)) {
       if(endptr==nseg.begin()) {
         WY_RETURN( Wym_ENOENT );
       }
       WY_RETURN( Wym_EBADMSG );
     }
     if(res>MultMax) {
       WY_RETURN( Wym_ERANGE );
     }
     const ResType tmp1=res*ns_radix;
     if(tmp1>std::numeric_limits<ResType>::max()-d0) {
       WY_RETURN( Wym_ERANGE );
     }
     res=tmp1+d0;
     if( (endptr=++sptr)==Zptr ) {
       return(Ok);
     }
   };
 }
 WY_TERMINATE("UNREACHABLE");
};

}; // end namespace Wy

#endif  // end WY_PARSE_NSTR_H__
