/* Copyright is licensed under GNU LGPL.                 by I.J.Wang 2007

  .NAN/INF is not accepted considering strnum(..) never returns an NAN/INF
  .instance such as "2147483648e-1" may fail of ERANGE but result is consistent
*/

#define WYLIB_SOURCE
#include "WyNum.h"        // for _charnum(char)
#include "WyTimeSpec.h"
#include <cstdlib>          // for lldiv
#include <assert.h>
#include <limits>

namespace Wy {

//-------------------------------------------------------------------
//          <S> ::= [+-][<p>](<S1>|<S2>|<S3>)[Exponent]
//          <S1>::= <ddd>[<S3>]
//          <S2>::= <ddd>.
//          <S3>::= .<ggg>
//          <Exponent>=<eEpP>[+-]<kkk>
//          
//          <ddd>::= <ggg> ::= <kkk> ::=non-empty combination of valid digit character
//  Note: ns_radix is implemented for only 10 and 16
//
// Return condidion: res is the successfully conversion of [nstr,endptr)
//          Ok        endptr= nseg.end()
//       Wym_EINVAL   endptr= nseg.begin()
//       Wym_ENOENT   endptr= nseg.begin()
//       Wym_EBADMSG  endptr= the bad character
//       Wym_ERANGE   endptr= the character that may cause conversion overflow
//
//  Note: endptr must not be from wy_begin_ref()
//
static WyRet wy_strnum(WyTimeSpec& res, const char*& endptr, const WyCSeg& nseg, Wy::WyRadix& ns_radix) WY__TSPC()
{
 typedef WyTimeSpec ResType;      // type of argument res
 typedef long long IFN_Type;      // signed integer type storing fraction
 typedef signed int IEX_Type;     // signed integer type storing exponential

 static const char DPChar='.';    // decimal point character

 assert((std::numeric_limits<WyTimeSpec::SecondType>::digits
         +std::numeric_limits<WyTimeSpec::NanoSecType>::digits)
           < std::numeric_limits<IFN_Type>::digits);

 // ifn_num is the number converted from the string prefix ignoring the dot, radix characters.
 //     e.g. IFN(0.00012345678e-4)= 12345[678] depending on the precision of type IFN_Type
 //
 static const IFN_Type IFN_MaxVal= (static_cast<IFN_Type>(
                              std::numeric_limits<WyTimeSpec::SecondType>::max())
                                   *WY__CONST_GIGA)+(WY__CONST_GIGA-1);
 static const IFN_Type IFN_MinVal= (static_cast<IFN_Type>(
                              std::numeric_limits<WyTimeSpec::SecondType>::min())
                                   *WY__CONST_GIGA)-(WY__CONST_GIGA-1);

 // note: ddd can overflow regardless of the possibly latter exponential may correct
 //       This is implement limits
 //
 static const IFN_Type IFN_MaxVal_ddd=std::numeric_limits<WyTimeSpec::SecondType>::max();
 static const IFN_Type IFN_MinVal_ddd=std::numeric_limits<WyTimeSpec::SecondType>::min();

 static const int MaxExp_10= 9;    // limits to one digit (easier to detect overflow)
 static const int MaxExp_16= 0xf;
 static const int MinExp_10= -9;
 static const int MinExp_16= -0xf;
 
 static const size_t Size_p10tab=19;
 static const IFN_Type p10tab[Size_p10tab]= {
  1LL,10LL,100LL,1000LL,10000LL,100000LL,1000000LL, 10000000LL, // 10^0 - 10^7
  100000000LL,          // 10^8
  1000000000LL,         // 10^9
  10000000000LL,        // 10^10
  100000000000LL,       // 10^11
  1000000000000LL,      // 10^12
  10000000000000LL,     // 10^13
  100000000000000LL,    // 10^14
  1000000000000000LL,   // 10^15
  10000000000000000LL,  // 10^16
  100000000000000000LL, // 10^17
  1000000000000000000LL // 10^18
 };

 res.reset();
 endptr=nseg.begin();        // always written before return. ns_radix may be written latter

 if(nseg.size()<=0) {
   WY_RETURN( Wym_ENOENT );
 }

 const char* const Zptr(nseg.end());
 IFN_Type ifn_num=0;                // value of mantissa (integer representation)
 int exp_base=10;                   // base value of exponent (2|10)
 IEX_Type iex_num=0;                // value of the exponent (signed integer)
 int ifn_digs=0;                    // number of significant digits that ifn_num is made of radix
 int ifn_sht=0;                     // the real mantissa is ifn_num*(radix^ifn_sht)
 bool is_neg(false);                // mantissa is negative or not
 int MaxExpVal=MaxExp_10;           // the max./min. value of <kkk> (may be biased temporarily
 int MinExpVal=MinExp_10;           //  in converting if <kkk> exists)
 const char* sptr(endptr);          // string scanning pointer
 WyRet r;

 // Resolve the sign symbol
 //
 if(*sptr=='+') {
   ++sptr;
   if(sptr==Zptr) {
     WY_RETURN(Wym_ENOENT);
   }
 } else if(*sptr=='-') {
   ++sptr;
   if(sptr==Zptr) {
     WY_RETURN(Wym_ENOENT);
   }
   is_neg=true;
 } else {
    // no sign symbol recognized
 };

 //  Resolve ns_radix
 //               V----------- first character after [+-]
 //  <S> ::= [+-][<p>]...
 //
 WY_ASSERT(sptr!=Zptr);
 
 if(ns_radix>Wy::Wy_MaxRadix) {
   WY_RETURN(Wym_EINVAL);
 } else if(ns_radix<Wy::Wy_MinRadix) {
   if(ns_radix!=0) {
     WY_RETURN(Wym_EINVAL);
   }

   // Determine effective ns_radix from the prefix and write back
   //
   if(sptr[0]=='0') {
     endptr=++sptr;
     if(sptr==Zptr) {
       ns_radix=10;
       return Ok;        // one digit string [+-]0
     } else {
       if((*sptr=='x')||(*sptr=='X')) {
         // [+-]0x
         ++sptr;
         if(sptr==Zptr) {
           ns_radix=10;
           WY_RETURN(Wym_EBADMSG);  // [+-]0x
         }
         if(*sptr==DPChar) {
           //  [+-]0x.?
           if(sptr+1==Zptr) {
             ns_radix=10;
             WY_RETURN(Wym_EBADMSG);  // [+-]0x.
           }
           const int tmp( Wy::_charnum(sptr[1]) );
            if((tmp<0)||(tmp>=16)) {
              // [+-]0x.<non-hex>  e.g. "0x.x"
              ns_radix=10;
              WY_RETURN(Wym_EBADMSG);
            }
            // FALLTHROUGH [+-]0x.<hex>  ---sptr pointed at '.'
         } else {
           // [+-]0x<non-dec-point>
           const int tmp( Wy::_charnum(*sptr) );
           if((tmp<0)||(tmp>=16)) {  
             // [+-]0x<non-hex>  e.g. "0xx" "0xg"
             ns_radix=10;
             WY_RETURN(Wym_EBADMSG);
           }
           // FALLTHROUGH [+-]0x<hex>???  ---sptr pointed at <hex>
         }
         ns_radix=16;
         MaxExpVal=MaxExp_16;
         MinExpVal=MinExp_16;
         exp_base=2;
       } else {
         // [+-]0<non-x>  e.g. "08" "0."
         ns_radix=10;
         if(*sptr==DPChar) {
           if(sptr+1==Zptr) {
             endptr=sptr+1;
             return(Ok); // [+-]0.
           }
           // [+-]0.? ?? ...sptr pointed  at ?
         } else {
           // [+-]<non-x-dot>
           // [+-]0??? ... sptr pointed at ?
         }
       };
     };
   } else {
     ns_radix=10;
   };
 } else {
   switch(ns_radix) {
     case 10: break;
     case 16:
             if(Zptr-sptr>=2) {
               if(sptr[0]=='0') {
                 endptr=++sptr;
                 if((*sptr=='x')||(*sptr=='X')) {
                   ++sptr;
                   if(sptr==Zptr) {
                     WY_RETURN(Wym_EBADMSG);  // [+-]0x
                   }
                   // FALL_THROUGH  [+-]0x? --- sptr pointed at ?
                 };
               };
             };
             exp_base=2;
             MaxExpVal=MaxExp_16;
             MinExpVal=MinExp_16;
             break;
      default:
             WY_RETURN(Wym_EINVAL);
   };
 };
 const IFN_Type IFN_MultMax=IFN_MaxVal/ns_radix;
 const IFN_Type IFN_MultMin=IFN_MinVal/ns_radix;
 const IFN_Type IFN_MultMax_ddd=IFN_MaxVal_ddd/ns_radix;
 const IFN_Type IFN_MultMin_ddd=IFN_MinVal_ddd/ns_radix;

 //               V----------- sptr= first character after <p>
 //        [+-]<p>?
 //
 //          <S1>::= <ddd>[.<ggg>]
 //          <S2>::= <ddd>.
 //          <S3>::= .<ggg>
 //
 WY_ASSERT(sptr!=Zptr);
 if(*sptr==DPChar) {
   ++sptr;
   if(sptr==Zptr) {
     if(endptr==nseg.begin()) {
       WY_RETURN( Wym_ENOENT );  //  [+-].
     } else {
       WY_RETURN( Wym_EBADMSG ); //  [+-]0x.
     }
   }
   if(is_neg) {
     goto Resolve_ggg_neg;
   } else {
     goto Resolve_ggg_pos;
   }
 }
 if(is_neg) {
   goto Resolve_S1S2_neg;
 } else {
   goto Resolve_S1S2_pos;
 }

//-----------
Precision_Skip_ggg:;
//-----------
 WY_ASSERT (sptr!=Zptr);
 for(;sptr!=Zptr; ++sptr) {
    const int d0= Wy::_charnum(*sptr);
    if((d0<0)||(d0>=ns_radix)) {
      goto Resolve_Exponent;
    }
 };
 endptr=sptr;
 r=Ok;
 goto AnsExit;     // [+-][<p>][<ddd>].<ggg>

//-----------
AnsExit:;
//-----------
 switch(ns_radix) {
   case 10: {
              WY_ASSERT(exp_base==ns_radix);
              int ee=ifn_sht+iex_num;
              if(ee>=0) {
                if(ee>=(int)Size_p10tab) {
                  // see code in Resolve_Exponent (should not occur)
                  WY_TERMINATE("");
                }
                res.reset(ifn_num*p10tab[ee], 0);
              } else {
                while(ee<=-(int)(Size_p10tab)) {
                  const std::lldiv_t dt=std::lldiv(ifn_num,WY__CONST_GIGA);
                  ifn_num=dt.quot;
                  ee+=9;
                }
                const std::lldiv_t dt=std::lldiv(ifn_num,p10tab[-ee]);
                if(ee>=-9) {
                  res.reset(dt.quot, dt.rem*p10tab[9+ee]);
                } else {
                  res.reset(dt.quot, dt.rem/p10tab[-9-ee]);
                }
              };
              //res= ifn_num*std::pow(10.0, ifn_sht+iex_num);
              return(r);
            }
   case 16:
            WY_ASSERT(exp_base==2);
            const int ee=(ifn_sht<<2)+iex_num;
            if(ee>=0) {
              res.reset(ifn_num*(1<<ee),0);
            } else {
              std::lldiv_t dt=std::lldiv(ifn_num,1<<-ee);
              res.reset(dt.quot, (dt.rem*WY__CONST_GIGA)/(1<<-ee));
            }
            return(r);
 };
 WY_TERMINATE("UNREACHABLE");

//----------------
Resolve_S1S2_pos:;
//----------------
 WY_ASSERT(sptr!=Zptr);
 for(;;) {
   const int d0= Wy::_charnum(*sptr);
   if((d0<0)||(d0>=ns_radix)) {
     if(endptr==nseg.begin()) {
       WY_RETURN( Wym_ENOENT );   // [+]?
     }
     break;
   }
   if(ifn_num>IFN_MultMax_ddd) {
     endptr=sptr;
     r.reset( Wym_ERANGE ); WY_HERE(r);
     goto AnsExit;
   }
   const IFN_Type tmp1=ifn_num*ns_radix;
   if(tmp1>IFN_MaxVal_ddd-d0) {
     endptr=sptr;
     r.reset( Wym_ERANGE ); WY_HERE(r);
     goto AnsExit;
   }
   if( (ifn_num=tmp1+d0)!=0 ) {
     ++ifn_digs;
   }
   if( (endptr=++sptr)==Zptr ) {
     r.reset(Ok);
     goto AnsExit;           // [+]<p><ddd>
   }
 };
 if(*sptr!=DPChar) {
   goto Resolve_Exponent;
 }
 ++sptr;
 if(sptr==Zptr) {
   endptr=sptr;
   r.reset(Ok);
   goto AnsExit;        // [+]<p><ddd>.
 }
 // FALLTHROUGH to Resolve_ggg_pos

//-------------------
Resolve_ggg_pos:;
//-------------------
 WY_ASSERT(sptr!=Zptr);
 for(const int MinSht= std::numeric_limits<int>::min()/8;;) {
   const int d0= Wy::_charnum(*sptr);
   if((d0<0)||(d0>=ns_radix)) {
     if(endptr==nseg.begin()) {
       WY_RETURN( Wym_ENOENT );   // [+]<p>.
     }
     break;
   }
   if(ifn_num>IFN_MultMax) {
     goto Precision_Skip_ggg;
   }
   const IFN_Type tmp1=ifn_num*ns_radix;
   if(tmp1>IFN_MaxVal-d0) {
     goto Precision_Skip_ggg;
   }
   if(ifn_sht<=MinSht) {      // prevent ifn_sht overflow
     goto Precision_Skip_ggg; //  e.g. 0.000..repeat 6553600 times or longer
   }
   if( (ifn_num=tmp1+d0)!=0 ) {
     ++ifn_digs;
   }
   --ifn_sht;
   if( (endptr=++sptr)==Zptr ) {
     r.reset(Ok);
     goto AnsExit;           // [+][<p>][<ddd>].<ggg>
   }
 };
 goto Resolve_Exponent;

//----------------
Resolve_S1S2_neg:;
//----------------
 WY_ASSERT(sptr!=Zptr);
 for(;;) {
   const int d0= Wy::_charnum(*sptr);
   if((d0<0)||(d0>=ns_radix)) {
     if(endptr==nseg.begin()) {
       WY_RETURN( Wym_ENOENT );   // -?
     }
     break;
   }
   if(ifn_num<IFN_MultMin_ddd) {
     endptr=sptr;
     r.reset( Wym_ERANGE ); WY_HERE(r);
     goto AnsExit;
   }
   const IFN_Type tmp1=ifn_num*ns_radix;
   if(tmp1<IFN_MinVal_ddd+d0) {
     endptr=sptr;
     r.reset( Wym_ERANGE ); WY_HERE(r);
     goto AnsExit;
   }
   if( (ifn_num=tmp1-d0)!=0 ) {
     ++ifn_digs;
   }
   if( (endptr=++sptr)==Zptr ) {
     r.reset(Ok);
     goto AnsExit;           // -<p><ddd>
   }
 };
 if(*sptr!=DPChar) {
   goto Resolve_Exponent;
 }
 ++sptr;
 if(sptr==Zptr) {
   endptr=sptr;
   r.reset(Ok);
   goto AnsExit;        // -<p><ddd>.
 }
 // FALLTHROUGH to Resolve_ggg_neg

//-------------------
Resolve_ggg_neg:;
//-------------------
 WY_ASSERT(sptr!=Zptr);
 for(const int MinSht= std::numeric_limits<int>::min()/8;;) {
   const int d0= Wy::_charnum(*sptr);
   if((d0<0)||(d0>=ns_radix)) {
      if(endptr==nseg.begin()) {
        WY_RETURN( Wym_ENOENT );   // -[<p>]?
      }
      break;
   }
   if(ifn_num<IFN_MultMin) {
     goto Precision_Skip_ggg;
   }
   const IFN_Type tmp1=ifn_num*ns_radix;
   if(tmp1<IFN_MinVal+d0) {
     goto Precision_Skip_ggg;
   }
   if(ifn_sht<=MinSht) {      // prevent ifn_sht overflow
     goto Precision_Skip_ggg; //  e.g. 0.000..repeat 6553600 times or longer
   }
   if( (ifn_num=tmp1-d0)!=0 ) {
      ++ifn_digs;
   }
   --ifn_sht;
   if( (endptr=++sptr)==Zptr ) {
     r.reset(Ok);
     goto AnsExit;           // -[<p>][<ddd>].<ggg>
   }
 };
 // FALLTHROUGH to Resolve_Exponent

//------------------------
//                    V ------------- sptr= first character of <Exponent>
//   [+-][<p>]......[<Exponent>]
//
Resolve_Exponent:;
//------------------------
 const int ExpStr_Radix=10;    // radix of exponent part of the string pointed by nseg

 // FIXME: There are corner cases that the number with exponent IEX_MaxVal+1 are really
 //        convertible. Because of impl. time problem (and proto is in evaluation), 
 //        limit to const value. Result is smaller range of accepted number string
 //        e.g. "123456789.7e1"
 //
 IEX_Type IEX_MaxVal= std::numeric_limits<WyTimeSpec::SecondType>::digits10;

 WY_ASSERT(sptr!=Zptr);
 switch(ns_radix) {
   case 10:
            switch(*sptr) {
              case 'e':
              case 'E': break;
              default:
                       endptr=sptr;
                       r.reset(Wym_EBADMSG); WY_HERE(r);
                       goto AnsExit;
            };
            exp_base=10;
            IEX_MaxVal-=(ifn_digs+ifn_sht);
            break;
   case 16:
            switch(*sptr) {
              case 'p':
              case 'P': break;
              default:
                       endptr=sptr;
                       r.reset(Wym_EBADMSG); WY_HERE(r);
                       goto AnsExit;
            };
            exp_base=2;

            // FIXME: There are corner cases that the number with exponent IEX_MaxVal+1 are really
            //        convertible. Because of impl. time problem (and proto is in evaluation), 
            //        limit to const value. Result is smaller range of accepted number string
            //
            IEX_MaxVal= std::numeric_limits<WyTimeSpec::SecondType>::digits-((ifn_digs+ifn_sht)<<2);
            break;
   default:
            WY_TERMINATE("");
 };
 
 {
   //
   //  Resolve the exponent sign symbol
   //
   ++sptr;
   if(sptr==Zptr) {
     r.reset( Wym_EBADMSG ); WY_HERE(r);
     goto AnsExit;
   }
   if(*sptr=='+') {
     ++sptr;
     if(sptr==Zptr) {
       r.reset( Wym_EBADMSG ); WY_HERE(r);
       goto AnsExit;         // <number string>E+
     }
   } else if(*sptr=='-') {
     ++sptr;
     if(sptr==Zptr) {
       r.reset( Wym_EBADMSG ); WY_HERE(r);
       goto AnsExit;         // <number string>E-
     }

     //  Resolve -<kkk>
     //
     for(;;) {
       const int d0= Wy::_charnum(*sptr);
       if((d0<0)||(d0>=ExpStr_Radix)) {
          r.reset( Wym_EBADMSG ); WY_HERE(r);
          break;                // <number string>E[+-][kkk]?
       }
       if(iex_num>std::numeric_limits<IEX_Type>::min()/8) {
         //
         // Because iex_num gets more negative, $res approaches toward zero.
         // We just need iex_num stop at a computable value.
         //
         iex_num= iex_num*ExpStr_Radix-d0;
       }
       if( (endptr=++sptr)==Zptr ) {
         r.reset(Ok);
         break;           // <number string>
       }
     };
     goto AnsExit; 
   } else {
      // no sign symbol recognized
   };

   //  Resolve [+]<kkk>
   //
   const IEX_Type IEX_MultMax=IEX_MaxVal/ExpStr_Radix;
   for(;;) {
     const int d0= Wy::_charnum(*sptr);
     if((d0<0)||(d0>=ExpStr_Radix)) {
        r.reset( Wym_EBADMSG ); WY_HERE(r);
        break;                // <number string>E[+-][kkk]?
     }
     if(iex_num>IEX_MultMax) {
        r.reset( Wym_ERANGE ); WY_HERE(r);
        break;
     }
     const IEX_Type tmp1=iex_num*ExpStr_Radix;
     if(tmp1>IEX_MaxVal-d0) {
       r.reset( Wym_ERANGE ); WY_HERE(r);
       break;
     }
     iex_num=tmp1+d0;
     if( (endptr=++sptr)==Zptr ) {
       r.reset(Ok); WY_HERE(r);
       break;           // <number string>
     }
   };
   goto AnsExit; 
 }
};

WyRet _strnum(WyTimeSpec& res, const WyCSeg& nseg, const Wy::WyRadix& ns_radix) WY__TSPC()
{
 Wy::WyRadix t_radix=ns_radix;
 const char* endptr;
 return Wy::wy_strnum(res,endptr,nseg,t_radix);
};

WyRet _strnum(WyTimeSpec& res, const WyCSeg& nseg, Wy::WyRadix& ns_radix) WY__TSPC()
{
 const char* endptr;
 return Wy::wy_strnum(res,endptr,nseg,ns_radix);
};

WyRet _strnum(WyTimeSpec& res, const WyCSeg& nseg) WY__TSPC()
{
 Wy::WyRadix t_radix=10;
 const char* endptr;
 return Wy::wy_strnum(res,endptr,nseg,t_radix);
};

WyRet _strnum(WyTimeSpec& res, WyCSeg& cseg, const Wy::WyRadix& ns_radix) WY__TSPC()
{
 Wy::WyRadix t_radix=ns_radix;
 const char* endptr;
 const WyRet r( Wy::wy_strnum(res, endptr, cseg, t_radix) );
 cseg.wy_set_begin(endptr);
 return(r);
};

WyRet _strnum(WyTimeSpec& res, WyCSeg& cseg, Wy::WyRadix& ns_radix) WY__TSPC()
{
 const char* endptr;
 const WyRet r( Wy::wy_strnum(res, endptr, cseg, ns_radix) );
 cseg.wy_set_begin(endptr);
 return(r);
};

WyRet _strnum(WyTimeSpec& res, WyCSeg& cseg) WY__TSPC()
{
 Wy::WyRadix t_radix=10;
 const char* endptr;
 const WyRet r( Wy::wy_strnum(res, endptr, cseg, t_radix) );
 cseg.wy_set_begin(endptr);
 return(r);
};

};  // end namespace Wy
